# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def calculate_fuel_costs(dict_ac_data, dict_operating_costs, runtime_output):
    """Calculate fuel costs.

    The fuel costs depend on the trip fuel mass and the number of annual flights. The trip fuel mass results from the
    takeoff mass minus the operating empty mass and the payload mass. As the latter is dependent on the range, the
    cruise fuel mass is also dependent on the stage length.

    This function adds the 'fuel_costs' dict to the 'dict_operating_costs' that contains the following parameters:
        - array fuel_costs: Fuel costs (range dependent) in EUR
        - float fuel_costs_design_point: Fuel costs at design point of current aircraft in EUR

    :param dict dict_ac_data: Dict containing parameters and values from aircraft exchange and module config file
    :param dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    """

    # Extract necessary variables from 'dict_ac_data'.
    fuel_price = dict_ac_data['fuel_price']
    m_operating_empty = dict_ac_data['m_operating_empty']
    m_takeoff = dict_ac_data['m_takeoff']
    # Extract necessary values from 'dict_operating_costs'.
    n_annual_flights = dict_operating_costs['dependent_parameter']['flights_per_year']
    idx_design_point = dict_operating_costs['payload_range_diagram_data']['idx_design_point']
    payloads_for_prd = dict_operating_costs['payload_range_diagram_data']['payloads_for_prd']

    # Calculate trip fuel mass.
    m_trip_fuel = m_takeoff - m_operating_empty - payloads_for_prd
    # Calculate fuel costs.
    fuel_costs = fuel_price * m_trip_fuel * n_annual_flights

    # Print.
    runtime_output.print('Fuel costs (per year): ' + f"{int(fuel_costs[idx_design_point]):,}" + ' €')

    # Prepare output data.
    dict_operating_costs['fuel_costs'] = {}
    dict_operating_costs['fuel_costs']['fuel_costs'] = fuel_costs
    dict_operating_costs['fuel_costs']['fuel_costs_design_point'] = fuel_costs[idx_design_point]

    return dict_operating_costs
