# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def calculate_landing_costs(dict_ac_data, dict_operating_costs, runtime_output):
    """Calculate landing costs.

    The annual landing costs depend on the maximum takeoff mass and the number of flights per year.

    This function adds the 'landing_costs' dict to the 'dict_operating_costs' that contains the following parameters:
        - float landing_costs: Landing costs per year in EUR

    :param dict dict_ac_data: Dict containing parameters and values from aircraft exchange and module config file
    :param dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    """

    # Extract necessary variables from 'dict_ac_data'.
    fees_landing = dict_ac_data['fees_landing']
    m_takeoff_max = dict_ac_data['m_takeoff_max']
    # Extract necessary values from 'dict_operating_costs'.
    n_annual_flights = dict_operating_costs['dependent_parameter']['flights_per_year']

    # Calculate landing costs.
    landing_costs = m_takeoff_max * fees_landing * n_annual_flights

    # Print.
    runtime_output.print('Landing costs (per year): ' + f"{int(landing_costs):,}" + ' €')

    # Prepare output data.
    dict_operating_costs['landing_costs'] = {}
    dict_operating_costs['landing_costs']['landing_costs'] = landing_costs

    return dict_operating_costs
