# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard libraries.
from numpy import array, nonzero


def calculate_corrected_seat_kilometer_costs(dict_ac_data, dict_operating_costs, runtime_output):
    """Calculate corrected seat kilometer costs.

    A method of freight equivalent passenger seats is applied.

    This function adds the 'corrected_seat_kilometer_costs' dict to the 'dict_operating_costs' that contains the
    following parameters:
        - array corrected_seat_kilometer_costs: Corrected seat kilometer costs (range dependent) in EUR/SKO

    :param dict dict_ac_data: Dict containing parameters and values from aircraft exchange and module configuration file
    :param dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict dict_operating_costs: Dict containing parameters from cost estimation calculation
    """

    # Extract necessary variables from 'dict_ac_data'.
    m_passenger_incl_luggage = dict_ac_data['m_passenger'] + dict_ac_data['m_luggage']
    revenue_per_freight_km = dict_ac_data['revenue_per_freight_km']
    # Extract necessary values from 'dict_operating_costs'.
    seat_kilometer_costs = dict_operating_costs['seat_kilometer_costs']['seat_kilometer_costs']
    n_pax_for_prd = dict_operating_costs['payload_range_diagram_data']['n_pax_for_prd']
    n_points = dict_operating_costs['payload_range_diagram_data']['n_points']
    payloads_for_prd = dict_operating_costs['payload_range_diagram_data']['payloads_for_prd']
    revenue_per_seat_and_flight = dict_operating_costs['dependent_parameter']['revenue_per_seat_and_flight']

    # Definition of function variables.
    ACCELERATION_OF_GRAVITY = 9.80665

    # # Calculate.
    # idx_prd_point_a = 0
    # idx_prd_point_b = n_points[0]
    # points = len(n_points)
    # max_number_of_pax = max(n_pax_for_prd)
    # max_weight_pax = ACCELERATION_OF_GRAVITY * m_passenger_incl_luggage * max_number_of_pax
    # # First n entries can be obtained using the payloads for PRD directly.
    # weight_payload = [ACCELERATION_OF_GRAVITY * payloads_for_prd[i] for i in range(points)]
    # delta_weight_payload = ACCELERATION_OF_GRAVITY * (payloads_for_prd[idx_prd_point_a]
    #                                                   - payloads_for_prd[idx_prd_point_b])
    # # Entries between points A and B must be calculated.
    # step = delta_weight_payload / points
    # idx = 0
    # while weight_payload[(points - 1) + idx] > max_weight_pax:
    #     weight_payload.append(weight_payload[(points - 1) + idx] - step)
    #     idx += 1
    # weight_payload[-1] = max_weight_pax
    # # All remaining points are constant.
    # array([weight_payload.append(max_weight_pax) for i in range((points * 3) - (points + idx))])
    # # Calculate revenue cargo.
    # revenue_cargo = revenue_per_freight_km * (weight_payload - max_weight_pax)
    # equivalent_seat_revenue = revenue_cargo / revenue_per_seat_and_flight
    # # Replace zeros to eliminate 'invalid value encountered in divide' warning.
    # zero_n_pax = nonzero(n_pax_for_prd == 0)[0]
    # n_pax_for_prd[zero_n_pax] = 1e-8

    # Calculate corrected seat kilometer costs.
    corrected_seat_kilometer_costs = 0

    # Prepare output data.
    dict_operating_costs['corrected_seat_kilometer_costs'] = {}
    dict_operating_costs['corrected_seat_kilometer_costs']['corrected_seat_kilometer_costs'] = \
        corrected_seat_kilometer_costs

    return dict_operating_costs
