# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing export functionalities for current calculation method."""
# Import standard libraries.
import xml.etree.ElementTree as ET


def method_xml_export(paths_and_names, routing_dict, data_dict, method_specific_output_dict, xml_export_tree,
                      path_to_results_file, runtime_output):
    """Export function.

    This function is responsible for the export of method-specific data to the corresponding method-specific XML. In
    detail, this includes the following steps:
        (1) Parse the XML file specified by the 'path_to_results_file' variable.
        (2) Find the 'calculation_results' element in the XML file. This is the element under which method-specific
        nodes will be added.
        (3) Extract method-related information from the configuration file, such as the method name and description and
        add method node.
        (4) Extract design mission data from the method-specific output dictionary and write it to the XML file.
        (5) If a study exists, extract study data from the method-specific output dict and write it to the XML file.
        (6) Attempt to write the modified XML data back to the 'costEstimation_results.xml' file. Handle an OSError
        exception in case an error occurs during this operation.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict data_dict: Dictionary containing results of module execution
    :param dict method_specific_output_dict: Dictionary containing method-specific output data
    :param ElementTree xml_export_tree: Element tree of method-specific XML tree
    :param str path_to_results_file: Path to method-specific output XML file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :raises OSError: Raised if writing to aircraft exchange file failed
    :return: None
    """
    runtime_output.print("Method-specific data are written to '" + routing_dict['module_name'] + "_results.xml'...")

    # Function to write data to 'cost_estimation_results.xml'.
    root_of_results_file = xml_export_tree.getroot()
    parent = root_of_results_file.find('calculation_results')
    # Add method node.
    root_of_module_config_tree = paths_and_names['root_of_module_config_tree']
    method_name = root_of_module_config_tree.find('./program_settings/configuration/method_name/value').text
    method_description = ("Empirical method to estimate the direct operating costs (DOC) and indirect operating costs"
                          " (IOC) of an aircraft.")
    child = ET.SubElement(parent, method_name)
    child.set("description", method_description)

    # Extract design mission data from 'method_specific_output_dict' and prepare ElementTree for export to
    # module-specific XML.
    specific_output_dict_design = method_specific_output_dict['specific_output_dict_design']
    prepare_element_tree_for_module_specific_export(root_of_results_file, specific_output_dict_design)

    # If study mission exists, extract study mission data from 'method_specific_output_dict' and prepare ElementTree
    # for export to module-specific XML.
    study_exists = ('specific_output_dict_study' in method_specific_output_dict.keys())
    if study_exists:
        specific_output_dict_study = method_specific_output_dict['specific_output_dict_study']
        prepare_element_tree_for_module_specific_export(root_of_results_file, specific_output_dict_study)

    # Write all parameters to export file.
    try:
        # Ensure proper indentation.
        ET.indent(root_of_results_file, space="    ", level=0)
        # Write data to file.
        xml_export_tree.write(path_to_results_file)
    # Exception handling for operating system error.
    except OSError:
        runtime_output.error('Error: Writing to aircraft exchange file failed. Program aborted!')


def prepare_element_tree_for_module_specific_export(root_of_results_file, specific_output_dict):
    """ Prepare ElementTree for module-specific results export.

    This function is responsible for preparing the ElementTree of the module-specific export file.
    In summary, the code dynamically updates an XML structure based on a list of paths and a dictionary
    ('specific_output_dict'). It ensures that the specified paths exist in the XML structure and creates the necessary
    sub-elements along the way, setting attributes and text values as specified in 'specific_output_dict'.

    :param ElementTree root_of_results_file: Root of method-specific export ElementTree
    :param dict specific_output_dict: Dictionary containing method-specific output data
    :return: None
    """
    # Extract 'list_of_paths' from 'specific_output_dict' and delete it from dictionary.
    list_of_paths = specific_output_dict['list_of_paths']
    del specific_output_dict['list_of_paths']
    # Iterate over paths.
    for current_path in list_of_paths:
        # Check, if 'current_path' exists in XML structure and generate path and sub-elements if not.
        if root_of_results_file.find(current_path) is None:
            # Split path into path_parts using '/' as delimiter, excluding first empty part.
            path_parts = current_path.split('/')[1:]
            # Initialize 'path_to_check' with the root element ('.').
            path_to_check = '.'
            # Iterate over 'path_parts'.
            for part in path_parts:
                # Find parent element corresponding to current 'path_to_check'.
                parent_path = root_of_results_file.find(path_to_check)
                # Update 'path_to_check' by appending the current part.
                path_to_check += ('/' + part)
                # Check, if updated 'path_to_check' does not exist in the XML structure.
                if root_of_results_file.find(path_to_check) is None:
                    # Check, if 'part' is in 'specific_output_dict'.
                    if specific_output_dict[part] is not None:
                        # Create sub-element.
                        new_node = ET.SubElement(parent_path, part)
                        # Add attributes (if necessary).
                        if 'attributes' in specific_output_dict[part]:
                            for key, value in specific_output_dict[part]['attributes'].items():
                                new_node.set(key, value)
                        # Add further sub-elements if defined.
                        if 'parameters' in specific_output_dict[part]:
                            current_path = root_of_results_file.find(path_to_check)
                            for key, value in specific_output_dict[part]['parameters'].items():
                                parameter_node = ET.SubElement(current_path, key)
                                parameter_node.text = str(value)
