# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing calculating functions for the operating cost estimation according to TU Berlin."""
# Import standard libraries.
import importlib
import os
import sys


def operating_cost_estimation(paths_and_names, routing_dict, dict_ac_data, runtime_output):
    """Operating cost estimation method.

    [Add some text here...]
    The output dictionary 'dict_operating_costs' contains the results of the cost estimation and is structured according
    to the following scheme:
        kerosene_output_dict = {'group_name_1': {'parameter_1': value, 'parameter_2': value, ...},
                                'group_name_2': {'parameter_1': value, 'parameter_2': value, ...},
                                 ...}
    
    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict dict_ac_data: Dictionary containing data from aircraft exchange and module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :raises ModuleNotFoundError: Raised if calculation module not found
    :return dict dict_operating_costs: Dictionary containing operating cost estimation results
    """

    # Initialize local parameter.
    function_names = [
        # Data preparation function.
        'determine_route_and_class_dependent_parameter',
        # Route independent costs functions.
        'calculate_capital_costs', 'calculate_crew_costs', 'calculate_route_independent_costs',
        # Route dependent costs functions.
        'prepare_payload_range_diagram_data', 'calculate_fuel_costs', 'calculate_handling_costs',
        'calculate_landing_costs', 'calculate_air_traffic_control_costs', 'calculate_maintenance_costs',
        'calculate_route_dependent_costs',
        # Calculate direct operating costs function.
        'calculate_direct_operating_costs',
        # Related DOC functions.
        'calculate_flight_kilometer_costs', 'calculate_seat_kilometer_costs',
        'calculate_corrected_seat_kilometer_costs', 'calculate_ton_kilometer_costs',
        'calculate_revenue_passenger_kilometer_costs',
        # Indirect operating costs function.
        'calculate_indirect_operating_costs'
    ]
    # Clean strings in 'function_names' list to remove underscores.
    cleaned_function_names = [name.replace('_', '') for name in function_names]
    dict_operating_costs = {}

    # If the script is compiled, sys._MEIPASS gives the temp directory for PyInstaller.
    if getattr(sys, 'frozen', False):
        # Read all python files from 'call_functions' directory, delete cache folder, and sort 'call_function_list'.
        call_function_list = os.listdir(sys._MEIPASS + '/'
                                        + routing_dict['module_import_name'].replace('.', '/') + '/general/call_functions')
    # If running as a regular Python script, use the current directory.
    else:
        # Read all python files from 'call_functions' directory, delete cache folder, and sort 'call_function_list'.
        call_function_list = os.listdir(paths_and_names['working_directory'] + '/'
                                        + routing_dict['module_import_name'].replace('.', '/') + '/general/call_functions')
        
    if '__pycache__' in call_function_list:
        call_function_list.remove('__pycache__')
    if '.DS_Store' in call_function_list:
        call_function_list.remove('.DS_Store')
    call_function_list = sorted(call_function_list, key=lambda x: (int(x.split('_')[1]), x))

    # Loop across all *.py files from 'call_function' directory and find matching entries in function_names.
    for call_function in call_function_list:
        try:
            # Generate function import string.
            import_call_function = (routing_dict['module_import_name']
                                    + '.general.call_functions.' + os.path.splitext(call_function)[0])
            # Extract function name from 'function_names' list.
            for idx, cleaned_name in enumerate(cleaned_function_names):
                # Check if the cleaned name is present in the call_function string.
                if cleaned_name in call_function:
                    function_name = function_names[idx]
                    break
            # Import calculation module.
            module = importlib.import_module(import_call_function)
            # Call function depending on routing parameter.
            dict_operating_costs = getattr(module, function_name)(dict_ac_data, dict_operating_costs, runtime_output)

        # Exception handling for module import error.
        except ModuleNotFoundError as module_import_error:
            runtime_output.print(str(module_import_error) + ' found in ' + routing_dict['module_name']
                                 + '. Program aborted.')
            sys.exit('Exit information: The own python module to be imported could not be found!')

    runtime_output.info('Starting operating cost estimation.')

    return dict_operating_costs
