# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing calculation functions provided by the user."""
# Import standard modules.
import os
import sys
import xml.etree.ElementTree as ET

# Import own modules.
from src.tube_and_wing.empirical.operating_cost_estimation_tu_berlin.general.operatingcostestimationtuberlin \
    import operating_cost_estimation


def method_kerosene(paths_and_names, routing_dict, dict_ac_exchange, dict_mod_config, runtime_output):
    """Operating cost estimation method according to TU Berlin for kerosene-powered aircraft.

    This function performs the operating cost estimation according to the TU Berlin method for kerosene-powered aircraft
    configurations.
    [Add more information here...]
    The output dictionary 'kerosene_output_dict' contains the results of the cost estimation and is structured according
    to the following scheme:
        kerosene_output_dict = {'design':
                                {'group_name_1': {'parameter_1': value, 'parameter_2': value, ...},
                                 'group_name_2': {'parameter_1': value, 'parameter_2': value, ...},
                                 ...},
                                'study':
                                 {'group_name_1': {'parameter_1': value, 'parameter_2': value, ...},
                                  'group_name_2': {'parameter_1': value, 'parameter_2': value, ...},
                                  ...}}
    Note that if the study mission does not exist in aircraft exchange file, the 'kerosene_output_dict' does only
    contain the 'design' sub-dictionary.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict dict_ac_exchange: Dict containing parameters and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameters and according values from module configuration file
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict kerosene_output_dict: Dictionary containing results from calculation for kerosene-powered aircraft
    """
    """ Add some additional values to dictionaries and merge them."""
    # Check if 'study_mission.xml' exists and try to read study mission data.
    path_to_study_mission_xml = paths_and_names['project_directory'] + '/mission_data/study_mission.xml'
    study_mission_exists = os.path.exists(path_to_study_mission_xml)
    if study_mission_exists:
        try:
            # Attempt to create an ElementTree and get the root element from the XML file.
            study_mission_xml_tree = ET.ElementTree(file=path_to_study_mission_xml)
            root_of_study_mission_xml_tree = study_mission_xml_tree.getroot()
            # Get values from 'study_mission.xml' file.
            m_payload_study_mission = root_of_study_mission_xml_tree.findall('./payload/value')
            # Check, if values exist.
            # If the value exists, set to aircraft exchange dictionary and start further estimations.
            if m_payload_study_mission is not None:
                dict_ac_exchange['m_payload_study'] = round(float(m_payload_study_mission[0].text), 4)
            # Else condition: The values are not given -> raise an error and abort program.
            else:
                runtime_output.critical(
                    'Error: The required payload value for the study mission is not found in the "study_mission.xml" '
                    + 'file. \n'
                    '                                            Program aborted!')
                sys.exit('Exit information: Missing required parameter in the "study_mission.xml" file!')

        # Exception handling for operating system (OS) error.
        except OSError:
            runtime_output.critical('Error: The study mission file: "study_mission.xml'
                                    + '" was found but could not be read. \n'
                                    '                                            Program aborted!')
            sys.exit('Exit information: Study mission file could not be read!')
    else:
        runtime_output.print('Attention! No "study_mission.xml" file found. '
                             + 'Direct operating costs only calculated for design mission.')

    # Calculate payload mass for design mission and add to dictionary.
    dict_ac_exchange['m_payload_design'] = (
        dict_ac_exchange['n_passengers']*(dict_ac_exchange['m_passenger'] + dict_ac_exchange['m_luggage'])
        + dict_ac_exchange['m_cargo'])

    # Merge both data dictionaries.
    dict_complete_data = dict_ac_exchange | dict_mod_config
    # Extract general parameters from 'complete_data_dict'.
    dict_general_data = {key: value for key, value in dict_complete_data.items() if
                         not key.endswith('_design') and not key.endswith('_study')}

    """Design mission."""
    # Extract design mission related parameters from 'complete_data_dict' and save in 'design_data' (w/o ending).
    dict_design_data = {key.replace('_design', ''): value for key, value in dict_complete_data.items()
                        if key.endswith('_design')}
    # Merge design and general data dictionary.
    dict_design = dict_general_data | dict_design_data
    # Iterate over dictionary.
    for key, value in dict_design.items():
        # Check, if any value is 'None'.
        if value is None:
            # If any value is 'None', the calculation of the costs is impossible and the program aborted.
            runtime_output.critical('Error: Operating cost estimation not possible due to missing parameter ('
                                    + key + ')! '
                                    + 'Program aborted.')
            sys.exit(1)
    # Calculate costs for design mission.
    runtime_output.print('----------------------------------------------------------')
    runtime_output.print('Operating cost estimation results for design mission.   ')
    runtime_output.print('----------------------------------------------------------')
    kerosene_output_dict_design = operating_cost_estimation(paths_and_names, routing_dict, dict_design, runtime_output)

    """Mission study."""
    if study_mission_exists:
        # Extract study mission related parameters from 'complete_data_dict' and save in 'study_data' (w/o ending).
        dict_study_data = {key.replace('_study', ''): value for key, value in dict_complete_data.items()
                        if key.endswith('_study')}
        # Check, if study mission data complete.
        # Assumption: If any value is 'None', costs cannot be calculated for mission study.
        mission_study_values_complete = not any(value is None for value in dict_study_data.values())
        if mission_study_values_complete:
            # Merge dictionaries.
            dict_study = dict_general_data | dict_study_data
            # Calculate costs for mission study.
            runtime_output.print('Operating cost estimation results for study mission.   ')
            runtime_output.print('----------------------------------------------------------')
            kerosene_output_dict_study = operating_cost_estimation(paths_and_names, routing_dict, dict_study,
                                                                runtime_output)
            # Save both dictionaries in 'kerosene_output_dict'.
            kerosene_output_dict = {'design': kerosene_output_dict_design, 'study': kerosene_output_dict_study}
        else:
            for key, value in dict_study_data.items():
                # Check, if any value is 'None'.
                if value is None:
                    # If any value is 'None', the calculation of the costs is impossible and the program aborted.
                    runtime_output.warning('Warning: Operating cost estimation of study mission not possible due to '
                                           + 'missing parameter (' + key + ')! '
                                            + 'Program aborted.')
                    runtime_output.print('----------------------------------------------------------')
                    kerosene_output_dict = {'design': kerosene_output_dict_design}

    return kerosene_output_dict
