/**
 * @file parameter_setter.cpp
 * @author Gerrit Pauls (gerrit.pauls@tuhh.de)
 * @brief 
 * @version 0.1
 * @date 2024-08-30
 * 
 * @copyright Copyright (c) 2024
 * 
 */

#include <aixml/endnode.h>
#include <string>
#include <moduleBasics/module.h>
#include <moduleBasics/modeSelector.h>
#include <unordered_map>
#include "standard_mission/standard_mission.h"

void standard_mission::set_module_settings() {
    /* Get settings from config and aircraft exchange file */
    std::string subnode("program_settings/" + this->mission_type);
    this->mission_xml_name =
        EndnodeReadOnly<std::string>("requirements_and_specifications/mission_files/" + this->mission_type + "_file").read(rtIO->acxml).value();
    this->auto_select_initial_cruise_altitude =
        EndnodeReadOnly<bool>(subnode + "/auto_select_initial_cruise_altitude").read(rtIO->moduleConfig).value();
    this->auto_select_flight_level =
        EndnodeReadOnly<bool>(subnode + "/auto_select_flight_level/enable").read(rtIO->moduleConfig).value();
    this->round_to_regular_flight_level =
        EndnodeReadOnly<bool>(subnode + "/round_to_regular_flight_level").read(rtIO->moduleConfig).value();
    this->no_steps =
        EndnodeReadOnly<bool>(subnode + "/auto_select_flight_level/no_steps").read(rtIO->moduleConfig).value();
    this->auto_climb_altitude_steps =
        EndnodeReadOnly<double>(subnode + "/auto_climb_altitude_steps").read(rtIO->moduleConfig).value();
    this->auto_rate_of_climb_steps =
        EndnodeReadOnly<double>(subnode + "/auto_rate_of_climb_steps").read(rtIO->moduleConfig).value();
    // Catch missing auto climb parameters if auto_select_flight_level is turned on
    if (this->auto_select_flight_level && (accuracyCheck(this->auto_climb_altitude_steps, 0., ACCURACY_LOW) ||
                                                accuracyCheck(this->auto_rate_of_climb_steps, 0., ACCURACY_LOW))) {
        myRuntimeInfo->warn << "auto_select_optimum_flight_level was switched on, but auto climb parameters were not set (auto_climb_altitude_steps, auto_rate_of_climb_steps)."
                            << "auto_select_optimum_flight_level will be ignored!" << std::endl;
        this->auto_select_flight_level = false;
    }
    // Correct contradiction on user input. no_steps has no effect when auto_select_flight_level is active
    if (this->no_steps && this->auto_select_flight_level)
        this->no_steps = false;
}

void standard_mission::set_range_and_payload() {
    std::string subpath("requirements_and_specifications");
    /* Set range for all mission types */
    this->range.set_value(
        EndnodeReadOnly<double>(subpath + "/requirements/top_level_aircraft_requirements/" + this->mission_type +  "/range").read(rtIO->acxml).value());
    subpath += "/design_specification/transport_task";
    /* Set cargo mass for all mission types */
    this->cargo_mass.set_value(EndnodeReadOnly<double>(subpath + "/cargo_definition/additional_cargo_mass").read(rtIO->acxml).value());
    /* Set number of passengers for design mission */
    this->pax_number.set_value(EndnodeReadOnly<int>(subpath + "/passenger_definition/total_number_passengers").read(rtIO->acxml).value());
    if (this->mission_type == "study_mission") { // alter masses via the stated fraction for study mission
        subpath = "requirements_and_specifications/requirements/top_level_aircraft_requirements/study_mission/payload_fractions";
        this->cargo_mass *= EndnodeReadOnly<double>(subpath + "/cargo_fraction").read(rtIO->acxml).value();
        this->pax_number.set_value( // ceil PAX to next integer value
            ceil(this->pax_number.value() * EndnodeReadOnly<double>(subpath + "/passenger_mass_fraction").read(rtIO->acxml).value()));
        subpath = "requirements_and_specifications/design_specification/transport_task";
    }
    /* Set payload for design mission */
    this->payload.set_value(this->cargo_mass.value() + this->pax_number.value() *
        (EndnodeReadOnly<double>(subpath + "/passenger_definition/mass_per_passenger").read(rtIO->acxml).value() +
        EndnodeReadOnly<double>(subpath + "/passenger_definition/luggage_mass_per_passenger").read(rtIO->acxml).value()));
    if (this->payload.value() > EndnodeReadOnly<double>(
            "requirements_and_specifications/requirements/top_level_aircraft_requirements/maximum_structrual_payload_mass").read(rtIO->acxml).value()) {
        throwError(__FILE__, __func__, __LINE__, "Payload mass exceeds overall structual payload limit. Abort program!");
    }
}

void standard_mission::set_cruise() {
    /* Get the maximum altitude for checks */
    std::string subpath("requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_operating_altitude");
    double maximum_altitude = convertUnit(NOPREFIX, METER, HECTO, FOOT, EndnodeReadOnly<double>(subpath).read(rtIO->acxml).value());

    /* Get initial cruise altitude and Mach number */
    subpath = "requirements_and_specifications/requirements/top_level_aircraft_requirements";
    if (this->mission_type == "design_mission"
        || (this->mission_type == "study_mission" && use_design_mission_ICO)) {// Check if study mission shall inherent design mission values
        subpath += "/design_mission";
    } else {
        subpath += "/study_mission";
    }

    /* Check if ICA exceeds maximum altitude */
    this->initial_cruise_altitude =
        convertUnit(NOPREFIX, METER, HECTO, FOOT, EndnodeReadOnly<double>(subpath + "/initial_cruise_altitude").read(rtIO->acxml).value());
    if (this->initial_cruise_altitude > maximum_altitude) {
        myRuntimeInfo->err << "Initial cruise altitude exceeds the aircrafts maximum altitude. Abort!" << std::endl;
        exit(1);
    }

    subpath = "requirements_and_specifications/requirements/top_level_aircraft_requirements"; // Reset subpath
    if (this->mission_type == "design_mission"
        || (this->mission_type == "study_mission" && use_design_mission_mach)) {// Check if study mission shall inherent design mission values
        subpath += "/design_mission";
    } else {
        subpath += "/study_mission";
    }
    this->initial_cruise_mach.set_value(
        EndnodeReadOnly<double>(subpath + "/initial_cruise_mach_number").read(rtIO->acxml).value());
    this->delta_mach_climb_to_mach_cruise =
        EndnodeReadOnly<double>(subpath + "/climb_speed_schedule/delta_mach_climb_cruise").read(rtIO->acxml).value();

    /* Set number of cruise steps and characterize them for all mission types */
    subpath = "analysis/mission/" + this->mission_type + "/cruise";
    if (!(rtIO->acxml.find(subpath + "/cruise_steps"))) { // if cruise_steps node doesn't define cruise steps
        double first_flight_level(0.); // in [FL]
        if (!this->auto_select_initial_cruise_altitude) {
            first_flight_level = this->initial_cruise_altitude;
            // Round to regular FL - first try to round towards the closer FL
            if (this->round_to_regular_flight_level && Rounding(first_flight_level, -1) <= maximum_altitude) {
                first_flight_level = Rounding(first_flight_level, -1);
            } else if (this->round_to_regular_flight_level) { // If FL gets rounded up above maximum, round down
                first_flight_level = RoundDown(first_flight_level, -1);
            }
        } else {
            if (this->mission_type == "design_mission")
                first_flight_level = 330.;
            else // Mission type: study mission
                first_flight_level = 350.;
        }
        if (convertUnit(METER, NAUTICALMILE, this->range.value()) <= 1000.) {
            this->cruise_step_number = 1;
        } else if (convertUnit(METER, NAUTICALMILE, this->range.value()) > 1000. && convertUnit(METER, NAUTICALMILE, this->range.value()) < 5000.) {
            if (!this->no_steps)
                this->cruise_step_number = 2;
            else
                this->cruise_step_number = 1;
        } else {
            if (!this->no_steps)
                this->cruise_step_number = 3;
            else
                this->cruise_step_number = 1;
        }
        if (this->cruise_step_number == 1) {
            this->cruise_step_segments.push_back({1., convertUnit(HECTO, FOOT, NOPREFIX, METER, first_flight_level)});
        } else {
            for (uint16_t i(0); i < this->cruise_step_number; i++) {
                this->cruise_step_segments.push_back({
                    1. / this->cruise_step_number * (i + 1), convertUnit(HECTO, FOOT, NOPREFIX, METER, first_flight_level + 20. * i)});
            }
        }
    } else { // Read existing Cruise Steps
        this->cruise_step_number = rtIO->acxml.at(subpath).getVector("cruise_steps/cruise_step").size();
        subpath += "/cruise_steps/cruise_step@";
        for (uint16_t i(0); i < this->cruise_step_number; i++) {
            this->cruise_step_segments.push_back({
                EndnodeReadOnly<double>(subpath + num2Str(i) + "/relative_end_of_cruise_step").read(rtIO->acxml).value(),
                EndnodeReadOnly<double>(subpath + num2Str(i) + "/altitude").read(rtIO->acxml).value()});
        }
    }
    if (fabs(this->cruise_step_segments.back().relative_segment_length - 1.) > ACCURACY_MEDIUM) {
        myRuntimeInfo->err << "Last mission segment of " << this->mission_type << " does not reach the end of cruise. Abort program!" << std::endl;
        exit(1);
    }
}

void standard_mission::set_climb_and_descent() {
    std::string subpath("/aircraft_exchange_file/requirements_and_specifications/requirements/top_level_aircraft_requirements/" + this->mission_type);
    this->mach_climb = this->initial_cruise_mach.value() + this->delta_mach_climb_to_mach_cruise;

    /* Calibrated airspeed limits*/
    this->CAS_ATC_limit_climb =
        EndnodeReadOnly<double>(subpath + "/climb_speed_schedule/climb_speed_below_FL100").read(rtIO->acxml).value();
    this->CAS_ATC_limit_descent =
        EndnodeReadOnly<double>(subpath + "/descent_speed_schedule/descent_speed_below_FL100").read(rtIO->acxml).value();
    this->CAS_over_flight_level_100_climb =
        EndnodeReadOnly<double>(subpath + "/climb_speed_schedule/climb_speed_above_FL100").read(rtIO->acxml).value();
    this->CAS_over_flight_level_100_descent =
        EndnodeReadOnly<double>(subpath + "/descent_speed_schedule/descent_speed_above_FL100").read(rtIO->acxml).value();
    if (adapt_climb_speed) {
        /* Initialize international standard atmosphere */
        atmosphere atm;
        double delta_ISA = EndnodeReadOnly<double>(subpath + "/delta_ISA").read(rtIO->acxml).value();
        atm.setAtmosphere(0., ISA_TEMPERATURE + delta_ISA, ISA_PRESSURE);
        /* Calculate the transition altitude [m] */
        double transition_altitude = this->get_transition_altitude(1.7 * this->CAS_over_flight_level_100_climb, this->mach_climb, atm);
        /* Calculate the calibrated airspeed at transition altitude specifications [m/s] */
        double CAS_at_transition_altitude =
            convertUnit(MACH, CALIBRATEDAIRSPEED, crossover_altitude_specs, atm, this->mach_climb);
        if (transition_altitude < crossover_altitude_specs) {
            /* Set the calibrated airspeed over flight level 100 during climb */
            myRuntimeInfo->out << "(CAS over FL100 adapted from " << this->CAS_over_flight_level_100_climb << " m/s to ";
            this->CAS_over_flight_level_100_climb =
                convertUnit(KNOTS, METERPERSECOND, Rounding((convertUnit(METERPERSECOND, KNOTS, CAS_at_transition_altitude) - 5) / 10., 0) * 10. + 5);
            myRuntimeInfo->out << this->CAS_over_flight_level_100_climb << " m/s due to unfavorable climb speed schedule.)" << std::endl;
        }
    }
    /* Check if the calibrated airspeed during climb and descent above flight level 100 is smaller than the calibrated airspeed below flight level 100 */
    if (this->CAS_over_flight_level_100_climb < this->CAS_ATC_limit_climb) {
        myRuntimeInfo->warn << "Attention: During climb CAS above FL100 smaller than CAS below FL100. "
                            << "This can cause problems in mission analysis if corresponding ROC is too small for decelaration!" << std::endl;
    }
    if (this->CAS_over_flight_level_100_descent < this->CAS_ATC_limit_descent) {
        myRuntimeInfo->err << "During descent CAS above FL100 smaller than CAS below FL100. This is not sensible. Exit program." << std::endl;
        exit(1);
    }
}

void standard_mission::set_start_and_landing() {
    std::string subpath("program_settings/" + this->mission_type);
    this->takeoff_procedure = Modeselector(EndnodeReadOnly<std::string>(subpath + "/takeoff_procedure").read(rtIO->moduleConfig).value()).get_mode_numeric();
    /* Set approach procedure */
    this->approach_procedure = Modeselector(EndnodeReadOnly<std::string>(subpath + "/approach_procedure").read(rtIO->moduleConfig).value()).get_mode_numeric();
    this->approach_speed =
        EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/maximum_approach_speed").read(rtIO->acxml).value();
    /* Set ground operations */
    this->taxi_time_origin.set_value(EndnodeReadOnly<double>(subpath + "/taxi_time_origin").read(rtIO->moduleConfig).value());
    this->taxi_time_destination.set_value(EndnodeReadOnly<double>(subpath + "/taxi_time_destination").read(rtIO->moduleConfig).value());
    this->terminal_operation_time.set_value(EndnodeReadOnly<double>(subpath + "/terminal_operation_time").read(rtIO->moduleConfig).value());
    this->engine_warmup_time.set_value(EndnodeReadOnly<double>(subpath + "/engine_warmup_time").read(rtIO->moduleConfig).value());
    this->origin_airport.set_value(EndnodeReadOnly<std::string>(subpath + "/origin_airport").read(rtIO->moduleConfig).value());
    this->destination_airport.set_value(EndnodeReadOnly<std::string>(subpath + "/destination_airport").read(rtIO->moduleConfig).value());

    /* Translate taxiing mode */
    Modeselector taxi_mode(EndnodeReadOnly<std::string>(subpath + "/taxiing_procedure").read(rtIO->moduleConfig).value());
    if (taxi_mode.get_mode_numeric() == 0) {
        this->taxiing_procedure.set_value("propulsion_taxiing");
    } else if (taxi_mode.get_mode_numeric() == 1) {
        this->taxiing_procedure.set_value("electric_taxiing");
    } else {// Unknown take-off procedure, error message
        myRuntimeInfo->err << "Unkown taxiing mode. Abort!" << std::endl;
        exit(1);
    }

    /* Distance to alternate aerodrome */
    this->alternate_distance.set_value(EndnodeReadOnly<double>(subpath + "/alternate_distance").read(rtIO->moduleConfig).value());
}

void standard_mission::set_performance_limits() {
    Modeselector thrust_settings(EndnodeReadOnly<std::string>("program_settings/climb_thrust_setting").read(rtIO->moduleConfig).value());
    /* Translate modes into their real names */
    if (thrust_settings.get_mode_numeric() == 0) {
        this->climb_thrust_setting = "takeoff";
    } else if (thrust_settings.get_mode_numeric() == 1) {
        this->climb_thrust_setting = "climb";
    } else if (thrust_settings.get_mode_numeric() == 2) {
        this->climb_thrust_setting = "maximum_continuous";
    } else if (thrust_settings.get_mode_numeric() == 3) {
        this->climb_thrust_setting = "cruise";
    }
    this->maximum_rate_of_climb = EndnodeReadOnly<double>("program_settings/maximum_rate_of_climb").read(rtIO->moduleConfig).value();
    if (this->maximum_rate_of_climb < 0) {
        this->maximum_rate_of_climb = NAN; // Set to NAN -> indicates that full thrust will be used (no preset limit)
    } else {
       this->maximum_rate_of_climb = EndnodeReadOnly<double>("program_settings/maximum_rate_of_climb").read(rtIO->moduleConfig).value();
    }
}
