/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <iostream>
#include <memory>
#include <string>
#include <unitConversion/unitConversion.h>
#include "calculateEmissions_schaefer.h"
#include "schaeferIOData.h"
#include <moduleBasics/plot.h>


void LifeCycleEmissionsSchaefer::run(const std::shared_ptr<RuntimeIO> &rtIO, std::shared_ptr<node> ecoAssXML) {
    if (rtIO->aircraft_energy_carrier_type() != "kerosene" || rtIO->aircraft_configuration_type() != "tube_and_wing") {
            myRuntimeInfo->warn << "Schaefer method can only be used for kerosene powered TAW aircraft. No LCA calculation will be performed!" << std::endl;
    } else {
        myRuntimeInfo->out << "  Calculation of life cycle emissions" << std::endl;
        schaeferIOData_ = std::make_shared<SchaeferIOData>(rtIO, ecoAssXML);
        /* Production phase */
        myRuntimeInfo->out << "     Calculation of production phase..." << std::endl;
        theProduction.calculateProductionResources(schaeferIOData_, rtIO);
        theProduction.calculateProductionEmissions(schaeferIOData_);

        /* Operation phase*/
        myRuntimeInfo->out << "     Calculation of operation phase..." << std::endl;
        theOperation.calculateOperationResources(schaeferIOData_, rtIO, theProduction);
        theOperation.calculateOperationEmissions(schaeferIOData_, theProduction);

        /* End-of-Life Phase */
        myRuntimeInfo->out << "     Calculation of end of life phase..." << std::endl;
        theEndOfLife.calculateEoLResources(schaeferIOData_);
        theEndOfLife.calculateEoLEmissions(schaeferIOData_, theProduction);

        /* Development Phase (calculation is executed last because of production/EoL of test aircraft)*/
        myRuntimeInfo->out << "     Calculation of development phase..." << std::endl;
        theDevelopment.calculateDevelopmentResources(schaeferIOData_);
        theDevelopment.calculateDevelopmentEmissions(schaeferIOData_, theProduction, theEndOfLife);
    }
}

