/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <vector>
#include <algorithm>

#include "development.h"
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <unitConversion/unitConversion.h>
#include <standardFiles/functions.h>

using std::pow;

Development::Development()
    :
    myEcoDatabase(ecoDatabase()),
    Energy_total(0.),
    fuel_total(0.),
    taxiTestFactor(0.),
    testACEnergy(0.),
    testEngineEnergy(0.) {
        //ctor
    }

void Development::calculateDevelopmentResources(const std::shared_ptr<SchaeferIOData>& data) {
    /* Tests at SL und ISA+deltaISA (assumed); max. Takeoff Thrust SL and ISA (see Ruijgrok, P. 298 and ICAO 2008 P. III-2-2*/
    // TODO(franz#1#): T142 check if with offtakes (see engineSizing output cpp, here if necessary with a part released out again) at which atm. conditions engine tests
    //max. Thrust
    /*BEGINN: from engineSizing output.cpp, to determine Max. Cont. Thrust for SL */
    /*Maximum Continuous */
    std::vector<double>  N1;
    /* For bucket curve without offtakes */
    std::vector<double>  Thrust;
    std::vector<double>  FuelFlow;
    std::vector<double>  SFC;
    int ID_maxThrust(0);
    int ID_bucketSFC(0);
//    int ID_maxThrustOfftakes(0);
    do {
        if (static_cast<int>(N1.size()) == 0) {
            N1.push_back(0.7);
        } else {
            N1.push_back(N1.back() + 0.001);
        }
        /* Set engine without offtakes */
        data->myEngine.calculate_N1_with_penalties(0, 0., myAtm, Rounding(N1.back(), 4), "maximum_continuous", 0., 0.);
        Thrust.push_back(data->myEngine.get_thrust_aircraft() / data->numberOfEngines);
        FuelFlow.push_back(data->myEngine.get_aircraft_fuelflow() / data->numberOfEngines);
        SFC.push_back(FuelFlow.back() / Thrust.back());
    } while (N1.back() <= 1.);
    // Determination of the bucket point (run through from the back, as discontinuity is possible at low rotation speeds with props)
    double tmp_SFC = SFC.back();
    for (unsigned int i(SFC.size() - 2); i > 0; i--) {
        if (SFC.at(i) <= tmp_SFC) {
            tmp_SFC = SFC.at(i);
            ID_bucketSFC = i;
        } else {
            break;
        }
    }
    /* max thrust*/
    std::vector<double>::iterator it1 = max_element(Thrust.begin(), Thrust.end());
    double MaxContThrustSL(*it1);// Thrust.at(ID_maxThrust); Maximum continuous thrust at Sea Level for one scaled engine [kN]
    ID_maxThrust = static_cast<int>(it1 - Thrust.begin());
//    maxThrust.FuelFlow = FuelFlow.at(ID_maxThrust);
    double SFCMaxContThrustSL = SFC.at(ID_maxThrust); // Thrust specific fuel consumption at continuous thrust at Sea Level for one scaled engine [(kg/s)/kN]
//    maxThrust.N1 = N1.at(ID_maxThrust);
    /* bucket thrust; assumption: = const. thrust*/
    // TODO(franz#1#): T143 check if you can get a rating for TW; BucketThrust assumption wrong (at FL = 0 and Ma = 0 there is no bucket, curve goes down constantly)
    double ContThrustSL = Thrust.at(ID_bucketSFC); //Continuous thrust at Sea Level for one scaled engine [kN]
    double SFCContThrustSL = SFC.at(ID_bucketSFC); //Thrust specific fuel consumption at continuous thrust at Sea Level for one scaled engine [(kg/s)/kN]
    /*END: from engineSizing output.cpp, to determine Max. Cont. Thrust for SL */

    /* calculation unit MJ! (1 kWh = 3.6 MJ) **/
    if (data->configuration.developmentEmissionSetting != "mode_0") {
        /* Calculation of personnel expenses **/
        theEngeneering.T_eng = 1.309 * 5.18 * pow(data->OME, 0.777) * pow(convertUnit(METERPERSECOND, KILOMETERPERHOUR, data->v_max), 0.894)
                                * pow(data->configuration.strucTestAC + data->configuration.flightTestAC, 0.163);
        theEngeneering.E_elec = E_perSqmYear_dev[0] * F_netFloor_dev / spaceEfficiency_dev * theEngeneering.T_eng / t_labourYear_dev
                                    * (1 + data->configuration.engineEngineeringFactor);
        theEngeneering.E_heat = E_perSqmYear_dev[1] * F_netFloor_dev / spaceEfficiency_dev * theEngeneering.T_eng / t_labourYear_dev
                                    * (1 + data->configuration.engineEngineeringFactor);
        theEngeneering.E_total = theEngeneering.E_elec + theEngeneering.E_heat;
        myRuntimeInfo->info << "F_netFloor_dev " << F_netFloor_dev << std::endl;
        myRuntimeInfo->info << "spaceEfficiency_dev " << spaceEfficiency_dev << std::endl;
        myRuntimeInfo->info << "T_eng " << theEngeneering.T_eng << std::endl;
        myRuntimeInfo->info << "t_labourYear_dev " << t_labourYear_dev << std::endl;
        /* calculation of Tests **/
        /* according to SA by Kuno Jandaurek **/
        /* Energy of the wind tunnel test **/
        testResources.Energy_windTunnelTest = convertUnit(HOUR, SECOND, data->configuration.windTunnelTestHours) * 1.782; //Power of subsonic wind tunnel: 1782kW
        if (data->configuration.developmentEmissionSetting == "mode_1") {
            myRuntimeInfo->info << "-----Test Emissions-----" << std::endl;
            myRuntimeInfo->info << "CAUTION: Test emissions without production and end-of-life emissions!" << std::endl;
        }
        myRuntimeInfo->info << "Energy_windTunnelTest: " << testResources.Energy_windTunnelTest << std::endl;
        /* Energy of Structure tests**/
        testResources.Energy_fatigueTest = data->configuration.strucTestCycles * data->MTOM / 1000. * 1.45;
        myRuntimeInfo->info << "Energy_fatigueTest: " << testResources.Energy_fatigueTest << std::endl;
        /* Energy of System tests**/
        testResources.Energy_systemTest = convertUnit(KILO, WATTHOUR, MEGA, JOULE, (data->configuration.sysIntTestHours + data->configuration.ironBirdTestHours)
                                                      * data->propDesignPowerTotal);
        myRuntimeInfo->info << "Energy_systemTest: " << testResources.Energy_systemTest << std::endl;
        if (data->configuration.doEngineTest) {
            /* Fuel consumption for engine tests on rig (assumption: Sea Level Conditions; LTO emissions according to
               ICAO Engine Emission Databank (also NOx!); DesignMission as reference) **/
            // TODO(franz#1#): T136 NOx from TW-Decks instead of from ICAO Databank (like mission-class!))
            double rigHours;
            if (data->configuration.testETOPS) {
                if (data->configuration.inclETOPS) {
                    rigHours = data->configuration.rigTestHours;
                } else {
                    rigHours = data->configuration.rigTestHours + 3000.*0.5;
                }
            } else {
                rigHours = data->configuration.rigTestHours;
            }
            testResources.fuel_engineTestRig = data->configuration.numberOfNewEngines * rigHours * 3600.
                                               *(SFCMaxContThrustSL * MaxContThrustSL * data->configuration.MCTpercentage
                                                 + SFCContThrustSL * ContThrustSL * (1. - data->configuration.MCTpercentage));
            myRuntimeInfo->debug << "SFCMaxContThrustSL: " << SFCMaxContThrustSL << std::endl;
            myRuntimeInfo->debug << "MaxContThrustSL: " << MaxContThrustSL << std::endl;
            myRuntimeInfo->debug << "ContThrustSL: " <<  ContThrustSL << std::endl;
            /* Fuel consumption in engine tests at the FTB (assumption: cruise conditions; emissions of the
               design mission as reference, only specifications of the new engine are integrated) **/
            testResources.fuel_engineFlightTest = data->configuration.numberOfNewEngines * convertUnit(HOUR, SECOND, data->configuration.FTBHours)
                                                  * ((data->configuration.FTBEngineNo * data->configuration.FTBFuelConsumptionPerEngine) + data->fuelFlowContThrustCruise);
            myRuntimeInfo->info << "fuel_engineFlightTest: " << testResources.fuel_engineFlightTest << std::endl;
        } else {
            myRuntimeInfo->info << "No engine tests are performed because doEngineTest switch is off." << std::endl;
        }
        /* Energy of ground tests */
        testResources.Energy_groundTest = convertUnit(KILO, WATTHOUR, MEGA, JOULE, 0.95 * data->configuration.flightTestHours * data->propDesignPowerTotal);
        myRuntimeInfo->info << "Energy_groundTest: " << testResources.Energy_groundTest << std::endl;
        /* Fuel consumption for flight tests */
        if (data->configuration.inclEngineOptions)
            testResources.flightHours_flightTest = data->configuration.flightTestHours;
        else
            testResources.flightHours_flightTest = (0.63 + 0.37 * data->configuration.engineOptions) * data->configuration.flightTestHours;
        myRuntimeInfo->info << "flightHours: " << testResources.flightHours_flightTest << std::endl;
        testResources.fuel_flightTest = data->blockfuelDesign / data->blocktimeDesign *
                                        testResources.flightHours_flightTest; //strongly simplified: Assumption: Const. Fuel consumption, determined from design mission.
        /* Test total: fuel consumption (without taxi, added at calcEmission) and energy (without Fatigue Test Airframe)**/
        testResources.Energy_totalTest = testResources.Energy_windTunnelTest + testResources.Energy_fatigueTest + testResources.Energy_systemTest + testResources.Energy_groundTest;
        testResources.fuel_engineTest_total =  testResources.fuel_engineTestRig + testResources.fuel_engineFlightTest;
        myRuntimeInfo->debug << "testResources.fuel_engineTestRig: " << testResources.fuel_engineTestRig << std::endl;
        myRuntimeInfo->debug << "testResources.fuel_engineFlightTest: " << testResources.fuel_engineFlightTest << std::endl;
        testResources.fuel_totalTest = testResources.fuel_engineTest_total + testResources.fuel_flightTest;
        myRuntimeInfo->debug << "testResources.fuel_engineTest_total: " << testResources.fuel_engineTest_total << std::endl;
        myRuntimeInfo->debug << "testResources.fuel_flightTest: " << testResources.fuel_flightTest << std::endl;
        Energy_total = theEngeneering.E_total + testResources.Energy_totalTest;
        fuel_total = testResources.fuel_totalTest;
        // TODO(franz#1#): T135 Possible improvement: Fuel consumption for various types of fuel. Determine missisonen (duration);
        myRuntimeInfo->info << "fuel_flightTest: " << testResources.fuel_flightTest << std::endl;
    }
}

void Development::calculateDevelopmentEmissions(const std::shared_ptr<SchaeferIOData>& data, const Production &theProduction, const EndOfLife &theEndOfLife) {
    if (data->configuration.developmentEmissionSetting != "mode_0") {
    double emissionIndexCO2 = 3.149; // EI for kersosene combustion
    double emissionIndexSO2 = 0.84*pow(10, -3); // EI for kersosene combustion
        /*Test phase*/
        /* Wind tunnel Test*/
        myRuntimeInfo->info << "      Wind tunnel Test Energy: " << testResources.Energy_windTunnelTest << " MJ" << std::endl;
        windtunnelTest.CO2 = testResources.Energy_windTunnelTest * myEcoDatabase.elecEmData.CO2;
        windtunnelTest.CH4 = testResources.Energy_windTunnelTest * myEcoDatabase.elecEmData.CH4;
        windtunnelTest.N2O = testResources.Energy_windTunnelTest * myEcoDatabase.elecN2O;
        windtunnelTest.PFC = testResources.Energy_windTunnelTest * myEcoDatabase.elecPFC;
        windtunnelTest.NOx = testResources.Energy_windTunnelTest * myEcoDatabase.elecEmData.NOx;
        windtunnelTest.SO2 = testResources.Energy_windTunnelTest * myEcoDatabase.elecEmData.SO2;
        windtunnelTest.CO = testResources.Energy_windTunnelTest * myEcoDatabase.elecEmData.CO;
        windtunnelTest.NMVOC = testResources.Energy_windTunnelTest * myEcoDatabase.elecNMVOC;
        windtunnelTest.CED = testResources.Energy_windTunnelTest * myEcoDatabase.elecEmData.specificEnergy;
        /* Structure Test*/
        myRuntimeInfo->info << "      Structure Test Energy (Fatigue Test): " << testResources.Energy_fatigueTest << " MJ" << std::endl;
        structureTest.CO2 = testResources.Energy_fatigueTest * myEcoDatabase.elecEmData.CO2;// - productionACEnginesEco.CO2);
        structureTest.CH4 = testResources.Energy_fatigueTest * myEcoDatabase.elecEmData.CH4;// - productionACEnginesEco.CH4);
        structureTest.N2O = testResources.Energy_fatigueTest * myEcoDatabase.elecN2O;// - productionACEnginesEco.N2O);
        structureTest.PFC = testResources.Energy_fatigueTest * myEcoDatabase.elecPFC;// - productionACEnginesEco.PFC);
        structureTest.NOx = testResources.Energy_fatigueTest * myEcoDatabase.elecEmData.NOx;// - productionACEnginesEco.NOx);
        structureTest.SO2 = testResources.Energy_fatigueTest * myEcoDatabase.elecEmData.SO2;// - productionACEnginesEco.SO2);
        structureTest.CO = testResources.Energy_fatigueTest * myEcoDatabase.elecEmData.CO;// - productionACEnginesEco.CO);
        structureTest.NMVOC = testResources.Energy_fatigueTest * myEcoDatabase.elecNMVOC;// - productionACEnginesEco.NMVOC);
        structureTest.CED = testResources.Energy_fatigueTest * myEcoDatabase.elecEmData.specificEnergy;
        /* System Test*/
        myRuntimeInfo->info << "      System Test Energy: " << testResources.Energy_systemTest << " MJ" << std::endl;
        systemTest.CO2 = testResources.Energy_systemTest * myEcoDatabase.elecEmData.CO2;
        systemTest.CH4 = testResources.Energy_systemTest * myEcoDatabase.elecEmData.CH4;
        systemTest.N2O = testResources.Energy_systemTest * myEcoDatabase.elecN2O;
        systemTest.PFC = testResources.Energy_systemTest * myEcoDatabase.elecPFC;
        systemTest.NOx = testResources.Energy_systemTest * myEcoDatabase.elecEmData.NOx;
        systemTest.SO2 = testResources.Energy_systemTest * myEcoDatabase.elecEmData.SO2;
        systemTest.CO = testResources.Energy_systemTest * myEcoDatabase.elecEmData.CO;
        systemTest.NMVOC = testResources.Energy_systemTest * myEcoDatabase.elecNMVOC;
        systemTest.CED = testResources.Energy_systemTest * myEcoDatabase.elecEmData.specificEnergy;
        /* Ground Test incl. fuel prod.*/
        taxiTestFactor = 0.05 * data->configuration.flightTestHours / (data->taxiTime / (3600.));
        myRuntimeInfo->info << "      Ground Test (System) Energy: " << testResources.Energy_groundTest << " MJ" << std::endl;
        myRuntimeInfo->info << "      Taxi Test Fuel: " << data->taxifuelDesign *taxiTestFactor << " kg" << std::endl;
        testResources.fuel_totalTest += data->taxifuelDesign * taxiTestFactor;
        fuel_total += data->taxifuelDesign * taxiTestFactor;
        groundTest.CO2 = testResources.Energy_groundTest * myEcoDatabase.elecEmData.CO2
                         + data->taxifuelDesign * taxiTestFactor * (emissionIndexCO2 + myEcoDatabase.fuelProdEmData.at(0).CO2);
        groundTest.CH4 = testResources.Energy_groundTest * myEcoDatabase.elecEmData.CH4
                         + data->taxifuelDesign * taxiTestFactor * myEcoDatabase.fuelProdEmData.at(0).CH4;
        groundTest.N2O = testResources.Energy_groundTest * myEcoDatabase.elecN2O;
        groundTest.PFC = testResources.Energy_groundTest * myEcoDatabase.elecPFC;
        groundTest.NOx = testResources.Energy_groundTest * myEcoDatabase.elecEmData.NOx
                         + data->theTestMission.LTO.NOx * taxiTestFactor
                         + data->taxifuelDesign * taxiTestFactor * myEcoDatabase.fuelProdEmData.at(0).NOx;;
        groundTest.SO2 = testResources.Energy_groundTest * myEcoDatabase.elecEmData.SO2
                         + data->taxifuelDesign * taxiTestFactor * emissionIndexSO2;
        groundTest.CO = testResources.Energy_groundTest * myEcoDatabase.elecEmData.CO
                        + data->theTestMission.LTO.CO * taxiTestFactor
                        + data->taxifuelDesign * taxiTestFactor * myEcoDatabase.fuelProdEmData.at(0).CO;
        groundTest.NMVOC = testResources.Energy_groundTest * myEcoDatabase.elecNMVOC;
        groundTest.HC = data->theTestMission.LTO.HC * taxiTestFactor;
        groundTest.H2O = data->theTestMission.LTO.H2O * taxiTestFactor
                         + data->taxifuelDesign * taxiTestFactor * myEcoDatabase.fuelProdEmData.at(0).H2O;;
        groundTest.PM10 = data->theTestMission.LTO.soot * taxiTestFactor
                          + data->taxifuelDesign * taxiTestFactor * myEcoDatabase.fuelProdEmData.at(0).PM10;
        groundTest.CED = testResources.Energy_groundTest * myEcoDatabase.elecEmData.specificEnergy
                         + data->taxifuelDesign * taxiTestFactor * myEcoDatabase.fuelSpecificEnergy.at(0);
        /* Engine Test incl. Fuel Production*/
        if (data->configuration.doEngineTest) {
            double engineTestFuelTotal = testResources.fuel_engineTest_total;
            engineTest.CO2 = engineTestFuelTotal * (emissionIndexCO2 + myEcoDatabase.fuelProdEmData.at(0).CO2);
            engineTest.CH4 = engineTestFuelTotal * myEcoDatabase.fuelProdEmData.at(0).CH4;
            engineTest.N2O = 0.;
            engineTest.PFC = 0.;
            engineTest.NOx = testResources.fuel_engineTestRig * (data->theTestMission.LTO.NOx / data->LTOfuelDesign)
                             + (data->theTestMission.Cruise.NOx + data->theTestMission.LTO.NOx) / data->blocktimeDesign * data->configuration.FTBHours
                             + engineTestFuelTotal * myEcoDatabase.fuelProdEmData.at(0).NOx;
            engineTest.SO2 = (testResources.fuel_engineTestRig + testResources.fuel_engineFlightTest) * emissionIndexSO2;
            engineTest.CO = testResources.fuel_engineTestRig * (data->theTestMission.LTO.CO / data->LTOfuelDesign)
                            + data->theTestMission.LTO.CO / data->blocktimeDesign * data->configuration.FTBHours
                            + engineTestFuelTotal * myEcoDatabase.fuelProdEmData.at(0).CO;; //no values for CO_Cruise, because from LTO data!
            engineTest.NMVOC = 0.;
            engineTest.HC = testResources.fuel_engineTestRig * (data->theTestMission.LTO.HC / data->LTOfuelDesign)
                            + (data->theTestMission.Cruise.HC + data->theTestMission.LTO.HC) / data->blocktimeDesign * data->configuration.FTBHours;
            engineTest.H2O = testResources.fuel_engineTestRig * (data->theTestMission.LTO.H2O / data->LTOfuelDesign)
                             + (data->theTestMission.Cruise.H2O + data->theTestMission.LTO.H2O) / data->blocktimeDesign * data->configuration.FTBHours
                             + engineTestFuelTotal * myEcoDatabase.fuelProdEmData.at(0).H2O;; //but actually no matter, only important for cirrus formation
            engineTest.PM10 = testResources.fuel_engineTestRig * (data->theTestMission.LTO.soot / data->LTOfuelDesign)
                              + (data->theTestMission.LTO.soot) / data->blocktimeDesign * data->configuration.FTBHours
                              + engineTestFuelTotal * myEcoDatabase.fuelProdEmData.at(0).PM10;; //PM10 interesting only during LTO, no greenhouse gas and not ozone-forming
            engineTest.CED = engineTestFuelTotal * myEcoDatabase.fuelSpecificEnergy.at(0);
        }
        /* Flight Test incl. fuel prod. */
        myRuntimeInfo->info << "      Flight Test Fuel: " << testResources.fuel_flightTest << " kg" << std::endl;
        flightTest.CO2 = testResources.fuel_flightTest * (emissionIndexCO2 + myEcoDatabase.fuelProdEmData.at(0).CO2);
        flightTest.CH4 = testResources.fuel_flightTest * myEcoDatabase.fuelProdEmData.at(0).CH4;
        flightTest.N2O = 0.;
        flightTest.PFC = 0.;
        flightTest.NOx = (data->theTestMission.Cruise.NOx + data->theTestMission.LTO.NOx) / data->blocktimeDesign * testResources.flightHours_flightTest
                         + testResources.fuel_flightTest * myEcoDatabase.fuelProdEmData.at(0).NOx;
        flightTest.SO2 = testResources.fuel_flightTest * emissionIndexSO2;
        flightTest.CO = data->theTestMission.LTO.CO / data->blocktimeDesign * testResources.flightHours_flightTest
                        + testResources.fuel_flightTest * myEcoDatabase.fuelProdEmData.at(0).CO;//no values for CO_Cruise, because from LTO data!
        flightTest.NMVOC = 0.;
        flightTest.HC = (data->theTestMission.Cruise.HC + data->theTestMission.LTO.HC) / data->blocktimeDesign * testResources.flightHours_flightTest; //ozone-forming
        flightTest.H2O = (data->theTestMission.Cruise.H2O + data->theTestMission.LTO.H2O) / data->blocktimeDesign * testResources.flightHours_flightTest
                         + testResources.fuel_flightTest * myEcoDatabase.fuelProdEmData.at(0).H2O;
        flightTest.PM10 = (data->theTestMission.LTO.soot) / data->blocktimeDesign * testResources.flightHours_flightTest
                          + testResources.fuel_flightTest
                          * myEcoDatabase.fuelProdEmData.at(0).PM10;//PM10 only interesting during LTO, no greenhouse gas and no ozone-forming
        flightTest.CED = testResources.fuel_flightTest * myEcoDatabase.fuelSpecificEnergy.at(0);
    }
    if (data->configuration.developmentEmissionSetting == "mode_2") {
        testACEnergy = theProduction.Energy_total * data->configuration.flightTestAC
                       + (theProduction.Energy_total - theProduction.Energy_engine)
                       * data->configuration.strucTestAC; //because Fuel from Testflight and FuelProd. does not contribute to Energy!
        testResources.Energy_totalTest += testACEnergy;
        Energy_total += testACEnergy;

        /*Production and EoL of test aircrafts*/
        myRuntimeInfo->info << "      Test AC Production Energy: " << testACEnergy << " MJ" << std::endl;
        testACProduction.CO2 = (theProduction.production_total.CO2 - theProduction.finalFlightTests.CO2 + theEndOfLife.endOfLife_total.CO2) * data->configuration.flightTestAC
                               + (theProduction.production_total.CO2 - theProduction.production_engine.CO2 - theProduction.finalFlightTests.CO2
                                  + theEndOfLife.endOfLife_total.CO2 - theEndOfLife.materialEoL_engine.CO2 - theEndOfLife.eoLTransport.CO2) *data->configuration.strucTestAC;
        testACProduction.H2O = (theProduction.production_total.H2O - theProduction.finalFlightTests.H2O + theEndOfLife.endOfLife_total.H2O) * data->configuration.flightTestAC
                               + (theProduction.production_total.H2O - theProduction.production_engine.H2O - theProduction.finalFlightTests.H2O
                                  + theEndOfLife.endOfLife_total.H2O - theEndOfLife.materialEoL_engine.H2O - theEndOfLife.eoLTransport.H2O) *data->configuration.strucTestAC;
        testACProduction.NOx = (theProduction.production_total.NOx - theProduction.finalFlightTests.NOx + theEndOfLife.endOfLife_total.NOx) * data->configuration.flightTestAC
                               + (theProduction.production_total.NOx - theProduction.production_engine.NOx - theProduction.finalFlightTests.NOx
                                  + theEndOfLife.endOfLife_total.NOx - theEndOfLife.materialEoL_engine.NOx - theEndOfLife.eoLTransport.NOx) *data->configuration.strucTestAC;
        testACProduction.SO2 = (theProduction.production_total.SO2 - theProduction.finalFlightTests.SO2 + theEndOfLife.endOfLife_total.SO2) * data->configuration.flightTestAC
                               + (theProduction.production_total.SO2 - theProduction.production_engine.SO2 - theProduction.finalFlightTests.SO2
                                  + theEndOfLife.endOfLife_total.SO2 - theEndOfLife.materialEoL_engine.SO2 - theEndOfLife.eoLTransport.SO2) *data->configuration.strucTestAC;
        testACProduction.CO = (theProduction.production_total.CO - theProduction.finalFlightTests.CO + theEndOfLife.endOfLife_total.CO) * data->configuration.flightTestAC
                              + (theProduction.production_total.CO - theProduction.production_engine.CO - theProduction.finalFlightTests.CO
                                 + theEndOfLife.endOfLife_total.CO - theEndOfLife.materialEoL_engine.CO - theEndOfLife.eoLTransport.CO) *data->configuration.strucTestAC;
        testACProduction.CH4 = (theProduction.production_total.CH4 - theProduction.finalFlightTests.CH4 + theEndOfLife.endOfLife_total.CH4) * data->configuration.flightTestAC
                               + (theProduction.production_total.CH4 - theProduction.production_engine.CH4 - theProduction.finalFlightTests.CH4
                                  + theEndOfLife.endOfLife_total.CH4 - theEndOfLife.materialEoL_engine.CH4 - theEndOfLife.eoLTransport.CH4) *data->configuration.strucTestAC;
        testACProduction.PM10 = (theProduction.production_total.PM10 - theProduction.finalFlightTests.PM10 + theEndOfLife.endOfLife_total.PM10) * data->configuration.flightTestAC
                                + (theProduction.production_total.PM10 - theProduction.production_engine.PM10 - theProduction.finalFlightTests.PM10
                                   + theEndOfLife.endOfLife_total.PM10 - theEndOfLife.materialEoL_engine.PM10 - theEndOfLife.eoLTransport.PM10) *data->configuration.strucTestAC;
        testACProduction.N2O = (theProduction.production_total.N2O - theProduction.finalFlightTests.N2O + theEndOfLife.endOfLife_total.N2O) * data->configuration.flightTestAC
                               + (theProduction.production_total.N2O - theProduction.production_engine.N2O - theProduction.finalFlightTests.N2O
                                  + theEndOfLife.endOfLife_total.N2O - theEndOfLife.materialEoL_engine.N2O - theEndOfLife.eoLTransport.N2O) *data->configuration.strucTestAC;
        testACProduction.PFC = (theProduction.production_total.PFC - theProduction.finalFlightTests.PFC + theEndOfLife.endOfLife_total.PFC) * data->configuration.flightTestAC
                               + (theProduction.production_total.PFC - theProduction.production_engine.PFC - theProduction.finalFlightTests.PFC
                                  + theEndOfLife.endOfLife_total.PFC - theEndOfLife.materialEoL_engine.PFC - theEndOfLife.eoLTransport.PFC) *data->configuration.strucTestAC;
        testACProduction.NMVOC = (theProduction.production_total.NMVOC - theProduction.finalFlightTests.NMVOC + theEndOfLife.endOfLife_total.NMVOC) *
                            data->configuration.flightTestAC + (theProduction.production_total.NMVOC - theProduction.production_engine.NMVOC - theProduction.finalFlightTests.NMVOC
                               + theEndOfLife.endOfLife_total.NMVOC - theEndOfLife.materialEoL_engine.NMVOC - theEndOfLife.eoLTransport.NMVOC) *data->configuration.strucTestAC;
        testACProduction.HC = (theProduction.production_total.HC - theProduction.finalFlightTests.HC + theEndOfLife.endOfLife_total.HC) * data->configuration.flightTestAC
                              + (theProduction.production_total.HC - theProduction.production_engine.HC - theProduction.finalFlightTests.HC
                                 + theEndOfLife.endOfLife_total.HC - theEndOfLife.materialEoL_engine.HC - theEndOfLife.eoLTransport.HC) *data->configuration.strucTestAC;
        testACProduction.CED = (theProduction.production_total.CED - theProduction.finalFlightTests.CED + theEndOfLife.endOfLife_total.CED) * data->configuration.flightTestAC
                               + (theProduction.production_total.CED - theProduction.production_engine.CED - theProduction.finalFlightTests.CED
                                  + theEndOfLife.endOfLife_total.CED - theEndOfLife.materialEoL_engine.CED - theEndOfLife.eoLTransport.CED) *data->configuration.strucTestAC;

        /* Production and EoL (EoL Scen only, without Disassembly+Dismant.) of the Test Engines */
        if (data->configuration.doEngineTest) {
            testEngineEnergy = theProduction.Energy_engine * data->configuration.numberOfTestEngines;
            testResources.Energy_totalTest += testEngineEnergy;
            Energy_total += testEngineEnergy;
            myRuntimeInfo->info << "      Test Engine Production Energy: " << testEngineEnergy << " MJ" << std::endl;
            testEngineProduction.CO2 = (theProduction.production_engine.CO2 + theEndOfLife.materialEoL_engine.CO2) * data->configuration.numberOfTestEngines;
            testEngineProduction.H2O = (theProduction.production_engine.H2O + theEndOfLife.materialEoL_engine.H2O) * data->configuration.numberOfTestEngines;
            testEngineProduction.NOx = (theProduction.production_engine.NOx + theEndOfLife.materialEoL_engine.NOx) * data->configuration.numberOfTestEngines;
            testEngineProduction.SO2 = (theProduction.production_engine.SO2 + theEndOfLife.materialEoL_engine.SO2) * data->configuration.numberOfTestEngines;
            testEngineProduction.CO = (theProduction.production_engine.CO + theEndOfLife.materialEoL_engine.CO) * data->configuration.numberOfTestEngines;
            testEngineProduction.CH4 = (theProduction.production_engine.CH4 + theEndOfLife.materialEoL_engine.CH4) * data->configuration.numberOfTestEngines;
            testEngineProduction.PM10 = (theProduction.production_engine.PM10 + theEndOfLife.materialEoL_engine.PM10) * data->configuration.numberOfTestEngines;
            testEngineProduction.N2O = (theProduction.production_engine.N2O + theEndOfLife.materialEoL_engine.N2O) * data->configuration.numberOfTestEngines;
            testEngineProduction.PFC = (theProduction.production_engine.PFC + theEndOfLife.materialEoL_engine.PFC) * data->configuration.numberOfTestEngines;
            testACProduction.NMVOC = (theProduction.production_engine.NMVOC + theEndOfLife.materialEoL_engine.NMVOC) * data->configuration.numberOfTestEngines;
            testACProduction.HC = (theProduction.production_engine.HC + theEndOfLife.materialEoL_engine.HC) * data->configuration.numberOfTestEngines;
            testEngineProduction.CED = (theProduction.production_engine.CED + theEndOfLife.materialEoL_engine.CED) * data->configuration.numberOfTestEngines;
        }
    }

    /* Addition all Test-Emissions */
    //EnginesTest-Emissions are Null, in case switch is off !
    myRuntimeInfo->info << "    Total Test Power: " << testResources.Energy_totalTest << " MJ" << std::endl;
    myRuntimeInfo->info << "    Total Test Fuel: " << testResources.fuel_totalTest << " kg" << std::endl;
    test_total.CO2  = windtunnelTest.CO2 + structureTest.CO2 + systemTest.CO2 + groundTest.CO2
    + engineTest.CO2 + flightTest.CO2 + testACProduction.CO2;//+testFuelProduction.CO2;
    test_total.H2O  = windtunnelTest.H2O + structureTest.H2O + systemTest.H2O + groundTest.H2O
    + engineTest.H2O + flightTest.H2O + testACProduction.H2O; //+testFuelProduction.H2O;
    test_total.NOx  = windtunnelTest.NOx + structureTest.NOx + systemTest.NOx + groundTest.NOx
    + engineTest.NOx + flightTest.NOx + testACProduction.NOx; //+testFuelProduction.NOx;
    test_total.SO2  = windtunnelTest.SO2 + structureTest.SO2 + systemTest.SO2 + groundTest.SO2
    + engineTest.SO2 + flightTest.SO2 + testACProduction.SO2; //+testFuelProduction.SO2;
    test_total.CO   = windtunnelTest.CO  + structureTest.CO + systemTest.CO + groundTest.CO
    + engineTest.CO + flightTest.CO + testACProduction.CO; //+testFuelProduction.CO;
    test_total.CH4  = windtunnelTest.CH4 + structureTest.CH4 + systemTest.CH4 + groundTest.CH4
    + engineTest.CH4 + flightTest.CH4 + testACProduction.CH4; //+testFuelProduction.CH4;
    test_total.PM10 = windtunnelTest.PM10 + structureTest.PM10 + systemTest.PM10 + groundTest.PM10 + engineTest.PM10 + flightTest.PM10 +
                      testACProduction.PM10; //+testFuelProduction.PM10;//LTO emissions only!
    test_total.N2O  = windtunnelTest.N2O + structureTest.N2O + systemTest.N2O + groundTest.N2O
    + engineTest.N2O + flightTest.N2O + testACProduction.N2O; //+testFuelProduction.N2O;
    test_total.PFC  = windtunnelTest.PFC + structureTest.PFC + systemTest.PFC + groundTest.PFC
    + engineTest.PFC + flightTest.PFC + testACProduction.PFC; //+testFuelProduction.PFC;
    test_total.NMVOC = windtunnelTest.NMVOC + structureTest.NMVOC + systemTest.NMVOC + groundTest.NMVOC
    + engineTest.NMVOC + flightTest.NMVOC + testACProduction.NMVOC; //+testFuelProduction.NMVOC;
    test_total.HC   = windtunnelTest.HC + structureTest.HC + systemTest.HC + groundTest.HC + engineTest.HC + flightTest.HC + testACProduction.HC; //+testFuelProduction.HC;
    test_total.CED   = windtunnelTest.CED + structureTest.CED + systemTest.CED + groundTest.CED + engineTest.CED + flightTest.CED + testACProduction.CED;

    /* Calculation of emissions from personnel expenses in the development phase*/
    //electricity
    myRuntimeInfo->info << "      Test Engineering Electric Power: " << theEngeneering.E_elec << " MJ" << std::endl;
    engineering_elec.CO2 = theEngeneering.E_elec * myEcoDatabase.elecEmData.CO2;
    engineering_elec.H2O = theEngeneering.E_elec * myEcoDatabase.elecEmData.H2O;
    engineering_elec.NOx = theEngeneering.E_elec * myEcoDatabase.elecEmData.NOx;
    engineering_elec.SO2 = theEngeneering.E_elec * myEcoDatabase.elecEmData.SO2;
    engineering_elec.CO = theEngeneering.E_elec * myEcoDatabase.elecEmData.CO;
    engineering_elec.CH4 = theEngeneering.E_elec * myEcoDatabase.elecEmData.CH4;
    engineering_elec.PM10 = theEngeneering.E_elec * myEcoDatabase.elecEmData.PM10;
    engineering_elec.CED = theEngeneering.E_elec * myEcoDatabase.elecEmData.specificEnergy;
    //heating
    myRuntimeInfo->info << "      Test Engineering Heat Power: " << theEngeneering.E_heat << " MJ" << std::endl;
    engineering_heat.CO2 = theEngeneering.E_heat * myEcoDatabase.heatEmData.CO2;
    engineering_heat.H2O = theEngeneering.E_heat * myEcoDatabase.heatEmData.H2O;
    engineering_heat.NOx = theEngeneering.E_heat * myEcoDatabase.heatEmData.NOx;
    engineering_heat.SO2 = theEngeneering.E_heat * myEcoDatabase.heatEmData.SO2;
    engineering_heat.CO = theEngeneering.E_heat * myEcoDatabase.heatEmData.CO;
    engineering_heat.CH4 = theEngeneering.E_heat * myEcoDatabase.heatEmData.CH4;
    engineering_heat.PM10 = theEngeneering.E_heat * myEcoDatabase.heatEmData.PM10;
    engineering_heat.CED = theEngeneering.E_heat * myEcoDatabase.heatEmData.specificEnergy;
    //Total personnel expenses
    myRuntimeInfo->info << "    Test Engineering Power Total: " << theEngeneering.E_total << " MJ" << std::endl;
    engineering_total.CO2 = engineering_elec.CO2 + engineering_heat.CO2;
    engineering_total.H2O = engineering_elec.H2O + engineering_heat.H2O;
    engineering_total.NOx = engineering_elec.NOx + engineering_heat.NOx;
    engineering_total.SO2 = engineering_elec.SO2 + engineering_heat.SO2;
    engineering_total.CO = engineering_elec.CO + engineering_heat.CO;
    engineering_total.CH4 = engineering_elec.CH4 + engineering_heat.CH4;
    engineering_total.PM10 = engineering_elec.PM10 + engineering_heat.PM10;
    engineering_total.N2O = engineering_elec.N2O + engineering_heat.N2O;
    engineering_total.PFC = engineering_elec.PFC + engineering_heat.PFC;
    engineering_total.NMVOC = engineering_elec.NMVOC + engineering_heat.NMVOC;
    engineering_total.HC = engineering_elec.HC + engineering_heat.HC;
    engineering_total.CED = engineering_elec.CED + engineering_heat.CED;

    /*Entire development phase*/
    myRuntimeInfo->info << "  Development Power Total: " << Energy_total << " MJ" << std::endl;
    myRuntimeInfo->info << "  Development Fuel Total: " << fuel_total << " kg" << std::endl;
    development_total.CO2 = engineering_total.CO2 + test_total.CO2;
    development_total.H2O = engineering_total.H2O + test_total.H2O;
    development_total.NOx = engineering_total.NOx + test_total.NOx;
    development_total.SO2 = engineering_total.SO2 + test_total.SO2;
    development_total.CO  = engineering_total.CO + test_total.CO;
    development_total.CH4 = engineering_total.CH4 + test_total.CH4;
    development_total.PM10 = engineering_total.PM10 + test_total.PM10;
    development_total.N2O = engineering_total.N2O + test_total.N2O;
    development_total.PFC = engineering_total.PFC + test_total.PFC;
    development_total.NMVOC = engineering_total.NMVOC + test_total.NMVOC;
    development_total.HC = engineering_total.HC + test_total.HC;
    development_total.CED = engineering_total.CED + test_total.CED;
    myRuntimeInfo->info << "  Development Primary Power Total: " << development_total.CED << " MJ" << std::endl;
}


Development::~Development() {
    //dtor
}

// TODO(franz#1#): Mantis ID: 0000325 Class structure even clearer?! by class Energy, class Fuel, and class:  Calculate taxi fuel already here

Development::test::test()
    :
    Energy_windTunnelTest(0.),
    Energy_fatigueTest(0.),
    Energy_systemTest(0.),
    fuel_engineTestRig(0.),
    fuel_engineFlightTest(0.),
    fuel_engineTest_total(0.),
    Energy_groundTest(0.),
    fuel_flightTest(0.),
    flightHours_flightTest(0.),
    Energy_totalTest(0.),
    fuel_totalTest(0.) {
}
Development::test::~test() {
    //dtor
}

Development::Engineering::Engineering()
    :
    T_eng(0.),
    E_elec(0.),
    E_heat(0.),
    E_total(0.) {
}
Development::Engineering::~Engineering() {
}
