/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_DEVELOPMENT_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_DEVELOPMENT_H_

#include "../../../emissionsClass.h"
#include "../../ecoDatabase.h"
#include "production.h"
#include "endOfLife.h"
#include "schaeferIOData.h"
#include <memory>

class SchaeferIOData;
class production;
class endOfLife;

/** \class Development
 * \brief Contains all properties of the development phase
 */
class Development{
 protected:
    const ecoDatabase myEcoDatabase; /**< Object contains data of ecoDatabase*/
    atmosphere myAtm;/**< Atmosphere object */

 public:
    /** \brief Constructor of class development
     */
    Development();
    /** \brief Destructor of class development
     */
    virtual ~Development();

    /** \class test
     * \brief Class contains the properties of the expenses for testing
     */
    class test {
     public:
        /* Necessary resources for the entire certification process (i.e. for the entire aircraft program) */
        double Energy_windTunnelTest;/**< Required energy for wind tunnel testing [MJ] */
        double Energy_fatigueTest;/**< Required energy for fatigue tests [MJ] */
        double Energy_systemTest;/**< Required energy for system tests [MJ] */
        double fuel_engineTestRig;/**< Required fuel for engine test rig [kg] */
        double fuel_engineFlightTest;/**< Required fuel for engine flight tests [kg]*/
        double fuel_engineTest_total;/**< Total required fuel for engine testing [kg]*/
        double Energy_groundTest;/**< Required energy for ground tests [MJ]*/
        double fuel_flightTest;/**< Total required fuel for flight tests [kg]*/
        double flightHours_flightTest;/**< Flight hours during flight test phase [h]*/
        double Energy_totalTest;/**< Total required energy for test phase [MJ]*/
        double fuel_totalTest;/**< Total required fuel for test phase [kg]*/

        /** \brief Constructor of class test
         */
        test();

        /** \brief Destructor of class test
         */
        virtual ~test();
    };
    test testResources; /**< Object of class test contains data of test resources*/

    /** \class Engineering
     * \brief Contains the properties of personnel expenses for engineering
     */
    class Engineering {
     public:
        double T_eng;/**< Airframe Engineering Hours [h] */
        double E_elec;/**< Test Engineering Electric Power [MJ] */
        double E_heat;/**< Test Engineering Heat Power [MJ] */
        double E_total;/** Test Engineering total Power [MJ] */

        /** \brief Constructor of class Engineering
         */
        Engineering();

        /** \brief Destructor of class Engineering
         */
        virtual ~Engineering();
    };
    Engineering theEngeneering; /**< Object of class engineering contains personnel expenses for engineering */

    double Energy_total;/**< Total energy demand for the development phase [MJ] */
    double fuel_total;/**< total fuel demand for the development phase [kg] */

    Emissions engineering_elec;/**< Emissions due to electricity generated by development personnel */
    Emissions engineering_heat;/**< Emissions due to heating of development personnel */
    Emissions engineering_total;/**< Total emissions from development personnel */
    Emissions windtunnelTest;/**< Emissions due to wind tunnel tests */
    Emissions structureTest;/**< Emissions due to structure tests */
    Emissions systemTest;/**< Emissions due to system tests */
    Emissions engineTest;/**< Emissions due to engine tests */
    Emissions groundTest;/**< Emissions due to ground test incl. fuel production */
    double taxiTestFactor;/**< Factor to calculate taxi fuel for test mission [-] */
    Emissions flightTest;/**< Emissions due to flight test incl. fuel production */
    Emissions test_total;/**< All test emissions (excluding fuel and material production) */
    Emissions testACProduction;/**< Emissions of production of the test aircraft */
    Emissions testEngineProduction;/**< Emissions of production of the test aircraft engines */
    double testACEnergy; /**< Needed energy for production of test aircraft [MJ]*/
    double testEngineEnergy;/**< Needed energy for production of test aircraft engine [MJ]*/
    /** \todo [T137] (https://unicado.ilr.rwth-aachen.de/T137): Put resources into respective class if possible
     */
    Emissions development_total;/**< Total emissions during the development phase */

    /** \brief Function to calculate the material and energy demand of the development phase
     * \param data IO Data pointer
     */
    void calculateDevelopmentResources(const std::shared_ptr<SchaeferIOData>& data);

    /** \brief Function to calculate the emissions of the development phase
     * \param data IO Data pointer
     * \param theProduction production object containing the production data
     * \param theEndOfLife production object containing the end of life data
     */
    void calculateDevelopmentEmissions(const std::shared_ptr<SchaeferIOData>& data, const Production&theProduction, const EndOfLife &theEndOfLife);
};


#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_DEVELOPMENT_H_
