/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "endOfLife.h"
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>

EndOfLife::EndOfLife()
    :
    myEcoDatabase(ecoDatabase()),
    cost_DD(0.),
    E_elec_DD(0.),
    E_EoL_total(0.),
    Energy_total(0.),
    transportFuel(0.) {
    }

void EndOfLife::calculateEoLResources(const std::shared_ptr<SchaeferIOData>& data) {
    /*Energy for Disassembly and Dismantling*/
    cost_DD = (218800 * data->OME) / (data->OME + 31600);
    E_elec_DD = E_PerWorker_pr[0] * cost_DD / (C_labourHour_pr * t_labourYear_pr);
    for (int i(0); i < data->configuration.partsNo; i++) {
        theAcComp.push_back(mainComponent(data->mainPartsName[i]));
        theAcComp.back().mass = data->mainPartsMass[i];
        myRuntimeInfo->debug << std::endl;
        myRuntimeInfo->debug << std::endl;
        myRuntimeInfo->debug << "Component: " << theAcComp.back().compName << std::endl;
        /* Energy calculation: distribution according to aircraft components (Main Parts Modus) */
        //Material distribution. For numeration see production.h: const double matDistr
        double endOfLifeEnergy = 0.;
        for (int j(0); j < numberOfMaterials; j++) {
            theAcComp.back().r_material.push_back(matDistr[i][j]);
            theAcComp.back().r_recycling.push_back(recyclingScen[i][j]);
            theAcComp.back().r_incineration.push_back(incinerationScen[i][j]);
            theAcComp.back().r_landfill.push_back(landfillScen[i][j]);
            double massMaterialComp = theAcComp.back().r_material.back() * theAcComp.back().mass;
            theAcComp.back().materialRecovered.push_back(massMaterialComp * theAcComp.back().r_recycling.back()*recoveryRate[j]);
            if (j != 1)
                theAcComp.back().E_recycling = 1 / recyclingEnergyRatio[j] * myEcoDatabase.matSpecificEnergy.at(j)
                                               / myEcoDatabase.elecEmData.specificEnergy * massMaterialComp * theAcComp.back().r_recycling.back();
            else
                theAcComp.back().E_recycling = 0.; //Composites will not be recycled, recyclingEnergyRatio[j] = 0
            theAcComp.back().E_incineration = myEcoDatabase.incinerationSpecificEnergy.at(j)
                                              / myEcoDatabase.elecEmData.specificEnergy * massMaterialComp * theAcComp.back().r_incineration.back();
            theAcComp.back().E_landfill = myEcoDatabase.disposalSpecificEnergy.at(j)
                                          / myEcoDatabase.elecEmData.specificEnergy * massMaterialComp * theAcComp.back().r_landfill.back();
            endOfLifeEnergy += theAcComp.back().E_recycling + theAcComp.back().E_incineration + theAcComp.back().E_landfill;
            myRuntimeInfo->debug << std::endl;
            myRuntimeInfo->debug << "Material " << j + 1 << "..." << std::endl;
            myRuntimeInfo->debug << "ER: " << recyclingEnergyRatio[j] << std::endl;
            myRuntimeInfo->debug << "Mass fraction: " << theAcComp.back().r_material.back() << std::endl;
            myRuntimeInfo->debug << "Recycling fraction: " << theAcComp.back().r_recycling.back() << std::endl;
            myRuntimeInfo->debug << "Incinaration fraction: " << theAcComp.back().r_incineration.back() << std::endl;
            myRuntimeInfo->debug << "Landfill fraction: " << theAcComp.back().r_landfill.back() << std::endl;
            myRuntimeInfo->debug << "Mass material: " << massMaterialComp << std::endl;
            myRuntimeInfo->debug << "energyRecycling: " << theAcComp.back().E_recycling << std::endl;
            myRuntimeInfo->debug << "energyIncinenartion: " << theAcComp.back().E_incineration << std::endl;
            myRuntimeInfo->debug << "energyLandfill: " << theAcComp.back().E_landfill << std::endl;
            myRuntimeInfo->debug << "endOfLifeEnergy: " << endOfLifeEnergy << std::endl;
        }
        theAcComp.back().E_endOfLife = endOfLifeEnergy;
        E_EoL_total += theAcComp.back().E_endOfLife;
        myRuntimeInfo->debug << "Mass:   " << theAcComp.back().mass << std::endl;
        myRuntimeInfo->debug << "Energy: " << theAcComp.back().E_endOfLife << std::endl;
    }
    myRuntimeInfo->debug << "E_EoL_total: " << E_EoL_total << std::endl;
//    getchar();
    /* Energy Calculation: Distribution by Materials (Material Mode) */
    //TO DO: delete?
    //E_EoL_total = 0.;
    for (int i(0); i < numberOfMaterials; i++) {
        theMaterial.push_back(EndOfLife::material());
        theMaterial.back().matName = data->materialName[i];
    }
    myRuntimeInfo->debug << "E_EoL_total: " << E_EoL_total << std::endl;
//    getchar();
    Energy_total = E_EoL_total + E_elec_DD;
}

void EndOfLife::calculateEoLEmissions(const std::shared_ptr<SchaeferIOData>& data, const Production&theProduction) {
    /*Transport to End-of-Life Site */
    /* emission calculation */
    double transportFactor = data->configuration.distanceToEoLSite / data->rangeDesign;
    transportFuel = data->blockfuelDesign * transportFactor; //strongly simplified: Assumption: Const. Fuel consumption, determined from design mission.
    myRuntimeInfo->info << "  EoL Fuel: " << transportFuel << " kg" << std::endl;
    eoLTransport.CO2 = data->theTestMission.LTO.CO2 + data->theTestMission.Cruise.CO2 * transportFactor + transportFuel * myEcoDatabase.fuelProdEmData.at(0).CO2;
    eoLTransport.H2O = data->theTestMission.LTO.H2O + data->theTestMission.Cruise.H2O * transportFactor + transportFuel * myEcoDatabase.fuelProdEmData.at(0).H2O;
    eoLTransport.NOx = data->theTestMission.LTO.NOx + data->theTestMission.Cruise.NOx * transportFactor + transportFuel * myEcoDatabase.fuelProdEmData.at(0).NOx;
    eoLTransport.SO2 = data->theTestMission.LTO.SO2 + data->theTestMission.Cruise.SO2 * transportFactor + transportFuel * myEcoDatabase.fuelProdEmData.at(0).SO2;
    eoLTransport.CO = data->theTestMission.LTO.CO + data->theTestMission.Cruise.CO * transportFactor + transportFuel * myEcoDatabase.fuelProdEmData.at(0).CO;
    eoLTransport.CH4 = transportFuel * myEcoDatabase.fuelProdEmData.at(0).CH4;
    eoLTransport.PM10 = data->theTestMission.LTO.soot + data->theTestMission.Cruise.soot * transportFactor + transportFuel * myEcoDatabase.fuelProdEmData.at(0).PM10;
    eoLTransport.HC = data->theTestMission.LTO.HC + data->theTestMission.Cruise.HC * transportFactor;
    eoLTransport.CED = transportFuel * myEcoDatabase.fuelSpecificEnergy.at(0);
//    // TODO (student#1#): Noch weitere Emissionen wichtig?
    /*Disassembly and Dismantling */
    //Elektrische Emissionen
    myRuntimeInfo->info << "    Disassembly and Dismantling Energy: " << E_elec_DD << " MJ" << std::endl;
    eoLDisassDism.CO2 = E_elec_DD * myEcoDatabase.elecEmData.CO2;
    eoLDisassDism.H2O = E_elec_DD * myEcoDatabase.elecEmData.H2O;
    eoLDisassDism.NOx = E_elec_DD * myEcoDatabase.elecEmData.NOx;
    eoLDisassDism.SO2 = E_elec_DD * myEcoDatabase.elecEmData.SO2;
    eoLDisassDism.CO =  E_elec_DD * myEcoDatabase.elecEmData.CO;
    eoLDisassDism.CH4 = E_elec_DD * myEcoDatabase.elecEmData.CH4;
    eoLDisassDism.PM10 = E_elec_DD * myEcoDatabase.elecEmData.PM10;
    eoLDisassDism.CED = E_elec_DD * myEcoDatabase.elecEmData.specificEnergy;
    /*End-of-Life Scenario (Energy consumption through processes and energy gain through material extraction*/

    if (data->configuration.productionMode == "mode_0") {
        //Material modus
        this->emissions_materialMode();
    } else if (data->configuration.productionMode == "mode_1") {
        //Main Parts Modus
        this->emissions_mainPartsMode(data, theProduction);
    } else {
        myRuntimeInfo->err << "Unknown production mode " << data->configuration.productionMode << ". (config-entry \"productionMode\") " << std::endl;
        myRuntimeInfo->err << "Abort program!" << std::endl;
        exit(1);
    }
}

void EndOfLife::emissions_materialMode() {
//    double materialEnergy=0.;
    for (int i(0); i < numberOfMaterials; i++) {
        std::string materialEntry(theMaterial.at(i).matName);
        std::vector<std::string> matName_tmp;
        for (unsigned int k(0.); k < myEcoDatabase.matEmData.size(); k++) {
            matName_tmp.push_back(myEcoDatabase.matEmData.at(k).specification);
        }
        std::vector<std::string>::iterator entry = find(matName_tmp.begin(), matName_tmp.end(), materialEntry); //Iterator to find the configuration entry
        if (entry == matName_tmp.end()) { //Entry not available!
            myRuntimeInfo->err << "Material " << materialEntry << " not found. Abort!" << std::endl;
            exit(1);
        } else {
            unsigned int pos(0);
            pos = static_cast<int>(entry - matName_tmp.begin());
            /* emission calculation */
            materialEoL.push_back(Emissions());
            materialEoL.back().CO2 = - theMaterial.at(i).massRecovered * myEcoDatabase.matEmData.at(pos).CO2 //Row material
                                     + theMaterial.at(i).E_endOfLife * myEcoDatabase.elecEmData.CO2; //manufacturing
            materialEoL.back().H2O = - theMaterial.at(i).massRecovered * myEcoDatabase.matEmData.at(pos).H2O
                                     + theMaterial.at(i).E_endOfLife * myEcoDatabase.elecEmData.H2O;
            materialEoL.back().NOx = - theMaterial.at(i).massRecovered * myEcoDatabase.matEmData.at(pos).NOx
                                     + theMaterial.at(i).E_endOfLife * myEcoDatabase.elecEmData.NOx;
            materialEoL.back().SO2 = - theMaterial.at(i).massRecovered * myEcoDatabase.matEmData.at(pos).SO2
                                     + theMaterial.at(i).E_endOfLife * myEcoDatabase.elecEmData.SO2;
            materialEoL.back().CO = - theMaterial.at(i).massRecovered * myEcoDatabase.matEmData.at(pos).CO
                                    + theMaterial.at(i).E_endOfLife * myEcoDatabase.elecEmData.CO;
            materialEoL.back().CH4 = - theMaterial.at(i).massRecovered * myEcoDatabase.matEmData.at(pos).CH4
                                     + theMaterial.at(i).E_endOfLife * myEcoDatabase.elecEmData.CH4;
            materialEoL.back().PM10 = - theMaterial.at(i).massRecovered * myEcoDatabase.matEmData.at(pos).PM10
                                      + theMaterial.at(i).E_endOfLife * myEcoDatabase.elecEmData.PM10;
            materialEoL.back().CED = - theMaterial.at(i).massRecovered * myEcoDatabase.matSpecificEnergy.at(pos)
                                     + theMaterial.at(i).E_endOfLife * myEcoDatabase.elecEmData.specificEnergy;;
        }
        materialEoL_total.CO2  += materialEoL.back().CO2;
        materialEoL_total.H2O  += materialEoL.back().H2O;
        materialEoL_total.NOx  += materialEoL.back().NOx;
        materialEoL_total.SO2  += materialEoL.back().SO2;
        materialEoL_total.CO   += materialEoL.back().CO;
        materialEoL_total.CH4  += materialEoL.back().CH4;
        materialEoL_total.PM10 += materialEoL.back().PM10;
        materialEoL_total.N2O += materialEoL.back().N2O;
        materialEoL_total.PFC += materialEoL.back().PFC;
        materialEoL_total.NMVOC += materialEoL.back().NMVOC;
        materialEoL_total.HC += materialEoL.back().HC;
        materialEoL_total.CED += materialEoL.back().CED;
    }
    myRuntimeInfo->info << "    EoL Scenario Primary Energy Total: " << materialEoL_total.CED << " MJ" << std::endl;
}

void EndOfLife::emissions_mainPartsMode(const std::shared_ptr<SchaeferIOData>& data, const Production&theProduction) {
    for (int i(0); i < data->configuration.partsNo; i++) {
        /* emission calculation */
        materialEoL.push_back(Emissions());
        materialEoL.back().CO2 = theAcComp.at(i).E_endOfLife * myEcoDatabase.elecEmData.CO2;
        materialEoL.back().H2O = theAcComp.at(i).E_endOfLife * myEcoDatabase.elecEmData.H2O;
        materialEoL.back().NOx = theAcComp.at(i).E_endOfLife * myEcoDatabase.elecEmData.NOx;
        materialEoL.back().SO2 = theAcComp.at(i).E_endOfLife * myEcoDatabase.elecEmData.SO2;
        materialEoL.back().CO = theAcComp.at(i).E_endOfLife * myEcoDatabase.elecEmData.CO;
        materialEoL.back().CH4 = theAcComp.at(i).E_endOfLife * myEcoDatabase.elecEmData.CH4;
        materialEoL.back().PM10 = theAcComp.at(i).E_endOfLife * myEcoDatabase.elecEmData.PM10;
        materialEoL.back().CED = theAcComp.at(i).E_endOfLife * myEcoDatabase.elecEmData.specificEnergy;
        for (int j(0); j < numberOfMaterials; j++) {
            std::string materialEntry(theProduction.theMaterial.at(j).matName);
            std::vector<std::string> matName_tmp;
            for (unsigned int k(0.); k < myEcoDatabase.matEmData.size(); k++) {
                matName_tmp.push_back(myEcoDatabase.matEmData.at(k).specification);
            }
            std::vector<std::string>::iterator entry = find(matName_tmp.begin(), matName_tmp.end(), materialEntry); //Iterator to find the configuration entry
            if (entry == matName_tmp.end()) { //Entry not available!
                myRuntimeInfo->err << "Material " << materialEntry << " in EcoDatabase not found. Abort!" << std::endl;
                exit(1);
            } else {
                unsigned int pos(0);
                pos = static_cast<int>(entry - matName_tmp.begin());
                /* emission calculation */
                materialEoL.back().CO2 -= theAcComp.back().materialRecovered.at(j) * myEcoDatabase.matEmData.at(pos).CO2;
                materialEoL.back().H2O -= theAcComp.back().materialRecovered.at(j) * myEcoDatabase.matEmData.at(pos).H2O;
                materialEoL.back().NOx -= theAcComp.back().materialRecovered.at(j) * myEcoDatabase.matEmData.at(pos).NOx;
                materialEoL.back().SO2 -= theAcComp.back().materialRecovered.at(j) * myEcoDatabase.matEmData.at(pos).SO2;
                materialEoL.back().CO -= theAcComp.back().materialRecovered.at(j) * myEcoDatabase.matEmData.at(pos).CO;
                materialEoL.back().CH4 -= theAcComp.back().materialRecovered.at(j) * myEcoDatabase.matEmData.at(pos).CH4;
                materialEoL.back().PM10 -= theAcComp.back().materialRecovered.at(j) * myEcoDatabase.matEmData.at(pos).PM10;
                materialEoL.back().CED -= theAcComp.back().materialRecovered.at(j) * myEcoDatabase.matSpecificEnergy.at(pos);
            }
        }
        materialEoL_total.CO2 += materialEoL.back().CO2;
        materialEoL_total.H2O += materialEoL.back().H2O;
        materialEoL_total.NOx += materialEoL.back().NOx;
        materialEoL_total.SO2 += materialEoL.back().SO2;
        materialEoL_total.CO += materialEoL.back().CO;
        materialEoL_total.CH4 += materialEoL.back().CH4;
        materialEoL_total.PM10 += materialEoL.back().PM10;
        materialEoL_total.N2O += materialEoL.back().N2O;
        materialEoL_total.PFC += materialEoL.back().PFC;
        materialEoL_total.NMVOC += materialEoL.back().NMVOC;
        materialEoL_total.HC += materialEoL.back().HC;
        materialEoL_total.CED += materialEoL.back().CED;
    }
    if (data->configuration.engineMode) {
        materialEoL_engine.CO2 = materialEoL.back().CO2;
        materialEoL_engine.H2O = materialEoL.back().H2O;
        materialEoL_engine.NOx = materialEoL.back().NOx;
        materialEoL_engine.SO2 = materialEoL.back().SO2;
        materialEoL_engine.CO = materialEoL.back().CO;
        materialEoL_engine.CH4 = materialEoL.back().CH4;
        materialEoL_engine.PM10 = materialEoL.back().PM10;
        materialEoL_engine.N2O = materialEoL.back().N2O;
        materialEoL_engine.PFC = materialEoL.back().PFC;
        materialEoL_engine.NMVOC = materialEoL.back().NMVOC;
        materialEoL_engine.HC = materialEoL.back().HC;
        materialEoL_engine.CED = materialEoL.back().CED;
    }
    myRuntimeInfo->info << "    EoL Scenario Primary Energy Total: " << materialEoL_total.CED << " MJ" << std::endl;

    /*EoL Phase Total*/
    myRuntimeInfo->info << "  EoL Energy Total: " << Energy_total << " MJ" << std::endl;
    endOfLife_total.CO2 = eoLTransport.CO2 + eoLDisassDism.CO2 + materialEoL_total.CO2;
    endOfLife_total.H2O = eoLTransport.H2O + eoLDisassDism.H2O + materialEoL_total.H2O;
    endOfLife_total.NOx = eoLTransport.NOx + eoLDisassDism.NOx + materialEoL_total.NOx;
    endOfLife_total.SO2 = eoLTransport.SO2 + eoLDisassDism.SO2 + materialEoL_total.SO2;
    endOfLife_total.CO = eoLTransport.CO + eoLDisassDism.CO + materialEoL_total.CO;
    endOfLife_total.CH4 = eoLTransport.CH4 + eoLDisassDism.CH4 + materialEoL_total.CH4;
    endOfLife_total.PM10 = eoLTransport.PM10 + eoLDisassDism.PM10 + materialEoL_total.PM10;
    endOfLife_total.N2O = eoLTransport.N2O + eoLDisassDism.N2O + materialEoL_total.N2O;
    endOfLife_total.PFC = eoLTransport.PFC + eoLDisassDism.PFC + materialEoL_total.PFC;
    endOfLife_total.NMVOC = eoLTransport.NMVOC + eoLDisassDism.NMVOC + materialEoL_total.NMVOC;
    endOfLife_total.HC = eoLTransport.HC + eoLDisassDism.HC + materialEoL_total.HC;
    endOfLife_total.CED = eoLTransport.CED + eoLDisassDism.CED + materialEoL_total.CED;
    myRuntimeInfo->info << "  EoL Primary Energy Total: " << endOfLife_total.CED << " MJ" << std::endl;
}

EndOfLife::~EndOfLife() {
}

EndOfLife::mainComponent::mainComponent(const std::string& name)
    :
    compName(name),
    mass(0.),
    E_recycling(0.),
    E_incineration(0.),
    E_landfill(0.),
    E_endOfLife(0.) {
}
EndOfLife::mainComponent::~mainComponent() {
}

EndOfLife::material::material()
    :
    matName(""),
    mass(0.),
    massRecovered(0.),
    E_recycling(0.),
    E_incineration(0.),
    E_landfill(0.),
    E_endOfLife(0.) {
}
EndOfLife::material::~material() {
}
