/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <algorithm>

#include "operation.h"
#include "production.h"

#include <aixml/node.h>
#include <standardFiles/functions.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>

using std::vector;
using std::string;

// TODO(franz#1#): T129 Emission calculation at Derate correct? Check! (Problem: Base curves with EI_ICAO (without Derate))
Operation::Operation()
    :
    E_spare_total(0.),
    E_labour_total(0.),
    E_total(0.),
    operationFuel(0.) {
    }

void Operation::calculateOperationResources(const std::shared_ptr<SchaeferIOData>& data, const std::shared_ptr<RuntimeIO>& rtIO,
                                                const Production &theProduction) {//Due to maintenance
    /* Maintenance Spare Parts */
    double engineCostRatioMaterial(0.75);//Assumption: Material Engine Maintenance 3 times as high as Labour (see PhD Schaefer)
    double conversionFactor = convertUnit(SECOND, HOUR, data->flightTimeStudy) * data->flightsPerYear * data->durationOperation; // Used for [$/FH] -> [$/LC]
    for (uint8_t i(0); i < data->configuration.sparePartsNo; i++) {
        /*Creating the object for the corresponding components*/
        theSpareParts.push_back(spareParts(data->sparePartATA[i], data->sparePartName[i]));
        /*RC components*/
        theSpareParts.back().RC_component = theProduction.theAcComp.at(i).RC;
        /*Material maintenance costs*/
        if (data->configuration.calculateMaintCosts) { // Calculation according to Americal Airlines model
            double CPI_factor = 3.833; // Converts 1976-USD to 2010-USD [-]
            if (data->sparePartName[i] == "Wing") { //ATA 57
                double timeFactor = (convertUnit(SECOND, HOUR, data->flightTimeStudy) * 0.49 + 0.51) / (2.5 * 0.49 + 0.51);
                theSpareParts.back().maintenance_material_costs = CPI_factor * (0.126 + 0.00506 * data->S_ref) * timeFactor * conversionFactor;
            } else if (data->sparePartName[i] == "Fuselage") { //Airframe
                for (uint8_t k(0); k < 4; k++) { // sparePartATAAirframe: ATA51, ATA52, ATA53, ATA56
                    /*Order of coefficients: ATA51, ATA52, ATA53, ATA56*/
                    vector<double> c{0., 0.387, 0.5833, 0.};// Method coefficient
                    vector<double> d{0., 0.00785, 0., 0.0362};// Method coefficient
                    vector<double> X{data->airframeMass/1000, data->seatsAvailable, data->airframeMass/1000, data->seatsAvailable};// Method coefficient
                    vector<double> Y{1., 1., 1., 1.}; // Method coefficient
                    vector<double> HR{1., 0.51, 0.50, 0.80};// Flight hour related cost ratio (ATA51, ATA52, ATA53, ATA56)
                    vector<double> CR{0., 0.49, 0.50, 0.20};// Flight cycle related cost ratio (ATA51, ATA52, ATA53, ATA56)
                    double timeFactor = (convertUnit(SECOND, HOUR, data->flightTimeStudy) * HR[k] + CR[k]) / (2.5 * HR[k] + CR[k]);
                    theSpareParts.back().maintenance_material_costs += CPI_factor * (c[k] + d[k]*X[k]*Y[k]) * timeFactor * conversionFactor;
                }
            } else if (data->sparePartName[i] == "Tailplane") {//ATA 55
                double timeFactor = (convertUnit(SECOND, HOUR, data->flightTimeStudy) * 0.49 + 0.51) / (2.5 *  0.49 + 0.51);
                theSpareParts.back().maintenance_material_costs = data->stabilizerMassRatio * CPI_factor * 0.3737 * timeFactor * conversionFactor;
            } else if (data->sparePartName[i] == "Fin") { //ATA 55
                double timeFactor = (convertUnit(SECOND, HOUR, data->flightTimeStudy) * 0.49 + 0.51) / (2.5 *  0.49 + 0.51);
                theSpareParts.back().maintenance_material_costs = data->finMassRatio * CPI_factor * 0.3737 * timeFactor * conversionFactor;
            } else if (data->sparePartName[i] == "LandingGear") { //ATA 32
                double timeFactor = (convertUnit(SECOND, HOUR, data->flightTimeStudy) * 0.18 + 0.82) / (2.5 *  0.18 + 0.82);
                theSpareParts.back().maintenance_material_costs = CPI_factor * (4.961 + .1810 * data->MTOM / 1000) * timeFactor *conversionFactor;
            } else if (data->sparePartName[i] == "Pylons") { //ATA 54
                double timeFactor = (convertUnit(SECOND, HOUR, data->flightTimeStudy) * 0.8 + 0.2) / (2.5 *  0.8 + 0.2);
                theSpareParts.back().maintenance_material_costs = data->pylonMassRatio * CPI_factor * 0.1391 * data->numberOfEngines * timeFactor * conversionFactor;
            } else if (data->sparePartName[i] == "PowerUnit") { //sparePartATAPowerUnit: ATA28, ATA36, ATA54
                for (uint8_t k(0); k < 3; k++) {
                    double airflow(0.454 * data->studyMissionPAX); // Air conditioning mass flow [kg/min]
                    /*Order of coefficients: ATA28, ATA36, ATA54*/
                    vector<double> c{0.595, 0., 0.};// Method coefficient
                    vector<double> d{0.0123, 0.0019, 0.1391};// Method coefficient
                    vector<double> X{data->blockfuelStudy/1000, airflow, static_cast<double>(data->numberOfEngines)};// Method coefficient
                    vector<double> Y{1., data->myEngine.get_scaled_SLST()/10000, 1.}; // Method coefficient
                    vector<double> HR{0.94, 0.26, 0.80};// Flight hour related cost ratio (ATA28, ATA36, ATA54)
                    vector<double> CR{0.06, 0.74, 0.20};// Flight cycle related cost ratio (ATA28, ATA36, ATA54)
                    double timeFactor = (convertUnit(SECOND, HOUR, data->flightTimeStudy) * HR[k] + CR[k]) / (2.5 * HR[k] + CR[k]);
                    theSpareParts.back().maintenance_material_costs += CPI_factor * (c[k] + d[k]*X[k]*Y[k]) * timeFactor * conversionFactor;
                }
            } else {
               myRuntimeInfo->err << "Invalid name of spare part." << std::endl;
            }
        } else {
            double costRatioDMC(1.); // Mass dependent cost ratio
            double DMCmatComp(0.); // Material costs of component [2010-USD / 100ASK]
            string maintPath;
            string maintPath2;
            string maintPathEngine;
            if (data->sparePartName[i] == "PowerUnit")
                maintPathEngine = "aircraft_exchange_file/MonetaryValues/LCC/OperatingCosts/DirectOperatingCosts/DOCComponents/Maintenance/MaintenanceComponents/Engine";
            else
                maintPath = string("aircraft_exchange_file/MonetaryValues/LCC/OperatingCosts/DirectOperatingCosts/DOCComponents/Maintenance/MaintenanceComponents/Airframe/")
                            + string("AirframeComponents/") + data->sparePartATA[i] + "/CostComponents/MaterialCosts";
            /*DMC Material components*/
            double nacelleRC = checkBoundaries(rtIO->acxml.at("aircraft_exchange_file/MonetaryValues/LCC/RecurringCosts/RCComponents/PowerUnit/PowerUnitComponents/Nacelles"),
                                0.0, true, 100000000., true);
            if (data->sparePartName[i] == "Fuselage") { //Airframe (ATA54) incl. Rest
                for (int k(0); k < 4; k++) { //s. sparePartATAAirframe number
                    maintPath2 =
                        "aircraft_exchange_file/MonetaryValues/LCC/OperatingCosts/DirectOperatingCosts/DOCComponents/Maintenance/MaintenanceComponents/Airframe/AirframeComponents/"
                        + data->sparePartATAAirframe[k] + "/CostComponents/MaterialCosts";
                    DMCmatComp += checkBoundaries(rtIO->acxml.at(maintPath2), -0.0, true, 10000000000., true);
                }
            } else if (data->sparePartName[i] == "Pylons") {
                costRatioDMC = theSpareParts.at(5).RC_component / (theSpareParts.at(5).RC_component + nacelleRC);
                DMCmatComp = costRatioDMC * checkBoundaries(rtIO->acxml.at(maintPath), -0.0, true, 10000000000., true);
            } else if (data->sparePartName[i] == "PowerUnit") {
                DMCmatComp = engineCostRatioMaterial * checkBoundaries(rtIO->acxml.at(maintPathEngine), -0.0, true, 10000000000., true);
                for (int k(0); k < 3; k++) { //s. sparePartATAPowerUnit number
                    maintPath2 =
                        "aircraft_exchange_file/MonetaryValues/LCC/OperatingCosts/DirectOperatingCosts/DOCComponents/Maintenance/MaintenanceComponents/Airframe/AirframeComponents/"
                            + data->sparePartATAPowerUnit[k] + "/CostComponents/MaterialCosts";
                    if (data->sparePartATAPowerUnit[k] == "ATA54")
                        costRatioDMC = nacelleRC / (theSpareParts.at(5).RC_component + nacelleRC);
                    DMCmatComp += costRatioDMC * checkBoundaries(rtIO->acxml.at(maintPath2), -0.0, true, 10000000000., true);
                }
            } else {
                if (data->sparePartName[i] == "Fin")
                    costRatioDMC = data->finMassRatio;
                else if (data->sparePartName[i] == "Tailplane")
                    costRatioDMC = data->stabilizerMassRatio;
                DMCmatComp = costRatioDMC * checkBoundaries(rtIO->acxml.at(maintPath), -0.0, true, 10000000000., true);
            }
            theSpareParts.back().maintenance_material_costs = DMCmatComp * (data->seatsAvailable / 100. * convertUnit(METER, KILO, METER, data->rangeStudy)) *
                                                    data->flightsPerYear * data->durationOperation;
        }
        theSpareParts.back().energyConsumptionRatio = theSpareParts.back().maintenance_material_costs / theSpareParts.back().RC_component;
        myRuntimeInfo->info << "sparePartName: " << data->sparePartName[i] << std::endl;
        myRuntimeInfo->info << "RC_component [$]: " << theSpareParts.back().RC_component << std::endl;
        myRuntimeInfo->info << "maintenance_material_costs [$]: " << theSpareParts.back().maintenance_material_costs << std::endl;
        myRuntimeInfo->info << "seatsAvailable [-]: " << data->seatsAvailable << std::endl;
        myRuntimeInfo->info << "rangeStudy [m]: " << data->rangeStudy << std::endl;
        myRuntimeInfo->info << "energyConsumptionRatio [-]: " << theSpareParts.back().energyConsumptionRatio << std::endl;
        string compEntry(data->sparePartName[i]);
        vector<string> compName_tmp;
        for (unsigned int k(0.); k < theProduction.theAcComp.size(); k++) {
            compName_tmp.push_back(theProduction.theAcComp.at(k).compName);
        }
        vector<string>::iterator entry = std::find(compName_tmp.begin(), compName_tmp.end(), compEntry); //Iterator to find the configuration input
        if (entry == compName_tmp.end()) { //Input not available!
            myRuntimeInfo->err << "components " << compEntry << " not found. Abort!" << std::endl;
            exit(1);
        } else {
            unsigned int pos(0);
            pos = static_cast<int>(entry - compName_tmp.begin());
            theSpareParts.back().E_manuf = theSpareParts.back().energyConsumptionRatio * theProduction.theAcComp.at(pos).E_manuf;
        }
        E_spare_total += theSpareParts.back().E_manuf;
    }
    /* Maintenance Labour (Cost calculation: DOC TUB)*/
    myRuntimeInfo->info << "  Maintenance labour starts" << std::endl;
    double convertEuroToDollar(1/0.72); // Conversion factor EUR -> USD
    double wageCostLabourTUB(50); // [€/h]
    double burden = 2.0; //Cost burden
    double SLST = data->myEngine.get_scaled_SLST()/1000/G_FORCE; // [t] converted thrust from N
    double engineDMC = convertEuroToDollar * (data->numberOfEngines * (1.5 * SLST + //[$/yr]
                30.5 * convertUnit(SECOND, HOUR, data->flightTimeStudy) + 10.6)) * data->flightsPerYear;
    double engineLabourCosts = (1 - engineCostRatioMaterial) * engineDMC; //[$/yr]
    double airframeLabourCosts = convertEuroToDollar * wageCostLabourTUB * (1 + burden) * ((0.655 + 0.01 * data->OME / 1000.) //[USD/yr]
                                     *convertUnit(SECOND, HOUR, data->flightTimeStudy) + 0.254 + 0.01 * data->OME / 1000.) * data->flightsPerYear;
    theMaintLabour.DMC_labourYear = (airframeLabourCosts + engineLabourCosts);
    myRuntimeInfo->debug << "Available Seats " << data->seatsAvailable << std::endl;
    myRuntimeInfo->debug << "flight time [s] " << data->flightTimeStudy << std::endl;
    myRuntimeInfo->debug << "range [km] " << convertUnit(NOPREFIX, METER, KILO, METER, data->rangeStudy) << std::endl;
    myRuntimeInfo->debug << "Flights per year " << data->flightsPerYear << std::endl;
    myRuntimeInfo->debug << "DMC_labourYear [$] " << theMaintLabour.DMC_labourYear << std::endl;
    theMaintLabour.E_elec = E_PerWorker_op[0] * theMaintLabour.DMC_labourYear / (C_labourHour_op * t_labourYear_op) * data->durationOperation;
    theMaintLabour.E_heat = E_PerWorker_op[1] * theMaintLabour.DMC_labourYear / (C_labourHour_op * t_labourYear_op) * data->durationOperation;
    E_labour_total += theMaintLabour.E_elec + theMaintLabour.E_heat;
    E_total = E_spare_total + E_labour_total;
    myRuntimeInfo->debug << "Check Labour Hours Maintenance: " << theMaintLabour.DMC_labourYear/(C_labourHour_op) << std::endl;
}

void Operation::calculateOperationEmissions(const std::shared_ptr<SchaeferIOData>& data, const Production &theProduction) {
    /* Missions in the operational phase */
    operationFuel = (data->blockfuelStudy) * data->flightsPerYear * data->durationOperation;
    myRuntimeInfo->info << "  Operation Fuel: " << operationFuel << " kg" << std::endl;
    operationMissions.CO2 = (data->theStudyMission.LTO.CO2 + data->theStudyMission.Cruise.CO2) * data->flightsPerYear * data->durationOperation;
    operationMissions.H2O = (data->theStudyMission.LTO.H2O + data->theStudyMission.Cruise.H2O) * data->flightsPerYear * data->durationOperation;
    operationMissions.NOx = (data->theStudyMission.LTO.NOx + data->theStudyMission.Cruise.NOx) * data->flightsPerYear * data->durationOperation;
    operationMissions.SO2 = (data->theStudyMission.LTO.SO2 + data->theStudyMission.Cruise.SO2) * data->flightsPerYear * data->durationOperation;
    operationMissions.HC = (data->theStudyMission.LTO.HC + data->theStudyMission.Cruise.HC) * data->flightsPerYear * data->durationOperation;
    operationMissions.CO = (data->theStudyMission.LTO.CO + data->theStudyMission.Cruise.CO) * data->flightsPerYear * data->durationOperation;
    operationMissions.CH4 = (data->theStudyMission.LTO.CH4 + data->theStudyMission.Cruise.CH4) * data->flightsPerYear * data->durationOperation;
    operationMissions.PM10 = (data->theStudyMission.LTO.soot + data->theStudyMission.Cruise.soot) * data->flightsPerYear * data->durationOperation;

    /*Fuel production in the operational phase*/
    operationFuelProduction.CO2 = operationFuel * myEcoDatabase.fuelProdEmData.at(0).CO2;
    operationFuelProduction.H2O = operationFuel * myEcoDatabase.fuelProdEmData.at(0).H2O;
    operationFuelProduction.NOx = operationFuel * myEcoDatabase.fuelProdEmData.at(0).NOx;
    operationFuelProduction.SO2 = operationFuel * myEcoDatabase.fuelProdEmData.at(0).SO2;
    operationFuelProduction.CO = operationFuel * myEcoDatabase.fuelProdEmData.at(0).CO;
    operationFuelProduction.CH4 = operationFuel * myEcoDatabase.fuelProdEmData.at(0).CH4;
    operationFuelProduction.PM10 = operationFuel * myEcoDatabase.fuelProdEmData.at(0).PM10;
    operationFuelProduction.CED = operationFuel * myEcoDatabase.fuelSpecificEnergy.at(0);

    /*Maintenance Labour*/
    //Electrical emissions
    myRuntimeInfo->info << "      Maintenance (Labour) Energy electrical: " << theMaintLabour.E_elec << " MJ" << std::endl;
    maintLabour_elec.CO2 = theMaintLabour.E_elec * myEcoDatabase.elecEmData.CO2;
    maintLabour_elec.H2O = theMaintLabour.E_elec * myEcoDatabase.elecEmData.H2O;
    maintLabour_elec.NOx = theMaintLabour.E_elec * myEcoDatabase.elecEmData.NOx;
    maintLabour_elec.SO2 = theMaintLabour.E_elec * myEcoDatabase.elecEmData.SO2;
    maintLabour_elec.CO =  theMaintLabour.E_elec * myEcoDatabase.elecEmData.CO;
    maintLabour_elec.CH4 = theMaintLabour.E_elec * myEcoDatabase.elecEmData.CH4;
    maintLabour_elec.PM10 = theMaintLabour.E_elec * myEcoDatabase.elecEmData.PM10;
    maintLabour_elec.CED = theMaintLabour.E_elec * myEcoDatabase.elecEmData.specificEnergy;
    //Heating emissions
    myRuntimeInfo->info << "      Maintenance (Labour) Energy heat: " << theMaintLabour.E_heat << " MJ" << std::endl;
    maintLabour_heat.CO2 = theMaintLabour.E_heat * myEcoDatabase.heatEmData.CO2;
    maintLabour_heat.H2O = theMaintLabour.E_heat * myEcoDatabase.heatEmData.H2O;
    maintLabour_heat.NOx = theMaintLabour.E_heat * myEcoDatabase.heatEmData.NOx;
    maintLabour_heat.SO2 = theMaintLabour.E_heat * myEcoDatabase.heatEmData.SO2;
    maintLabour_heat.CO = theMaintLabour.E_heat * myEcoDatabase.heatEmData.CO;
    maintLabour_heat.CH4 = theMaintLabour.E_heat * myEcoDatabase.heatEmData.CH4;
    maintLabour_heat.PM10 = theMaintLabour.E_heat * myEcoDatabase.heatEmData.PM10;
    maintLabour_heat.CED = theMaintLabour.E_heat * myEcoDatabase.heatEmData.specificEnergy;
    //Total maintenance Labour Emissions
    myRuntimeInfo->info << "    Maintenance (Labour) Energy: " << E_labour_total << " MJ" << std::endl;
    maintLabour_total.CO2 = maintLabour_elec.CO2 + maintLabour_heat.CO2;
    maintLabour_total.H2O = maintLabour_elec.H2O + maintLabour_heat.H2O;
    maintLabour_total.NOx = maintLabour_elec.NOx + maintLabour_heat.NOx;
    maintLabour_total.SO2 = maintLabour_elec.SO2 + maintLabour_heat.SO2;
    maintLabour_total.CO = maintLabour_elec.CO + maintLabour_heat.CO;
    maintLabour_total.CH4 = maintLabour_elec.CH4 + maintLabour_heat.CH4;
    maintLabour_total.PM10 = maintLabour_elec.PM10 + maintLabour_heat.PM10;
    maintLabour_total.CED = maintLabour_elec.CED + maintLabour_heat.CED;
    //are zero, nevertheless perform:
    maintLabour_total.N2O = maintLabour_elec.N2O + maintLabour_heat.N2O;
    maintLabour_total.PFC = maintLabour_elec.PFC + maintLabour_heat.PFC;
    maintLabour_total.NMVOC = maintLabour_elec.NMVOC + maintLabour_heat.NMVOC;
    maintLabour_total.HC = maintLabour_elec.HC + maintLabour_heat.HC;
    /*Maintenance Spare Parts*/
    for (int i(0); i < data->configuration.sparePartsNo; i++) {
        /* emission calculation */
        //Electrical energy of production
        sparePartsVector.push_back(Emissions());
        sparePartsVector.back().CO2 = theSpareParts.at(i).E_manuf * myEcoDatabase.elecEmData.CO2;
        sparePartsVector.back().H2O = theSpareParts.at(i).E_manuf * myEcoDatabase.elecEmData.H2O;
        sparePartsVector.back().NOx = theSpareParts.at(i).E_manuf * myEcoDatabase.elecEmData.NOx;
        sparePartsVector.back().SO2 = theSpareParts.at(i).E_manuf * myEcoDatabase.elecEmData.SO2;
        sparePartsVector.back().CO =  theSpareParts.at(i).E_manuf * myEcoDatabase.elecEmData.CO;
        sparePartsVector.back().CH4 = theSpareParts.at(i).E_manuf * myEcoDatabase.elecEmData.CH4;
        sparePartsVector.back().PM10 = theSpareParts.at(i).E_manuf * myEcoDatabase.elecEmData.PM10;
        sparePartsVector.back().CED = theSpareParts.at(i).E_manuf * myEcoDatabase.elecEmData.specificEnergy;
        //Find components from production with the same name
        string compEntry(data->sparePartName[i]);
        vector<string> compName_tmp;
        for (unsigned int j(0.); j < theProduction.theAcComp.size(); j++) {
            compName_tmp.push_back(theProduction.theAcComp.at(j).compName);
        }
        vector<string>::iterator entry = find(compName_tmp.begin(), compName_tmp.end(), compEntry); //Iterator to find the configuration entry
        if (entry == compName_tmp.end()) { //Entry not available!
            myRuntimeInfo->err << "component " << compEntry << " not found. Abort!" << std::endl;
            exit(1);
        } else {
            unsigned int compPos(0);
            compPos = static_cast<int>(entry - compName_tmp.begin());
            //raw material extraction
            for (int k(0); k < numberOfMaterials; k++) {
                string materialEntry(theProduction.theMaterial.at(k).matName);
                vector<string> matName_tmp;
                for (unsigned int m(0.); m < myEcoDatabase.matEmData.size(); m++) {
                    matName_tmp.push_back(myEcoDatabase.matEmData.at(m).specification);
                }
                entry = find(matName_tmp.begin(), matName_tmp.end(), materialEntry); //Iterator to find the configuration entry
                if (entry == matName_tmp.end()) { //Entry not available!
                    myRuntimeInfo->err << "Material " << materialEntry << " in EcoDatabase not found!" << std::endl;
                    exit(1);
                } else {
                    unsigned int matPos(0);
                    matPos = static_cast<int>(entry - matName_tmp.begin());
                    /* emission calculation */
                    //The factor energyConsumptionRatio also defines here as the relationship between production and maintenance.
                    //The mass of the total component is therefore taken here with this factor
                    sparePartsVector.back().CO2 += theSpareParts.at(i).energyConsumptionRatio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).r_material.at(k) * myEcoDatabase.matEmData.at(matPos).CO2;
                    sparePartsVector.back().H2O += theSpareParts.at(i).energyConsumptionRatio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).r_material.at(k) * myEcoDatabase.matEmData.at(matPos).H2O;
                    sparePartsVector.back().NOx += theSpareParts.at(i).energyConsumptionRatio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).r_material.at(k) * myEcoDatabase.matEmData.at(matPos).NOx;
                    sparePartsVector.back().SO2 += theSpareParts.at(i).energyConsumptionRatio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).r_material.at(k) * myEcoDatabase.matEmData.at(matPos).SO2;
                    sparePartsVector.back().CO += theSpareParts.at(i).energyConsumptionRatio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).r_material.at(k) * myEcoDatabase.matEmData.at(matPos).CO;
                    sparePartsVector.back().CH4 += theSpareParts.at(i).energyConsumptionRatio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).r_material.at(k) * myEcoDatabase.matEmData.at(matPos).CH4;
                    sparePartsVector.back().PM10 += theSpareParts.at(i).energyConsumptionRatio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).r_material.at(k) * myEcoDatabase.matEmData.at(matPos).PM10;
                    sparePartsVector.back().CED += theSpareParts.at(i).energyConsumptionRatio * (1 + clippings[k])
                    * theProduction.theAcComp.at(compPos).mass * theProduction.theAcComp.at(compPos).r_material.at(k) * myEcoDatabase.matSpecificEnergy.at(matPos);
                }
            }
        }
        //Spare part emissions total
        spareParts_total.CO2 += sparePartsVector.back().CO2;
        spareParts_total.H2O += sparePartsVector.back().H2O;
        spareParts_total.NOx += sparePartsVector.back().NOx;
        spareParts_total.SO2 += sparePartsVector.back().SO2;
        spareParts_total.CO += sparePartsVector.back().CO;
        spareParts_total.CH4 += sparePartsVector.back().CH4;
        spareParts_total.PM10 += sparePartsVector.back().PM10;
        spareParts_total.N2O += sparePartsVector.back().N2O;
        spareParts_total.PFC += sparePartsVector.back().PFC;
        spareParts_total.NMVOC += sparePartsVector.back().NMVOC;
        spareParts_total.HC += sparePartsVector.back().HC;
        spareParts_total.CED += sparePartsVector.back().CED;
    }
    myRuntimeInfo->info << "    Maintenance (Spare Parts) Energy: " << E_spare_total << " MJ" << std::endl;

    /*Operation phase total*/
    myRuntimeInfo->info << "  Operating Energy Total: " << E_total << " MJ" << std::endl;
    operation_total.CO2 = operationMissions.CO2 + operationFuelProduction.CO2 + maintLabour_elec.CO2 + maintLabour_heat.CO2 + spareParts_total.CO2;
    operation_total.H2O = operationMissions.H2O + operationFuelProduction.H2O + maintLabour_elec.H2O + maintLabour_heat.H2O + spareParts_total.H2O;
    operation_total.NOx = operationMissions.NOx + operationFuelProduction.NOx + maintLabour_elec.NOx + maintLabour_heat.NOx + spareParts_total.NOx;
    operation_total.SO2 = operationMissions.SO2 + operationFuelProduction.SO2 + maintLabour_elec.SO2 + maintLabour_heat.SO2 + spareParts_total.SO2;
    operation_total.CO = operationMissions.CO + operationFuelProduction.CO + maintLabour_elec.CO + maintLabour_heat.CO + spareParts_total.CO;
    operation_total.CH4 = operationMissions.CH4 + operationFuelProduction.CH4 + maintLabour_elec.CH4 + maintLabour_heat.CH4 + spareParts_total.CH4;
    operation_total.PM10 = operationMissions.PM10 + operationFuelProduction.PM10 + maintLabour_elec.PM10 + maintLabour_heat.PM10 + spareParts_total.PM10;
    operation_total.N2O = operationMissions.N2O + operationFuelProduction.N2O + maintLabour_elec.N2O + maintLabour_heat.N2O + spareParts_total.N2O;
    operation_total.PFC = operationMissions.PFC + operationFuelProduction.PFC + maintLabour_elec.PFC + maintLabour_heat.PFC + spareParts_total.PFC;
    operation_total.NMVOC = operationMissions.NMVOC + operationFuelProduction.NMVOC + maintLabour_elec.NMVOC + maintLabour_heat.NMVOC + spareParts_total.NMVOC;
    operation_total.HC = operationMissions.HC + operationFuelProduction.HC + maintLabour_elec.HC + maintLabour_heat.HC + spareParts_total.HC;
    operation_total.CED = operationFuelProduction.CED + maintLabour_elec.CED + maintLabour_heat.CED + spareParts_total.CED;
    myRuntimeInfo->info << "     operationMissions.HC: " << operationMissions.HC << " kg" << std::endl;
    myRuntimeInfo->info << "     studyMission.LTO.HC: " << data->theStudyMission.LTO.HC << " kg" << std::endl;
    myRuntimeInfo->info << "     studyMission.Cruise.HC: " << data->theStudyMission.Cruise.HC << " kg" << std::endl;
    myRuntimeInfo->info << "     maintLabour_heat.CO2: " << maintLabour_heat.CO2 << " kg" << std::endl;
    myRuntimeInfo->info << "     operationMissions.CH4: " << operationMissions.CH4 << " kg" << std::endl;
    myRuntimeInfo->info << "     operationFuelProduction.CH4: " << operationFuelProduction.CH4 << " kg" << std::endl;
    myRuntimeInfo->info << "     maintLabour_elec.CH4: " << maintLabour_elec.CH4 << " kg" << std::endl;
    myRuntimeInfo->info << "     maintLabour_heat.CH4: " << maintLabour_heat.CH4 << " kg" << std::endl;
    myRuntimeInfo->info << "  Operating Primary Energy Total: " << operation_total.CED << " MJ" << std::endl;
}


Operation::~Operation() {
    //dtor
}

Operation::Labour::Labour()
    :
    DMC_labourYear(0.),
    E_elec(0.),
    E_heat(0.) {
}
Operation::Labour::~Labour() {
}

Operation::spareParts::spareParts(const string& anATA_ID, const string& name)
    :
    ATA_ID(anATA_ID),
    compName(name),
    RC_component(0.),
    maintenance_material_costs(0.),
    energyConsumptionRatio(0.),
    E_manuf(0.) {
}
Operation::spareParts::~spareParts() {
}

