/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_OPERATION_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_OPERATION_H_

#include <string>
#include <vector>
#include <memory>
#include <moduleBasics/runtimeIO.h>
#include "../../../emissionsClass.h"
#include "../../ecoDatabase.h"
#include "schaeferIOData.h"

class Production;

/** \class Operation
 * \brief Class contains the properties of the emissions for the operation phase
 */
class Operation {
 private:
    const ecoDatabase myEcoDatabase; /**< ecodatabase object*/

 public:
    /** \brief Constructor of class Operation */
    Operation();
    /** \brief Destructor of class Operation */
    virtual ~Operation();


    /*Maintenance*/
    /** \class Labour
     * \brief Class contains the properties for the maintenance labour
     */
    class Labour {
     public:
        double DMC_labourYear;/**< Direct maintenance costs of the labour per year [USD]  */
        double E_elec;/**< Electrical energy demand of the labour (offices, computers) [MJ] */
        double E_heat;/**< Heat energy demand of the labour (offices) [MJ] */

        /** \brief Constructor of class Labour */
        Labour();
        /** \brief Destructor of class Labour */
        virtual ~Labour();
    };
    Labour theMaintLabour; /** Maintenace labour object */

    /** \class spareParts
     * \brief Class contains the emission properties for spare parts that have to be acquired during the operational phase
     */
    class spareParts {
     public:
        std::string ATA_ID; /**< ID of ATA chapter */
        std::string compName; /** Name of component */

        double RC_component;/**< Recurring cost of the component [USD] */
        double maintenance_material_costs;/**< Material maintenance costs for operating phase [USD] */

        double energyConsumptionRatio;/**< Relationship between energy for spare part production and production energy of the entire new part [-]*/
        double E_manuf;/**< Energy expenditure for spare parts production [MJ] */

        /** \brief Constructor of class spareParts
         * \param anATA_ID ID of the ATA chapter
         * \param name Name of the spare part
         */
        spareParts(const std::string& anATA_ID, const std::string& name);
        /** \brief Destructor of class spareParts */
        virtual ~spareParts();
    };
    std::vector<spareParts> theSpareParts; /**< SpareParts object */

    double E_spare_total;/**<Total energy for spare parts [MJ]*/
    double E_labour_total;/**<Total energy for labour [MJ]*/
    double E_total;/**<Total energy [MJ]*/

    /* Emissions */
    Emissions maintLabour_elec;/**< Emissions of Maintenance Power consumption */
    Emissions maintLabour_heat;/**< Emissions of Maintenance Heating */
    Emissions maintLabour_total;/**< Emissions of Maintenance total */
    std::vector<Emissions> sparePartsVector;/**< Emissions of Spare part production divided according to various components (material only) */
    Emissions spareParts_total;/**< Emissions of Complete spare parts production (material only) */
    Emissions operationMissions;/**< Emissions due to operation */
    double operationFuel;/**< Fuel needed for operation [kg] */
    Emissions operationFuelProduction;/**< Emissions of fuel production for the operating phase */
    Emissions operation_total;/**< Total emissions during the operating phase */

    /** \brief Calculates the energy demand due to maintance during operation phase
     * \param data Access to IO data
     * \param rtIO Access to rtIO
     * \param theProduction Access to production data
     */
    void calculateOperationResources(const std::shared_ptr<SchaeferIOData>& data, const std::shared_ptr<RuntimeIO>& rtIO, const Production &theProduction);

    /** \brief Calculates the emissions during operation phase
     * \param data Access to IO data
     * \param theProduction Access to production data
     */
    void calculateOperationEmissions(const std::shared_ptr<SchaeferIOData>& data, const Production &theProduction);
};

#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_OPERATION_H_
