/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "schaeferConfig.h"
#include "../../ecoDatabase.h"
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <standardFiles/functions.h>

using std::string;

SchaeferConfig::SchaeferConfig()
    :
    engineMode(true),
    /*Development*/
    developmentEmissionSetting(""),
    testETOPS(true),
    engineOptions(0),
    windTunnelTestHours(0),
    strucTestCycles(0),
    sysIntTestHours(0),
    ironBirdTestHours(0),
    doEngineTest(true),
    numberOfNewEngines(0),
    numberOfTestEngines(0),
    rigTestHours(0),
    inclETOPS(true),
    MCTpercentage(0.),
    FTBHours(0),
    FTBEngineNo(0),
    FTBFuelConsumptionPerEngine(0),
    flightTestHours(0),
    inclEngineOptions(true),
    strucTestAC(0.),
    flightTestAC(0.),
    engineEngineeringFactor(0.),
    /*Production*/
    productionMode(""),
    primaryMaterialRecycling(true),
    calculateRC(true),
    producedAC(0),
    /*Operation*/
    calculateMaintCosts(true),
    /*End of life*/
    distanceToEoLSite(0.),
    /*Derived settings*/
    partsNo(0),
    compTransportNo(0),
    sparePartsNo(0) {}

void SchaeferConfig::readAllConfigData(const node& configXML) {
    engineMode = EndnodeReadOnly<bool>("program_settings/standard_strategy/emission_calculation/engine_mode_switch").read(configXML).value();
    engineEngineeringFactor = EndnodeReadOnly<double>(std::string("program_settings/standard_strategy/emission_calculation/life_cycle_emissions_methods/")
                                                        + std::string("schaefer/engine_engineering_ratio")).read(configXML).value();
    /* LCA phases */
    readDevelopmentConfigData(configXML.at("program_settings/standard_strategy/emission_calculation/development_phase/"));
    readProductionConfigData(configXML.at("program_settings/standard_strategy/emission_calculation/production_phase/"));
    readOperationConfigData(configXML.at("program_settings/standard_strategy/emission_calculation/operating_phase/"));
    readEndOfLifeConfigData(configXML.at("program_settings/standard_strategy/emission_calculation/end_of_life_phase/"));
    /* Settings depending on engine mode */
    if (engineMode) { //doEngineTest may be false if no new engines will be developed
        partsNo = numberOfMainParts;
        compTransportNo = numberOfCompTransported;
        sparePartsNo = numberOfSpareComponents;
        //productionMode = 2; //because engines from static test ac must be subtracted
    } else {
        doEngineTest = false;
        partsNo = numberOfMainParts - 1;
        compTransportNo = numberOfCompTransported - 1;
        sparePartsNo = numberOfSpareComponents - 1;
    }
    if (numberOfNewEngines > engineOptions) {
        myRuntimeInfo->err << "numberOfNewEngines is bigger than engineOptions. Program abort!" << std::endl;
        exit(1);
    }
}

void SchaeferConfig::readDevelopmentConfigData(const node& development) {
    developmentEmissionSetting = EndnodeReadOnly<string>("test_phase/development_emission_setting").read(development).value();
    testETOPS = EndnodeReadOnly<bool>("test_phase/aircraft_specs/ETOPS_switch").read(development).value();
    engineOptions = EndnodeReadOnly<int>("test_phase/aircraft_specs/engine_options").read(development).value();
    windTunnelTestHours = EndnodeReadOnly<int>("test_phase/wind_tunnel_test/test_hours").read(development).value_xml();
    strucTestCycles = EndnodeReadOnly<int>("test_phase/structural_test/test_cycles").read(development).value();
    strucTestAC = EndnodeReadOnly<int>("test_phase/structural_test/number_of_structural_test_aircraft").read(development).value();
    sysIntTestHours = EndnodeReadOnly<int>("test_phase/system_test/system_integration_test_hours").read(development).value_xml();
    ironBirdTestHours = EndnodeReadOnly<int>("test_phase/system_test/iron_bird_test_hours").read(development).value_xml();
    doEngineTest = EndnodeReadOnly<bool>("test_phase/engine_test/enable").read(development).value();
    numberOfNewEngines = EndnodeReadOnly<int>("test_phase/engine_test/number_of_new_engines").read(development).value();
    numberOfTestEngines = EndnodeReadOnly<int>("test_phase/engine_test/number_of_test_engines").read(development).value();
    rigTestHours = EndnodeReadOnly<int>("test_phase/engine_test/rig_test/test_hours").read(development).value_xml();
    inclETOPS = EndnodeReadOnly<bool>("test_phase/engine_test/rig_test/test_hours/incl_ETOPS_switch").read(development).value();
    MCTpercentage = EndnodeReadOnly<double>("test_phase/engine_test/rig_test/max_continuous_thrust_percentage").read(development).value();
    FTBHours = EndnodeReadOnly<int>("test_phase/engine_test/flying_testbed/test_hours").read(development).value_xml();
    FTBEngineNo = EndnodeReadOnly<int>("test_phase/engine_test/flying_testbed/flying_testbed_engines/engine_number").read(development).value();
    FTBFuelConsumptionPerEngine = EndnodeReadOnly<double>(string("test_phase/engine_test/flying_testbed/flying_testbed_engines/") +
                                    string("flying_testbed_fuel_consumption_per_engine")).read(development).value();
    flightTestHours = EndnodeReadOnly<int>("test_phase/flight_test/test_hours").read(development).value_xml();
    inclEngineOptions = EndnodeReadOnly<bool>("test_phase/flight_test/test_hours/incl_engine_options").read(development).value();
    flightTestAC = EndnodeReadOnly<int>("test_phase/flight_test/number_of_flight_test_aircraft").read(development).value();
}

void SchaeferConfig::readProductionConfigData(const node& production) {
    productionMode = EndnodeReadOnly<string>("production_mode").read(production).value();
    primaryMaterialRecycling = EndnodeReadOnly<bool>("primary_material_recycling_switch").read(production).value();
    calculateRC = EndnodeReadOnly<bool>("calculate_recurring_costs_switch").read(production).value();
    producedAC = EndnodeReadOnly<int>("number_produced_aircraft").read(production).value();
}

void SchaeferConfig::readOperationConfigData(const node& operation) {
    calculateMaintCosts = EndnodeReadOnly<bool>("calculate_maintenance_costs_switch").read(operation).value();
}

void SchaeferConfig::readEndOfLifeConfigData(const node& endOfLife) {
    distanceToEoLSite = EndnodeReadOnly<double>("distance_to_end_of_life_site").read(endOfLife).value();
}

SchaeferConfig::~SchaeferConfig() {
    //dtor
}

