/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_SCHAEFERCONFIG_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_SCHAEFERCONFIG_H_

#include <string>
#include <unordered_map>
#include <utility>
#include <vector>
#include <array>
#include <aixml/endnode.h>
#include "../../../emissionsClass.h"

/** \class SchaeferConfig
 * \brief Class for the config settings needed for schaefer strategy
*/
class SchaeferConfig {
 public:
    /*Program parameter */
    bool engineMode; /**<true: with Engine Production, false: without */
    /*Development phase inputs*/
    std::string developmentEmissionSetting;/**<mode_0: no development emissions, mode_1: calculate w/o production/recycling, mode_2: include production/recycling*/
    bool testETOPS;/**<true: Engine are tested for ETOPS approval, additional 3000 test cycles (a 0.5h), false: no test for ETOPS */
    int engineOptions;/**<Number of engine options for aircraft [-]*/
    int windTunnelTestHours;/**<Number of wind tunnel hours in aircraft development [h]*/
    int strucTestCycles;/**<Tested flight cycles in aircraft development; twice the number of flight cycles for which the aircraft will be certified [-]*/
    int sysIntTestHours;/**<Number of test hours at the system integration test rig [h]*/
    int ironBirdTestHours;/**<Number of test hours at Iron-Bird test rig [h]*/
    bool doEngineTest;/**<true: new engine will be certified for this aircraft, false: only engine tests */
    int numberOfNewEngines;/**<Number of new to be certified engines [-]*/
    int numberOfTestEngines;/**<Number of engies tested (rig + flight test) [-]*/
    int rigTestHours;/**<Number of test hours on the engine-Rig in aircraft development [h]*/
    bool inclETOPS;/**<true: ETOPS hours already in rigTestHours, false: not included */
    double MCTpercentage;/**<Percentage of test hours on the engine rig with Maximum Continuous Thrust in aircraft development [-]*/
    int FTBHours;/**<Number of test hours on Flying Test Bed aircraft development [h]*/
    int FTBEngineNo;/**<Number of engines on Flying Test Bed (without tested engine) [-]*/
    double FTBFuelConsumptionPerEngine;/**<Fuel Consumption per hour von Test Bed AC (pro Engine) [kg/s]*/
    int flightTestHours; /**<Number of flight test hours in aircraft development [h]*/
    bool inclEngineOptions;/**< true: flightTestHours incl. all Engine options, false: only for one. */
    int strucTestAC;/**<Number of aircraft destroyed during structural tests [-]*/
    int flightTestAC;  /**<Number of aircraft used only for flight tests [-]*/
    double engineEngineeringFactor;/**< Ratio of engine engineering effort to whole aircraft, derived from Micado CSR-02 costs [-] */
    /*Production phase inputs*/
    std::string productionMode; /**<mode_0: material mode, mode_1: main parts mode*/
    bool primaryMaterialRecycling;/**<'true': primary material will be recycled, 'false': no recyling*/
    bool calculateRC;/**<'true': recurring costs will be calculated, 'false': costs will be read from xml (old schema implemented, needs to be adapted)*/
    int producedAC; /**<Number of aircraft produced per program*/
    /*Operating phase inputs*/
    bool calculateMaintCosts;/**<'true': recurring costs will be calculated, 'false': costs will be read from xml (old schema implemented, needs to be adapted)*/
    /*End of life phase inputs*/
    double distanceToEoLSite;/**<Distance to be flown to the wreck location [m] */
    /*Additional settings*/
    int partsNo;/**< Number of main part [-]*/
    int compTransportNo; /**< Number of components transported [-]*/
    int sparePartsNo;/**< Number of spare components [-]*/


    /** \brief Default constructor of class SchaeferIOData
     */
    SchaeferConfig();


    /** \brief Constructor of class SchaeferIOData
     */
    virtual ~SchaeferConfig();

    /** \brief Function to read schaefer specific data from ecological_assessment.xml
     * \param config Node of the data structure defined by config file
     */
    void readAllConfigData(const node& config);

 private:
    /** \brief Function to read development data from config.xml
     * \param development Node of the development data structure defined in config file
     */
    void readDevelopmentConfigData(const node& development);

    /** \brief Function to read production data from config.xml
     * \param production Node of the production data structure defined by config file
     */
    void readProductionConfigData(const node& production);

    /** \brief Function to read operation data from config.xml
     * \param operation Node of the operation data structure defined by config file
     */
    void readOperationConfigData(const node& operation);

    /** \brief Function to read enf of life data from config.xml
     * \param endOfLife node of the endOfLife data structure defined by config file
     */
    void readEndOfLifeConfigData(const node& endOfLife);
};

#endif // ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_EMISSION_CALCULATION_LIFE_CYCLE_EMISSIONS_LCA_SCHAEFER_SCHAEFERCONFIG_H_
