/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "schaeferIOData.h"
#include "schaeferConfig.h"
#include <unordered_map>
#include <string>
#include <aixml/node.h>
#include <runtimeInfo/runtimeInfo.h>
#include <moduleBasics/runtimeIO.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>

using std::endl;
using std::ostream;
using std::string;
using std::vector;

SchaeferIOData::SchaeferIOData(std::shared_ptr<RuntimeIO> rtIO, std::shared_ptr<node> ecoAssXML)
    :
    configuration(SchaeferConfig()),
    myEngine(rtIO->getEngineDataDir() + "/" +
            EndnodeReadOnly<std::string>("component_design/propulsion/specific/propulsion/engine/model").read(rtIO->acxml).value(),
            EndnodeReadOnly<double>("component_design/propulsion/specific/propulsion/engine/scale_factor").read(rtIO->acxml).value()),
    theTestMission(mission()),
    /*Requirements and specifications*/
    v_max(EndnodeReadOnly<double>(string("requirements_and_specifications/requirements/top_level_aircraft_requirements/flight_envelope/")
                                    + string("maximum_operating_velocity")).read(rtIO->acxml).value()),
    flightsPerYear(EndnodeReadOnly<int>("assessment/cost_estimation/operating_cost/direct_operating_cost/flights_per_year_study_mission").read(rtIO->acxml).value()),
    durationOperation(EndnodeReadOnly<int>("requirements_and_specifications/assessment_scenario/duration_operation").read(rtIO->acxml).value_xml()),
    seatsAvailable(EndnodeReadOnly<double>(string("requirements_and_specifications/design_specification/transport_task/passenger_definition/")
                                            + string("total_number_passengers")).read(rtIO->acxml).value()),
    /*Aerodynamics*/
    S_ref(EndnodeReadOnly<double>("analysis/aerodynamics/reference_values/S_ref").read(rtIO->acxml).value()),
    /*Masses*/
    MTOM(EndnodeReadOnly<double>("analysis/masses_cg_inertia/maximum_takeoff_mass/mass_properties/mass").read(rtIO->acxml).value()),
    OME(EndnodeReadOnly<double>("analysis/masses_cg_inertia/operating_mass_empty/mass_properties/mass").read(rtIO->acxml).value()),
    sparePartsMass(6),
    mainPartsMass(7),
    transportedCompMass(6),
    airframeMass(0.),
    finMassRatio(0.),
    stabilizerMassRatio(0.),
    pylonMassRatio(0.),
    nacelleMassRatio(0.),
    engineMass(0.),
    nacelleMass(0.),
    pylonMass(0.),
    bleedAirSystemMass(0.),
    fuelSystemMass(0.),
    otherEngineSystemsMass(0.),
    /*Propulsion*/
    numberOfEngines(rtIO->acxml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion/").size()),
    fuelFlowContThrustCruise(0.),
    /*Systems*/
    propDesignPowerTotal(0.),
    /*Design Mission*/
    rangeDesign(EndnodeReadOnly<double>("analysis/mission/design_mission/range").read(rtIO->acxml).value()),
    blocktimeDesign(convertUnit(SECOND, HOUR, EndnodeReadOnly<double>("analysis/mission/design_mission/block_time").read(rtIO->acxml).value())),
    flightTimeDesign(convertUnit(SECOND, HOUR, EndnodeReadOnly<double>("analysis/mission/design_mission/flight_time").read(rtIO->acxml).value())),
    taxifuelDesign(0.),
    tripfuelDesign(0.),
    blockfuelDesign(0.),
    LTOfuelDesign(0.),
    /*Study Mission*/
    rangeStudy(EndnodeReadOnly<double>("requirements_and_specifications/requirements/top_level_aircraft_requirements/study_mission/range").read(rtIO->acxml).value()),
    flightTimeStudy(EndnodeReadOnly<double>("analysis/mission/study_mission/flight_time").read(rtIO->acxml).value()),
    tripfuelStudy(0.),
    studyMissionPAX(0.),
    blockfuelStudy(0.) {
        configuration.readAllConfigData(rtIO->moduleConfig);
        readAircraft(rtIO);
        getMissionFileData(rtIO);
        readEcologicalAssessmentXML(ecoAssXML);
    }

/* Input data methods*/
void SchaeferIOData::readAircraft(std::shared_ptr<RuntimeIO> rtIO) {
    acftName = std::string(rtIO->acxml.at("aircraft_exchange_file/general/type/value")) + std::string("-") + std::string(rtIO->acxml.at("general/model/value"));
    double propDesignPowerElecHyd(EndnodeReadOnly<double>("component_design/systems/specific/maximium_power_demand").read(rtIO->acxml).value());
    propDesignPowerTotal = (fabs(propDesignPowerElecHyd)) * 1 / 0.27 /1000; //kW; assumption: 27% electrical+hydr. power, rest: pneumatically:
    //...s. _Wheeler_MoreElectricAircraft (under AircraftDesign/Systems/...)
    numberOfEngines = rtIO->acxml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion/").size();
    myEngine.set_same_engine_type_quantity(numberOfEngines);
    double buckt_point_thrust(EndnodeReadOnly<double>("component_design/propulsion/specific/propulsion/engine/bucket_point/thrust").read(rtIO->acxml).value());
    double buckt_point_tsfc(EndnodeReadOnly<double>("component_design/propulsion/specific/propulsion/engine/bucket_point/tsfc").read(rtIO->acxml).value());
    fuelFlowContThrustCruise = buckt_point_thrust * buckt_point_tsfc;
    studyMissionPAX = std::round(EndnodeReadOnly<double>(std::string("requirements_and_specifications/requirements/top_level_aircraft_requirements/study_mission/")
                                                + std::string("payload_fractions/passenger_mass_fraction")).read(rtIO->acxml).value()
                                    * EndnodeReadOnly<double>(std::string("requirements_and_specifications/design_specification/transport_task/")
                                                + std::string("passenger_definition/total_number_passengers")).read(rtIO->acxml).value());
    getMasses(rtIO);
}

void SchaeferIOData::getMissionFileData(std::shared_ptr<RuntimeIO> rtIO) {
    std::string missionFile = std::string(rtIO->acxml.at("/aircraft_exchange_file/requirements_and_specifications/mission_files/design_mission_file/value"));
    const node& missionXML(aixml::openDocument(rtIO->getMissionDataDir() + "/" + missionFile));
    taxiTime = EndnodeReadOnly<double>("mission/taxi_time_origin").read(missionXML).value() + EndnodeReadOnly<double>("mission/taxi_time_destination").read(missionXML).value();
}

void SchaeferIOData::getMasses(std::shared_ptr<RuntimeIO> rtIO) {
    /* Aircraft masses */
    double MME(EndnodeReadOnly<double>("analysis/masses_cg_inertia/manufacturer_mass_empty/mass_properties/mass").read(rtIO->acxml).value());
    double systemsMass(EndnodeReadOnly<double>("component_design/systems/mass_properties/mass").read(rtIO->acxml).value());
    double wingMass(EndnodeReadOnly<double>("component_design/wing/mass_properties/mass").read(rtIO->acxml).value());
    double fuselageMass(EndnodeReadOnly<double>("component_design/fuselage/mass_properties/mass").read(rtIO->acxml).value());
    double landingGearMass(EndnodeReadOnly<double>("component_design/landing_gear/mass_properties/mass").read(rtIO->acxml).value());

    /* Propulsion masses and ratio */
    double powerUnitMass(0.);
    int numberOfPropulsors(rtIO->acxml.getVector("aircraft_exchange_file/component_design/propulsion/specific/propulsion/").size());
    for (int i(0); i < numberOfPropulsors; i++) {
       nacelleMass += EndnodeReadOnly<double>("component_design/propulsion/specific/nacelle/mass_properties/mass").read(rtIO->acxml).value();
       pylonMass += EndnodeReadOnly<double>("component_design/propulsion/specific/pylon/mass_properties/mass").read(rtIO->acxml).value();
       engineMass += EndnodeReadOnly<double>("component_design/propulsion/specific/engine/mass_properties/mass").read(rtIO->acxml).value();
    }
    bleedAirSystemMass = EndnodeReadOnly<double>("component_design/systems/specific/geometry/mass_properties/bleed_air_system").read(rtIO->acxml).value();
    fuelSystemMass = EndnodeReadOnly<double>("component_design/systems/specific/geometry/mass_properties/fuel_system").read(rtIO->acxml).value();
    //otherEngineSystemsMass = todo;
    /*TODO: check if engine systems are part of propulsion mass for powerUnitMass, if not add it from systems block*/
    powerUnitMass = EndnodeReadOnly<double>("component_design/propulsion/mass_properties/mass").read(rtIO->acxml).value() - pylonMass;
    pylonMassRatio = pylonMass / (pylonMass + nacelleMass);
    nacelleMassRatio = nacelleMass / (pylonMass + nacelleMass);

    /* Empennage masses and ratios */
    double finMass(0.);
    double stabiliserMass(0.);
    node empennageGeometry(rtIO->acxml.at("aircraft_exchange_file/component_design/empennage/specific/geometry/"));
    int numberOfSurfaces(empennageGeometry.getVector("aerodynamic_surface").size());
    for (int i = 0; i < numberOfSurfaces; i++) {
        std::string name(EndnodeReadOnly<std::string>("aerodynamic_surface@" + num2Str(i) +"/name").read(empennageGeometry).value());
        double mass(EndnodeReadOnly<double>("aerodynamic_surface@" + num2Str(i) +"/mass_properties/mass").read(empennageGeometry).value());
        if (name == "vertical_stabiliser")
            finMass += mass;
        else if (name == "horizontal_stabiliser")
            stabiliserMass += mass;
    }
    empennageMass = finMass + stabiliserMass;
    finMassRatio = finMass / (empennageMass);
    stabilizerMassRatio = stabiliserMass / (empennageMass);

    /* Airframe mass */
    airframeMass =  MME - systemsMass - engineMass - nacelleMass;
    /* Spare parts masses*/
    sparePartsMass = {wingMass, fuselageMass, stabiliserMass, finMass, landingGearMass, powerUnitMass};
    /* Main parts masses*/
    mainPartsMass = {wingMass, fuselageMass, stabiliserMass, finMass, landingGearMass, pylonMass, powerUnitMass};
    /* Transported components masses*/
    transportedCompMass = {wingMass, fuselageMass, stabiliserMass, finMass, landingGearMass, powerUnitMass};

    /* Fuel masses */
    tripfuelDesign = EndnodeReadOnly<double>("analysis/mission/design_mission/trip_energy").read(rtIO->acxml).value()/rtIO->get_fuel_gravimetric_energy_density(0);;
    LTOfuelDesign = (EndnodeReadOnly<double>("analysis/mission/design_mission/takeoff_energy").read(rtIO->acxml).value()
                    + EndnodeReadOnly<double>("analysis/mission/design_mission/landing_energy").read(rtIO->acxml).value())
                    /rtIO->get_fuel_gravimetric_energy_density(0);
    taxifuelDesign = EndnodeReadOnly<double>("analysis/mission/design_mission/taxi_energy").read(rtIO->acxml).value()/rtIO->get_fuel_gravimetric_energy_density(0);;
    blockfuelDesign = taxifuelDesign + tripfuelDesign;
    taxifuelStudy = EndnodeReadOnly<double>("analysis/mission/study_mission/taxi_energy").read(rtIO->acxml).value()/rtIO->get_fuel_gravimetric_energy_density(0);;
    tripfuelStudy = EndnodeReadOnly<double>("analysis/mission/study_mission/trip_energy").read(rtIO->acxml).value()/rtIO->get_fuel_gravimetric_energy_density(0);;
    blockfuelStudy = tripfuelStudy + taxifuelStudy;
}

void SchaeferIOData::readEcologicalAssessmentXML(std::shared_ptr<node> ecologicalAssessmentXML) {
    const node mission_emissions(ecologicalAssessmentXML->at("ecological_data_file/mission_emissions"));
    theTestMission.LTO.CO2 = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/CO2").read(mission_emissions).value();
    theTestMission.LTO.H2O = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/H2O").read(mission_emissions).value();
    theTestMission.LTO.SO2 = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/SO2").read(mission_emissions).value();
    theTestMission.LTO.NOx = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/NOx").read(mission_emissions).value();
    theTestMission.LTO.HC = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/HC").read(mission_emissions).value();
    theTestMission.LTO.CH4 = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/CH4").read(mission_emissions).value();
    theTestMission.LTO.CO = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/CO").read(mission_emissions).value();
    theTestMission.LTO.soot = EndnodeReadOnly<double>("design_mission/emissions/LTO_cycle/soot").read(mission_emissions).value();
    theTestMission.Cruise.CO2 = EndnodeReadOnly<double>("design_mission/emissions/cruise/CO2").read(mission_emissions).value();
    theTestMission.Cruise.H2O = EndnodeReadOnly<double>("design_mission/emissions/cruise/H2O").read(mission_emissions).value();
    theTestMission.Cruise.SO2 = EndnodeReadOnly<double>("design_mission/emissions/cruise/SO2").read(mission_emissions).value();
    theTestMission.Cruise.NOx = EndnodeReadOnly<double>("design_mission/emissions/cruise/NOx").read(mission_emissions).value();
    theTestMission.Cruise.HC = EndnodeReadOnly<double>("design_mission/emissions/cruise/HC").read(mission_emissions).value();
    theTestMission.Cruise.CH4 = EndnodeReadOnly<double>("design_mission/emissions/cruise/CH4").read(mission_emissions).value();
    theTestMission.Cruise.CO = EndnodeReadOnly<double>("design_mission/emissions/cruise/CO").read(mission_emissions).value();
    theTestMission.Cruise.soot = EndnodeReadOnly<double>("design_mission/emissions/cruise/soot").read(mission_emissions).value();
    theStudyMission.LTO.CO2 = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/CO2").read(mission_emissions).value();
    theStudyMission.LTO.H2O = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/H2O").read(mission_emissions).value();
    theStudyMission.LTO.SO2 = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/SO2").read(mission_emissions).value();
    theStudyMission.LTO.NOx = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/NOx").read(mission_emissions).value();
    theStudyMission.LTO.HC = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/HC").read(mission_emissions).value();
    theStudyMission.LTO.CH4 = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/CH4").read(mission_emissions).value();
    theStudyMission.LTO.CO = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/CO").read(mission_emissions).value();
    theStudyMission.LTO.soot = EndnodeReadOnly<double>("study_mission/emissions/LTO_cycle/soot").read(mission_emissions).value();
    theStudyMission.Cruise.CO2 = EndnodeReadOnly<double>("study_mission/emissions/cruise/CO2").read(mission_emissions).value();
    theStudyMission.Cruise.H2O = EndnodeReadOnly<double>("study_mission/emissions/cruise/H2O").read(mission_emissions).value();
    theStudyMission.Cruise.SO2 = EndnodeReadOnly<double>("study_mission/emissions/cruise/SO2").read(mission_emissions).value();
    theStudyMission.Cruise.NOx = EndnodeReadOnly<double>("study_mission/emissions/cruise/NOx").read(mission_emissions).value();
    theStudyMission.Cruise.HC = EndnodeReadOnly<double>("study_mission/emissions/cruise/HC").read(mission_emissions).value();
    theStudyMission.Cruise.CH4 = EndnodeReadOnly<double>("study_mission/emissions/cruise/CH4").read(mission_emissions).value();
    theStudyMission.Cruise.CO = EndnodeReadOnly<double>("study_mission/emissions/cruise/CO").read(mission_emissions).value();
    theStudyMission.Cruise.soot = EndnodeReadOnly<double>("study_mission/emissions/cruise/soot").read(mission_emissions).value();
}

/* Output data methods*/
// void SchaeferIOData::updateEcologicalValuesXml(std::shared_ptr<node> ecologicalAssessmentXML, const mission& theStudyMission, const mission& theTestMission,
//                     const development& theDevelopment, const production& theProduction, const operation& theOperation, const endOfLife& theEndOfLife) {
//     std::string subPath("EcologicalValues/Emissions/");
//     /*LCA results*/
//     if (configuration.LCA) {
//         /*Development Block*/
//         if (configuration.developmentEmissionSetting != "mode_0") {
//             ecologicalAssessmentXML->at(subPath + "Development/CO2") = Rounding(theDevelopment.development_total.CO2, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/H2O") = Rounding(theDevelopment.development_total.H2O, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/NOx") = Rounding(theDevelopment.development_total.NOx, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/CO") = Rounding(theDevelopment.development_total.CO, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/SO2") = Rounding(theDevelopment.development_total.SO2, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/CH4") = Rounding(theDevelopment.development_total.CH4, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/PM10") = Rounding(theDevelopment.development_total.PM10, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/N2O") = Rounding(theDevelopment.development_total.N2O, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/PFC") = Rounding(theDevelopment.development_total.PFC, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/NMVOC") = Rounding(theDevelopment.development_total.NMVOC, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/HC") = Rounding(theDevelopment.development_total.HC, 3);
//             ecologicalAssessmentXML->at(subPath + "Development/CO2") = Rounding(theDevelopment.development_total.CO2, 3);
//         }
//         /*Production Block*/
//         ecologicalAssessmentXML->at(subPath + "Production/CO2") = Rounding(theProduction.production_total.CO2, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/H2O") = Rounding(theProduction.production_total.H2O, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/NOx") = Rounding(theProduction.production_total.NOx, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/CO") = Rounding(theProduction.production_total.CO, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/SO2") = Rounding(theProduction.production_total.SO2, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/CH4") = Rounding(theProduction.production_total.CH4, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/PM10") = Rounding(theProduction.production_total.PM10, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/N2O") = Rounding(theProduction.production_total.N2O, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/PFC") = Rounding(theProduction.production_total.PFC, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/NMVOC") = Rounding(theProduction.production_total.NMVOC, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/HC") = Rounding(theProduction.production_total.HC, 3);
//         ecologicalAssessmentXML->at(subPath + "Production/CO2") = Rounding(theProduction.production_total.CO2, 3);

//         /*End Of Life Block*/
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/CO2") = Rounding(theEndOfLife.endOfLife_total.CO2, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/H2O") = Rounding(theEndOfLife.endOfLife_total.H2O, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/NOx") = Rounding(theEndOfLife.endOfLife_total.NOx, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/CO") = Rounding(theEndOfLife.endOfLife_total.CO, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/SO2") = Rounding(theEndOfLife.endOfLife_total.SO2, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/CH4") = Rounding(theEndOfLife.endOfLife_total.CH4, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/PM10") = Rounding(theEndOfLife.endOfLife_total.PM10, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/N2O") = Rounding(theEndOfLife.endOfLife_total.N2O, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/PFC") = Rounding(theEndOfLife.endOfLife_total.PFC, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/NMVOC") = Rounding(theEndOfLife.endOfLife_total.NMVOC, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/HC") = Rounding(theEndOfLife.endOfLife_total.HC, 3);
//         ecologicalAssessmentXML->at(subPath + "EndOfLife/CO2") = Rounding(theEndOfLife.endOfLife_total.CO2, 3);

//         /*CED*/
//         subPath = "EcologicalValues/CED/";
//         double CEDtotal = theDevelopment.development_total.CED + theProduction.production_total.CED +
//                         theOperation.operation_total.CED + theEndOfLife.endOfLife_total.CED;
//         ecologicalAssessmentXML->at(subPath + "CEDTotal") = Rounding(CEDtotal, 3);
//         if (configuration.developmentEmissionSetting != "mode_0") {
//             ecologicalAssessmentXML->at(subPath + "CEDComponents/Development") = Rounding(theDevelopment.development_total.CED, 3);
//         }
//         ecologicalAssessmentXML->at(subPath + "CEDComponents/Production") = Rounding(theProduction.production_total.CED, 3);
//         ecologicalAssessmentXML->at(subPath + "CEDComponents/Operations") = Rounding(theOperation.operation_total.CED, 3);
//         ecologicalAssessmentXML->at(subPath + "CEDComponents/EndOfLife") = Rounding(theEndOfLife.endOfLife_total.CED, 3);
//     }

//     /* Operation Block */
//     subPath = "EcologicalValues/Emissions/Operations/";
//     /*Total emissions during the operational phase*/
//     ecologicalAssessmentXML->at(subPath + "CO2") = Rounding(theOperation.operation_total.CO2, 3);
//     ecologicalAssessmentXML->at(subPath + "H2O") = Rounding(theOperation.operation_total.H2O, 3);
//     ecologicalAssessmentXML->at(subPath + "NOx") = Rounding(theOperation.operation_total.NOx, 3);
//     ecologicalAssessmentXML->at(subPath + "CO") = Rounding(theOperation.operation_total.CO, 3);
//     ecologicalAssessmentXML->at(subPath + "SO2") = Rounding(theOperation.operation_total.SO2, 3);
//     ecologicalAssessmentXML->at(subPath + "CH4") = Rounding(theOperation.operation_total.CH4, 3);
//     ecologicalAssessmentXML->at(subPath + "PM10") = Rounding(theOperation.operation_total.PM10, 3);
//     ecologicalAssessmentXML->at(subPath + "N2O") = Rounding(theOperation.operation_total.N2O, 3);
//     ecologicalAssessmentXML->at(subPath + "PFC") = Rounding(theOperation.operation_total.PFC, 3);
//     ecologicalAssessmentXML->at(subPath + "NMVOC") = Rounding(theOperation.operation_total.NMVOC, 3);
//     ecologicalAssessmentXML->at(subPath + "HC") = Rounding(theOperation.operation_total.HC, 3);
//     ecologicalAssessmentXML->at(subPath + "CO2") = Rounding(theOperation.operation_total.CO2, 3);
//     /*ICAO LTO*/
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/CO2") = Rounding(theStudyMission.ICAO_LTO.CO2, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/H2O") = Rounding(theStudyMission.ICAO_LTO.H2O, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/SO2") = Rounding(theStudyMission.ICAO_LTO.SO2, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/HC") = Rounding(theStudyMission.ICAO_LTO.HC, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/CO") = Rounding(theStudyMission.ICAO_LTO.CO, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/NOx") = Rounding(theStudyMission.ICAO_LTO.NOx, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/ICAO_LTO/PM10") = Rounding(theStudyMission.ICAO_LTO.soot, 3);
//     /*LTO Times*/
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/Total") = Rounding(theStudyMission.LTOtime, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/TaxiTime") = Rounding(theStudyMission.taxiTime / 60., 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/TakeOffTime") = Rounding(theStudyMission.takeOffTime / 60., 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/ClimbOutTime") = Rounding(theStudyMission.climbOutTime / 60., 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTOTimes/ApproachTime") = Rounding(theStudyMission.approachTime / 60., 3);
//     /*LTO*/
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTO/CO2") = Rounding(theStudyMission.LTO.CO2, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTO/H2O") = Rounding(theStudyMission.LTO.H2O, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTO/SO2") = Rounding(theStudyMission.LTO.SO2, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTO/HC") = Rounding(theStudyMission.LTO.HC, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTO/CO/") = Rounding(theStudyMission.LTO.CO, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTO/NOx") = Rounding(theStudyMission.LTO.NOx, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/LTO/PM10") = Rounding(theStudyMission.LTO.soot, 3);
//     /*Cruise*/
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/Cruise/CO2") = Rounding(theStudyMission.Cruise.CO2, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/Cruise/H2O") = Rounding(theStudyMission.Cruise.H2O, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/Cruise/SO2") = Rounding(theStudyMission.Cruise.SO2, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/Cruise/HC") = Rounding(theStudyMission.Cruise.HC, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/Cruise/CO") = Rounding(theStudyMission.Cruise.CO, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/Cruise/NOx") = Rounding(theStudyMission.Cruise.NOx, 3);
//     ecologicalAssessmentXML->at(subPath + "OperationComponents/Mission/Cruise/PM10") = Rounding(theStudyMission.Cruise.soot, 3);
// }
