/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "mission.h"

void Mission::calculate_hydrogen_combustion_emissions() {
    emissionFlow.resize(numberOfSteps);
    /* H2O - Stoichiometric factor for the combustion of hydrogen: 8.94 kg/kgFuel */
    this->LTO.H2O = m_fuel_LTO * EMISSION_INDEX_H2O_HYDROGEN;
    this->Cruise.H2O = m_fuel_Cruise * EMISSION_INDEX_H2O_HYDROGEN;
    this->LTO_Total.H2O = this->LTO.H2O * flightsPerLC;
    this->Cruise_Total.H2O = this->Cruise.H2O * flightsPerLC;
    for (int16_t i(0); i < numberOfSteps; i++) {
        emissionFlow[i].H2O = missionInfo[i].fuelFlow * EMISSION_INDEX_H2O_HYDROGEN;
    }
    /* NOx - Non proportional to fuel flow */
    if (missionIOData_->config.hydrogenMethodNOx == "mode_0") { // assume during different mission phases constant emission indices [Kos22]
        hydrogen_combustion_NOx_constants();
    } else if (missionIOData_->config.hydrogenMethodNOx == "mode_1") { // use kerosene emissions and adapt it to hydrogen
        hydrogen_combustion_NOx_P3T3();
    } else {
        myRuntimeInfo->out << "NOx emission method " << missionIOData_->config.hydrogenMethodNOx << " not implemented. Check settings in configuration file!" << std::endl;
    }
}

void Mission::hydrogen_combustion_NOx_constants() {
    double EI_NOx_taxi(1.1e-3);
    double EI_NOx_TO(8.02e-3);
    double EI_NOx_climb(6.17e-3);
    double EI_NOx_cruise(3.14e-3);
    double EI_NOx_app(2.4e-3);
    double EI_NOx_mom(0.);
    /* Taxiing NOx */
    this->LTO.NOx = EI_NOx_taxi * m_fuel_Taxi; //kg
    /* Non-taxiing NOx*/
    for (int missionPoint = 0; missionPoint < numberOfSteps ; missionPoint++) {
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4
            && (missionInfo.at(missionPoint).engineRating == "takeoff" || missionInfo.at(missionPoint).engineRating == "maximum_continuous")) {
            EI_NOx_mom = EI_NOx_TO;
        } else if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4 && missionInfo.at(missionPoint).engineRating == "climb") {
            EI_NOx_mom = EI_NOx_climb;
        } else if (this->myFlightPath.pathPoints.at(missionPoint).altitude > 914.4) {
            EI_NOx_mom = EI_NOx_cruise;
        } else if (this->myFlightPath.pathPoints.at(missionPoint).altitude < 914.4
            && (missionInfo.at(missionPoint).engineRating == "idle" || missionInfo.at(missionPoint).engineRating == "cruise")) {
            EI_NOx_mom = EI_NOx_app;
        }
        emissionFlow[missionPoint].NOx = EI_NOx_mom * missionInfo[missionPoint].fuelFlow;
        /* Calculate NOx mass */
        if (missionPoint == 0) {
            timeStep = missionInfo.at(missionPoint).time;
        } else {
            timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.NOx += EI_NOx_mom * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        } else {
            this->Cruise.NOx += EI_NOx_mom * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        }
    }
}

void Mission::hydrogen_combustion_NOx_P3T3() {
    kerosene_NOx_P3T3();
    double hydrogenNOxFactor = calculateNOxHydrogenFactor();
    if (hydrogenNOxFactor > ACCURACY_HIGH) {
        this->LTO.NOx = hydrogenNOxFactor * this->LTO.NOx;
        this->Cruise.NOx = hydrogenNOxFactor * this->Cruise.NOx;
        this->LTO_Total.NOx = this->LTO.NOx * flightsPerLC;
        this->Cruise_Total.NOx = this->Cruise.NOx * flightsPerLC;
        for (int16_t i(0); i < missionInfo.size(); i++) {
            emissionFlow[i].NOx = missionInfo[i].fuelFlow * hydrogenNOxFactor * EINOxKerosene[i];
        }

        /* Emissions calculated via ICAO Fuel and ICAO Times and ICAO EI (if available, otherwise constants) */
        double ICAO_LTO_Fuel =  missionIOData_->numberOfEngines * ((missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * myEcoDatabase.taxiTime_ICAO)
                                + (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::takeoff) * myEcoDatabase.takeoffTime_ICAO)
                                + (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::climb) * myEcoDatabase.climboutTime_ICAO)
                                + (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::approach) * myEcoDatabase.appTime_ICAO));
        this->ICAO_LTO.H2O = ICAO_LTO_Fuel * EMISSION_INDEX_H2O_HYDROGEN;
        this->ICAO_LTO.NOx = missionIOData_->numberOfEngines * hydrogenNOxFactor
            * ((missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx)
                * myEcoDatabase.taxiTime_ICAO)
            + (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::takeoff) * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::NOx)
                * myEcoDatabase.takeoffTime_ICAO)
            + (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::climb) * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::NOx)
                * myEcoDatabase.climboutTime_ICAO)
            + (missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::approach) * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::NOx)
                * myEcoDatabase.appTime_ICAO));
    } else { // use mode_0 if Marek correlation is not suitable
        hydrogen_combustion_NOx_constants();
    }
}

double Mission::calculateNOxHydrogenFactor() {
    std::vector<double> H2Factor;
    for (int missionPoint = 0; missionPoint < numberOfSteps ; missionPoint++) {
        double T3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature; //[K]
        double P3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3TotalPressure / pow(1. + (1.4 - 1.)
                    / 2.*pow(this->myFlightPath.pathPoints.at(missionPoint).St3Mach, 2.), 1.4 / (1.4 - 1.)); //[kPa]
        if (588.7 < T3_mom && T3_mom < 802.6 && 689.4 < P3_mom && P3_mom < 1043.2 //calculate EI according to [Mar05] in validity range of correlation
            && missionInfo.at(missionPoint).modeName == "Cruise") {
            double FA_H2 = this->myFlightPath.pathPoints.at(missionPoint).St4FuelAir;
            double Phi_H2 = FA_H2/0.0292; // equivalence ration (fa/fa_stoc)
            double residenceTime = 2.0; //[ms]
            /*correlation constants for emission index based on expermental data (C3 config)*/
            double constant_A(101);
            double constant_a(.0);
            double constant_b(2.99);
            double constant_c(0.439);
            double constant_d(547);
            double constant_e(0.165);
            double pressureLoss(20);// dp/p [%]
            /*Calculate NOx emission from hydrogen correlation*/
            double ppmNOxH2 = constant_A * pow(143+P3_mom, constant_a) * pow(Phi_H2, constant_b) *
                pow(residenceTime, constant_c) * exp((1.8 * T3_mom - 460) / constant_d) * pow(pressureLoss, constant_e);
            double EINOxH2 = convertUnit(GRAM, KILO, GRAM, (1+FA_H2)/(FA_H2*533) * ppmNOxH2);
            /*Calculate scaling factor EI_H2:EI_Kerosene in current missionPoint*/
            H2Factor.push_back(EINOxH2 / EINOxKerosene.at(missionPoint));
            //myRuntimeInfo->out << "T3: " << T3_mom << " - p3: " << P3_mom << " - EI H2: " << EINOxH2*1000 <<
            //" - EI Kerosene: " <<  EINOxKerosene.at(missionPoint)*1000 << std::endl;
        }
    }
    /* Return value */
    if (H2Factor.size() == 0) {
        myRuntimeInfo->warn << "No cruise point matches EINOx correlation range! mode_0 will be used to calculate emissions." << std::endl;
        return 0;
    } else {
        return std::accumulate(H2Factor.begin(), H2Factor.end(), 0.0) / H2Factor.size();
    }
}
