/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "mission.h"

void Mission::calculate_kerosene_emissions() {
    emissionFlow.resize(numberOfSteps);
    /* CO2 - Stoichiometric factor for the combustion of kerosene: 3.149 kg/kgFuel */
    this->LTO.CO2 = m_fuel_LTO * EMISSION_INDEX_CO2_KEROSENE;
    this->Cruise.CO2 = m_fuel_Cruise * EMISSION_INDEX_CO2_KEROSENE;
    this->LTO_Total.CO2 = this->LTO.CO2 * flightsPerLC;
    this->Cruise_Total.CO2 = this->Cruise.CO2 * flightsPerLC;

    /* H2O - Stoichiometric factor for the combustion of kerosene: 1.2 kg/kgFuel */
    this->LTO.H2O = m_fuel_LTO * EMISSION_INDEX_H2O_KEROSENE;
    this->Cruise.H2O = m_fuel_Cruise * EMISSION_INDEX_H2O_KEROSENE;
    this->LTO_Total.H2O = this->LTO.H2O * flightsPerLC;
    this->Cruise_Total.H2O = this->Cruise.H2O * flightsPerLC;

    /* SO2 - Stoichiometric factor for the combustion of kerosene: 0.84e-3 kg/kgFuel */
    this->LTO.SO2 = m_fuel_LTO * EMISSION_INDEX_SO2_KEROSENE;
    this->Cruise.SO2 = m_fuel_Cruise * EMISSION_INDEX_SO2_KEROSENE;
    this->LTO_Total.SO2 = this->LTO.SO2 * flightsPerLC;
    this->Cruise_Total.SO2 = this->Cruise.SO2 * flightsPerLC;

    /* SO4 - Stoichiometric factor for the combustion of kerosene:2e-4 kg/kgFuel */
    this->LTO.SO4 = m_fuel_LTO * EMISSION_INDEX_SO4_KEROSENE;
    this->Cruise.SO4 = m_fuel_Cruise * EMISSION_INDEX_SO4_KEROSENE;
    this->LTO_Total.SO4 = this->LTO.SO4 * flightsPerLC;
    this->Cruise_Total.SO4 = this->Cruise.SO4 * flightsPerLC;

    /* Fill mission vector with emissions proportional to fuel flow*/
    for (int16_t i(0); i < numberOfSteps; i++) {
        emissionFlow[i].CO2 = missionInfo[i].fuelFlow * EMISSION_INDEX_CO2_KEROSENE;
        emissionFlow[i].H2O = missionInfo[i].fuelFlow * EMISSION_INDEX_H2O_KEROSENE;
        emissionFlow[i].SO2 = missionInfo[i].fuelFlow * EMISSION_INDEX_SO2_KEROSENE;
        emissionFlow[i].SO4 = missionInfo[i].fuelFlow * EMISSION_INDEX_SO4_KEROSENE;
    }

    /* HC - Non proportional to fuel flow */
    if (missionIOData_->config.keroseneMethodHC == "mode_0") { //DLR Omega
        kerosene_HC_DLR();
    } else if (missionIOData_->config.keroseneMethodHC == "mode_1") { //BFM2;
        kerosene_HC_BFM2();
    } else { //ICAO
        this->LTO.HC = (m_fuel_Taxi * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC))
                        + (m_fuel_TO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC))
                        + (m_fuel_ClimbOut * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::HC))
                        + (m_fuel_App * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::HC));
        this->Cruise.HC = 0.;//not calculable with ICAO
    }
    this->LTO_Total.HC = this->LTO.HC * flightsPerLC;
    this->Cruise_Total.HC = this->Cruise.HC * flightsPerLC;

    /* CO - Non proportional to fuel flow */
    if (missionIOData_->config.kersoseneMethodCO == "mode_0") { //DLR Omega
        kerosene_CO_DLR();
    } else if (missionIOData_->config.kersoseneMethodCO == "mode_1") { //BFM2;
        kerosene_CO_BFM2();
    } else { //ICAO
        this->LTO.CO = (m_fuel_Taxi * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO))
                        + (m_fuel_TO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::CO))
                        + (m_fuel_ClimbOut * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::CO))
                        + (m_fuel_App * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::CO));
        this->Cruise.CO = 0.;//not calculable with ICAO
    }
    this->LTO_Total.CO = this->LTO.CO * flightsPerLC;
    this->Cruise_Total.CO = this->Cruise.CO * flightsPerLC;

    /* NOx - Non proportional to fuel flow */
    if (missionIOData_->config.keroseneMethodNOx == "mode_0") { //P3T3
        kerosene_NOx_P3T3();
    } else if (missionIOData_->config.keroseneMethodNOx == "mode_1") { //BFM2;
        kerosene_NOx_BFM2();
    } else { //GasTurb
        kerosene_NOx_gasTurb();
    }
    this->LTO_Total.NOx = this->LTO.NOx * flightsPerLC;
    this->Cruise_Total.NOx = this->Cruise.NOx * flightsPerLC;

    /* Soot - Non proportional to fuel flow */
    if (rtIOPt->aircraft_energy_carrier_type() == "kerosene") {
        if (missionIOData_->config.keroseneMethodSoot == "mode_0") { //DLR
            kerosene_soot_DLR();
        } else if (missionIOData_->config.keroseneMethodSoot == "mode_1") { //Whyte;
            kerosene_soot_whyte();
        } else { //const. EI = 0.025e-3 kg/kgFuel
            this->LTO.soot = m_fuel_LTO * EMISSION_INDEX_SO4_KEROSENE;
            this->Cruise.soot = m_fuel_Cruise * EMISSION_INDEX_SO4_KEROSENE;
        }
    }
    this->Cruise_Total.soot = this->Cruise.soot * flightsPerLC;
    this->LTO_Total.soot = this->LTO.soot * flightsPerLC;

    /* Emissions calculated via ICAO Fuel and ICAO Times and ICAO EI (if available, otherwise constants) */
    double ICAO_LTO_Fuel =  m_fuel_Taxi_ICAO + m_fuel_TO_ICAO + m_fuel_ClimbOut_ICAO + m_fuel_App_ICAO;
    this->ICAO_LTO.CO2 = ICAO_LTO_Fuel * EMISSION_INDEX_CO2_KEROSENE;
    this->ICAO_LTO.H2O = ICAO_LTO_Fuel * EMISSION_INDEX_H2O_KEROSENE;
    this->ICAO_LTO.SO2 = ICAO_LTO_Fuel * EMISSION_INDEX_SO2_KEROSENE;
    this->ICAO_LTO.HC = missionIOData_->numberOfEngines *
                        (m_fuel_Taxi_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC)
                        + m_fuel_TO_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC)
                        + m_fuel_App_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::HC)
                        + m_fuel_App_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::HC));
    this->ICAO_LTO.CO = missionIOData_->numberOfEngines *
                        (m_fuel_Taxi_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO)
                        + m_fuel_TO_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::CO)
                        + m_fuel_App_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::CO)
                        + m_fuel_App_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::CO));
    this->ICAO_LTO.NOx = missionIOData_->numberOfEngines *
                        (m_fuel_Taxi_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx)
                        + m_fuel_TO_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::NOx)
                        + m_fuel_App_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::NOx)
                        + m_fuel_App_ICAO * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::NOx));
    this->ICAO_LTO.soot = ICAO_LTO_Fuel * EMISSION_INDEX_SO4_KEROSENE;
}

void Mission::kerosene_NOx_P3T3() {
    /* Taxiing NOx */
    this->LTO.NOx = missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx) * m_fuel_Taxi; //kg
    /* Non-taxiing NOx*/
    for (int missionPoint = 0; missionPoint < numberOfSteps ; missionPoint++) {
        double T3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature;
        double P3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3TotalPressure / pow(1. + (1.4 - 1.)
                    / 2.*pow(this->myFlightPath.pathPoints.at(missionPoint).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double n = 0.4; //exponent of the altitude correction for P3T3 Method, can be changed for best fit (0.4 recommended; p.29 Norman 2003)
        double beta = 7.90298 * (1. - 373.16 / (this->myFlightPath.pathPoints.at(missionPoint).Tamb)) + 3.00571 + 5.02808
                * log10(373.16 / (this->myFlightPath.pathPoints.at(missionPoint).Tamb)) + 1.3816 * pow(10., -7.)
                * (1. - pow(10., 11.344 * (1. - (this->myFlightPath.pathPoints.at(missionPoint).Tamb)
                / 373.16))) + 8.1328 * pow(10., -3.) * (pow(10., 3.49149 * (1. - 373.16 / (this->myFlightPath.pathPoints.at(missionPoint).Tamb))) - 1.);
        double P_v = 0.014504 * pow(10., beta); //psi
        //(psi = Pa * 0.00014504); AEM3 Correction Implemented!!!
        double omega = 0.62198 * missionIOData_->config.relativeHumidity * P_v / (this->myFlightPath.pathPoints.at(missionPoint).Pamb *
            0.00014504 - 0.37802 * missionIOData_->config.relativeHumidity * P_v);
        double H = -19.0 * (omega - 0.00634);
        double EINOx_SL = quadraticFit(T3_mom, T3_ID, T3_TO, T3_CL, T3_AP,
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx),
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::NOx),
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::NOx),
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::NOx), 0);// [kg/kg]
        double P3_SL = potentialFit(T3_mom, T3_ID, T3_TO, T3_CL, T3_AP, P3_ID, P3_TO, P3_CL, P3_AP, 1); //[kPa]
        if (EINOx_SL < 0.) {
            EINOx_SL = 0.;
        }
        EINOxKerosene.push_back(EINOx_SL * pow(P3_mom / P3_SL, n) * exp(H)); //kg/kg
        emissionFlow[missionPoint].NOx = EINOxKerosene.back() * missionInfo[missionPoint].fuelFlow;

        /* calculate NOx mass */
        if (missionPoint == 0) {
            timeStep = missionInfo.at(missionPoint).time;
        } else {
            timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.NOx += EINOxKerosene.at(missionPoint) * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        } else {
            this->Cruise.NOx += EINOxKerosene.at(missionPoint) * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        }
    }
}

void Mission::kerosene_NOx_BFM2() {
    this->LTO.NOx = missionIOData_->numberOfEngines * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx)
                    * missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * taxiTime; //kg
    for (int missionPoint = 0; missionPoint < numberOfSteps ; missionPoint++) {
        double delta_amb = this->myFlightPath.pathPoints.at(missionPoint).Pamb / 101325.;
        double theta_amb = this->myFlightPath.pathPoints.at(missionPoint).Tamb / 288.15;
        double beta = 7.90298 * (1. - 373.16 / (this->myFlightPath.pathPoints.at(missionPoint).Tamb)) + 3.00571 + 5.02808 * log10(373.16
            / (this->myFlightPath.pathPoints.at(missionPoint).Tamb)) + 1.3816 * pow(10., -7.) * (1. - pow(10., 11.344
            * (1. - (this->myFlightPath.pathPoints.at(missionPoint).Tamb) / 373.16))) + 8.1328 * pow(10., -3.)
            * (pow(10., 3.49149 * (1. - 373.16 / (this->myFlightPath.pathPoints.at(missionPoint).Tamb))) - 1.);
        double P_v = 0.014504 * pow(10., beta);
        //(psi = Pa * 0.00014504); AEM3 Correction Implemented!!!
        double omega = 0.62198 * missionIOData_->config.relativeHumidity * P_v / (this->myFlightPath.pathPoints.at(missionPoint).Pamb *
            0.00014504 - 0.37802 * missionIOData_->config.relativeHumidity * P_v);
        double H = -19.0 * (omega - 0.0063);
        ///For EINOx: Choose between LinearLeastSquareFit() or PointToPointFit()
        EINOxKerosene.push_back(LinearLeastSquareFit(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::NOx),
                missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::NOx),
                missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::NOx),
                missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx), missionPoint)
                * exp(H) * pow(pow(delta_amb, 1.02) / pow(theta_amb, 3.3), 0.5)); //[kg/kg];
        emissionFlow[missionPoint].NOx = EINOxKerosene.back() * missionInfo[missionPoint].fuelFlow;
        if (missionPoint == 0) {
            timeStep = missionInfo.at(missionPoint).time;
        } else {
            timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.NOx += EINOxKerosene.at(missionPoint) * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        } else {
            this->Cruise.NOx += EINOxKerosene.at(missionPoint) * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        }
    }
}

void Mission::kerosene_NOx_gasTurb() {
    this->LTO.NOx = missionIOData_->numberOfEngines * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::NOx)
                    * missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * taxiTime; //kg
    for (int missionPoint = 1; missionPoint < numberOfSteps; missionPoint++) {
        /* First mission point */
        double altitude(missionInfo.at(missionPoint-1).alt); // Flight Level [m]
        double Mach(missionInfo.at(missionPoint-1).Mach); // Mach number
        double N1 =  missionInfo.at(missionPoint-1).N1; // Current low pressure spool speed [-]
        missionIOData_->myEngine.set_operating_point({N1, altitude, Mach});
        double NOx_EI_1(missionIOData_->myEngine.get_engine_NOx_emission_index());
        /* Second mission point */
        altitude = missionInfo.at(missionPoint).alt;// Flight Level [m]
        Mach = missionInfo.at(missionPoint).Mach;// Mach number
        N1 =  missionInfo.at(missionPoint-1).N1; // Current low pressure spool speed [-]
        missionIOData_->myEngine.set_operating_point({N1, altitude, Mach});
        double NOx_EI_2(missionIOData_->myEngine.get_engine_NOx_emission_index());
        /* Average EI */
        EINOxKerosene.push_back((NOx_EI_2 + NOx_EI_1) / 2.);
        /* Nox masses */
        timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.NOx += EINOxKerosene.at(missionPoint-1) * this->myFlightPath.pathPoints.at(missionPoint-1).fuelFlow * timeStep; //kg
        } else {
            this->Cruise.NOx += EINOxKerosene.at(missionPoint-1) * this->myFlightPath.pathPoints.at(missionPoint-1).fuelFlow * timeStep; //kg
        }
        /* Nox flow */
        emissionFlow[missionPoint].NOx = EINOxKerosene.back() * missionInfo[missionPoint].fuelFlow;
    }
}

void Mission::kerosene_HC_BFM2() {
    this->LTO.HC = missionIOData_->numberOfEngines * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC)
                * missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * taxiTime; //kg
    for (int missionPoint = 0; missionPoint < numberOfSteps ; missionPoint++) {
        double delta_amb = this->myFlightPath.pathPoints.at(missionPoint).Pamb / 101325.;
        double theta_amb = this->myFlightPath.pathPoints.at(missionPoint).Tamb / 288.15;
        double EIHC = biLinearLeastSquareFit(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC),
                        missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::HC),
                        missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::HC),
                        missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC), missionPoint)
                        * pow(theta_amb, 3.3) / pow(delta_amb, 1.02); //[kg/kg];
        if (missionPoint == 0) {
            timeStep = missionInfo.at(missionPoint).time;
        } else {
            timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.HC += EIHC * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        } else {
            this->Cruise.HC += EIHC * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        }
        emissionFlow[missionPoint].HC = EIHC * missionInfo[missionPoint].fuelFlow;
    }
}

void Mission::kerosene_HC_DLR() {
    double c = 0.4; //exponent of the altitude correction for Omega Method, can be changed for best fit (0.4 recommended p.39 Petzold 2002)
    //Pressure in bar according to Doepelheuer 2002! but no matter, because potential fit!
    double omegaVC_ID = St3MassFlow_ID / pow(P3_ID / 100., 1.8) / exp(T3_ID / 300.);
    double omegaVC_TO = St3MassFlow_TO / pow(P3_TO / 100., 1.8) / exp(T3_TO / 300.);
    double omegaVC_CL = St3MassFlow_CL / pow(P3_CL / 100., 1.8) / exp(T3_CL / 300.);
    double omegaVC_AP = St3MassFlow_AP / pow(P3_AP / 100., 1.8) / exp(T3_AP / 300.);
    this->LTO.HC = missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC) * m_fuel_Taxi; //kg
    for (int missionPoint = 1; missionPoint < numberOfSteps ; missionPoint++) {
        double T3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature;
        double P3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3TotalPressure / pow(1. + (1.4 - 1.)
                 / 2.*pow(this->myFlightPath.pathPoints.at(missionPoint).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double omegaVC_mom = this->myFlightPath.pathPoints.at(missionPoint).St3MassFlow / pow(P3_mom / 100., 1.8) / exp(T3_mom / 300.);
        missionIOData_->myEngine.set_operating_point({this->myFlightPath.pathPoints.at(missionPoint).N1, 0, 0});
        double EIHC_SL = quadraticFit(omegaVC_mom, omegaVC_ID, omegaVC_TO, omegaVC_CL, omegaVC_AP,
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::HC),
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC),
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::HC),
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::HC), 1);
        double T3_SL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
        double P3t_SL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
        double St3Mach_SL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
        double P3_SL = P3t_SL / (1. + 1.4 / 2.*pow(St3Mach_SL, 2));
        double EIHC = EIHC_SL * pow(this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature / T3_SL * P3_SL / P3_mom, c); //kg/kg
        if (missionPoint == 0) {
            timeStep = missionInfo.at(missionPoint).time;
        } else {
            timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.HC += EIHC_SL * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        } else if (this->myFlightPath.pathPoints.at(missionPoint).altitude < 12192.) {
            this->Cruise.HC += EIHC_SL * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        } else {
            this->Cruise.HC += EIHC * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        }
        emissionFlow[missionPoint].HC = EIHC * missionInfo[missionPoint].fuelFlow;
    }
}

void Mission::kerosene_CO_BFM2() {
    this->LTO.CO = missionIOData_->numberOfEngines * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO)
                    * missionIOData_->myEngine.get_LTO_fuelflow(LTOPhases::taxi) * taxiTime; //kg
    for (int missionPoint = 0; missionPoint < numberOfSteps ; missionPoint++) {
        double delta_amb = this->myFlightPath.pathPoints.at(missionPoint).Pamb / 101325.;
        double theta_amb = this->myFlightPath.pathPoints.at(missionPoint).Tamb / 288.15;
        double EICO = biLinearLeastSquareFit(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::HC),
                        missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::CO),
                        missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::CO),
                        missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO), missionPoint)
                        * pow(theta_amb, 3.3) / pow(delta_amb, 1.02); //[kg/kg];
        if (missionPoint == 0) {
            timeStep = missionInfo.at(missionPoint).time;
        } else {
            timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.CO += EICO * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        } else {
            this->Cruise.CO += EICO * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        }
        emissionFlow[missionPoint].CO = EICO * missionInfo[missionPoint].fuelFlow;
    }
}

void Mission::kerosene_CO_DLR() {
    double c = 0.4; //exponent of the altitude correction for Omega Method, can be changed for best fit (0.4 recommended p.39 Petzold 2002)
    //Pressure in bar according to Doepelheuer 2002! but no matter, because potential fit!
    double omegaVC_ID = St3MassFlow_ID / pow(P3_ID / 100., 1.8) / exp(T3_ID / 300.);
    double omegaVC_TO = St3MassFlow_TO / pow(P3_TO / 100., 1.8) / exp(T3_TO / 300.);
    double omegaVC_CL = St3MassFlow_CL / pow(P3_CL / 100., 1.8) / exp(T3_CL / 300.);
    double omegaVC_AP = St3MassFlow_AP / pow(P3_AP / 100., 1.8) / exp(T3_AP / 300.);
    this->LTO.CO = missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO) * m_fuel_Taxi; //kg
    for (int missionPoint = 1; missionPoint < numberOfSteps ; missionPoint++) {
        double T3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature;
        double P3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3TotalPressure / pow(1. + (1.4 - 1.)
                 / 2.*pow(this->myFlightPath.pathPoints.at(missionPoint).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double omegaVC_mom = this->myFlightPath.pathPoints.at(missionPoint).St3MassFlow / pow(P3_mom / 100., 1.8) / exp(T3_mom / 300.);
        double EICO_SL = quadraticFit(omegaVC_mom, omegaVC_ID, omegaVC_TO, omegaVC_CL, omegaVC_AP,
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::CO),
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::CO),
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::CO),
                            missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::CO), 1); // [kg/kg]
        //Set temporary engine rating to SL
        missionIOData_->myEngine.set_operating_point({this->myFlightPath.pathPoints.at(missionPoint).N1, 0, 0});
        //Use manual engine rating for sea-level engine data
        double T3_SL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::StaticTemperature, EngineStage::St3);
        double P3t_SL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
        double St3Mach_SL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
        double P3_SL = P3t_SL / (1. + 1.4 / 2.*pow(St3Mach_SL, 2));
        double EICO = EICO_SL * pow(this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature / T3_SL * P3_SL / P3_mom, c); //kg/kg
        if (missionPoint == 0) {
            timeStep = missionInfo.at(missionPoint).time;
        } else {
            timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.CO += EICO_SL * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        } else if (this->myFlightPath.pathPoints.at(missionPoint).altitude < 12192.) {
            this->Cruise.CO += EICO_SL * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        } else {
            this->Cruise.CO += EICO * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        }
        emissionFlow[missionPoint].CO = EICO * missionInfo[missionPoint].fuelFlow;
    }
}

void Mission::kerosene_soot_DLR() {
    double EI_soot = 0.; //[kg/kg]
    //DLR Correlation
    double c_sootRef_ID = 0.0000026156 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::SN), 4.)
                            - 0.00010998 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::SN), 3.)
                            + 0.0022367 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::SN), 3.)
                            + 0.10955 * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::taxi, EngineEmissions::SN) + 0.0012842; //mg/kg
    double c_sootRef_TO = 0.0000026156 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::SN), 4.)
                            - 0.00010998 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::SN), 3.)
                             + 0.0022367 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::SN), 3.)
                             + 0.10955 * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::takeoff, EngineEmissions::SN) + 0.0012842; //mg/kg
    double c_sootRef_CL = 0.0000026156 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::SN), 4.)
                            - 0.00010998 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::SN), 3.)
                            + 0.0022367 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::SN), 3.)
                            + 0.10955 * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::climb, EngineEmissions::SN) + 0.0012842; //mg/kg
    double c_sootRef_AP = 0.0000026156 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::SN), 4.)
                            - 0.00010998 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::SN), 3.)
                            + 0.0022367 * pow(missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::SN), 3.)
                            + 0.10955 * missionIOData_->myEngine.get_LTO_emission_index(LTOPhases::approach, EngineEmissions::SN) + 0.0012842; //mg/kg
    EI_soot = c_sootRef_ID * St3MassFlow_ID / (fuelFlow_ID / missionIOData_->numberOfEngines) / 1.225 / 1000000.; //kg/kg
    this->LTO.soot = EI_soot * m_fuel_Taxi; //kg
    for (int missionPoint = 1; missionPoint < numberOfSteps ; missionPoint++) {
        double fuelFlow_mom = this->myFlightPath.pathPoints.at(missionPoint).fuelFlow;//[kg/s]
        double T3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature;//[K]
        double P3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3TotalPressure / pow(1. + (1.4 - 1.)
                 / 2.*pow(this->myFlightPath.pathPoints.at(missionPoint).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double ro_air_mom = this->myFlightPath.pathPoints.at(missionPoint).Pamb / this->myFlightPath.pathPoints.at(missionPoint).Tamb / 286.7;//[kg/m^3]
        double AFR = this->myFlightPath.pathPoints.at(missionPoint).St3MassFlow / fuelFlow_mom;
        double T_flame_mom = 2281.*pow(P3_mom / 100., 0.009375) + 0.000178 * pow(P3_mom / 100, 0.055)
                             * (this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature - 298.);//[K]
        double c_sootRef_mom = linearInterpolation((fuelFlow_mom / missionIOData_->numberOfEngines),
                                            (fuelFlow_ID / missionIOData_->numberOfEngines), (fuelFlow_TO / missionIOData_->numberOfEngines),
                                            (fuelFlow_CL / missionIOData_->numberOfEngines),
                                            (fuelFlow_AP / missionIOData_->numberOfEngines), c_sootRef_ID, c_sootRef_TO, c_sootRef_CL, c_sootRef_AP);//[mg/m^3]
        //Set temporary engine rating to SL
        missionIOData_->myEngine.set_operating_point({this->myFlightPath.pathPoints.at(missionPoint).N1, 0, 0});
        //Use manual engine rating for sea-level engine data
        double P3t_SL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::TotalPressure, EngineStage::St3);
        double St3Mach_SL = missionIOData_->myEngine.get_physical_properties_stage(StageProperties::MachNumber, EngineStage::St3);
        double P3_SL = P3t_SL / (1. + 1.4 / 2.*pow(St3Mach_SL, 2));//[kPa]
        double T_flameSL = 2281.*pow(P3_mom / 100., 0.009375) + 0.000178 * pow(P3_mom / 100., 0.055) * (T3_mom - 298.);//[K]
        double c_soot_alt = c_sootRef_mom * pow((P3_mom / P3_SL), 1.35) * exp(-20000 / T_flame_mom) / exp(-20000 / T_flameSL); //mg/kg
        EI_soot = c_soot_alt * AFR / ro_air_mom / 1000000.;//kg/kg
        if (missionPoint == 0) {
            timeStep = missionInfo.at(missionPoint).time;
        } else {
            timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.soot += EI_soot * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
            if (c_soot_alt > c_soot_LTO_max)
                c_soot_LTO_max = c_soot_alt;
        } else {
            this->Cruise.soot += EI_soot * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        }
        emissionFlow[missionPoint].soot = EI_soot * missionInfo[missionPoint].fuelFlow;
    }
}

void Mission::kerosene_soot_whyte() {
    /**< \todo [T2150] (https://unicado.ilr.rwth-aachen.de/T2150): Correct Soot Correlation
     */
    double EI_soot = 0.;//[kg/kg]
    EI_soot = 1.08 * pow(10., 27.) * pow(P3_ID / 1000. / (St3MassFlow_ID / (fuelFlow_ID / missionIOData_->numberOfEngines)
              * 167. / 28.85 * 0.42 / 12.), 2.7) / pow(1.91667, 5.45) / pow(T3_ID, 8.66) / 1000.;
    this->LTO.soot = EI_soot * m_fuel_Taxi; //kg
    for (int missionPoint = 1; missionPoint < numberOfSteps ; missionPoint++) {
        double AFR = this->myFlightPath.pathPoints.at(missionPoint).St3MassFlow / (this->myFlightPath.pathPoints.at(missionPoint).fuelFlow / missionIOData_->numberOfEngines);
        double OCRatio = AFR * 167. / 28.85 * 0.42 / 12.;
        if (OCRatio < 11.)
            OCRatio = 11.;
        //myRuntimeInfo->debug<< "AFR = " << AFR << std::endl;
        //myRuntimeInfo->debug<< "X = " << X << std::endl;
        double P3_mom = this->myFlightPath.pathPoints.at(missionPoint).St3TotalPressure / pow(1. + (1.4 - 1.) / 2.*pow(this->myFlightPath.pathPoints.at(missionPoint).St3Mach, 2.),
                 1.4 / (1.4 - 1.)) / pow(1. + (1.4 - 1.) / 2.*pow(this->myFlightPath.pathPoints.at(missionPoint).St3Mach, 2.), 1.4 / (1.4 - 1.)); //kPa
        double ro_air_mom = this->myFlightPath.pathPoints.at(missionPoint).Pamb / this->myFlightPath.pathPoints.at(missionPoint).Tamb / 286.7;//[kg/m^3]
        //EI_soot = 1.08*pow(10.,29.)*pow(P3_mom/1000./X,2.7) / pow(1.91667,5.45) / pow(this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature,8.66)/1000.;
        EI_soot = 1.08 * pow(10., 27.) * pow(P3_mom / 1000. / OCRatio, 2.7) / pow(1.91667, 5.45)
                 / pow(this->myFlightPath.pathPoints.at(missionPoint).St3StaticTemperature, 8.66) / 1000.; //kg/kg
        double c_soot_alt = EI_soot * ro_air_mom / AFR * 1000000.;//[mg/m^3]
        //myRuntimeInfo->debug<< "EI_soot = " << EI_soot << std::endl;
        if (missionPoint == 0) {
            timeStep = missionInfo.at(missionPoint).time;
        } else {
            timeStep = missionInfo.at(missionPoint).time - missionInfo.at(missionPoint - 1).time;
        }
        if (this->myFlightPath.pathPoints.at(missionPoint).altitude <= 914.4) {
            this->LTO.soot += EI_soot * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
            if (c_soot_alt > c_soot_LTO_max)
                c_soot_LTO_max = c_soot_alt;
        } else {
            this->Cruise.soot += EI_soot * this->myFlightPath.pathPoints.at(missionPoint).fuelFlow * timeStep; //kg
        }
        emissionFlow[missionPoint].soot = EI_soot * missionInfo[missionPoint].fuelFlow;
    }
}
