/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "../../standard_strategy.h"
#include "../../../toolinfo.h"
#include <vector>
#include <unitConversion/unitConversion.h>

void StandardStrategy::setHtmlBody_mission(const std::shared_ptr<RuntimeIO> &rtIO) {// cppcheck-suppress unusedFunction
       double totalCO2(0.);
       double totalHC(0.);
       double totalNOx(0.);
       double totalSO2(0.);
       double totalCO(0.);
       double totalH2O(0.);
       double totalPM10(0.);

       /* Table of operating scenario */
       report_->htmlReportStream()
              << html::h2("Operating Scenario")
              << "<br>" << "\n"
              << html::table_begin({"Parameter", "Comment", "Unit", "Value"})
              << html::table_row("Aircraft name", "", "-", std::string(rtIO->aircraft_type()+"-"+rtIO->aircraft_model()))
              << html::table_row("Range", "", "NM", Rounding(theStudyMission.rangeMission, 2))
              << html::table_row("Blockfuel", "<font size=\"-1\">Fuel from break release to end of taxiing after landing</font>", "kg", Rounding(theStudyMission.m_fuel_LTO +
                     theStudyMission.m_fuel_Cruise, 2))
              << html::table_row("LTO time", "<font size=\"-1\">Time for all mission segments below 3000 ft</font>", "min", Rounding(theStudyMission.LTOtime, 2))
              << html::table_row("Taxi time", "", "min", Rounding(theStudyMission.taxiTime / 60., 2))
              << html::table_row("Takeoff time", "<font size=\"-1\">from 35 ft to 3000 ft</font>", "min", Rounding(theStudyMission.takeOffTime / 60., 2))
              << html::table_row("Climb out time", "<font size=\"-1\">from 35 ft to 3000 ft</font>", "min", Rounding(theStudyMission.climbOutTime / 60., 2))
              << html::table_row("Approach time", "<font size=\"-1\">from 3000 ft to 0 ft</font>", "min", Rounding(theStudyMission.approachTime / 60., 2))
              << html::table_row("Flights per year", "", "-", theStudyMission.missionIOData_->flightsPerYear)
              << html::table_row("Operating years", "", "yr", theStudyMission.missionIOData_->durationOperation)
              << html::table_end;

       report_->htmlReportStream() << html::h2("Mission based emissions")
                                   << html::div_start("style=\"display: flex; justify-content: space-between;\"");

       /* Table of study mission emissions */
       totalCO2 = theStudyMission.LTO.CO2 + theStudyMission.Cruise.CO2;
       totalHC = theStudyMission.LTO.HC + theStudyMission.Cruise.HC;
       totalNOx = theStudyMission.LTO.NOx + theStudyMission.Cruise.NOx;
       totalSO2 = theStudyMission.LTO.SO2 + theStudyMission.Cruise.SO2;
       totalCO = theStudyMission.LTO.CO + theStudyMission.Cruise.CO;
       totalH2O = theStudyMission.LTO.H2O + theStudyMission.Cruise.H2O;
       totalPM10 = theStudyMission.LTO.soot + theStudyMission.Cruise.soot;
       report_->htmlReportStream()
              << html::div_start("style=\"margin: 0 10px; width: 48%;\"")
              << html::table_begin({{"Parameter"}, {"LTO", "colspan=\"2\" style=\"text-align: center;\""}, {"Cruise", "colspan=\"2\" style=\"text-align: center;\""}},
                                           "Emissions of one study mission")
              << html::table_row("", "[kg]", "[%]", "[kg]", "[%]")
              << html::table_row("CO2",
                     Rounding(theStudyMission.LTO.CO2, 3),
                     Rounding(theStudyMission.LTO.CO2 / totalCO2 * 100, 2),
                     Rounding(theStudyMission.Cruise.CO2, 3),
                     Rounding(theStudyMission.Cruise.CO2 / totalCO2 * 100, 2))
              << html::table_row("H2O",
                     Rounding(theStudyMission.LTO.H2O, 3),
                     Rounding(theStudyMission.LTO.H2O / totalH2O * 100, 2),
                     Rounding(theStudyMission.Cruise.H2O, 3),
                     Rounding(theStudyMission.Cruise.H2O / totalH2O * 100, 2))
              << html::table_row("SO2",
                     Rounding(theStudyMission.LTO.SO2, 3),
                     Rounding(theStudyMission.LTO.SO2 / totalSO2 * 100, 2),
                     Rounding(theStudyMission.Cruise.SO2, 3),
                     Rounding(theStudyMission.Cruise.SO2 / totalSO2 * 100, 2))
              << html::table_row("HC",
                     Rounding(theStudyMission.LTO.HC, 3),
                     Rounding(theStudyMission.LTO.HC / totalHC * 100, 2),
                     Rounding(theStudyMission.Cruise.HC, 3),
                     Rounding(theStudyMission.Cruise.HC / totalHC * 100, 2))
              << html::table_row("CO",
                     Rounding(theStudyMission.LTO.CO, 3),
                     Rounding(theStudyMission.LTO.CO / totalCO * 100, 2),
                     Rounding(theStudyMission.Cruise.CO, 3),
                     Rounding(theStudyMission.Cruise.CO / totalCO * 100, 2))
              << html::table_row("NOx",
                     Rounding(theStudyMission.LTO.NOx, 3),
                     Rounding(theStudyMission.LTO.NOx / totalNOx * 100, 2),
                     Rounding(theStudyMission.Cruise.NOx, 3),
                     Rounding(theStudyMission.Cruise.NOx / totalNOx * 100, 2))
              << html::table_row("PM10",
                     Rounding(theStudyMission.LTO.soot, 3),
                     Rounding(theStudyMission.LTO.soot / totalPM10 * 100, 2),
                     Rounding(theStudyMission.Cruise.soot, 3),
                     Rounding(theStudyMission.Cruise.soot / totalPM10 * 100, 2))
              << html::table_end
              << "<br>";
       if (rtIO->plotOn) {
              report_->htmlReportStream()
                     << html::image(relativePath(rtIO->getPlotDir() + FILESEPERATOR + std::string(TOOL_NAME) + "_mission_study.svg", rtIO->getHtmlDir(), false), "");
       }
       report_->htmlReportStream() << html::div_end;

       /* Table of design mission emissions */
       totalCO2 = theDesignMission.LTO.CO2 + theDesignMission.Cruise.CO2;
       totalHC = theDesignMission.LTO.HC + theDesignMission.Cruise.HC;
       totalNOx = theDesignMission.LTO.NOx + theDesignMission.Cruise.NOx;
       totalSO2 = theDesignMission.LTO.SO2 + theDesignMission.Cruise.SO2;
       totalCO = theDesignMission.LTO.CO + theDesignMission.Cruise.CO;
       totalH2O = theDesignMission.LTO.H2O + theDesignMission.Cruise.H2O;
       totalPM10 = theDesignMission.LTO.soot + theDesignMission.Cruise.soot;
       report_->htmlReportStream()
              << html::div_start("style=\"margin: 0 10px; width: 48%;\"")
              << html::table_begin({{"Parameter"}, {"LTO", "colspan=\"2\" style=\"text-align: center;\""}, {"Cruise", "colspan=\"2\" style=\"text-align: center;\""}},
                                          "Emissions of one design mission")
              << html::table_row("", "[kg]", "[%]", "[kg]", "[%]")
              << html::table_row("CO2",
                     Rounding(theDesignMission.LTO.CO2, 3),
                     Rounding(theDesignMission.LTO.CO2 / totalCO2 * 100, 2),
                     Rounding(theDesignMission.Cruise.CO2, 3),
                     Rounding(theDesignMission.Cruise.CO2 / totalCO2 * 100, 2))
              << html::table_row("H2O",
                     Rounding(theDesignMission.LTO.H2O, 3),
                     Rounding(theDesignMission.LTO.H2O / totalH2O * 100, 2),
                     Rounding(theDesignMission.Cruise.H2O, 3),
                     Rounding(theDesignMission.Cruise.H2O / totalH2O * 100, 2))
              << html::table_row("SO2",
                     Rounding(theDesignMission.LTO.SO2, 3),
                     Rounding(theDesignMission.LTO.SO2 / totalSO2 * 100, 2),
                     Rounding(theDesignMission.Cruise.SO2, 3),
                     Rounding(theDesignMission.Cruise.SO2 / totalSO2 * 100, 2))
              << html::table_row("HC",
                     Rounding(theDesignMission.LTO.HC, 3),
                     Rounding(theDesignMission.LTO.HC / totalHC * 100, 2),
                     Rounding(theDesignMission.Cruise.HC, 3),
                     Rounding(theDesignMission.Cruise.HC / totalHC * 100, 2))
              << html::table_row("CO",
                     Rounding(theDesignMission.LTO.CO, 3),
                     Rounding(theDesignMission.LTO.CO / totalCO * 100, 2),
                     Rounding(theDesignMission.Cruise.CO, 3),
                     Rounding(theDesignMission.Cruise.CO / totalCO * 100, 2))
              << html::table_row("NOx",
                     Rounding(theDesignMission.LTO.NOx, 3),
                     Rounding(theDesignMission.LTO.NOx / totalNOx * 100, 2),
                     Rounding(theDesignMission.Cruise.NOx, 3),
                     Rounding(theDesignMission.Cruise.NOx / totalNOx * 100, 2))
              << html::table_row("PM10",
                     Rounding(theDesignMission.LTO.soot, 3),
                     Rounding(theDesignMission.LTO.soot / totalPM10 * 100, 2),
                     Rounding(theDesignMission.Cruise.soot, 3),
                     Rounding(theDesignMission.Cruise.soot / totalPM10 * 100, 2))
              << html::table_end
              << "<br>"
              << "<br>";
              if (rtIO->plotOn) {
                     report_->htmlReportStream()
                            << html::image(relativePath(rtIO->getPlotDir() + FILESEPERATOR + std::string(TOOL_NAME) + "_mission_design.svg", rtIO->getHtmlDir(), false), "");
              }
       report_->htmlReportStream()
              << html::div_end
              << html::div_end;
}

void StandardStrategy::setTexBody_mission() {
    double totalCO2 = theStudyMission.LTO.CO2 + theStudyMission.Cruise.CO2;
    double totalHC = theStudyMission.LTO.HC + theStudyMission.Cruise.HC;
    double totalNOx = theStudyMission.LTO.NOx + theStudyMission.Cruise.NOx;
    double totalSO2 = theStudyMission.LTO.SO2 + theStudyMission.Cruise.SO2;
    double totalCO = theStudyMission.LTO.CO + theStudyMission.Cruise.CO;
    double totalH2O = theStudyMission.LTO.H2O + theStudyMission.Cruise.H2O;
    double totalPM10 = theStudyMission.LTO.soot + theStudyMission.Cruise.soot;
    report_->texReportStream() << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << std::endl;
    report_->texReportStream() << "\\subsection{Emissions}" << std::endl;
    report_->texReportStream() << "\\begin{frame}" << std::endl;
    report_->texReportStream() << " \\frametitle{Emissions}" << std::endl;
    report_->texReportStream() << "     \\begin{table}" << std::endl;
    report_->texReportStream() << "     \\centering" << std::endl;
    report_->texReportStream() << "     \\scriptsize" << std::endl;
    report_->texReportStream() << "     \\begin{tabularx}{250pt}{Xllr}" << std::endl;
    report_->texReportStream() << "     \\toprule" << std::endl;
    report_->texReportStream() << "         \\textcolor{ilrdark}{Parameter} & \\textcolor{ilrdark}{Comment} & \\textcolor{ilrdark}{Unit} &"
                                                 << "\\textcolor{ilrdark}{Value} \\\\" << std::endl;
    report_->texReportStream() << "         \\midrule[0.5pt]" << std::endl;
    report_->texReportStream() << "         \\multicolumn{4}{l}{\\textbf{Operating Scenario}}\\\\   " << std::endl;
    report_->texReportStream() << "         \\midrule[0.5pt]" << std::endl;
    report_->texReportStream() << "             Range & & $\\nauticalmiles$ & " << Rounding(theStudyMission.rangeMission, 2) << "\\\\" << std::endl;
    report_->texReportStream() << "             Block Fuel &  & $\\kilogram$ & " << Rounding(theStudyMission.m_fuel_LTO
                                                                         + theStudyMission.m_fuel_Cruise, 2) << "\\\\" << std::endl;
    report_->texReportStream() << "             LTO Time & \\tiny{all mission segments below 3000 $\\feet$} & $\\min$ & " << Rounding(theStudyMission.LTOtime, 2)
           << " \\\\" << std::endl;
    report_->texReportStream() << "             Taxi Time &  & $\\min$ & " << Rounding(theStudyMission.taxiTime / 60., 2) << " \\\\" << std::endl;
    report_->texReportStream() << "             Takeoff Time & \\tiny{to 35 $\\feet$} & $\\min$ & " << Rounding(theStudyMission.takeOffTime / 60., 2) << " \\\\" << std::endl;
    report_->texReportStream() << "             Climb Out Time & \\tiny{from 35 $\\feet$ to 3000 $\\feet$} & $\\min$ & " << Rounding(theStudyMission.climbOutTime / 60., 2)
           << " \\\\" << std::endl;
    report_->texReportStream() << "             Approach Time & \\tiny{from 3000 $\\feet$ to 0 $\\feet$} & $\\min$ & " << Rounding(theStudyMission.approachTime / 60., 2)
           << " \\\\" << std::endl;
    report_->texReportStream() << "             Flights Per Year &  & $-$ & " << theStudyMission.missionIOData_->flightsPerYear << "\\\\" << std::endl;
    report_->texReportStream() << "             Operating Years &  & $\\yr$ & " << theStudyMission.missionIOData_->durationOperation << "\\\\" << std::endl;
    report_->texReportStream() << "     \\bottomrule" << std::endl;
    report_->texReportStream() << "     \\end{tabularx}" << std::endl;
    report_->texReportStream() << "     \\end{table}" << std::endl;
    report_->texReportStream() << "     \\begin{table}" << std::endl;
    report_->texReportStream() << "     \\centering" << std::endl;
    report_->texReportStream() << "     \\scriptsize" << std::endl;
    report_->texReportStream() << "     \\begin{tabularx}{250pt}{Xrrrr}" << std::endl;
    report_->texReportStream() << "     \\toprule" << std::endl;
    report_->texReportStream() << "         \\textcolor{ilrdark}{Parameter} & \\multicolumn{2}{c}{\\textcolor{ilrdark}{~~~~~~~~~LTO}}  "
              "& \\multicolumn{2}{c}{\\textcolor{ilrdark}{~~~~~~~~~Cruise}} \\\\" << std::endl;
    report_->texReportStream() << "        & \\textcolor{ilrdark}{$\\kilogram$} & \\textcolor{ilrdark}{\\% } & \\textcolor{ilrdark}{$\\kilogram$} & "
                                                 << "\\textcolor{ilrdark}{\\% }\\\\" << std::endl;
    report_->texReportStream() << "         \\midrule[0.5pt]" << std::endl;
    report_->texReportStream() << "              $CO_2$ & " << Rounding(theStudyMission.LTO.CO2, 3) << "&"
           << Rounding(theStudyMission.LTO.CO2 / totalCO2 * 100, 2);
    report_->texReportStream() <<                  " & " << Rounding(theStudyMission.Cruise.CO2, 3) << "&"
           << Rounding(theStudyMission.Cruise.CO2 / totalCO2 * 100, 2) << " \\\\" << std::endl;
    report_->texReportStream() << "              $H_2O$ & ";
    report_->texReportStream() << Rounding(theStudyMission.LTO.H2O, 3) << "&" << Rounding(theStudyMission.LTO.H2O / totalH2O * 100, 2);
    report_->texReportStream() << " & " << Rounding(theStudyMission.Cruise.H2O, 3) << "&" << Rounding(theStudyMission.Cruise.H2O / totalH2O * 100, 2)
           << " \\\\" << std::endl;
    report_->texReportStream() << "              $SO_2$ & ";
    report_->texReportStream() << Rounding(theStudyMission.LTO.SO2, 3) << "&" << Rounding(theStudyMission.LTO.SO2 / totalSO2 * 100, 2);
    report_->texReportStream() << " & " << Rounding(theStudyMission.Cruise.SO2, 3) << "&" << Rounding(theStudyMission.Cruise.SO2 / totalSO2 * 100, 2)
           << " \\\\" << std::endl;
    report_->texReportStream() << "              $HC$ & ";
    report_->texReportStream() << Rounding(theStudyMission.LTO.HC, 3) << "&" << Rounding(theStudyMission.LTO.HC / totalHC * 100, 2);
    report_->texReportStream() << " & " << Rounding(theStudyMission.Cruise.HC, 3) << "&" << Rounding(theStudyMission.Cruise.HC / totalHC * 100, 2)
           << " \\\\" << std::endl;
    report_->texReportStream() << "              $CO$ & ";
    report_->texReportStream() << Rounding(theStudyMission.LTO.CO, 3) << "&" << Rounding(theStudyMission.LTO.CO / totalCO * 100, 2);
    report_->texReportStream() << " & " << Rounding(theStudyMission.Cruise.CO, 3) << "&" << Rounding(theStudyMission.Cruise.CO / totalCO * 100, 2)
           << " \\\\" << std::endl;
    report_->texReportStream() << "              $NO_x$ & ";
    report_->texReportStream() << Rounding(theStudyMission.LTO.NOx, 3) << "&" << Rounding(theStudyMission.LTO.NOx / totalNOx * 100, 2);
    report_->texReportStream() << " & " << Rounding(theStudyMission.Cruise.NOx, 3) << "&" << Rounding(theStudyMission.Cruise.NOx / totalNOx * 100, 2)
           << " \\\\" << std::endl;
    report_->texReportStream() << "              $PM_{10}$ & ";
    report_->texReportStream() << Rounding(theStudyMission.LTO.soot, 3) << "&" << Rounding(theStudyMission.LTO.soot / totalPM10 * 100, 2);
    report_->texReportStream() << " & " << Rounding(theStudyMission.Cruise.soot, 3) << "&" << Rounding(theStudyMission.Cruise.soot / totalPM10 * 100,
            2) << " \\\\" << std::endl;
    report_->texReportStream() << "     \\bottomrule" << std::endl;
    report_->texReportStream() << "     \\end{tabularx}" << std::endl;
    report_->texReportStream() << "     \\end{table}" << std::endl;
    report_->texReportStream() << "\\end{frame}" << std::endl;
}


