/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <iostream>
#include <memory>
#include <algorithm>
#include <string>
#include <unitConversion/unitConversion.h>
#include <moduleBasics/runtimeIO.h>
#include <aixml/node.h>
#include "dallara.h"


void ClimateModelDallara::run(const std::shared_ptr<RuntimeIO> &rtIO, std::shared_ptr<node> ecoAssXML) {
    myRuntimeInfo->out << "  Calculation of climate impact" << std::endl;
    dallaraIOData_ = std::make_shared<DallaraIOData>(rtIO->acxml, rtIO->moduleConfig, ecoAssXML);
    calculateClimateImpact_dallara(rtIO->aircraft_energy_carrier_type());
}

void ClimateModelDallara::calculateClimateImpact_dallara(const std::string &fuelType) {
    /* ATR calculation s. Gl. (20) Reference Dallara*/
    /* Fill std::vector s_i_quer:*/
    double s_i_cruise[8]{0., 0., 0., 0., 0., 0., 0., 0.};//Cruise average forcing factors [-]
    double s_i_climb[8]{0., 0., 0., 0., 0., 0., 0., 0.};//Climb average forcing factors [-]
    double s_i_descent[8]{0., 0., 0., 0., 0., 0., 0., 0.};//Descent average forcing factors [-]
    double s_i_quer[8]{0., 0., 0., 0., 0., 0., 0., 0.};// Mission average forcing factors [-] \cite Dal11 page 1607 equation(18) or modified version \cite Ste13 is used.
    for (int i = 0; i < 8; i++) {
        s_i_cruise[i] = 0.;
        s_i_climb[i] = 0.;
        s_i_descent[i] = 0.;
        s_i_quer[i] = 0.;
        /* s_i_cruise (summation of weighted cruise steps)*/
        double relStepLength;
        for (unsigned int j = 0; j < dallaraIOData_->flightLevel.size(); j++) {
            if (j == 0) {
                relStepLength = dallaraIOData_->endOfCruiseStep.at(j);
            } else {
                relStepLength = dallaraIOData_->endOfCruiseStep.at(j) - dallaraIOData_->endOfCruiseStep.at(j-1);
            }
            s_i_cruise[i] = s_i_cruise[i] + (relStepLength * get_s_i_height(i, dallaraIOData_->flightLevel.at(j) * 100.));
        }
        myRuntimeInfo->info << "s_i_cruise[i] FINAL: " << s_i_cruise[i] << std::endl;
        /* s_i_climb und s_i_descent (integrate)*/
        uint16_t stepSize(100);
        double deltaHclimb = (dallaraIOData_->flightLevel.front() - 0.) * 100. / stepSize;
        double deltaHdescent = (dallaraIOData_->flightLevel.back() - 0) * 100. / stepSize;
        double HclimbTmp = 0.;
        double HdescentTmp = dallaraIOData_->flightLevel.back() * 100.;
        for (uint16_t j = 0; j <= stepSize; j++) {
            // trapezoidal rule
            if (j == 0 || j == stepSize) {
                s_i_climb[i] += 0.5 * get_s_i_height(i, HclimbTmp) * deltaHclimb;
                s_i_descent[i] += 0.5 * get_s_i_height(i, HdescentTmp) * deltaHdescent;
                myRuntimeInfo->info << "get_s_i_height(i,HdescentTmp): " << get_s_i_height(i, HdescentTmp) << std::endl;
                myRuntimeInfo->info << "HdescentTmp: " << HdescentTmp << std::endl;
                myRuntimeInfo->info << "s_i_descent[i]: " << s_i_descent[i] << std::endl;
            } else {
                s_i_climb[i] += get_s_i_height(i, HclimbTmp) * deltaHclimb;
                s_i_descent[i] += get_s_i_height(i, HdescentTmp) * deltaHdescent;
                myRuntimeInfo->info << "get_s_i_height(i,HdescentTmp): " << get_s_i_height(i, HdescentTmp) << std::endl;
                myRuntimeInfo->info << "HdescentTmp: " << HdescentTmp << std::endl;
                myRuntimeInfo->info << "s_i_descent[i]: " << s_i_descent[i] << std::endl;
            }
            HclimbTmp += deltaHclimb;
            HdescentTmp -= deltaHdescent;
        }
        s_i_climb[i] = s_i_climb[i] / ((dallaraIOData_->flightLevel.front() - 0.) * 100.);
        s_i_descent[i] = s_i_descent[i] / ((dallaraIOData_->flightLevel.back() - 0.) * 100.);
        myRuntimeInfo->info << "s_i_climb[i] FINAL: " << s_i_climb[i] << std::endl;
        myRuntimeInfo->info << "s_i_descent[i] FINAL: " << s_i_descent[i] << std::endl;
        double climbRatio = dallaraIOData_->topOfClimbRange / dallaraIOData_->rangeStudy;
        double cruiseRatio = (dallaraIOData_->topOfDescentRange - dallaraIOData_->topOfClimbRange) / dallaraIOData_->rangeStudy;
        double descentRatio = (dallaraIOData_->rangeStudy - dallaraIOData_->topOfDescentRange) / dallaraIOData_->rangeStudy;
        s_i_quer[i] = cruiseRatio*s_i_cruise[i] + climbRatio*s_i_climb[i] + descentRatio*s_i_descent[i];
        myRuntimeInfo->info << "s_i_quer[i] FINAL: " << s_i_quer[i] << std::endl;
    }
    for (int i = 0; i < 8; i++) {
        int sigma;//s. Dallara eq. (17)
        /* Calculate RF und RF_modified (s. boundary conditions eq.(19))*/
        std::vector<double> RadForcingTmp;
        std::vector<double> normRadForcingTmp;
        if (i == 0 || i == 1 || i == 2) { //RF to be calculated via time integral at CO2, CH4 and O3L
            for (uint16_t t = 0; t <= dallaraIOData_->configuration.tmax; t++) {
                RadForcingTmp.push_back(0.);//Create std::vector entry, then calculate
                normRadForcingTmp.push_back(0.);
                for (int tau = 0; tau <= t; tau++) {
                    if (tau < dallaraIOData_->durationOperation)
                        sigma = 1;
                    else
                        sigma = 0;
                    if (tau == 0 || tau == t) { //stepSize is 1 year!
                        RadForcingTmp.back() += 0.5 * get_response_function(i, (t - tau), fuelType) * sigma;
                        normRadForcingTmp.back() += 0.5 * f_i[i] *  get_response_function(i, (t - tau), fuelType) * sigma / RF_2CO2;
                    } else {
                        RadForcingTmp.back() += get_response_function(i, (t - tau), fuelType) * sigma;
                        normRadForcingTmp.back() += f_i[i] *  get_response_function(i, (t - tau), fuelType) * sigma / RF_2CO2;
                    }
                }
                myRuntimeInfo->info << "i: " << i << ", RF: " << RadForcingTmp.back() << std::endl;
            }
        } else {
            for (uint16_t t = 0; t <= dallaraIOData_->configuration.tmax; t++) {
                RadForcingTmp.push_back(0.);//Create std::vector entry, then calculate
                normRadForcingTmp.push_back(0.);
                if (t < dallaraIOData_->durationOperation)
                    sigma = 1;
                else
                    sigma = 0;
                RadForcingTmp.back() += get_response_function(i, (t), fuelType) * sigma;
                normRadForcingTmp.back() += f_i[i] *  get_response_function(i, (t), fuelType) * sigma / RF_2CO2;
                myRuntimeInfo->info << "i: " << i << ", RF: " << RadForcingTmp.back() << std::endl;
            }
        }
        RadForcing.push_back(RadForcingTmp); //calculate normalized RadForcing
        normRadForcing.push_back(normRadForcingTmp);
        /*calculate delta T  (s. Dallara eq.(14))*/
        std::vector<double> TempChangeTmp;
        std::vector<double> AGWPTmp;
        for (uint16_t t = 0; t <= dallaraIOData_->configuration.tmax; t++) {
            TempChangeTmp.push_back(0.);//Create std::vector entry, then calculate
            AGWPTmp.push_back(0.);
            for (int tau = 0; tau <= t; tau++) {
                double Gt = S * (alphaT / tauT1 * exp(-(t - tau) / tauT1) + (1 - alphaT) / tauT2 * exp(-(t - tau) / tauT2));//eq. (15) but with boundary conditions s. eq. (19)+(20)
                if (tau == 0 || tau == t) { //stepSize is 1 year!
                    TempChangeTmp.back() += 0.5 * Gt * f_i[i] * RadForcing.at(i).at(tau) / RF_2CO2; //do not summate here, since calculated via uATR (s.. eq (20))
                    AGWPTmp.back() += 0.5 * RadForcing.at(i).at(tau); // for die AGWP calculation
                } else {
                    TempChangeTmp.back() += Gt * f_i[i] * RadForcing.at(i).at(tau) / RF_2CO2;
                    AGWPTmp.back() += RadForcing.at(i).at(tau);
                }
            }
            myRuntimeInfo->info << "i: " << i << ", deltaT: " << TempChangeTmp.back() << std::endl;
            myRuntimeInfo->info << "i: " << i << ", AGWP: " << AGWPTmp.back() << std::endl;
        }
        TempChange.push_back(TempChangeTmp);
        AGWP.push_back(AGWPTmp);
        /* calculate uATR , (s.Dallara ql (19) and (1)*/
        double weightFunc;// s. Dallara eq. (2)
        double uATRTmp = 0.;
        std::vector<double> absTempChangeTmp;
        std::vector<double> absTempChangeweightedTmp;
        std::vector<double> TotalTempChangeTmp;
        std::vector<double> absRadForcingTmp;
        std::vector<double> absnormRadForcingTmp;
        std::vector<double> absAGWPTmp;
        for (uint16_t t = 0; t <= dallaraIOData_->configuration.tmax; t++) {
            if (t < dallaraIOData_->durationOperation)
                weightFunc = 1;
            else
                weightFunc = 1.0 / (pow(1 + dallaraIOData_->configuration.devRate, (t - dallaraIOData_->durationOperation)));
            if (t == 0 || t == dallaraIOData_->configuration.tmax) {
                uATRTmp += 0.5 * TempChange.at(i).at(t) * weightFunc;
            } else {
                uATRTmp += TempChange.at(i).at(t) * weightFunc;
            }
            absTempChangeTmp.push_back(0.);
            absTempChangeTmp.back() = TempChange.at(i).at(t)
                                      * dallaraIOData_->flightsPerYear * s_i_quer[i] * dallaraIOData_->emissionMass[i]; //TempChange (total annual emission quantity) per emission
            absTempChangeweightedTmp.push_back(0.);
            absTempChangeweightedTmp.back() = TempChange.at(i).at(t) * weightFunc * dallaraIOData_->flightsPerYear * s_i_quer[i]
                * dallaraIOData_->emissionMass[i]; //TempChange (total annual emission quantity) per emission
            absRadForcingTmp.push_back(0.);
            absRadForcingTmp.back() = RadForcing.at(i).at(t) * dallaraIOData_->flightsPerYear * s_i_quer[i] * dallaraIOData_->emissionMass[i];
            absnormRadForcingTmp.push_back(0.);
            absnormRadForcingTmp.back() = normRadForcing.at(i).at(t) * dallaraIOData_->flightsPerYear * s_i_quer[i] * dallaraIOData_->emissionMass[i];
            absAGWPTmp.push_back(0.);
            absAGWPTmp.back() = AGWP.at(i).at(t) * dallaraIOData_->flightsPerYear * s_i_quer[i] * dallaraIOData_->emissionMass[i];
            TotalTempChangeTmp.push_back(0.);
            if (t == 0) {
                TotalTempChangeTmp.back() = TempChange.at(i).at(t) * weightFunc
                                            * dallaraIOData_->flightsPerYear * s_i_quer[i] * dallaraIOData_->emissionMass[i]; //TotalTempChange, Added tempChange per emission
            } else {
                TotalTempChangeTmp.back() = TotalTempChangeTmp.at(t - 1) + TempChange.at(i).at(t) * weightFunc * dallaraIOData_->flightsPerYear
                    * s_i_quer[i] * dallaraIOData_->emissionMass[i]; //TotalTempChange, Added tempChange per emission
            }
        }
        uATR[i] = uATRTmp / dallaraIOData_->durationOperation;
        myRuntimeInfo->info << "i: " << i << ", uATR: " << uATR[i] << std::endl;
        myRuntimeInfo->info << "i: " << i << ", s_quer: " << s_i_quer[i] << std::endl;
        myRuntimeInfo->info << "i: " << i << ", dallaraIOData_->emissionMass: " << dallaraIOData_->emissionMass[i] << std::endl;
        /* ATR per year for all emissions (s. Dallara eq (20))*/
        dallaraIOData_->ATR += dallaraIOData_->flightsPerYear * s_i_quer[i] * dallaraIOData_->emissionMass[i] * uATR[i];
        myRuntimeInfo->info << "ATRtmp: " << dallaraIOData_->ATR << std::endl;
        /* ATR for different components*/
        ATRyear_i.push_back(0.);
        ATRyear_i.back() = dallaraIOData_->flightsPerYear * s_i_quer[i] * dallaraIOData_->emissionMass[i] * uATR[i];
        myRuntimeInfo->info << "ATR_i: " << ATRyear_i.at(i) << std::endl;
        /* annual emissions for different components*/
        e_year_i.push_back(0.);
        e_year_i.back() = dallaraIOData_->flightsPerYear * dallaraIOData_->emissionMass[i];
        absTempChange.push_back(absTempChangeTmp);
        absTempChangeweighted.push_back(absTempChangeweightedTmp);
        TotalTempChange.push_back(TotalTempChangeTmp);
        absRadForcing.push_back(absRadForcingTmp);
        absnormRadForcing.push_back(absnormRadForcingTmp);
        absAGWP.push_back(absAGWPTmp);
    }
    myRuntimeInfo->info << "ATR pro Jahr: " << dallaraIOData_->ATR << std::endl;
    for (uint16_t t = 0; t <= dallaraIOData_->configuration.tmax; t++) { //Summing up the component values
        absTempChangeTotal.push_back(0.);
        TotalTempChangeTotal.push_back(0.);
        absTempChangeweightedTotal.push_back(0.);
        absTempChangeNoxCooling.push_back(0.);
        absTempChangeH20SO4Soot.push_back(0.);
        absRadForcingTotal.push_back(0.);
        absnormRadForcingTotal.push_back(0.);
        absAGWPTotal.push_back(0.);
        for (unsigned int i = 0; i < 8; i++) {
            absRadForcingTotal.back() += absRadForcing.at(i).at(t);
            absnormRadForcingTotal.back() += absnormRadForcing.at(i).at(t);
            absAGWPTotal.back() += absAGWP.at(i).at(t);
            absTempChangeTotal.back() += absTempChange.at(i).at(t);
            TotalTempChangeTotal.back() += TotalTempChange.at(i).at(t);
            absTempChangeweightedTotal.back() += absTempChangeweighted.at(i).at(t);
            if (i == 1 || i == 2) { //NOX Cooling (CH4 and O3L)
                absTempChangeNoxCooling.back() += absTempChange.at(i).at(t);
            }
            if (i == 4 || i == 5 || i == 6) { //NOX Cooling (CH4 and O3L)
                absTempChangeH20SO4Soot.back() += absTempChange.at(i).at(t);
            }
        }
    }
}

double ClimateModelDallara::get_s_i_height(int emissionNo, double flightHeight) {
    myRuntimeInfo->info << "emissionNo: " << emissionNo << " flightHeight: " << flightHeight << std::endl;
    double forcingFac = 0.;
    uint16_t i_max = 0; // index of last row in dataset
    double s_i_height[68][5]; // size of largest dataset
    if (dallaraIOData_->configuration.s_i_dataSet == "mode_0") {
        std::copy(&s_i_height_Dall[0][0], &s_i_height_Dall[0][0]+13*5, &s_i_height[0][0]); // copy dallara dataset [Dal11] to s_i_height
        i_max = 12;
    } else if (dallaraIOData_->configuration.s_i_dataSet == "mode_1") {
        std::copy(&s_i_height_Dahl[0][0], &s_i_height_Dahl[0][0]+68*5, &s_i_height[0][0]); // copy dahlmann dataset [Kos22] to s_i_height
        i_max = 67;
    } else {
        myRuntimeInfo->err << "Forcing factor data set doesn't exist. Check s_i_dataSet in config." << std::endl;
        exit(1);
    }
    for (uint16_t i = 0; i <= i_max; i++) { // for all altitudes
        if (i < i_max && s_i_height[i][0] < flightHeight) {
        } else {
            if (emissionNo == 1 || emissionNo == 2) { //CH4 and O3L
                if (i == 0)
                    forcingFac = s_i_height[i][2] * dallaraIOData_->configuration.s_i_heightVariation_CH4O3L;
                else if (i == i_max)
                    forcingFac = s_i_height[i][2] * dallaraIOData_->configuration.s_i_heightVariation_CH4O3L;
                else
                    forcingFac = (s_i_height[i - 1][2] + (s_i_height[i][2] - s_i_height[i - 1][2]) / (s_i_height[i][0] - s_i_height[i - 1][0])
                                  * (flightHeight - s_i_height[i - 1][0])) * dallaraIOData_->configuration.s_i_heightVariation_CH4O3L;
                myRuntimeInfo->info << "i: " << i << " forcingFac: " << forcingFac << std::endl;
            } else if (emissionNo == 3) { //O3S
                if (i == 0)
                    forcingFac = s_i_height[i][3] * dallaraIOData_->configuration.s_i_heightVariation_O3S;
                else if (i == i_max)
                    forcingFac = s_i_height[i][3] * dallaraIOData_->configuration.s_i_heightVariation_O3S;
                else
                    forcingFac = (s_i_height[i - 1][3] + (s_i_height[i][3] - s_i_height[i - 1][3]) / (s_i_height[i][0] - s_i_height[i - 1][0])
                                  * (flightHeight - s_i_height[i - 1][0])) *  dallaraIOData_->configuration.s_i_heightVariation_O3S;
                myRuntimeInfo->info << "i: " << i << " forcingFac: " << forcingFac << std::endl;
            } else if (emissionNo == 4) { //H2O
                if (i == 0)
                    forcingFac = s_i_height[i][4];
                else if (i == i_max)
                    forcingFac = s_i_height[i][4];
                else
                    forcingFac = (s_i_height[i - 1][4] + (s_i_height[i][4] - s_i_height[i - 1][4]) / (s_i_height[i][0] - s_i_height[i - 1][0])
                                  * (flightHeight - s_i_height[i - 1][0]));
                myRuntimeInfo->info << "i: " << i << " forcingFac: " << forcingFac << std::endl;
            } else if (emissionNo == 7) { //AIC
                if (i == 0)
                    forcingFac = s_i_height[i][1] * dallaraIOData_->configuration.s_i_heightVariation_AIC;
                else if (i == i_max)
                    forcingFac = s_i_height[i][1] * dallaraIOData_->configuration.s_i_heightVariation_AIC;
                else
                    forcingFac = (s_i_height[i - 1][1] + (s_i_height[i][1] - s_i_height[i - 1][1]) / (s_i_height[i][0] - s_i_height[i - 1][0])
                                  * (flightHeight - s_i_height[i - 1][0])) * dallaraIOData_->configuration.s_i_heightVariation_AIC;
                myRuntimeInfo->info << "i: " << i << " forcingFac: " << forcingFac << std::endl;
            } else {
                forcingFac = 1.;
                myRuntimeInfo->info << "i: " << i << " forcingFac: " << forcingFac << std::endl;
            }
            break;
        }
    }
    return forcingFac;
}

double ClimateModelDallara::get_response_function(int emissionNo, double time, const std::string &fuelType) {
    if (emissionNo == 0) { //CO2, eq. (6)
        return A_CO2 * (1 + (alphaC1 * (expm1(-time / tauC1)) + alphaC2 * (expm1(-time / tauC2)) + alphaC3 * (expm1(-time / tauC3))));
    } else if (emissionNo == 1) { //CH4, eq. (7)
        return A_CH4 * exp(-time / tauN);
    } else if (emissionNo == 2) { //O3L, eq. (7)
        return A_O3L * exp(-time / tauN);
    } else if (emissionNo == 3) { //O3S, eq. (10)
        return RFrefEref_O3S;
    } else if (emissionNo == 4) { //H2O, eq. (10)
        return RFrefEref_H2O;
    } else if (emissionNo == 5) { //soot, eq. (10)
        return RFrefEref_soot;
    } else if (emissionNo == 6) { //SO4, eq. (10)
        return RFrefEref_SO4;
    } else if (emissionNo == 7) { //AIC, eq. (10)
        double AICfactor(1.);
        if (fuelType == "liquid_hydrogen")
            AICfactor = dallaraIOData_->configuration.fuelAICfactor;
        return AICfactor * RFrefLref_AIC;
    } else {
        myRuntimeInfo->err << "program error!" << std::endl;
        exit(1);
    }
}
