/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <string>
#include "../../standard_strategy.h"
#include "../../../toolinfo.h"
#include <moduleBasics/plot.h>

void StandardStrategy::generatePlots_dallara() {
    std::string pltDataFileName(rtIO_->getCsvFilesDir() + "/ecological_assessment_climate_impact.csv");
    this->generatePlotData_dallara(pltDataFileName);
    auto fig = matplot::figure(true);
    fig->size(1200, 1600);
    /* Define x axis */
    std::vector<uint16_t> time;
    for (uint16_t i = 0; i <= theClimateModelDallara.dallaraIOData_->configuration.tmax; ++i) {
        time.push_back(i);
    }
    /* Read csv data */
    std::vector<double> emissions_CO2;
    std::vector<double> emissions_NOx;
    std::vector<double> emissions_H2O;
    std::vector<double> emissions_soot;
    std::vector<double> emissions_SO4;
    std::vector<double> ATR_CO2;
    std::vector<double> ATR_CH4;
    std::vector<double> ATR_O3L;
    std::vector<double> ATR_O3S;
    std::vector<double> ATR_H2O;
    std::vector<double> ATR_soot;
    std::vector<double> ATR_SO4;
    std::vector<double> ATR_AIC;
    std::vector<double> ATR_total;
    std::ifstream plotDataFile(pltDataFileName);
    if (!plotDataFile) {
        myRuntimeInfo->err << "File " << pltDataFileName << " can not be opened!" << std::endl;
    }
    std::vector<std::string> lineVector;
    std::string line;
    std::vector<std::string> lineData;
    int row = 0;
    while (std::getline(plotDataFile, line)) {
        lineVector.push_back(line);
        split(lineVector.back(), ';', &lineData);
        if (row > 1) {
            for (unsigned int column= 0; column < lineData.size(); column++) {
                if (column == 64) {
                        emissions_CO2.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 65) {
                        emissions_NOx.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 66) {
                        emissions_H2O.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 67) {
                        emissions_soot.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 68) {
                        emissions_SO4.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 2) {
                        ATR_CO2.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 3) {
                        ATR_CH4.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 4) {
                        ATR_O3L.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 5) {
                        ATR_O3S.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 6) {
                        ATR_H2O.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 7) {
                        ATR_soot.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 8) {
                        ATR_SO4.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 9) {
                        ATR_AIC.push_back(atof(lineData.at(column).c_str()));
                } else if (column == 1) {
                        ATR_total.push_back(atof(lineData.at(column).c_str()));
                }
            }
        }
        lineData.clear();
        row++;
    }
    plotDataFile.close();

    /* Subplot Emissions */
    auto ax1 = matplot::subplot(4, 2, 0);
    matplot::plot(time, emissions_CO2);
    matplot::hold(matplot::on);
    matplot::plot(time, emissions_NOx);
    matplot::plot(time, emissions_H2O);
    matplot::plot(time, emissions_soot);
    matplot::plot(time, emissions_SO4);
    matplot::hold(matplot::off);
    matplot::title("Emissions");
    ax1->xlabel("Time [yr]");
    ax1->ylabel("Emissions  [kg]");
    double y_max = std::max({*std::max_element(emissions_CO2.begin(), emissions_CO2.end()),
                            *std::max_element(emissions_NOx.begin(), emissions_NOx.end()),
                            *std::max_element(emissions_H2O.begin(), emissions_H2O.end()),
                            *std::max_element(emissions_soot.begin(), emissions_soot.end()),
                            *std::max_element(emissions_SO4.begin(), emissions_SO4.end())});
    ax1->ylim({ACCURACY_LOW, y_max*1.1+1});
    ax1->y_axis().scale(matplot::axis_type::axis_scale::log);
    auto lgdEmissions = matplot::legend({"CO2", "NOx", "H2O", "Soot", "SO4"});

    /* Subplot ATR*/
    auto ax2 = matplot::subplot(4, 2, 1);
    matplot::plot(time, ATR_CO2);
    matplot::hold(matplot::on);
    matplot::plot(time, ATR_CH4);
    matplot::plot(time, ATR_O3L);
    matplot::plot(time, ATR_O3S);
    matplot::plot(time, ATR_H2O);
    matplot::plot(time, ATR_soot);
    matplot::plot(time, ATR_SO4);
    matplot::plot(time, ATR_AIC);
    matplot::plot(time, ATR_total);
    matplot::hold(matplot::off);
    matplot::title("Average Temperature Response");
    ax2->xlabel("Time [yr]");
    ax2->ylabel("ATR [K]");
    std::vector<std::string> legend_labels = {"CO2", "CH4", "O3L", "O3S", "H2O", "Soot", "SO4", "AIC", "Total"};
    auto lgd_ATR = matplot::legend(legend_labels);
    lgd_ATR->num_columns(2);
    lgd_ATR->location(matplot::legend::general_alignment::topright);

    /* Subplot RF */
    auto ax3 = matplot::subplot(4, 2, 2);
    matplot::plot(time, theClimateModelDallara.absRadForcing.at(0));
    matplot::hold(matplot::on);
    matplot::plot(time, theClimateModelDallara.absRadForcing.at(1));
    matplot::plot(time, theClimateModelDallara.absRadForcing.at(2));
    matplot::plot(time, theClimateModelDallara.absRadForcing.at(3));
    matplot::plot(time, theClimateModelDallara.absRadForcing.at(4));
    matplot::plot(time, theClimateModelDallara.absRadForcing.at(5));
    matplot::plot(time, theClimateModelDallara.absRadForcing.at(6));
    matplot::plot(time, theClimateModelDallara.absRadForcing.at(6));
    matplot::plot(time, theClimateModelDallara.absRadForcingTotal);
    matplot::hold(matplot::off);
    matplot::title("Radiative Forcing");
    ax3->xlabel("Time [yr]");
    ax3->ylabel("RF [W/m^2]");
    auto lgd_RF = matplot::legend(legend_labels);
    lgd_RF->num_columns(2);
    lgd_RF->location(matplot::legend::general_alignment::topright);

    /* Subplot RF modified */
    auto ax4 = matplot::subplot(4, 2, 3);
    matplot::plot(time, theClimateModelDallara.absnormRadForcing.at(0));
    matplot::hold(matplot::on);
    matplot::plot(time, theClimateModelDallara.absnormRadForcing.at(1));
    matplot::plot(time, theClimateModelDallara.absnormRadForcing.at(2));
    matplot::plot(time, theClimateModelDallara.absnormRadForcing.at(3));
    matplot::plot(time, theClimateModelDallara.absnormRadForcing.at(4));
    matplot::plot(time, theClimateModelDallara.absnormRadForcing.at(5));
    matplot::plot(time, theClimateModelDallara.absnormRadForcing.at(6));
    matplot::plot(time, theClimateModelDallara.absnormRadForcing.at(6));
    matplot::plot(time, theClimateModelDallara.absnormRadForcingTotal);
    matplot::title("Normalized Radiative Forcing");
    matplot::hold(matplot::off);
    ax4->xlabel("Time [yr]");
    ax4->ylabel("RF normalized [-]");
    auto lgd_nRF = matplot::legend(legend_labels);
    lgd_nRF->num_columns(2);
    lgd_nRF->location(matplot::legend::general_alignment::topright);

    /* Subplot temperature change */
    auto ax5 = matplot::subplot(4, 2, 4);
    matplot::plot(time, theClimateModelDallara.absTempChange.at(0));
    matplot::hold(matplot::on);
    matplot::plot(time, theClimateModelDallara.absTempChange.at(1));
    matplot::plot(time, theClimateModelDallara.absTempChange.at(2));
    matplot::plot(time, theClimateModelDallara.absTempChange.at(3));
    matplot::plot(time, theClimateModelDallara.absTempChange.at(4));
    matplot::plot(time, theClimateModelDallara.absTempChange.at(5));
    matplot::plot(time, theClimateModelDallara.absTempChange.at(6));
    matplot::plot(time, theClimateModelDallara.absTempChange.at(6));
    matplot::plot(time, theClimateModelDallara.absTempChangeTotal);
    matplot::hold(matplot::off);
    matplot::title("Temperature Change");
    ax5->xlabel("Time [yr]");
    ax5->ylabel("Temperature Change [K]");
    auto lgd_T = matplot::legend(legend_labels);
    lgd_T->num_columns(2);
    lgd_T->location(matplot::legend::general_alignment::topright);

    /* Subplot temperature change weighted*/
    auto ax6 = matplot::subplot(4, 2, 5);
    matplot::plot(time, theClimateModelDallara.absTempChangeweighted.at(0));
    matplot::hold(matplot::on);
    matplot::plot(time, theClimateModelDallara.absTempChangeweighted.at(1));
    matplot::plot(time, theClimateModelDallara.absTempChangeweighted.at(2));
    matplot::plot(time, theClimateModelDallara.absTempChangeweighted.at(3));
    matplot::plot(time, theClimateModelDallara.absTempChangeweighted.at(4));
    matplot::plot(time, theClimateModelDallara.absTempChangeweighted.at(5));
    matplot::plot(time, theClimateModelDallara.absTempChangeweighted.at(6));
    matplot::plot(time, theClimateModelDallara.absTempChangeweighted.at(6));
    matplot::plot(time, theClimateModelDallara.absTempChangeweightedTotal);
    matplot::hold(matplot::off);
    matplot::title("Weighted Temperature Change");
    ax6->xlabel("Time [yr]");
    ax6->ylabel("Weighted Temperature Change [K]");
    auto lgd_T_w = matplot::legend(legend_labels);
    lgd_T_w->num_columns(2);
    lgd_T_w->location(matplot::legend::general_alignment::topright);

    /* Subplot AGWP */
    auto ax7 = matplot::subplot(4, 2, 6);
    matplot::plot(time, theClimateModelDallara.absAGWP.at(0));
    matplot::hold(matplot::on);
    matplot::plot(time, theClimateModelDallara.absAGWP.at(1));
    matplot::plot(time, theClimateModelDallara.absAGWP.at(2));
    matplot::plot(time, theClimateModelDallara.absAGWP.at(3));
    matplot::plot(time, theClimateModelDallara.absAGWP.at(4));
    matplot::plot(time, theClimateModelDallara.absAGWP.at(5));
    matplot::plot(time, theClimateModelDallara.absAGWP.at(6));
    matplot::plot(time, theClimateModelDallara.absAGWP.at(6));
    matplot::plot(time, theClimateModelDallara.absAGWP);
    matplot::hold(matplot::off);
    matplot::title("Absolute Global Warming Potential");
    ax7->xlabel("Time [yr]");
    ax7->ylabel("AGWP [W/m^2]");
    auto lgd_AGWP = matplot::legend(legend_labels);
    lgd_AGWP->num_columns(2);
    lgd_AGWP->location(matplot::legend::general_alignment::topright);

    /* Subplot AGTP */
    auto ax8 = matplot::subplot(4, 2, 7);
    matplot::plot(time, theClimateModelDallara.TotalTempChange.at(0));
    matplot::hold(matplot::on);
    matplot::plot(time, theClimateModelDallara.TotalTempChange.at(1));
    matplot::plot(time, theClimateModelDallara.TotalTempChange.at(2));
    matplot::plot(time, theClimateModelDallara.TotalTempChange.at(3));
    matplot::plot(time, theClimateModelDallara.TotalTempChange.at(4));
    matplot::plot(time, theClimateModelDallara.TotalTempChange.at(5));
    matplot::plot(time, theClimateModelDallara.TotalTempChange.at(6));
    matplot::plot(time, theClimateModelDallara.TotalTempChange.at(6));
    matplot::plot(time, theClimateModelDallara.TotalTempChangeTotal);
    matplot::hold(matplot::off);
    matplot::title("Absolute Global Temperature Potential");
    ax8->xlabel("Time [yr]");
    ax8->ylabel("AGTP [K]");
    auto lgd_AGTP = matplot::legend(legend_labels);
    lgd_AGTP->num_columns(2);
    lgd_AGTP->location(matplot::legend::general_alignment::topright);
    /* Save plot */
    std::string plotFilePath = rtIO_->getPlotDir() + "/" + std::string(TOOL_NAME) + "_climate_impact.svg";
    matplot::save(plotFilePath);
}

void StandardStrategy::generatePlotData_dallara(const std::string& pltDataFile) {
    std::ofstream plot;
    plot.open(pltDataFile.c_str());
    if (!plot) {
        myRuntimeInfo->err << pltDataFile + " could not be opened!" << std::endl;
        exit(1);
    } else {
        plot << "# " + pltDataFile << std::endl;
        /*Adjust data structure: **/
        plot << "# (1)Year; (2)ATR_Total; (3)ATR_CO2; (4)ATR_CH4; (5)ATR_O3L; (6)ATR_O3S; (7)ATR_H2O; (8)ATR_Soot; (9)ATR_SO4; (10)ATR_AIC;"
                " (11)RadiativeForcing_Total[W/m^2]; (12)RadiativeForcing_CO2[W/m^2]; (13)RadiativeForcing_CH4[W/m^2]; (14)RadiativeForcing_O3L[W/m^2];"
                " (15)RadiativeForcing_O3S[W/m^2]; (16)RadiativeForcing_H2O[W/m^2]; (17)RadiativeForcing_Soot[W/m^2]; (18)RadiativeForcing_SO4[W/m^2];"
                " (19)RadiativeForcing_AIC[W/m^2]; (20)normRadiativeForcing_Total[]; (21)normRadiativeForcing_CO2[]; (22)normRadiativeForcing_CH4[];"
                " (23)normRadiativeForcing_O3L[]; (24)normRadiativeForcing_O3S[]; (25)normRadiativeForcing_H2O[]; (26)normRadiativeForcing_Soot[];"
                " (27)normRadiativeForcing_SO4[]; (28)normRadiativeForcing_AIC[]; (29)AGWP_Total[W/m^2]; (30)AGWP_CO2[W/m^2]; (31)AGWP_CH4[W/m^2];"
                " (32)AGWP_O3L[W/m^2]; (33)AGWP_O3S[W/m^2]; (34)AGWP_H2O[W/m^2]; (35)AGWP_Soot[W/m^2]; (36)AGWP_SO4[W/m^2]; (37)AGWP_AIC[W/m^2];"
                " (38)TempChange_Total[K]; (39)TempChange_CO2[K]; (40)TempChange_CH4[K]; (41)TempChange_O3L[K]; (42)TempChange_O3S[K]; (43)TempChange_H2O[K];"
                " (44)TempChange_Soot[K]; (45)TempChange_SO4[K]; (46)TempChange_AIC[K]; (47)TempChangeweighted_Total[K]; (48)TempChangeweighted_CO2[K];"
                " (49)TempChangeweighted_CH4[K]; (50)TempChangeweighted_O3L[K]; (51)TempChangeweighted_O3S[K]; (52)TempChangeweighted_H2O[K];"
                " (53)TempChangeweighted_Soot[K]; (54)TempChangeweighted_SO4[K]; (55)TempChangeweighted_AIC[K]; (56)AbsoluteGTP_Total[K]; (57)AbsoluteGTP_CO2[K];"
                " (58)AbsoluteGTP_CH4[K]; (59)AbsoluteGTP_O3L[K]; (60)AbsoluteGTP_O3S[K]; (61)AbsoluteGTP_H2O[K]; (62)AbsoluteGTP_Soot[K]; (63)AbsoluteGTP_SO4[K];"
                " (64)AbsoluteGTP_SO4[K]; (65)EmissionsPerYear_CO2[kg]; (66)EmissionsPerYear_NOx[kg]; (67)EmissionsPerYear_H2O[kg]; (68)EmissionsPerYear_Soot[kg];"
                " (69)EmissionsPerYear_SO4[kg]; (70)NauticalMilesPerYear_forAIC[nm];  " << std::endl;
        std::vector<double> ATRyearTmp;
        std::vector<double> ATRyear_CO2Tmp;
        std::vector<double> ATRyear_CH4Tmp;
        std::vector<double> ATRyear_O3LTmp;
        std::vector<double> ATRyear_O3STmp;
        std::vector<double> ATRyear_H2OTmp;
        std::vector<double> ATRyear_sootTmp;
        std::vector<double> ATRyear_SO4Tmp;
        std::vector<double> ATRyear_AICTmp;
        std::vector<double> e_year_CO2Tmp;
        std::vector<double> e_year_NOxTmp;
        std::vector<double> e_year_H2OTmp;
        std::vector<double> e_year_sootTmp;
        std::vector<double> e_year_SO4Tmp;
        std::vector<double> e_year_AICTmp;
        for (uint16_t t = 0; t <= theClimateModelDallara.dallaraIOData_->configuration.tmax; t++) {
            double sigma = 0;
            if (t < theClimateModelDallara.dallaraIOData_->durationOperation) {
                sigma = 1;
            } else {
                sigma = 0;
            }
            /* write out the ATR values */
            ATRyearTmp.push_back(0.);
            ATRyear_CO2Tmp.push_back(0.);
            ATRyear_CH4Tmp.push_back(0.);
            ATRyear_O3LTmp.push_back(0.);
            ATRyear_O3STmp.push_back(0.);
            ATRyear_H2OTmp.push_back(0.);
            ATRyear_sootTmp.push_back(0.);
            ATRyear_SO4Tmp.push_back(0.);
            ATRyear_AICTmp.push_back(0.);
            ATRyearTmp.back() = sigma * theClimateModelDallara.dallaraIOData_->ATR;
            ATRyear_CO2Tmp.back() = sigma * theClimateModelDallara.ATRyear_i.at(0);
            ATRyear_CH4Tmp.back() = sigma * theClimateModelDallara.ATRyear_i.at(1);
            ATRyear_O3LTmp.back() = sigma * theClimateModelDallara.ATRyear_i.at(2);
            ATRyear_O3STmp.back() = sigma * theClimateModelDallara.ATRyear_i.at(3);
            ATRyear_H2OTmp.back() = sigma * theClimateModelDallara.ATRyear_i.at(4);
            ATRyear_sootTmp.back() = sigma * theClimateModelDallara.ATRyear_i.at(5);
            ATRyear_SO4Tmp.back() = sigma * theClimateModelDallara.ATRyear_i.at(6);
            ATRyear_AICTmp.back() = sigma * theClimateModelDallara.ATRyear_i.at(7);
            /* Write out the annual emission values */
            e_year_CO2Tmp.push_back(0.);
            e_year_NOxTmp.push_back(0.);
            e_year_H2OTmp.push_back(0.);
            e_year_sootTmp.push_back(0.);
            e_year_SO4Tmp.push_back(0.);
            e_year_AICTmp.push_back(0.);
            e_year_CO2Tmp.back() = sigma * theClimateModelDallara.e_year_i.at(0);
            e_year_NOxTmp.back() = sigma * theClimateModelDallara.e_year_i.at(1);
            e_year_H2OTmp.back() = sigma * theClimateModelDallara.e_year_i.at(4);
            e_year_sootTmp.back() = sigma * theClimateModelDallara.e_year_i.at(5);
            e_year_SO4Tmp.back() = sigma * theClimateModelDallara.e_year_i.at(6);
            e_year_AICTmp.back() = sigma * theClimateModelDallara.e_year_i.at(7);
            plot << t << "; " <<
                 ATRyearTmp.at(t) << ";" <<
                 ATRyear_CO2Tmp.at(t) << ";" <<
                 ATRyear_CH4Tmp.at(t) << ";" <<
                 ATRyear_O3LTmp.at(t) << ";" <<
                 ATRyear_O3STmp.at(t) << ";" <<
                 ATRyear_H2OTmp.at(t) << ";" <<
                 ATRyear_sootTmp.at(t) << ";" <<
                 ATRyear_SO4Tmp.at(t) << ";" <<
                 ATRyear_AICTmp.at(t) << ";" <<
                 theClimateModelDallara.absRadForcingTotal.at(t) << ";";
            for (unsigned int i = 0; i <= 7; i++) {
                plot << theClimateModelDallara.absRadForcing.at(i).at(t) << ";";
            }
            plot << theClimateModelDallara.absnormRadForcingTotal.at(t) << ";";
            for (unsigned int i = 0; i <= 7; i++) {
                plot << theClimateModelDallara.absnormRadForcing.at(i).at(t) << ";";
            }
            plot << theClimateModelDallara.absAGWPTotal.at(t) << ";";
            for (unsigned int i = 0; i <= 7; i++) {
                plot << theClimateModelDallara.absAGWP.at(i).at(t) << ";";
            }
            plot << theClimateModelDallara.absTempChangeTotal.at(t) << ";";
            for (unsigned int i = 0; i <= 7; i++) {
                plot << theClimateModelDallara.absTempChange.at(i).at(t) << ";";
            }
            plot << theClimateModelDallara.absTempChangeweightedTotal.at(t) << ";";
            for (unsigned int i = 0; i <= 7; i++) {
                plot << theClimateModelDallara.absTempChangeweighted.at(i).at(t) << ";";
            }
            plot << theClimateModelDallara.TotalTempChangeTotal.at(t) << ";";
            for (unsigned int i = 0; i <= 7; i++) {
                plot << theClimateModelDallara.TotalTempChange.at(i).at(t) << ";";
            }
            plot << e_year_CO2Tmp.at(t) << ";" <<
                 e_year_NOxTmp.at(t) << ";" <<
                 e_year_H2OTmp.at(t) << ";" <<
                 e_year_sootTmp.at(t) << ";" <<
                 e_year_SO4Tmp.at(t) << ";" <<
                 e_year_AICTmp.at(t) << ";" <<
                 std::endl;
        }
        plot.close();
    }
}
