/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <iostream>
#include <memory>
#include <algorithm>
#include <string>
#include <filesystem>
#include <unitConversion/unitConversion.h>
#include <aixml/node.h>
#include "standard_strategy.h"
#include "../toolinfo.h"


void StandardStrategy::initialize() {
    myRuntimeInfo->out << "Initialize standard strategy" << std::endl;
    /* Read calculation methods from config */
    lifeCycleEmissionsMethod = EndnodeReadOnly<std::string>("method").read(rtIO_->moduleConfig.at("program_settings/"
                                + std::string("standard_strategy/emission_calculation/life_cycle_emissions_methods/"))).value();
    myRuntimeInfo->out << "  Method " << lifeCycleEmissionsMethod << " will be used for life cycle emission calculation" << std::endl;
    climateModelMethod = EndnodeReadOnly<std::string>("method").read(rtIO_->moduleConfig.at("program_settings/"
                                + std::string("standard_strategy/impact_calculation/climate_model_methods/"))).value();
    myRuntimeInfo->out << "  Method " << climateModelMethod << " will be used for climate model" << std::endl;
    airQualityMethod = EndnodeReadOnly<std::string>("method").read(rtIO_->moduleConfig.at("program_settings/"
                                + std::string("standard_strategy/impact_calculation/air_quality_methods/"))).value();
    myRuntimeInfo->out << "  Method " << airQualityMethod << " will be used for air quality calculation" << std::endl;

    /* Prepare ecological_assessment.xml */
    if (!std::filesystem::exists(rtIO_->getXmlDir())) std::filesystem::create_directory(rtIO_->getXmlDir());
    std::string filepath = rtIO_->getXmlDir() + FILESEPERATOR + TOOL_NAME + "_results.xml";
    if (fileExists(filepath)) { // Delete old calculation results of this phase
        ecologicalAssessmentXML_ = aixml::openDocument(std::filesystem::path(filepath));
        node* node = ecologicalAssessmentXML_->find("/ecological_data_file");
        if (node != nullptr) {
            node->deleteChildren();
        }
    } else {
        ecologicalAssessmentXML_ = std::make_shared<node>(filepath);
        ecologicalAssessmentXML_->appendChild("ecological_data_file");
    }

    /* Initialize Report */
    report_ = std::shared_ptr<Report>(new Report(rtIO_));
}

void StandardStrategy::run() {
    myRuntimeInfo->out << "Run standard strategy" << std::endl;

    /* Mission emission calculation */
    theDesignMission.run("design", rtIO_, ecologicalAssessmentXML_);
    theStudyMission.run("study", rtIO_, ecologicalAssessmentXML_);

    /* Life cycle emission calculation */
    if (lifeCycleEmissionsMethod == "NO") {
    } else if (lifeCycleEmissionsMethod == "SCHAEFER") {
        theLifeCycleEmissionsSchaefer.run(rtIO_, ecologicalAssessmentXML_);
    } else {
        myRuntimeInfo->warn << "Life cycle emission method" << lifeCycleEmissionsMethod << " unknow. Please check settings in config file!" << std::endl;
    }

    /* Climate model */
    if (climateModelMethod == "NO") {
    } else if (climateModelMethod == "DALLARA") {
        theClimateModelDallara.run(rtIO_, ecologicalAssessmentXML_);
    } else {
        myRuntimeInfo->warn << "Climate model method" << climateModelMethod << " unknow. Please check settings in config file!" << std::endl;
    }

    /* Air quality */
    if (airQualityMethod == "NO") {
    } else if (airQualityMethod == "SCHAEFER") {
        theAQISchaefer.run(rtIO_, ecologicalAssessmentXML_);
    } else {
        myRuntimeInfo->warn << "Air quality calculation method " << airQualityMethod << " unknow. Please check settings in config file!" << std::endl;
    }
}

void StandardStrategy::update() {
    myRuntimeInfo->out << "Update " + rtIO_->acxmlAccess.filename().string() << std::endl;
    theClimateModelDallara.dallaraIOData_->updateAircraftXML(rtIO_);
}

void StandardStrategy::report() {
    /* Plots */
    if (rtIO_->plotOn) {
        myRuntimeInfo->out << "Generate plots" << std::endl;
        myRuntimeInfo->out << "  Generate study mission plots" << std::endl;
        generateMissionEmissionsPlot("study");
        myRuntimeInfo->out << "  Generate design mission plots" << std::endl;
        generateMissionEmissionsPlot("design");
        if (lifeCycleEmissionsMethod == "SCHAEFER" && rtIO_->aircraft_energy_carrier_type() == "kerosene" && rtIO_->aircraft_configuration_type() == "tube_and_wing") {
            myRuntimeInfo->out << "  Generate " << lifeCycleEmissionsMethod << " emissions plots" << std::endl;
            generatePlots_schaefer();
        }
        if (climateModelMethod == "DALLARA") {
            myRuntimeInfo->out << "  Generate " << climateModelMethod << " plots " << std::endl;
            generatePlots_dallara();
        }
    }

    /* Reports */
    myRuntimeInfo->out << "Write reports" << std::endl;
    report_->htmlReportStream() << html::div_start("class=\"box\""); // start html box
    myRuntimeInfo->out << "  Write mission report body" << std::endl;
    setHtmlBody_mission(rtIO_);
    setTexBody_mission();
    if (lifeCycleEmissionsMethod == "SCHAEFER" && rtIO_->aircraft_energy_carrier_type() == "kerosene" && rtIO_->aircraft_configuration_type() == "tube_and_wing") {
        myRuntimeInfo->out << "  Write " << lifeCycleEmissionsMethod << " emissions report body" << std::endl;
        setHtmlBody_LCA_schaefer(rtIO_);
        setTexBody_LCA_schaefer(rtIO_);
        generateInflightEmissionsCSV_schaefer();
        generateLifeCycleEmissionsCSV_schaefer();
    }
    if (climateModelMethod == "DALLARA") {
        myRuntimeInfo->out << "  Write " << climateModelMethod << " climate report body" << std::endl;
        setHtmlBody_dallara(rtIO_);
        setTexBody_dallara();
    }
    if (airQualityMethod == "SCHAEFER") {
        myRuntimeInfo->out << "  Write " << airQualityMethod << " air quality report body" << std::endl;
        setHtmlBody_AQISchaefer();
        setTexBody_AQISchaefer(rtIO_);
    }
    report_->htmlReportStream() << html::div_end; // end html box
    report_->generateReports();
}

void StandardStrategy::save() {
    aixml::saveDocument(*ecologicalAssessmentXML_);
}

