/* UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#ifndef ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_STANDARD_STRATEGY_H_
#define ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_STANDARD_STRATEGY_H_

#include <memory>
#include <vector>
#include <string>
#include <matplot/matplot.h>
#include <moduleBasics/strategySelector.h>
#include <moduleBasics/report.h>
#include <moduleBasics/html.h>
#include <moduleBasics/plot.h>
#include <moduleBasics/runtimeIO.h>
#include <aixml/node.h>
#include "emission_calculation/ecoDatabase.h"
#include "emission_calculation/mission_emissions/mission.h"
#include "emission_calculation/life_cycle_emissions/LCA_schaefer/calculateEmissions_schaefer.h"
#include "impact_calculation/climate_model_dallara/dallara.h"
#include "impact_calculation/air_quality_index_schaefer/AQI_schaefer.h"

/** \class StandardStrategy
 * \brief Class for emission and ecological impact calculation
 *
 */
class StandardStrategy : public Strategy {
 public:
  /** \brief Constructor of class StandardStrategy
  * \param rtIO Access to runtimeIO
  */
  explicit StandardStrategy(std::shared_ptr<RuntimeIO> rtIO): rtIO_(rtIO) {}
  /** \brief Destructor of class StandardStrategy */
  ~StandardStrategy() = default;

  /* Standard functions */
  void initialize();
  void run();
  void update();
  void report();
  void save();

  /* Reporting functions mission emission*/
  /** \brief Define the content of the HTML Report body
   */
  void setHtmlBody_mission(const std::shared_ptr<RuntimeIO> &rtIO);
  /** \brief Define the content of the Tex Report body
   */
  void setTexBody_mission();

  /* Plotting function for mission*/
  /** \brief Plots emission flows over mission
   * \param missionType Type of the mission (study or design)
   */
  void generateMissionEmissionsPlot(const std::string& missionType);

  /* Reporting functions LCA_Schaefer*/
  /** \brief Define the content of the HTML Report body
   * \param rtIO Access to runtime IO
   */
  void setHtmlBody_LCA_schaefer(const std::shared_ptr<RuntimeIO> &rtIO);
  /** \brief Define the content of the Tex Report body
   * \param rtIO Access to runtime IO
   */
  void setTexBody_LCA_schaefer(const std::shared_ptr<RuntimeIO> &rtIO);

  /* Plotting functions LCA_Schaefer*/
  /** \brief Generates plot of inflight emissions
   */
  void generatePlots_schaefer();
  /** \brief Generates csv file for inflight emissions
   * \param pltName Name of the plot
   */
  void generateInflightEmissionsCSV_schaefer();
  /** \brief Generates inflight emissions csv
   */
  void generateInflightEmissionsPlot_schaefer(const std::string& pltName);
  /** \brief Generates plots of life cycle emissions
   */
  void generateLifeCycleEmissionsPlot_schaefer(const std::string& pltName);
  /** \brief Generates csv file for life cycle emissions
   */
  void generateLifeCycleEmissionsCSV_schaefer();

  /* Reporting functions dallara climate model*/
  /** \brief Define the content of the HTML Report body
   * \param rtIO Access to runtime IO
   */
  void setHtmlBody_dallara(const std::shared_ptr<RuntimeIO> &rtIO);
  /** \brief Define the content of the Tex Report body
   */
  void setTexBody_dallara();

  /* Plotting functions climate model dallara */
  /** \brief Generated plots for dallara climate model
   */
  void generatePlots_dallara();
  /** \brief Writes csv with plot data of dallara climate model
   * \param pltDataFile Name of plot data csv
   */
  void generatePlotData_dallara(const std::string& pltDataFile);

  /* Reporting functions schaefer AQI */
  /** \brief Define the content of the HTML Report body
   */
  void setHtmlBody_AQISchaefer();
  /** \brief Define the content of the Tex Report body
   * \param rtIO Access to runtime IO
   */
  void setTexBody_AQISchaefer(const std::shared_ptr<RuntimeIO> &rtIO);



 private:
  std::shared_ptr<RuntimeIO> rtIO_; /**< RuntimeIO object*/
  std::shared_ptr<Report> report_; /**< Pointer to report object*/
  std::shared_ptr<node> ecologicalAssessmentXML_; /**< Pointer to ecological values node*/

  /* Methods */
  std::string lifeCycleEmissionsMethod; /**< Chosen method for life cycle emission calulation (NO, SCHAEFER)*/
  std::string climateModelMethod; /**< Chosen method for climate model (NO, DALLARA)*/
  std::string airQualityMethod; /**< Chosen method for air quality calculation (NO, SCHAEFER)*/

  /* Method objects*/
  Mission theStudyMission;/**< Object contains data of one study mission */
  Mission theDesignMission;/**< Object contains data of one design mission which can also be used for flight tests in LCA*/
  LifeCycleEmissionsSchaefer theLifeCycleEmissionsSchaefer; /**< Schaefer life cycle emissions object */
  ClimateModelDallara theClimateModelDallara; /**< Dallara climate model object */
  AQISchaefer theAQISchaefer;/**< Air quality index Schaefer object*/
};

#endif //ECOLOGICAL_ASSESSMENT_SRC_STANDARD_STRATEGY_STANDARD_STRATEGY_H_
