/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <aircraftGeometry2/airfoil_surface.h>
#include <aircraftGeometry2/geometry/surface.h>
#include <aixml/node.h>

#include "bwbVerticalTailsData.h"

namespace bwb {
namespace low {
VerticalTailsData::VerticalTailsData() :
  wing_reference_position(PositionIO("component_design/wing/position", "wing")),
  maximum_takeoff_mass(MassPropertiesIO("analysis/masses_cg_inertia/maximum_takeoff_mass/","maximum_takeoff_mass")),
  empennage_mass(MassPropertiesIO("component_design/empennage/", "Vertical tails mass")),
  empennage_position(PositionIO("component_design/empennage/position", "Vertical tails position")) {}

void VerticalTailsData::read(const node& xml) {
  wing_reference_position.read(xml);
  maximum_takeoff_mass.read(xml);
}
void VerticalTailsData::read_wing(std::filesystem::path path, std::filesystem::path airfoil_data_dir) {
  std::shared_ptr<node> acxml = aixml::openDocument(path);
  geom2::WingFactory wing_factory{acxml, airfoil_data_dir};
  wing = wing_factory.create("wing/specific/geometry/aerodynamic_surface@0");
}
void VerticalTailsData::update(node& xml) {
  update_vertical_tails(xml);
  empennage_mass.update(xml);
  empennage_position.update(xml);
  
}
void VerticalTailsData::update_vertical_tails(node& xml) {
  for (size_t id = 0; id < tails.size(); ++id) {
    geom2::io::SurfaceType io = geom2::io::Wing(tails[id]);
    std::visit(geom2::io::AixmlConverter(xml.at("component_design/empennage/specific/geometry/"),
                                         {"aerodynamic_surface", std::to_string(id), tails[id].name}),
               io);
    tails_mass[id].update(xml);
  }
}
}  // namespace low
}  // namespace bwb
