/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>
#include <matplot/matplot.h>

#include <array>
#include <cstdlib>
#include <deque>
#include <map>
#include <utility>
#include <algorithm>

#include "lowConventionalEmpennageDesign.h"

/**
 * \brief Generate polygon from spar or controldevice
 *
 * \param surface aerodynamic surface object
 * \param object spar or control device object
 * \return std::tuple<std::vector<double>,std::vector<double>,std::string> coordinates and object name
 */
std::tuple<std::vector<double>, std::vector<double>, std::string> gen_polygon(
    const geom2::MultisectionSurface<geom2::AirfoilSection>& surface,
    const geom2::MultisectionSurface<geom2::PolygonSection>& object);

/**
 * \brief RGB and alpha values to Matplot++ format
 *
 * \param r red value between 0 - 255
 * \param b blue value between 0 - 255
 * \param g green alue between 0 - 255
 * \param a alpha value between 0 (invisible) - 1 (solid)
 * \return std::array<float,4UL> matplot++ conform color format
 */
std::array<float, 4UL> rgba(float r, float b, float g, float a) { return {1 - a, r / 255, b / 255, g / 255}; }

/* */
std::array<float, 4UL> spar_color = rgba(99, 110, 114, 1.0);

std::map<std::string, std::array<float, 4UL>> control_device_colormap = {{"elevator", rgba(108, 92, 231, 1.0)},
                                                                         {"rudder", rgba(232, 67, 147, 1.0)}};

std::tuple<std::vector<double>, std::vector<double>, std::string> gen_polygon(
    const geom2::MultisectionSurface<geom2::AirfoilSection>& surface,
    const geom2::MultisectionSurface<geom2::PolygonSection>& object) {
  double half_span = surface.is_symmetric ? 0.5 * geom2::measure::span(surface) : geom2::measure::span(surface);
  /* Spar coordinates positive values normally*/
  double from = object.sections.at(0).origin.z() * half_span;
  double to = object.sections.at(1).origin.z() * half_span;

  /* Chordwise coordinates */
  double chord_from_fwd = object.sections.at(0).get_contour().vertex(0).x();
  double chord_from_aft = object.sections.at(0).get_contour().vertex(1).x();
  double chord_to_fwd = object.sections.at(1).get_contour().vertex(0).x();
  double chord_to_aft = object.sections.at(1).get_contour().vertex(1).x();

  auto f_chord_fwd = [chord_from_fwd, chord_to_fwd, half_span](double span_position) {
    return std::lerp(chord_from_fwd, chord_to_fwd, fabs(span_position) / half_span);
  };
  auto f_chord_aft = [chord_from_aft, chord_to_aft, half_span](double span_position) {
    return std::lerp(chord_from_aft, chord_to_aft, fabs(span_position) / half_span);
  };
  /* Add object from */
  std::vector<double> spanwise_coordinates{-fabs(from)};

  /* Add intermediate sections */
  for (auto& section : surface.sections) {
    if (fabs(section.origin.z()) > fabs(from) && fabs(section.origin.z()) < fabs(to)) {
      spanwise_coordinates.push_back(section.origin.z());
    }
  }
  /* Add object to */
  spanwise_coordinates.push_back(-fabs(to));

  /**/
  std::vector<double> x;
  std::vector<double> y;
  /* From in to out*/
  for (auto it = spanwise_coordinates.begin(); it != spanwise_coordinates.end(); ++it) {
    y.push_back(fabs(*it));
    x.push_back(geom2::measure::offset_LE(surface, *it).x() + f_chord_fwd(*it) * geom2::measure::chord(surface, *it));
  }
  /* From out to in*/
  for (auto rit = spanwise_coordinates.rbegin(); rit != spanwise_coordinates.rend(); ++rit) {
    y.push_back(fabs(*rit));
    x.push_back(geom2::measure::offset_LE(surface, *rit).x() +
                f_chord_aft(*rit) * geom2::measure::chord(surface, *rit));
  }

  /* Close loop */
  y.push_back(y.front());
  x.push_back(x.front());

  return {x, y, object.name};
}
namespace taw {
namespace low {

fs::path Conventional::plot_stabilizer_planform(
    const geom2::MultisectionSurface<geom2::AirfoilSection>& surface,
    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& spars,
    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& devices) {
  /* Plot active */
  if (!rtIO->plotOn) {
    return "";
  }
  auto coordinates_xy = geom2::transform::outline_2d(surface, geom2::Direction_3(0, 0, 1));
  std::deque<double> spanwise;
  std::deque<double> chordwise;
  auto span = geom2::measure::span(surface);
  auto half_span = surface.is_symmetric ? span * 0.5 : span;
  auto sections = surface.sections;
  auto fuselage_max_half = geom2::measure::width_max(data->fuselage) * 0.5;

  /* Check if control surface is oriented vertical */
  bool is_vertical = std::fabs(surface.normal.dz()) > 0.95;

  /* Loop coordinates from out to inside for LE and TE */
  for (auto it = sections.rbegin(); it != sections.rend(); ++it) {
    spanwise.push_front(-it->origin.z());
    spanwise.push_back(-it->origin.z());
    chordwise.push_front(it->origin.x());
    chordwise.push_back(it->origin.x() + it->get_chord_length());
  }

  /* Generate plot */
  std::string name = surface.name;
  std::replace(name.begin(), name.end(), '_', ' ');
  std::vector<std::string> legend_strings;
  auto f = matplot::figure(true);
  f->title(std::format("{} planform", name));
  auto ax = f->add_axes(false);
  if (is_vertical) {
    ax->plot(std::vector<double>(chordwise.begin(), chordwise.end()),
             std::vector<double>(spanwise.begin(), spanwise.end()))
        ->color("black")
        .line_width(1.5);
  } else {
    ax->plot(std::vector<double>(spanwise.begin(), spanwise.end()),
             std::vector<double>(chordwise.begin(), chordwise.end()))
        ->color("black")
        .line_width(1.5);
  }

  legend_strings.push_back("contour");
  matplot::hold(true);

  /* Plot spars */
  for (auto& spar : spars) {
    const auto [x, y, _] = gen_polygon(surface, spar);
    (void)_;
    if (is_vertical) {
      ax->plot(x, y)->color(spar_color).line_width(1).line_style("--");
    } else {
      ax->plot(y, x)->color(spar_color).line_width(1).line_style("--");
    }
    legend_strings.push_back("spar");
  }
  for (auto& device : devices) {
    const auto [x, y, name] = gen_polygon(surface, device);
    if (is_vertical) {
      ax->plot(x, y)->color(control_device_colormap[name]).line_width(1.5);
    } else {
      ax->plot(y, x)->color(control_device_colormap[name]).line_width(1.5);
    }
    legend_strings.push_back(device.name);
  }

  /* Select if vertical or horizontal element */
  if (std::fabs(surface.normal.dz()) > 0.7) {
    ax->xlabel("x [m]");
    ax->ylabel("z [m]");
  } else {
    ax->xlabel("y [m]");
    ax->ylabel("x [m]");
  }

  ax->limits_mode(matplot::manual);
  ax->axis(matplot::equal);
  if (is_vertical) {
    ax->ylim({0, *std::max_element(spanwise.begin(), spanwise.end()) * 1.1});
  } else {
    ax->xlim({0, *std::max_element(spanwise.begin(), spanwise.end()) * 1.1});
  }
  ax->legend(legend_strings);
  ax->legend()->location(matplot::legend::general_alignment::bottomright);
  ax->legend()->inside(true);
  ax->legend()->box(true);
  ax->legend()->font_size(8);

  fs::path plot_path = rtIO->getPlotDir() + "/empennage_design_" + surface.name + "_planform.svg";
  matplot::save(f, plot_path.string(), "svg");
  std::this_thread::sleep_for(std::chrono::milliseconds(500));
  return fs::relative(plot_path, fs::path(rtIO->getHtmlDir()));
}

fs::path Conventional::plot_thickness_and_twist_distribution(
    const geom2::MultisectionSurface<geom2::AirfoilSection>& surface) {
  /* Plot active */
  if (!rtIO->plotOn) {
    return "";
  }
  std::vector<double> eta;
  std::vector<double> thickness_to_chord_ratio;
  std::vector<double> twist;
  double half_span = surface.is_symmetric ? geom2::measure::span(surface) * 0.5 : geom2::measure::span(surface);
  for (auto section : surface.sections) {
    eta.push_back(-section.origin.z() / half_span);
    thickness_to_chord_ratio.push_back(geom2::measure::thickness_max(section) / section.get_chord_length());
    twist.push_back(section.get_twist_angle());
  }
  const auto [t_to_c_min,t_to_c_max] = std::ranges::minmax_element(thickness_to_chord_ratio);
  const auto [twist_min,twist_max] = std::ranges::minmax_element(twist);

  auto f = matplot::figure(true);

  auto ax1 = f->add_subplot(2, 1, 0);
  ax1->plot(eta, thickness_to_chord_ratio, "g-")->color("black").line_width(1.5);
  // ax1->xlabel("\u03B7");
  ax1->x_axis().visible(true);
  ax1->grid(true);
  ax1->ylabel("t / c [1]");
  ax1->ylim({*t_to_c_min - 0.01,*t_to_c_max + 0.01});

  auto ax2 = f->add_subplot(2, 1, 1);
  ax2->plot(eta, twist, "b-")->color("black").line_style("--").line_width(1.5);
  ax2->ylabel("\u03F5 [\u00b0]");
  ax2->xlabel("\u03B7 [1]");
  ax2->grid(true);
  ax2->ylim({*twist_min - 0.01, *twist_max + 0.01});

  fs::path plot_path = rtIO->getPlotDir() + "/empennage_design_" + surface.name + "_thickness_and_twist_distribution.svg";
  matplot::save(f, plot_path.string(), "svg");
  std::this_thread::sleep_for(std::chrono::milliseconds(500));
  return fs::relative(plot_path, fs::path(rtIO->getHtmlDir()));
}

/**
 * \brief Plot airfoils
 *
 * \return fs::path
 */
std::vector<fs::path> Conventional::plot_airfoils(const geom2::MultisectionSurface<geom2::AirfoilSection>& surface) {
  /* Plot active */
  if (!rtIO->plotOn) {
    return {""};
  }
  std::map<std::string, std::vector<geom2::Point_2>> airfoils_data{};
  std::vector<fs::path> airfoil_plots_path;

  for (auto section : surface.sections) {
    if (airfoils_data.find(section.name) != airfoils_data.end()) {
      continue;
    } else {
      airfoils_data.insert({section.name, section.get_contour().vertices()});
    }
  }
  /* loop over airfoils */

  for (const auto& airfoil_data : airfoils_data) {
    std::vector<double> x{}, y{};
    for (auto coord : airfoil_data.second) {
      x.push_back(coord.x());
      y.push_back(coord.y());
    }

    {
      auto f = matplot::figure(true);
      auto h = f->current_axes()->plot(x, y);
      h->line_width(1.5).color("black");

      auto ax = f->current_axes();
      ax->title(std::format("Airfoil - {}", airfoil_data.first));
      matplot::xlabel("\u03B7");
      ax->ylabel("z/c");
      ax->xlabel("\u03B7 [1]");
      ax->grid(true);
      ax->ylim({-0.2, 0.2});
      ax->xlim({0, 1});
      ax->axis(matplot::equal);

      fs::path plot_path = rtIO->getPlotDir() + "/empennage_design_" + surface.name + "_" + airfoil_data.first + "_airfoil.svg";
      matplot::save(f, plot_path.string(), "svg");
      airfoil_plots_path.push_back(fs::relative(plot_path, fs::path(rtIO->getHtmlDir())));
      std::this_thread::sleep_for(std::chrono::milliseconds(500));
    }
  }
  return airfoil_plots_path;
}

}  // namespace low
}  // namespace taw
