# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import math


def payload_tube_reference_point_calculation(dict_ac_exchange, dict_mod_config, payload_tube_description,
                                             runtime_output):
    """ Calculates the reference points of each payload tube.

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict payload_tube_description: Dict containing parameter and values of current blended wing body design
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict payload_tube_center_reference_points: Dict containing center reference points of each payload tube
            dict payload_tube_wall_reference_points: Dict containing wall reference points of each payload tube
            dict payload_tube_length: Dict containing the length of each payload tube
    """
    ''' initialize local parameter '''
    total_length = 0
    cabin_delta_in_x = []
    payload_tube_length = {}
    payload_tube_wall_reference_points = {}
    payload_tube_wall_front_reference_points = {}
    payload_tube_wall_aft_reference_points = {}
    payload_tube_center_reference_points = {}
    payload_tube_center_front_reference_points = {}
    payload_tube_center_aft_reference_points = {}
    # ToDo: ersetzen mit constants from lib
    convert_meter_to_inch = 39.3701

    ''' estimation of delta lengths in x-direction '''
    i = 0
    for payload_tube, payload_tube_width in payload_tube_description['payload_tube_width'].items():
        if dict_ac_exchange['number_of_payload_tubes'] % 2 != 0 and i == 0:
            cabin_delta_in_x.append(
                round((payload_tube_width / 2 + payload_tube_description['payload_tube_wall_thickness'][payload_tube])
                      * math.tan(math.radians(dict_ac_exchange['center_body_wing_sweep'])), 3))
        elif dict_ac_exchange['number_of_payload_tubes'] % 2 == 0 and i == 0 \
                or dict_ac_exchange['number_of_payload_tubes'] % 2 == 0 and i == 1:
            cabin_delta_in_x.append(
                round((payload_tube_width + payload_tube_description['payload_tube_wall_thickness'][payload_tube] * 1.5)
                      * math.tan(math.radians(dict_ac_exchange['center_body_wing_sweep'])), 3))
        else:
            cabin_delta_in_x.append(
                round((payload_tube_width + payload_tube_description['payload_tube_wall_thickness'][payload_tube])
                      * math.tan(math.radians(dict_ac_exchange['center_body_wing_sweep'])), 3))
        i += 1

    # check if the delta length in x direction is greater than the minimum required cockpit length
    #  -> if false: -> add offset to delta length
    if cabin_delta_in_x[0] < dict_mod_config['length_of_flight_deck'] * convert_meter_to_inch:
        delta_cockpit_length = \
            round(dict_mod_config['length_of_flight_deck'] * convert_meter_to_inch - cabin_delta_in_x[0], 3)

        # add delta cockpit length to all payload tubes
        for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
            cabin_delta_in_x[tube] = round(cabin_delta_in_x[tube] + delta_cockpit_length, 3)

    ''' estimation of reference points in x-, y- and z-direction of each payload tube '''
    for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
        max_height = 0.0
        max_width_at_base = 0.0
        payload_tube_center_front_reference_points['payload_tube_' + str(tube)] = {}
        payload_tube_wall_front_reference_points['payload_tube_' + str(tube)] = {}

        # check if the number of payload tubes is not even
        #  -> if true: -> set reference points for an uneven number of payload tube
        if tube == 0 and dict_ac_exchange['number_of_payload_tubes'] % 2 != 0:
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x'] = \
                round(dict_ac_exchange['global_reference_point_in_x_direction'] * convert_meter_to_inch
                      + cabin_delta_in_x[tube], 3)
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y'] = \
                round(dict_ac_exchange['global_reference_point_in_y_direction'] * convert_meter_to_inch, 3)
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z'] = \
                round(dict_ac_exchange['global_reference_point_in_z_direction'] * convert_meter_to_inch, 3)

        elif dict_ac_exchange['number_of_payload_tubes'] % 2 != 0 and tube % 2 != 0:
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x'] = \
                round(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                          'center_reference_point_in_x']
                      + cabin_delta_in_x[tube], 3)
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y'] = \
                round(-1 * payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                    'center_reference_point_in_y']
                      - payload_tube_description['payload_tube_width']['payload_tube_' + str(tube - 1)] / 2
                      - payload_tube_description['payload_tube_width']['payload_tube_' + str(tube)] / 2
                      + payload_tube_description['payload_tube_wall_thickness']['payload_tube_' + str(tube)], 3)
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z'] = \
                round(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                          'center_reference_point_in_z']
                      + (math.tan(math.radians(dict_mod_config['floor_angle_between_passenger_tubes']))
                         * (abs(
                            payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                                'center_reference_point_in_y'])
                            - abs(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                                      'center_reference_point_in_y']))), 3)

        elif dict_ac_exchange['number_of_payload_tubes'] % 2 != 0 and tube % 2 == 0:
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x'] = \
                payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                    'center_reference_point_in_x']
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y'] = \
                -1 * payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                    'center_reference_point_in_y']
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z'] = \
                payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                    'center_reference_point_in_z']

        # check if the number of payload tubes is even
        #  -> if true: -> set reference points for an even number payload tube
        if tube == 0 and dict_ac_exchange['number_of_payload_tubes'] % 2 == 0:
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x'] = \
                round(dict_ac_exchange['global_reference_point_in_x_direction'] * convert_meter_to_inch
                      + cabin_delta_in_x[tube], 3)
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y'] = \
                round(dict_ac_exchange['global_reference_point_in_y_direction'] * convert_meter_to_inch
                      - payload_tube_description['payload_tube_width']['payload_tube_' + str(tube)] / 2
                      + payload_tube_description['payload_tube_wall_thickness']['payload_tube_' + str(tube)] / 2, 3)
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z'] = \
                round(dict_ac_exchange['global_reference_point_in_z_direction'] * convert_meter_to_inch, 3)

        elif dict_ac_exchange['number_of_payload_tubes'] % 2 == 0 and tube % 2 != 0:
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x'] = \
                payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                    'center_reference_point_in_x']
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y'] = \
                -1 * payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                    'center_reference_point_in_y']
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z'] = \
                payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                    'center_reference_point_in_z']

        elif dict_ac_exchange['number_of_payload_tubes'] % 2 == 0 and tube % 2 == 0:
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x'] = \
                round(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                          'center_reference_point_in_x']
                      + cabin_delta_in_x[tube], 3)
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y'] = \
                round(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 2)][
                          'center_reference_point_in_y']
                      - payload_tube_description['payload_tube_width']['payload_tube_' + str(tube - 2)] / 2
                      - payload_tube_description['payload_tube_wall_thickness']['payload_tube_' + str(tube - 2)] / 2
                      - payload_tube_description['payload_tube_width']['payload_tube_' + str(tube)] / 2
                      + payload_tube_description['payload_tube_wall_thickness']['payload_tube_' + str(tube)], 3)
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z'] = \
                round(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 2)][
                          'center_reference_point_in_z']
                      + (math.tan(math.radians(dict_mod_config['floor_angle_between_passenger_tubes']))
                         * (abs(
                            payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                                'center_reference_point_in_y'])
                            - abs(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 2)][
                                      'center_reference_point_in_y']))), 3)

        # calculate upper z center reference point of each payload tube
        payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_upper_z'] = \
            round(payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z']
                  - dict_mod_config['cabin_floor_offset']
                  + dict_mod_config['passenger_aisle_standing_height']
                  + (dict_mod_config['system_bay_height_above_cabin'] * convert_meter_to_inch)
                  + payload_tube_description['payload_tube_wall_thickness']['payload_tube_' + str(tube)], 3)

        # calculate lower z center reference point of each payload tube
        uld_type = dict_mod_config['uld_type']['uld_type_ID' + str(tube)]
        # check which type of cargo accommodation is selected and extract dimensions depending on
        if uld_type == 'mixed':
            uld_types = dict_ac_exchange['relevant_cargo_accommodation']['uld_type_ID' + str(tube)]
            for key, value in uld_types.items():
                max_height = max(max_height, round(float(uld_types[key]['height']['value']), 3))
                max_width_at_base = max(max_width_at_base, round(float(uld_types[key]['width_at_base']['value']), 3))
        if uld_type is None:
            max_height = max(max_height,
                             round(dict_mod_config['system_bay_height_below_cargo'] * convert_meter_to_inch, 3))
            max_width_at_base = max(max_width_at_base, round(max_height * 2/3, 3))
        else:
            uld_definition = dict_ac_exchange['relevant_cargo_accommodation']['uld_type_ID' + str(tube)].items()
            container_name = next(iter(uld_definition))[0]
            max_height = round(float(dict_ac_exchange['relevant_cargo_accommodation']['uld_type_ID' + str(tube)]
                                     [container_name]['height']['value']), 3)
            max_width_at_base = round(float(dict_ac_exchange['relevant_cargo_accommodation']['uld_type_ID' + str(tube)]
                                            [container_name]['width_at_base']['value']), 3)

        # Calculation of fuselage height in negative z-direction according to [1].
        distance_to_cargo_bottom = \
            round(max_height + payload_tube_description['payload_tube_floor_thickness']['payload_tube_' + str(tube)]
                  + float(dict_ac_exchange['dict_cs_requirements']['fuselage_design_cs_requirements'][
                              'container_distances']['container_to_ceiling_distance']['value'])
                  + dict_mod_config['cabin_floor_offset'], 3)
        distance_to_lower_compartment_edge = \
            round(float(dict_ac_exchange['dict_cs_requirements']['fuselage_design_cs_requirements'][
                            'container_distances']['container_to_wall_distance']['value'])
                  + 1 / 2 * max_width_at_base, 3)
        payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_lower_z'] = \
            round(-1 * (pow(distance_to_lower_compartment_edge ** 2 + distance_to_cargo_bottom ** 2, 0.5)
                        + payload_tube_description['payload_tube_wall_thickness']['payload_tube_' + str(tube)]), 3)
        payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['distance_to_cargo_bottom'] = \
            distance_to_cargo_bottom

        # check if the force circular cross-section mode is selected
        #  -> if true: -> set lower and upper z to the maximum of both values
        if dict_mod_config['force_circular_cross_section']:
            max_diameter = max(payload_tube_description['payload_tube_width']['payload_tube_' + str(tube)],
                               (payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                                    'center_reference_point_lower_z']
                                + abs(payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                                    'center_reference_point_upper_z'])))
            payload_tube_description['payload_tube_width']['payload_tube_' + str(tube)] = max_diameter
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_upper_z'] =\
                round(max_diameter / 2, 3)
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_lower_z'] =\
                round(-max_diameter / 2, 3)
            runtime_output.warning('Attention! The force circle cross-section mode is selected. \n'
                                   '                                    Payload tube height and width are set to the '
                                   'maximum of both.')

        # Check if the number of selected payload tubes is even
        #  -> if true: -> set the height of opposite payload tubes to the same sizes
        if tube > 0 and dict_ac_exchange['number_of_payload_tubes'] % 2 != 0 and tube % 2 == 0:
            max_tube_height = max(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                                      'center_reference_point_lower_z'],
                                  payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                                      'center_reference_point_lower_z'])
            payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                'center_reference_point_lower_z'] = max_tube_height
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_lower_z'] =\
                max_tube_height
            max_tube_height = max(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                                      'center_reference_point_upper_z'],
                                  payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                                      'center_reference_point_upper_z'])
            payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                'center_reference_point_upper_z'] = max_tube_height
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_upper_z'] =\
                max_tube_height

        # Check if the number of selected payload tubes is even
        #  -> if true: -> set the height of opposite payload tubes to the same sizes
        if tube > 0 and dict_ac_exchange['number_of_payload_tubes'] % 2 == 0 and tube % 2 != 0:
            max_tube_height = max(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                                      'center_reference_point_lower_z'],
                                  payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                                      'center_reference_point_lower_z'])
            payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                'center_reference_point_lower_z'] = max_tube_height
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_lower_z'] =\
                max_tube_height
            max_tube_height = max(payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                                      'center_reference_point_upper_z'],
                                  payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                                      'center_reference_point_upper_z'])
            payload_tube_center_front_reference_points['payload_tube_' + str(tube - 1)][
                'center_reference_point_upper_z'] = max_tube_height
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_upper_z'] =\
                max_tube_height

        # save wall reference points witch are equal to center reference points
        payload_tube_wall_front_reference_points['payload_tube_' + str(tube)]['wall_reference_point_in_x'] = \
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x']
        payload_tube_wall_front_reference_points['payload_tube_' + str(tube)]['wall_reference_point_left_in_y'] = \
            round(payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y']
                  - payload_tube_description['payload_tube_width']['payload_tube_' + str(tube)] / 2, 3)
        payload_tube_wall_front_reference_points['payload_tube_' + str(tube)]['wall_reference_point_right_in_y'] = \
            round(payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y']
                  + payload_tube_description['payload_tube_width']['payload_tube_' + str(tube)] / 2, 3)
        payload_tube_wall_front_reference_points['payload_tube_' + str(tube)]['wall_reference_point_in_z'] = \
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z']

    # calculation of payload tube reference points in the aft
    for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
        payload_tube_center_aft_reference_points['payload_tube_' + str(tube)] = {}
        payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x'] = \
            round(payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x']
                  + payload_tube_description['required_cabin_length_per_tube']['payload_tube_' + str(tube)], 3)

    for _, aft_x_position in payload_tube_center_aft_reference_points.items():
        total_length = total_length + aft_x_position['center_reference_point_in_x']

    # calculation of the x-, y-, z-position and the length of each payload tube
    for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
        payload_tube_wall_aft_reference_points['payload_tube_' + str(tube)] = {}
        payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x'] = \
            round(total_length / dict_ac_exchange['number_of_payload_tubes'], 3)
        payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y'] = \
            payload_tube_center_front_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_y']
        payload_tube_length['payload_tube_' + str(tube)] = \
            round(payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x']
                  - payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                      'center_reference_point_in_x'], 3)
        payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z'] = \
            round(payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                      'center_reference_point_in_z']
                  + (math.tan(math.radians(dict_mod_config['floor_angle_of_passenger_tubes']))
                     * payload_tube_length['payload_tube_' + str(tube)]), 3)
        payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_upper_z'] = \
            round(payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z']
                  + payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                      'center_reference_point_upper_z'], 3)
        payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_lower_z'] = \
            round(payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z']
                  + payload_tube_center_front_reference_points['payload_tube_' + str(tube)][
                      'center_reference_point_lower_z'], 3)

        # wall reference points of the end of payload tube
        payload_tube_wall_aft_reference_points['payload_tube_' + str(tube)]['wall_reference_point_in_x'] = \
            payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_x']
        payload_tube_wall_aft_reference_points['payload_tube_' + str(tube)]['wall_reference_point_left_in_y'] = \
            payload_tube_wall_front_reference_points['payload_tube_' + str(tube)]['wall_reference_point_left_in_y']
        payload_tube_wall_aft_reference_points['payload_tube_' + str(tube)]['wall_reference_point_right_in_y'] = \
            payload_tube_wall_front_reference_points['payload_tube_' + str(tube)]['wall_reference_point_right_in_y']
        payload_tube_wall_aft_reference_points['payload_tube_' + str(tube)]['wall_reference_point_in_z'] = \
            payload_tube_center_aft_reference_points['payload_tube_' + str(tube)]['center_reference_point_in_z']

    ''' save front and aft reference points dictionaries to output dictionary '''
    payload_tube_center_reference_points['payload_tube_center_front_reference_points'] = \
        payload_tube_center_front_reference_points
    payload_tube_center_reference_points['payload_tube_center_aft_reference_points'] = \
        payload_tube_center_aft_reference_points
    payload_tube_wall_reference_points['payload_tube_wall_front_reference_points'] = \
        payload_tube_wall_front_reference_points
    payload_tube_wall_reference_points['payload_tube_wall_aft_reference_points'] = \
        payload_tube_wall_aft_reference_points

    ''' debug print '''
    runtime_output.debug('Debug: The payload_tube_reference_point_calculation function was successfully executed!')

    return payload_tube_center_reference_points, payload_tube_wall_reference_points, payload_tube_length
