# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import sys


def payload_tube_width_calculation(dict_ac_exchange, passenger_cabin_width, payload_tube_wall_thickness,
                                   runtime_output):
    """ Calculates the width of each payload tube including structural wall thicknesses.

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict passenger_cabin_width: Dict containing the passenger cabin with of each payload tube
    :param dict payload_tube_wall_thickness: Dict containing the payload tube structural wall thickness of each tube
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict payload_tube_width: Dict containing the payload tube widths
            float total_width_of_passenger_compartments: Float value containing the total width of passenger compartment
    """
    ''' initialize local parameter '''
    payload_tube_width = {}

    # Estimate the width of each payload tube including structural parts.
    for tube in range(0, dict_ac_exchange['number_of_payload_tubes']):
        max_cabin_width_per_tube = 0
        for key, value in passenger_cabin_width.items():
            max_cabin_width_per_tube = max(max_cabin_width_per_tube, value[tube])

        payload_tube_width['payload_tube_' + str(tube)] = \
            (max_cabin_width_per_tube
             + 2 * payload_tube_wall_thickness['payload_tube_' + str(tube)])

    # Check if the resulting payload tubes widths are not in descending order
    #  -> if true: -> raise an error an abort program.
    if not list(payload_tube_width.values()) == sorted(list(payload_tube_width.values()), reverse=True):
        runtime_output.critical('Error: The selected payload tube abreast numbers lead to alternating payload tube '
                                'sizes. Not a valid configuration. '
                                '\n                                     Program aborted!')
        sys.exit('Alternating payload tube sizes!')

    # Estimate total width of passenger compartment including all required structural wall thicknesses.
    if dict_ac_exchange['number_of_payload_tubes'] == 1:
        reduction_value = 0
    elif dict_ac_exchange['number_of_payload_tubes'] == 2:
        # Check if the resulting payload tube configuration is not symmetrical to the aircraft longitudinal axis
        #  -> if true: -> raise an error and abort Program.
        if not payload_tube_width['payload_tube_0'] == payload_tube_width['payload_tube_1']:
            runtime_output.critical('Error: The payload tube width of payload_tube @ID=0 and payload_tube @ID=1 are not'
                                    ' identical. \n'
                                    '                                     This results in an asymmetrical configuration'
                                    ' to the longitudinal axis. Please change payload tube accommodation parameter in'
                                    ' the module configuration file. \n'
                                    '                                     Program aborted!')
            sys.exit('Unsymmetrical blended wing body configuration!')
        # Else condition: the resulting payload tube configuration is not symmetrical to the aircraft longitudinal axis
        #  -> estimate reduction value for payload tube overlap.
        else:
            reduction_value = list(payload_tube_wall_thickness.items())[-1][-1]
    elif dict_ac_exchange['number_of_payload_tubes'] == 3:
        # Check if the resulting payload tube configuration is not symmetrical to the aircraft longitudinal axis
        #  -> if true: -> raise an error and abort Program.
        if not payload_tube_width['payload_tube_1'] == payload_tube_width['payload_tube_2']:
            runtime_output.critical('Error: The payload tube width of payload_tube @ID=1 and payload_tube @ID=2 are not'
                                    ' identical. \n'
                                    '                                     This results in an asymmetrical configuration'
                                    ' to the longitudinal axis. Please change payload tube accommodation parameter in'
                                    ' the module configuration file. \n'
                                    '                                     Program aborted!')
            sys.exit('Unsymmetrical blended wing body configuration!')
        # Else condition: the resulting payload tube configuration is not symmetrical to the aircraft longitudinal axis
        #  -> estimate reduction value for payload tube overlap.
        else:
            reduction_value = 2 * list(payload_tube_wall_thickness.items())[-1][-1]
    elif dict_ac_exchange['number_of_payload_tubes'] == 4:
        # Check if the resulting payload tube configuration is not symmetrical to the aircraft longitudinal axis
        #  -> if true: -> raise an error and abort Program.
        if not payload_tube_width['payload_tube_0'] == payload_tube_width['payload_tube_1'] \
                or not payload_tube_width['payload_tube_2'] == payload_tube_width['payload_tube_3']:
            runtime_output.critical('Error: The payload tube width of payload_tube @ID=0 and payload_tube @ID=1 or'
                                    ' payload_tube @ID=2 and payload_tube @ID=3 are not identical. \n'
                                    '                                     This results in an asymmetrical configuration'
                                    ' to the longitudinal axis. Please change payload tube accommodation parameter in'
                                    ' the module configuration file. \n'
                                    '                                     Program aborted!')
            sys.exit('Unsymmetrical blended wing body configuration!')
        # Else condition: the resulting payload tube configuration is not symmetrical to the aircraft longitudinal axis
        #  -> estimate reduction value for payload tube overlap.
        else:
            reduction_value = (2 * list(payload_tube_wall_thickness.items())[-1][-1]
                               + list(payload_tube_wall_thickness.items())[0][-1])
    elif dict_ac_exchange['number_of_payload_tubes'] == 5:
        # Check if the resulting payload tube configuration is not symmetrical to the aircraft longitudinal axis
        #  -> if true: -> raise an error and abort Program.
        if not payload_tube_width['payload_tube_1'] == payload_tube_width['payload_tube_2'] \
                or not payload_tube_width['payload_tube_3'] == payload_tube_width['payload_tube_4']:
            runtime_output.critical('Error: The payload tube width of payload_tube @ID=1 and payload_tube @ID=2 or'
                                    ' payload_tube @ID=3 and payload_tube @ID=4 are not identical. \n'
                                    '                                     This results in an asymmetrical configuration'
                                    ' to the longitudinal axis. Please change payload tube accommodation parameter in'
                                    ' the module configuration file. \n'
                                    '                                     Program aborted!')
            sys.exit('Unsymmetrical blended wing body configuration!')
        # Else condition: the resulting payload tube configuration is not symmetrical to the aircraft longitudinal axis
        #  -> estimate reduction value for payload tube overlap.
        else:
            reduction_value = (2 * list(payload_tube_wall_thickness.items())[-1][-1]
                               + 2 * list(payload_tube_wall_thickness.items())[1][-1])
    else:
        runtime_output.critical('Error: The number of selected payload tubes is greater than 5. \n'
                                '                                     The selected blended wing body method for the'
                                ' payload tube design is developed for a number of payload tubes between one and five.'
                                '\n                                     Program aborted!')
        sys.exit('Selected number of payload tubes does not match the selected fuselage design method!')

    total_width_of_passenger_compartments = round(sum(list(payload_tube_width.values())) - reduction_value, 3)

    ''' debug print '''
    runtime_output.debug('Debug: The payload_tube_wall_thickness_calculation function was successfully executed!')

    return payload_tube_width, total_width_of_passenger_compartments
