# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import numpy as np


def constant_section_calculation(dict_ac_exchange, geometry_description, dict_fuselage_deign_output, runtime_output):
    """ Generates ellipse geometry parameter from given dictionary for constant section of current aircraft project.
           Calculation method:
                [1] ellipse equation:
                    * x = xm + a * cos(t) * cos(th) + b * sin(t) * sin(th)
                    * y = ym + a * cos(t) * sin(th) + b * sin(t) * cos(th)

        :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
        :param dict geometry_description: Dict containing geometry parameter and values of current fuselage design
        :param dict dict_fuselage_deign_output: Dict containing output parameter and values of current fuselage design
        :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
        :return dict constant_section: Dict containing the description of the constant section ellipses
    """
    ''' initialize local parameter '''
    convert_inch_to_meter = 0.0254
    constant_section = {"x_direction_coordinates": {},
                        "y_direction_coordinates": {},
                        "z_upward_coordinates": {},
                        "z_downward_coordinates": {}
                        }

    # generate lists for script execution
    center_values = []
    x_direction_values = []
    y_direction_values = []
    z_upwards_values = []
    z_downwards_values = []
    center_coordinates = []
    x_coordinates = []
    y_coordinates = []
    z_upward_coordinates = []
    z_downward_coordinates = []

    ''' convert input values to inch '''
    [center_values.append(round(float(value) / 10 / 2.54, 4))
     for value in dict_ac_exchange['reference_ellipses']['constant_section']['center']]
    [x_direction_values.append(round(float(value) / 10 / 2.54, 4))
     for value in dict_ac_exchange['reference_ellipses']['constant_section']['x_direction']]
    [y_direction_values.append(round(float(value) / 10 / 2.54, 4))
     for value in dict_ac_exchange['reference_ellipses']['constant_section']['y_direction']]
    [z_upwards_values.append(round(float(value) / 10 / 2.54, 4))
     for value in dict_ac_exchange['reference_ellipses']['constant_section']['z_upwards']]
    [z_downwards_values.append(round(float(value) / 10 / 2.54, 4))
     for value in dict_ac_exchange['reference_ellipses']['constant_section']['z_downwards']]

    # Calculate length of cockpit section.
    if dict_ac_exchange['additional_fuselage_length'] is not None:
        constant_section_length = \
            round(geometry_description['fuselage_dimension']['fuselage_length']
                  - dict_fuselage_deign_output['cockpit_section']['total_length']
                  - dict_fuselage_deign_output['tail_section']['total_length'], 2)
    else:
        constant_section_length = \
            round(float(geometry_description['fuselage_dimension']['fuselage_length'])
                  * float(
                dict_ac_exchange['reference_ellipses']['scaling_factors']['percentage_of_constant_section'][0]),
                  4)

    ''' calculate scale factors for cockpit section '''
    # scale for x-direction
    scale_factor_x_direction = constant_section_length / (x_direction_values[-1] - x_direction_values[0])

    # scale for y-direction
    scale_factor_y_direction = \
        (float(geometry_description['fuselage_dimension']['fuselage_width'])
         / (float(dict_ac_exchange['reference_ellipses']['reference_diameter']['reference_y_diameter'][0]) / 10 / 2.54))

    # scale for upper z-direction
    scale_factor_upper_z_direction = \
        2 * (float(geometry_description['fuselage_dimension']['fuselage_upper_height'])
             / (float(dict_ac_exchange['reference_ellipses']['reference_diameter']['reference_z_diameter'][0])
                / 10 / 2.54))

    # scale for lower z-direction
    scale_factor_lower_z_direction = \
        2 * (float(geometry_description['fuselage_dimension']['fuselage_lower_height'])
             / (float(dict_ac_exchange['reference_ellipses']['reference_diameter']['reference_z_diameter'][0])
                / 10 / 2.54))

    ''' calculate new coordinates for constant section ellipses '''
    # center coordinates
    [center_coordinates.append(value * scale_factor_upper_z_direction) for value in center_values]

    # coordinates in x-direction
    [x_coordinates.append(value * scale_factor_x_direction) for value in x_direction_values]

    # coordinates in y-direction
    [y_coordinates.append(value * scale_factor_y_direction) for value in y_direction_values]

    # coordinates in z-upward direction
    [z_upward_coordinates.append(value * scale_factor_upper_z_direction) for value in z_upwards_values]

    # coordinates in z-upward direction
    [z_downward_coordinates.append(value * scale_factor_lower_z_direction) for value in z_downwards_values]

    ''' calculations for graphical visualization '''
    # generate support elements for ellipses
    support_elements_for_ellipses = np.linspace(0, np.pi, 100)

    # calculation of ellipses for constant section in according to [1]
    i = 0
    while i < len(center_coordinates):
        # x-direction coordinates for ellipses
        ellipse = []
        [ellipse.append(x_coordinates[i]) for _ in support_elements_for_ellipses]
        constant_section['x_direction_coordinates'].update({'ellipse_number_' + str(i + 1): ellipse})

        # y-direction coordinates for ellipses
        ellipse = []
        [ellipse.append(0 + y_coordinates[i] * np.cos(support_element) * np.cos(0) + z_upward_coordinates[i]
                        * np.sin(support_element) * np.sin(0)) for support_element in support_elements_for_ellipses]
        constant_section['y_direction_coordinates'].update({'ellipse_number_' + str(i + 1): ellipse})

        # z-upward coordinates for ellipses
        ellipse = []
        [ellipse.append(center_coordinates[i] + y_coordinates[i] * np.cos(support_element) * np.sin(0)
                        + z_upward_coordinates[i] * np.sin(support_element) * np.cos(0))
         for support_element in support_elements_for_ellipses]
        constant_section['z_upward_coordinates'].update({'ellipse_number_' + str(i + 1): ellipse})

        # z-downward coordinates for ellipses
        ellipse = []
        [ellipse.append(center_coordinates[i] + y_coordinates[i] * np.cos(support_element) * np.sin(0)
                        + z_downward_coordinates[i] * np.sin(support_element) * np.cos(0))
         for support_element in support_elements_for_ellipses]
        constant_section['z_downward_coordinates'].update({'ellipse_number_' + str(i + 1): ellipse})

        i += 1

    constant_section['total_length'] = constant_section_length

    ''' prepare data for fuselage sections of aircraft exchange file '''
    constant_section['output_data'] = {'origin': {},
                                       'upper_height': {},
                                       'lower_height': {},
                                       'width': {}}

    # Loop across all cockpit ellipsis
    i = 0
    while i < len(center_coordinates):
        constant_section['output_data']['origin']['ellipses_' + str(i)] = {'x': float(), 'y': float(), 'z': float()}
        constant_section['output_data']['origin']['ellipses_' + str(i)]['x'] = \
            round(x_coordinates[i] * convert_inch_to_meter, 9)
        constant_section['output_data']['origin']['ellipses_' + str(i)]['y'] = 0.0
        constant_section['output_data']['origin']['ellipses_' + str(i)]['z'] = \
            round(center_coordinates[i] * convert_inch_to_meter, 9)
        constant_section['output_data']['upper_height']['ellipses_' + str(i)] = \
            round(z_upward_coordinates[i] * convert_inch_to_meter, 9)
        constant_section['output_data']['lower_height']['ellipses_' + str(i)] = \
            round(abs(z_downward_coordinates[i] * convert_inch_to_meter), 9)
        constant_section['output_data']['width']['ellipses_' + str(i)] = \
            round(abs(2 * y_coordinates[i] * convert_inch_to_meter), 9)
        i += 1

    ''' debug print '''
    runtime_output.debug('Debug: The constant_section_calculation function was successfully executed!')

    return constant_section
