# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import sys


def estimate_cabin_width(dict_ac_exchange, dict_mod_config, geometry_description, runtime_output):
    """ Estimate cabin width from given class definition.

        :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
        :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
        :param dict geometry_description: Dict containing geometry parameter and values of current fuselage design
        :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
        :return dict geometry_description: Dict containing geometry parameter and values of current fuselage design
    """
    ''' initialize local parameter '''
    # Class description parameter.
    lowest_passenger_class = dict_ac_exchange['relevant_classes']['passenger_classes']['class_names'][-1]
    lowest_class_abreast_string = \
        dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class]['seat_abreast'].split('_')
    lowest_class_abreast = [int(element) for element in lowest_class_abreast_string]
    lowest_class_seat_space = \
        dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class]['seat_space']

    # Aisle description parameter.
    passenger_aisle_width = dict_mod_config['passenger_aisle_width']
    if passenger_aisle_width < int(
            dict_ac_exchange['dict_cs_requirements']['fuselage_design_cs_requirements'][
                'aisle']['passenger_aisle']['minimum_required_width_below_25_inch']['value']):
        runtime_output.critical('Error: The selected aisle width of the passenger compartment is smaller than required '
                                'by the CS-25.\n                                     Program aborted!')
        sys.exit('Selected passenger aisle width is too low.')

    ''' estimate total passenger cabin width '''
    # Get passenger seat parameter from left side.
    lowest_class_seat_width_left_side =\
        round(float(dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class][
                        'passenger_seats']['seat_left_side']['width']['value']), 2)
    lowest_class_inner_armrest_width_left_side = \
        round(float(dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class][
                        'passenger_seats']['seat_left_side']['width_inner_armrest']['value']), 2)
    lowest_class_outer_armrest_width_left_side = \
        round(float(dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class][
                        'passenger_seats']['seat_left_side']['width_outer_armrest']['value']), 2)

    # Get passenger seat parameter from right side.
    lowest_class_seat_width_right_side = \
        round(float(dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class][
                        'passenger_seats']['seat_right_side']['width']['value']), 2)
    lowest_class_inner_armrest_width_right_side = \
        round(float(dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class][
                        'passenger_seats']['seat_right_side']['width_inner_armrest']['value']), 2)
    lowest_class_outer_armrest_width_right_side = \
        round(float(dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class][
                        'passenger_seats']['seat_right_side']['width_outer_armrest']['value']), 2)

    # Number of seats per aircraft side.
    number_of_seats_left_side = lowest_class_abreast[0]
    number_of_seats_right_side = lowest_class_abreast[-1]

    # Width of seat row per aircraft side.
    seat_bench_width_left_side = round(number_of_seats_left_side * lowest_class_seat_width_left_side
                                       + 2 * lowest_class_outer_armrest_width_left_side, 2)
    if number_of_seats_left_side > 1:
        seat_bench_width_left_side = round(seat_bench_width_left_side
                                           + ((number_of_seats_left_side - 1)
                                              * lowest_class_inner_armrest_width_left_side), 2)

    seat_bench_width_right_side = round(number_of_seats_right_side * lowest_class_seat_width_right_side
                                        + 2 * lowest_class_outer_armrest_width_right_side, 2)
    if number_of_seats_right_side > 1:
        seat_bench_width_right_side = round(seat_bench_width_right_side
                                            + ((number_of_seats_right_side - 1)
                                               * lowest_class_inner_armrest_width_right_side), 2)

    # Calculates cabin width of current aircraft project.
    total_cabin_width = round(passenger_aisle_width + seat_bench_width_left_side + seat_bench_width_right_side
                              + 2 * lowest_class_seat_space, 2)

    # Check for wide-body-aircraft
    #  -> if true -> add width of center row and one additional passenger aisle to total cabin width.
    if dict_ac_exchange['fuselage_type'] == 'wide_body':
        if len(lowest_class_abreast) < 3:
            runtime_output.critical('Error: The selected fuselage type is a wide-body aircraft, but the selected number'
                                    ' of seats abreast in the lowest passenger class is for a single-aisle aircraft.\n'
                                    '                                     Program aborted!')
            sys.exit('Selected number of seat abreast does not match the selected fuselage type.')

        # Get passenger seat parameter from right side.
        lowest_class_seat_width_in_the_center = \
            round(float(dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class][
                            'passenger_seats']['seat_in_the_center']['width']['value']), 2)
        lowest_class_inner_armrest_width_in_the_center = \
            round(float(dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class][
                            'passenger_seats']['seat_in_the_center']['width_inner_armrest']['value']), 2)
        lowest_class_outer_armrest_width_in_the_center = \
            round(float(dict_ac_exchange['relevant_classes']['passenger_classes'][lowest_passenger_class][
                            'passenger_seats']['seat_in_the_center']['width_outer_armrest']['value']), 2)

        number_of_seats_in_the_center = lowest_class_abreast[1]
        seat_bench_width_in_the_center = round(number_of_seats_in_the_center * lowest_class_seat_width_in_the_center
                                               + 2 * lowest_class_outer_armrest_width_in_the_center, 2)
        if number_of_seats_in_the_center > 1:
            seat_bench_width_in_the_center = round(seat_bench_width_in_the_center
                                                   + ((number_of_seats_in_the_center - 1)
                                                      * lowest_class_inner_armrest_width_in_the_center), 2)

        total_cabin_width = round(total_cabin_width + passenger_aisle_width + seat_bench_width_in_the_center, 2)

    ''' store total cabin width to geometry description dictionary '''
    geometry_description['total_cabin_width'] = total_cabin_width

    ''' debug print '''
    runtime_output.debug('Debug: The estimate_cabin_width function was successfully executed!')

    return geometry_description
