# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing functions for data preprocessing."""
# Import standard modules.
import importlib
import sys

# Import own modules.
from src.readlayertext import read_energy_carrier
from datapreprocessingmodule import get_paths_and_names
from datapreprocessingmodule import read_routing_values_from_xml


def data_preprocessing(module_configuration_file, argv):
    """Conduct data preprocessing.

    This function provides data preprocessing functionalities. It sets up the necessary data and imports
    relevant modules. The importlib module is used to dynamically import necessary modules.

    The output dictionary 'preprocessing_dict' contains the following values:
        - 'layer_1': First routing layer (main gear mounting position) (str)
        - 'layer_2': Second routing layer (calculation method fidelity) (str)
        - 'layer_3': Third routing layer (calculation method) (str)
        - 'user_layer': Last routing layer (fuel type) (user layer) (str)
        - 'tool_level': Tool level of current tool (str)
        - 'module_import_name': Dynamic string for dynamically generated module import name based on layers (str)
        - 'module_name': Module name (name of the module configuration file without its file extension) (str)
        - 'func_run_method': Reference to 'run_method' function
        - 'func_method_data_preprocessing': Reference to 'run_method' function
        - 'func_method_data_postprocessing': Reference to 'method_data_preprocessing' function
        - 'func_user_method_data_input_preparation': Reference to 'user_method_data_input_preparation' function
        - 'func_user_method_data_output_preparation': Reference to 'user_method_data_output_preparation' function

    :param str module_configuration_file: Name of module configuration file
    :param list argv: List with optional input arguments
    :returns:
        - dict paths_and_names: Dictionary containing system paths and ElementTrees
        - dict preprocessing_dict: Dictionary containing data preprocessing results
        - logging.Logger runtime_output: Logging object used for capturing log messages in the module
    """

    """Get paths, names, and xml trees for module configuration and aircraft exchange file."""
    # Call 'get_paths_and_names' function to obtain various paths and names.
    paths_and_names, runtime_output = get_paths_and_names(module_configuration_file, argv)
    # Note: It is the exclusive responsibility of the module manager to modify the following information!
    # Create layer description dictionary according to the number of individual layers. The dictionary associates
    # layers with their respective XML paths and expected data types according to the following scheme:
    #   layer_description_dict = {'layer_1': [path, expected data type], 'layer_2': [...]}
    # If any information cannot be directly extracted from a specific aircraft exchange file path, please write 'None'
    # and manually add the missing value afterward.
    aircraft_exchange_tmp_path = 'aircraft_exchange_file/requirements_and_specifications/design_specification/'
    module_configuration_tmp_path = 'module_configuration_file/program_settings/configuration/'
    layer_description_dict = {
        'layer_1': [aircraft_exchange_tmp_path + 'configuration/undercarriage_definition/main_gear_mounting/value', float],  # noPep8 e501
        'layer_2': [module_configuration_tmp_path + 'fidelity_name/value', str],
        'layer_3': [module_configuration_tmp_path + 'method_name/value', str],
        'user_layer': [None, str]
    }

    """ Extract data from aircraft exchange and module configuration file."""
    # Extract root and path to aircraft exchange file and write key data to aircraft exchange file.
    root_of_aircraft_exchange_tree = paths_and_names['root_of_aircraft_exchange_tree']
    root_of_module_configuration_file = paths_and_names['root_of_module_config_tree']
    # Extract data from *.xml files based on the provided layer description (if no path information given ('None'),
    # the entry has to be specified manually afterward). The result is stored in the 'preprocessing_dict' dictionary.
    # It has the following output format (all values are strings):
    # dict_out = {'layer_1': value, 'layer_2': value, 'layer_3': value, 'user_layer': value, 'tool_level': value}
    preprocessing_dict = read_routing_values_from_xml(layer_description_dict, root_of_aircraft_exchange_tree,
                                                      root_of_module_configuration_file, runtime_output)

    # Manual specification of missing layer values ('None' entry layer).
    preprocessing_dict['user_layer'] = read_energy_carrier(root_of_aircraft_exchange_tree, runtime_output)

    """Prepare and import modules."""
    # Generate a dynamic import name 'module_import_name' for the selected calculation method modules based on the
    # provided layer values according to the following scheme:
    # 'src.[value of layer_1].[value of layer_2].[value of layer_3]'
    module_import_name = 'src'
    for _, value in list(preprocessing_dict.items())[:-2]:
        module_import_name += '.' + value
    # Create import commands by appending the python file name (incl. sub-folders, if necessary) to the generated
    # 'module_import_name'. E.g., the import command for the module import from the 'usermethoddatapreparation.py' file
    # is as follows:
    #   'src.[value of layer_1].[value of layer_2].[value of layer_3].usermethoddatapreparation'
    # The import command for the module import from the 'methodplot.py' file in the 'general' folder is as follows:
    #   'src.[value of layer_1].[value of layer_2].[value of layer_3].general.methodplot'
    # This step is executed for the following python files:
    #   * 'usermethoddatapreparation.py'
    #   * 'methodplot.py'
    #   * 'methodhtmlreport.py'
    #   * 'methodxmlreport.py'
    #   * 'methodtexoutput'.py'
    import_command_user_method_data_preparation = module_import_name + '.usermethoddatapreparation'
    import_command_user_method_plot = module_import_name + '.general.methodplot'
    import_command_user_method_html_report = module_import_name + '.general.methodhtmlreport'
    import_command_user_method_xml_export = module_import_name + '.general.methodxmlexport'
    import_command_user_method_tex_output = module_import_name + '.general.methodtexoutput'

    # Add module name and tool level to the preprocessing_dict.
    preprocessing_dict['module_import_name'] = module_import_name
    preprocessing_dict['module_name'] = module_configuration_file[:-9]

    # Dynamically import modules and functions based on the generated import commands.
    try:
        # Import functions from the specified modules.
        import_user_method_data_preparation = importlib.import_module(import_command_user_method_data_preparation)
        import_user_method_plot = importlib.import_module(import_command_user_method_plot)
        import_user_method_html_report = importlib.import_module(import_command_user_method_html_report)
        import_user_method_xml_export = importlib.import_module(import_command_user_method_xml_export)
        import_user_method_tex_output = importlib.import_module(import_command_user_method_tex_output)
        # Save the imported functions as variables in the 'preprocessing_dict' dictionary.
        preprocessing_dict['func_user_method_data_input_preparation'] \
            = import_user_method_data_preparation.user_method_data_input_preparation
        preprocessing_dict['func_user_method_data_output_preparation'] \
            = import_user_method_data_preparation.user_method_data_output_preparation
        preprocessing_dict['func_user_method_plot'] = import_user_method_plot.method_plot
        preprocessing_dict['func_user_method_html_report'] = import_user_method_html_report.method_html_report
        preprocessing_dict['func_user_method_xml_export'] = import_user_method_xml_export.method_xml_export
        preprocessing_dict['func_user_method_tex_output'] = import_user_method_tex_output.method_tex_output
    # Exception handling for module import error.
    except ModuleNotFoundError as module_import_error:
        runtime_output.critical('Error: ' + str(module_import_error) + ' found in ' + preprocessing_dict['module_name']
                                + '.\n'
                                + '                                     Program aborted.')
        sys.exit(1)

    return paths_and_names, preprocessing_dict, runtime_output
