# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

def estimate_positions(dict_ac_exchange, dict_mod_config, landing_gear_description, runtime_output):
    """ This function estimates start values of the x-positions for nose and main landing gear in meter.

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    """
    # Initialize local parameter.
    landing_gear_positions = {}
    landing_gear_limitations = {}

    ''' Estimate nose gear starting position in x-direction. '''
    # check if the minimum distance between main landing gear and aft most center of gravity is not existing
    #  -> if true: -> set initial distance from statistic (2 meter)
    if not ('landing_gear_limitations' in landing_gear_description) and not \
            ('minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity' in landing_gear_limitations):
        landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity'] = 2.0
    else:
        landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity'] = \
            landing_gear_description['landing_gear_limitations']['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity']  # noPep8 e501

    # check if the key 'nose_gear_x_position' not exists in the landing gear description dictionary
    #  -> if true: -> Estimate starting point for nose gear position
    if not ('landing_gear_positions' in landing_gear_description) and \
            not ('nose_gear_x_position' in landing_gear_positions):
        # check if the use specific nose gear starting x position switch is selected
        #  -> if true: -> use specific nose gear x position from module configuration file as starting point
        if dict_mod_config['use_specific_nose_gear_starting_x_position']:
            landing_gear_positions['nose_gear_x_position'] = dict_mod_config['specific_nose_gear_starting_x_position']
        # else condition: The use specific nose gear starting x position switch is not selected
        #  -> estimate the nose gear starting position in x-direction
        else:
            # Get the first key-value pair
            _, payload_tube_value = \
                next(iter(dict_ac_exchange['payload_tube_front_reference_point_in_x_direction'].items()), (None, None))
            _, landing_gear_position_value = \
                next(iter(dict_ac_exchange['landing_gear_positions_in_x_direction'].items()), (None, None))
            # Check if the nose gear position and the payload tube reference point in x direction are None.
            #  -> if true: -> Set starting point of nose gear to 5.0 meter.
            if payload_tube_value is None and landing_gear_position_value is None:
                landing_gear_positions['nose_gear_x_position'] = 5.0

            # Else if condition: Check if the nose gear position is None.
            #  -> if true: -> Use passenger tube front reference coord as starting point.
            elif landing_gear_position_value is None:
                if 5.0 <= payload_tube_value <= 6.0:
                    landing_gear_positions['nose_gear_x_position'] = payload_tube_value
                else:
                    landing_gear_positions['nose_gear_x_position'] = 5.0

            # Else condition: The nose gear position is not zero -> Use as starting point for further estimation.
            else:
                landing_gear_positions['nose_gear_x_position'] = landing_gear_position_value

    # Else condition: The key 'nose_gear_x_position' already exists in the landing gear description dictionary
    #  -> Use for further estimation
    else:
        landing_gear_positions['nose_gear_x_position'] = \
            landing_gear_description['landing_gear_positions']['nose_gear_x_position']

    ''' Estimate main gear starting position in x-direction. '''
    # check if the key 'main_gear_x_position' not exists in the landing gear description dictionary
    #  -> if true: -> Estimate starting point for main gear position
    if not ('landing_gear_positions' in landing_gear_description) and \
            not ('main_gear_x_position' in landing_gear_positions):
        _, landing_gear_position_value = \
            next(reversed(list(dict_ac_exchange['landing_gear_positions_in_x_direction'].items())), (None, None))
        _, wing_reference_point_value = \
            next(reversed(list(dict_ac_exchange['wing_section_position_in_x'].items())), (None, None))
        _, wing_chord_length_value = \
            next(reversed(list(dict_ac_exchange['wing_chord_length_on_section'].items())), (None, None))
        # Check if the main gear position and the wing center reference point
        #  and the o most aft CoG position in x direction are None.
        #   -> if true: -> Check which kind of fuselage type is selected and set type specific start values.
        if wing_reference_point_value is None and landing_gear_position_value is None and \
                dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] is None:
            # Check if the fuselage type of current project is single aisle
            #  -> if true: -> set starting point of main gear position to 20 meter
            if dict_ac_exchange['fuselage_type'] == 'single_aisle':
                landing_gear_positions['main_gear_x_position'] = 19.0
                landing_gear_positions['temporary_center_of_gravity'] = \
                    landing_gear_positions['main_gear_x_position'] \
                    - landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity']  # noPep8 e501
            # Else condition: The fuselage type of current project is weight body
            #  -> set starting point of main gear position to 35 meter
            else:
                landing_gear_positions['main_gear_x_position'] = 32.0
                landing_gear_positions['temporary_center_of_gravity'] = \
                    landing_gear_positions['main_gear_x_position'] \
                    - landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity']  # noPep8 e501

        # Else if condition: Check if the main gear position and the wing reference point in x direction are None.
        #  -> if true: -> Set starting point of main gear to most aft CoG position plus 3 meter.
        elif wing_reference_point_value is None and landing_gear_position_value is None:
            landing_gear_positions['main_gear_x_position'] =\
                dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] \
                + landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity']

        # Else if condition: Check if the main gear and the aft most center of gravity position is None.
        #  -> if true: -> Use wing center reference coord plus 75 percent of center wing chord length as starting point.
        elif landing_gear_position_value is None \
                and dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] is None:
            landing_gear_positions['main_gear_x_position'] = (wing_reference_point_value
                                                              + wing_chord_length_value
                                                              * 0.75)
            landing_gear_positions['temporary_center_of_gravity'] = \
                landing_gear_positions['main_gear_x_position'] \
                - landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity']

        # Else if condition: Check if the main gear position is None.
        #  -> if true: -> use position of aft most center of gravity position plus safety margin for starting point
        elif landing_gear_position_value is None:
            landing_gear_positions['main_gear_x_position'] = \
                dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] \
                + landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity']

        # Else if condition: Check if the most aft center of gravity position is None.
        #  -> if true: -> use position of main landing gear minus safety margin for starting point
        elif dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] is None:
            landing_gear_positions['main_gear_x_position'] = landing_gear_position_value
            if not ('temporary_center_of_gravity' in landing_gear_positions):
                landing_gear_positions['temporary_center_of_gravity'] = \
                    landing_gear_positions['main_gear_x_position'] - 1  # noPep8 e501
        # Else condition: The main gear position is not zero -> Use as starting point for further estimation.
        else:
            landing_gear_positions['main_gear_x_position'] = landing_gear_position_value
            if not ('temporary_center_of_gravity' in landing_gear_positions):
                landing_gear_positions['temporary_center_of_gravity'] = \
                    dict_ac_exchange['most_aft_center_of_gravity_in_x_direction']

    # Else condition: The key 'main_gear_x_position' already exists in the landing gear description dictionary
    #  -> Use for further estimation
    else:
        landing_gear_positions['main_gear_x_position'] = \
            landing_gear_description['landing_gear_positions']['main_gear_x_position']
        if 'temporary_center_of_gravity' in landing_gear_description['landing_gear_positions']:
            landing_gear_positions['temporary_center_of_gravity'] = \
                landing_gear_description['landing_gear_positions']['temporary_center_of_gravity']

    # Prepare output dictionary.
    landing_gear_positions = {key: round(value, 4) for key, value in landing_gear_positions.items()}
    landing_gear_description['landing_gear_positions'] = landing_gear_positions

    runtime_output.debug('estimate_positions successfully executed!')

    return landing_gear_description
