# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import math
import sys


def estimate_distances(dict_ac_exchange, dict_mod_config, landing_gear_description, runtime_output):
    """ This function estimates lever arms for load estimation of nose and main landing gear in meter.

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    """
    # initialize local parameter
    lever_arms = {}
    tail_tipping_point = {}
    fuselage_parameter = {}
    # ToDo: gravitational_acceleration should be in a separate python standard package
    gravitational_acceleration = float(9.80665)
    delta_height_fuselage_center_line_to_center_of_gravity = 0

    # check if the tail tipping point dictionary not exist -> if true: -> initialize dictionary with input values
    if not ('tail_tipping_point' in landing_gear_description):
        tail_tipping_point['maximum_rotation_angle_landing'] = dict_mod_config['tail_strike_limit']
        tail_tipping_point['maximum_rotation_angle_take_off'] = dict_mod_config['tail_strike_limit']
    # Else condition: The tail tipping point dictionary exist -> use content for further estimation
    else:
        tail_tipping_point['maximum_rotation_angle_landing'] \
            = landing_gear_description['tail_tipping_point']['maximum_rotation_angle_landing']
        tail_tipping_point['maximum_rotation_angle_take_off'] \
            = landing_gear_description['tail_tipping_point']['maximum_rotation_angle_take_off']

    ''' Estimation of the lever arms. '''
    try:
        # Check if the nose gear position in x direction not equal to the x-position of main landing gear.
        #  -> if true: -> Estimate lever arms for landing gear load estimation.
        if not landing_gear_description['landing_gear_positions']['nose_gear_x_position'] == \
               landing_gear_description['landing_gear_positions']['main_gear_x_position']:
            lever_arms['distance_between_nose_and_main_gear'] = \
                abs(landing_gear_description['landing_gear_positions']['main_gear_x_position']
                    - landing_gear_description['landing_gear_positions']['nose_gear_x_position'])

            # Check if the most forward position of the center of gravity is zero.
            #  -> if true: -> Use main gear x-position minus 5 meter for further estimation.
            if dict_ac_exchange['most_forward_center_of_gravity_in_x_direction'] is None:
                lever_arms['distance_between_nose_gear_and_foremost_center_of_gravity'] = \
                    abs((landing_gear_description['landing_gear_positions']['main_gear_x_position'] - 2)
                        - landing_gear_description['landing_gear_positions']['nose_gear_x_position'])

            # Else condition: The most forward position of the center of gravity is not zero.
            #  -> Use for further estimation.
            else:
                lever_arms['distance_between_nose_gear_and_foremost_center_of_gravity'] = \
                    abs(dict_ac_exchange['most_forward_center_of_gravity_in_x_direction']
                        - landing_gear_description['landing_gear_positions']['nose_gear_x_position'])

            # Check if the rearmost position of the center of gravity is zero.
            #  -> if true: -> Use main gear x-position minus 3 meter for further estimation.
            if dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] is None:
                lever_arms['distance_between_nose_gear_and_rearmost_center_of_gravity'] = \
                    abs((landing_gear_description['landing_gear_positions']['main_gear_x_position'] - 1)
                        - landing_gear_description['landing_gear_positions']['nose_gear_x_position'])
                lever_arms['distance_between_main_gear_and_rearmost_center_of_gravity'] = 1

            # Else condition: The rearmost position of the center of gravity is not zero.
            #  -> Use for further estimation.
            else:
                lever_arms['distance_between_nose_gear_and_rearmost_center_of_gravity'] = \
                    abs(dict_ac_exchange['most_aft_center_of_gravity_in_x_direction']
                        - landing_gear_description['landing_gear_positions']['nose_gear_x_position'])

                if not ('lever_arms' in landing_gear_description) and \
                        not ('distance_between_main_gear_and_rearmost_center_of_gravity' in lever_arms):
                    lever_arms['distance_between_main_gear_and_rearmost_center_of_gravity'] = 1
                else:
                    lever_arms['distance_between_main_gear_and_rearmost_center_of_gravity'] = \
                        landing_gear_description['lever_arms']['distance_between_main_gear_and_rearmost_center_of_gravity']  # noPep8 e501

        # Else condition: Nose gear and main landing gear have the same position in x-direction.
        #  -> Raise an error and abort program.
        else:
            raise ValueError('The nose and main gear are on the same x-position. \n '
                             '      Lever arm estimation for landing gear design not possible. Program aborted!')

        ''' Estimation of vertical distance between ground and center of gravity in accordance to Norman S. Currey. '''
        # Check if the key 'tail_tipping_point' is existing in the landing_gear_description dictionary
        #   -> if not: -> Empirical estimation of distances and tipping point
        if not ('tail_tipping_point' in landing_gear_description):
            # Get the first key-value pair
            _, payload_tube_value = \
                next(iter(dict_ac_exchange['payload_tube_aft_reference_point_in_x_direction'].items()), (None, None))
            # Check if the x-position of aft-position of payload tube is equal to None
            #   -> if true: -> set initial values for the tail tipping distance depending on the selected fuselage type.
            if payload_tube_value is None:
                # Check which kind of fuselage is selected and set initial values for the distance depending on.
                if dict_ac_exchange['fuselage_type'] == 'single_aisle':
                    lever_arms['distance_between_main_landing_gear_and_tail_tipping'] = 15
                else:
                    lever_arms['distance_between_main_landing_gear_and_tail_tipping'] = 25

            # Else condition: The x-position of aft-position of payload tube is not equal to zero
            #  -> use for estimation of tail tipping distance
            else:
                lever_arms['distance_between_main_landing_gear_and_tail_tipping'] = \
                    abs(payload_tube_value - landing_gear_description['landing_gear_positions']['main_gear_x_position'])

            # Estimate tail tipping point in x-direction
            tail_tipping_point['tail_tipping_point_in_x_direction'] = \
                round(lever_arms['distance_between_main_landing_gear_and_tail_tipping']
                      + landing_gear_description['landing_gear_positions']['main_gear_x_position'], 4)

            # Estimate tail tipping point in z-direction.
            # Get the first key-value pair
            _, fuselage_height_value = \
                next(iter(dict_ac_exchange['fuselage_section_total_height'].items()), (None, None))
            if fuselage_height_value is None and dict_ac_exchange['fuselage_type'] == 'single_aisle':
                fuselage_parameter['fuselage_height'] = 3.8
            elif fuselage_height_value is None and dict_ac_exchange['fuselage_type'] == 'wide_body':
                fuselage_parameter['fuselage_height'] = 5.8
            else:
                fuselage_parameter['fuselage_height'] = max(dict_ac_exchange['fuselage_section_total_height'].values())

            # Empirical estimation of tail tipping z-position
            #  -> statistical: z-position is 30 percent of fuselage diameter below the fuselage center line
            tail_tipping_point['tail_tipping_point_in_z_direction'] = \
                round(-0.3 * fuselage_parameter['fuselage_height'], 4)

        # Else condition: The tail tipping point dictionary exist -> use content for further estimation
        else:
            lever_arms['distance_between_main_landing_gear_and_tail_tipping'] \
                = (landing_gear_description['tail_tipping_point']['tail_tipping_point_in_x_direction']
                   - landing_gear_description['landing_gear_positions']['main_gear_x_position'])
            fuselage_parameter['fuselage_height'] = landing_gear_description['fuselage_parameter']['fuselage_height']
            tail_tipping_point['tail_tipping_point_in_x_direction'] = \
                landing_gear_description['tail_tipping_point']['tail_tipping_point_in_x_direction']
            tail_tipping_point['tail_tipping_point_in_z_direction'] =\
                landing_gear_description['tail_tipping_point']['tail_tipping_point_in_z_direction']

        # Estimate distance between tipping point in z-direction and ground
        #  -> Trigonometrical function: delta_z = tan(alpha) * adjacent leg
        delta_height_on_tipping_point = \
            round(abs(math.tan(math.radians(tail_tipping_point['maximum_rotation_angle_landing']))
                      * lever_arms['distance_between_main_landing_gear_and_tail_tipping']), 4)

        # check if a strut suspension system should be used
        #  -> if true: -> estimate length la and lb according to Sforza figure 7.10
        if dict_mod_config['strut_suspension']:
            length_a = (dict_mod_config['strut_suspension_travel']
                        / math.sin(math.radians(tail_tipping_point['maximum_rotation_angle_take_off'])))
            length_b = (delta_height_on_tipping_point
                        / math.sin(math.radians(tail_tipping_point['maximum_rotation_angle_take_off'])))

        # Else condition: A strut suspension system should not be used
        #  -> estimate only the length lb according to Sforza figure 7.10
        else:
            length_a = 0.0
            length_b = (delta_height_on_tipping_point
                        / math.sin(math.radians(tail_tipping_point['maximum_rotation_angle_take_off'])))

        # check if all necessary inputs are given in the aircraft exchange file
        # -> if true: -> estimate maximum rotation time during take-off in seconds according to Torenbeek equation 10-6
        # (bracket term)
        if dict_ac_exchange['velocity_lift_off'] is not None \
                and dict_ac_exchange['lift_coefficient_lift_off'] is not None \
                and dict_ac_exchange['lift_curve_slope_take_off'] is not None:
            tail_tipping_point['maximum_rotation_time_take_off'] = \
                round((2 * length_a / dict_ac_exchange['velocity_lift_off'])
                      + (math.sqrt((length_b * dict_ac_exchange['lift_coefficient_lift_off'])
                                   / (gravitational_acceleration * dict_ac_exchange['lift_curve_slope_take_off']))), 4)

        # Else condition: Not all necessary inputs are given in the aircraft exchange file
        #  -> use initial value for the maximum rotation time in seconds in accordance to Sforza equation 7.17
        else:
            tail_tipping_point['maximum_rotation_time_take_off'] = 1.5

        # set maximum allowed pitch rate during lift-off in degrees per second according to Torenbeek equation 10-6
        if dict_ac_exchange['fuselage_type'] == 'single_aisle':
            tail_tipping_point['maximum_allowed_pitch_rate_in_degree_per_second_lift_off'] = 3.5

        if dict_ac_exchange['fuselage_type'] == 'wide_body':
            tail_tipping_point['maximum_allowed_pitch_rate_in_degree_per_second_lift_off'] = 4.5

        # estimate maximum rotation angle during take-off according to Torenbeek equation 10-6
        tail_tipping_point['maximum_rotation_angle_take_off'] = \
            round(dict_ac_exchange['angle_off_attack_lift_off']
                  + tail_tipping_point['maximum_allowed_pitch_rate_in_degree_per_second_lift_off']
                  * tail_tipping_point['maximum_rotation_time_take_off'], 4)

        # Estimate distance between rearmost global center of gravity in z-direction and fuselage center line
        # Get the first key-value pair
        fuselage_reference_point_z_key, fuselage_reference_point_z_value = \
            next(iter(dict_ac_exchange['fuselage_reference_point_in_z_direction'].items()), (None, None))
        if fuselage_reference_point_z_value is not None \
                and dict_ac_exchange['center_of_gravity_in_z'] is not None:
            if fuselage_reference_point_z_value <= dict_ac_exchange['center_of_gravity_in_z']:
                delta_height_fuselage_center_line_to_center_of_gravity = \
                    abs(dict_ac_exchange['center_of_gravity_in_z']
                        - fuselage_reference_point_z_value)
            else:
                delta_height_fuselage_center_line_to_center_of_gravity = \
                    abs(fuselage_reference_point_z_value
                        - dict_ac_exchange['center_of_gravity_in_z'])
        elif dict_ac_exchange['fuselage_type'] == 'single_aisle':
            delta_height_fuselage_center_line_to_center_of_gravity = 0.5
        elif dict_ac_exchange['fuselage_type'] == 'wide_body':
            delta_height_fuselage_center_line_to_center_of_gravity = 1

        # Check if the limitation flag is currently set in the limitation dictionary
        #  -> if true: -> use existing vertical distance between ground and center of gravity
        if 'landing_gear_limitations' in landing_gear_description \
                and 'limitation_flag' in landing_gear_description['landing_gear_limitations']:
            lever_arms['vertical_distance_between_ground_and_aft_center_of_gravity'] = \
                landing_gear_description['lever_arms']['vertical_distance_between_ground_and_aft_center_of_gravity']
            if dict_mod_config['strut_suspension']:
                tail_tipping_point['delta_height_on_tipping_point'] = (delta_height_on_tipping_point
                                                                       - dict_mod_config['strut_suspension_travel'])
            else:
                tail_tipping_point['delta_height_on_tipping_point'] = delta_height_on_tipping_point
        # Else condition: The limitation flag is currently not set in the limitation dictionary
        #  -> estimate vertical distance between ground and center of gravity
        else:
            # Estimation of vertical distance between ground and center of gravity in accordance to Norman S. Currey
            # Check if a strut suspension system should be used
            #  -> if true: -> subtract the suspension travel from vertical distance between ground and center of gravity
            if dict_mod_config['strut_suspension']:
                lever_arms['vertical_distance_between_ground_and_aft_center_of_gravity'] = \
                    delta_height_on_tipping_point + delta_height_fuselage_center_line_to_center_of_gravity \
                    - dict_mod_config['strut_suspension_travel']
                tail_tipping_point['delta_height_on_tipping_point'] = (delta_height_on_tipping_point
                                                                       - dict_mod_config['strut_suspension_travel'])
            else:
                lever_arms['vertical_distance_between_ground_and_aft_center_of_gravity'] = \
                    delta_height_on_tipping_point + delta_height_fuselage_center_line_to_center_of_gravity
                tail_tipping_point['delta_height_on_tipping_point'] = delta_height_on_tipping_point

        lever_arms['vertical_distance_between_ground_and_fuselage_center_line'] = \
            (tail_tipping_point['delta_height_on_tipping_point']
             + abs(tail_tipping_point['tail_tipping_point_in_z_direction']))

        lever_arms['vertical_distance_between_aft_center_of_gravity_and_fuselage_bottom_edge'] = \
            abs(delta_height_fuselage_center_line_to_center_of_gravity - fuselage_parameter['fuselage_height']/2)

    # Exception handling for value error.
    except ValueError as e:
        runtime_output.critical('Error:', str(e))
        sys.exit('Lever arm estimation failed for landing gear design!')

    # prepare output dictionary
    lever_arms = {key: round(value, 4) for key, value in lever_arms.items()}
    landing_gear_description['lever_arms'] = lever_arms
    landing_gear_description['tail_tipping_point'] = tail_tipping_point
    landing_gear_description['fuselage_parameter'] = fuselage_parameter

    runtime_output.debug('estimate_distances successfully executed!')

    return landing_gear_description
