# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import sys
import math

# import own modules.
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.call_functions.sub_function.wingtipclearance import wing_tip_clearance  # noPep8 e501
from src.wing_mounted.empirical.landing_gear_design_tu_berlin.general.call_functions.sub_function.nacellegroundclearance import nacelle_ground_clearance  # noPep8 e501


def estimate_limitations(dict_ac_exchange, dict_mod_config, landing_gear_description, runtime_output):
    """ Hier beschreibung hin: Estimation of ground strike limitations in accordance to Sforza and CS-25
        [ref. P. M. Sforza; 2014]
        [ref. CS-25 Amendment 27; 2022]

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    """
    ''' initialize local parameter '''
    limitation_flag = False
    convert_inch_to_meter = 0.0254
    current_turn_over_angle = None
    main_gear_positioning_flag = True
    nacelle_ground_clearance_flag = True
    max_outside_diameter_in_inch = float()
    minimum_required_wheel_base = 0.0
    main_gear_tires = landing_gear_description['main_gear_description']['main_gear_tires']
    main_gear_arrangement = landing_gear_description['main_gear_description']['main_gear_arrangement']
    nose_gear_arrangement = landing_gear_description['nose_gear_description']['nose_gear_arrangement']
    tail_tipping_point = landing_gear_description['tail_tipping_point']
    landing_gear_positions = landing_gear_description['landing_gear_positions']

    ''' Estimate maximum rotation angel during take-off '''
    if dict_ac_exchange['maximum_angle_off_attack_landing'] is not None:
        maximum_rotation_angle_landing = max(dict_ac_exchange['maximum_angle_off_attack_landing'],
                                             dict_mod_config['tail_strike_limit'])
    else:
        maximum_rotation_angle_landing = dict_mod_config['tail_strike_limit']

    # Estimate maximum possible rotation tail tipping angle
    maximum_tail_tipping_angle = max(maximum_rotation_angle_landing,
                                     tail_tipping_point['maximum_rotation_angle_take_off'])

    landing_gear_limitations = landing_gear_description['landing_gear_limitations']
    landing_gear_limitations['maximum_tail_tipping_angle'] = maximum_tail_tipping_angle
    landing_gear_description['tail_tipping_point']['maximum_rotation_angle_landing'] = maximum_rotation_angle_landing

    ''' Tail strike take-off and landing limitation '''
    i = 0
    # while loop to iterate the landing gear strut length to avoid turn over
    while i < 1000 and not limitation_flag or nacelle_ground_clearance_flag:
        # Reset nacelle ground clearance flag for next iteration
        nacelle_ground_clearance_flag = False
        # Estimate minimum forward main landing gear position for tail strike take-off and landing limitation
        # check if strut suspension system should be used for limit estimation
        #  -> if true: -> add maximum strut suspension travel to vertical distance between center of gravity and ground
        if dict_mod_config['strut_suspension']:
            landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity'] = \
                round((landing_gear_description['lever_arms'][
                           'vertical_distance_between_ground_and_aft_center_of_gravity']
                       + dict_mod_config['strut_suspension_travel'])
                      * (math.atan(math.radians(maximum_tail_tipping_angle))), 4)

        # Else condition: Strut suspension system should not be used
        #  -> only vertical distance between center of gravity and ground is used for limit estimation
        else:
            landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity'] = \
                round(landing_gear_description['lever_arms'][
                          'vertical_distance_between_ground_and_aft_center_of_gravity']
                      * (math.atan(math.radians(maximum_tail_tipping_angle))), 4)

        # Estimate minimum forward main landing gear position in x-direction
        # check if the most aft center of gravity position is equal to 0.0 -> if true:
        #  -> use temporary center of gravity for minimum forward position of main gear estimation
        if dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] is None:
            landing_gear_limitations['minimum_forward_position_of_main_gear'] = \
                round(landing_gear_description['landing_gear_positions']['temporary_center_of_gravity']
                      + landing_gear_limitations[
                          'minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity'], 4)  # noPep8 e501
        # Else condition: The most aft center of gravity position is not equal to 0.0
        #  -> use for minimum forward position of main gear estimation
        else:
            landing_gear_limitations['minimum_forward_position_of_main_gear'] = \
                round(landing_gear_limitations[
                          'minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity']
                      + dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'], 4)

        # Check if the landing gear configuration is retractable
        #  -> if true: -> estimate minimum required distance between rear spar of center wing box and main gear position
        if dict_ac_exchange['undercarriage_retractability'] == 'retractable':
            max_diameter_values = [float(value['max_outside_diameter_in_inch'])
                                   for key, value in main_gear_tires.items() if 'max_outside_diameter_in_inch' in value]
            if max_diameter_values:
                max_outside_diameter_in_inch = max(max_diameter_values)

            # estimate minimum required distance to the rear spar of center wing box for retractable main landing gear
            if main_gear_arrangement['number_of_main_gear_axis_per_outer_strut'] > 1:
                if dict_mod_config['minimum_required_wheel_base'] < (1.10 * max_outside_diameter_in_inch
                                                                     * convert_inch_to_meter):
                    minimum_required_wheel_base = round(1.10 * (max_outside_diameter_in_inch * convert_inch_to_meter),
                                                        4)

                else:
                    minimum_required_wheel_base = round(dict_mod_config['minimum_required_wheel_base'], 4)
                landing_gear_limitations[
                    'minimum_distance_between_main_landing_gear_and_rear_spar_of_center_wing_box'] = \
                    round((main_gear_arrangement['number_of_main_gear_axis_per_outer_strut']
                           * max_outside_diameter_in_inch * convert_inch_to_meter + minimum_required_wheel_base)
                          + (0.15 * max_outside_diameter_in_inch * convert_inch_to_meter), 4)
            else:
                landing_gear_limitations[
                    'minimum_distance_between_main_landing_gear_and_rear_spar_of_center_wing_box'] = \
                    round(main_gear_arrangement['number_of_main_gear_axis_per_outer_strut']
                          * max_outside_diameter_in_inch * convert_inch_to_meter
                          + (0.15 * max_outside_diameter_in_inch * convert_inch_to_meter), 4)

            # Check if the current minimum forward position is lower than minimum required distance for retractability
            #  -> if true: -> estimate new distance between main landing gear and aft most center of gravity
            if landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity'] \
                    < landing_gear_limitations['minimum_distance_between_main_landing_gear_and_rear_spar_of_center_wing_box']:  # noPep8 e501
                if dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] is not None and\
                        dict_ac_exchange['rear_spar_position_center_wing_box_in_x_direction'] is not None:
                    # Check if the most aft center of gravity position is in front of the center wing rear spar position
                    #  -> if true: -> estimate new distance between rear most center of gravity and main gear position
                    if dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] \
                            < dict_ac_exchange['rear_spar_position_center_wing_box_in_x_direction']:
                        delta_between_rear_spar_and_aft_center_of_gravity = \
                            (dict_ac_exchange['rear_spar_position_center_wing_box_in_x_direction']
                             - dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'])
                        landing_gear_description['lever_arms'][
                            'distance_between_main_gear_and_rearmost_center_of_gravity']\
                            = round(landing_gear_limitations['minimum_distance_between_main_landing_gear_and_rear_spar_of_center_wing_box']  # noPep8 e501
                              + delta_between_rear_spar_and_aft_center_of_gravity, 4)

                    # Elif condition: Check if most aft center of gravity position is behind the center wing rear spar
                    #  position and the distance between is lower as the minimum distance between main landing gear
                    #   and center wing rear spar
                    #    -> if true: -> estimate new distance between rear most center of gravity and main gear position
                    elif dict_ac_exchange['most_aft_center_of_gravity_in_x_direction'] \
                            - dict_ac_exchange['rear_spar_position_center_wing_box_in_x_direction'] \
                            < landing_gear_limitations['minimum_distance_between_main_landing_gear_and_rear_spar_of_center_wing_box']:  # noPep8 e501
                        landing_gear_description['lever_arms'][
                            'distance_between_main_gear_and_rearmost_center_of_gravity']\
                            = round(landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity']  # noPep8 e501
                               + dict_ac_exchange['most_aft_center_of_gravity_in_x_direction']
                               - dict_ac_exchange['rear_spar_position_center_wing_box_in_x_direction'], 4)

                # Else condition: The distance between the most aft center of gravity position is behind the center wing
                #  rear spar position and the distance between is greater as minimum distance between main landing gear
                #   and center wing rear spar -> use minimum distance between main landing gear and aft most center of
                #    gravity as distance between main gear and rearmost center of gravity position
                else:
                    landing_gear_description['lever_arms']['distance_between_main_gear_and_rearmost_center_of_gravity']\
                        = landing_gear_limitations['minimum_distance_between_main_landing_gear_and_aft_most_center_of_gravity']  # noPep8 e501

        _, wing_reference_point_value = \
            next(reversed(list(dict_ac_exchange['wing_section_position_in_x'].items())), (None, None))
        _, wing_chord_length_value = \
            next(reversed(list(dict_ac_exchange['wing_chord_length_on_section'].items())), (None, None))
        # Check if the resulting minimum forward main gear position is not within 90 percent of the wing depth
        #  but retractable is selected -> if true: -> Set main gear positioning flag to False
        if wing_chord_length_value is not None and dict_ac_exchange['undercarriage_retractability'] == 'retractable':
            if landing_gear_limitations['minimum_forward_position_of_main_gear'] \
                    > wing_reference_point_value \
                    + 0.9 * wing_chord_length_value:
                main_gear_positioning_flag = False

        ''' estimate nose and main gear strut lengths '''
        # main gear strut length
        min_tire_diameter = min([float(value['min_outside_diameter_in_inch'])
                                 for key, value in main_gear_tires.items() if 'min_outside_diameter_in_inch' in value])

        _, fuselage_section_width_value = \
            next(iter(dict_ac_exchange['fuselage_section_width'].items()), (None, None))
        if dict_ac_exchange['dihedral_angle_wing'] is not None \
                and dict_ac_exchange['rear_spar_position_center_wing_box_in_z_direction'] is not None \
                and fuselage_section_width_value is not None and 'main_gear_outer_strut_y_position' \
                in landing_gear_positions:
            main_gear_arrangement['main_landing_gear_wing_mounting_position_in_z_direction'] =\
                round(dict_ac_exchange['rear_spar_position_center_wing_box_in_z_direction']
                      + ((landing_gear_positions['main_gear_outer_strut_y_position']
                          - max(dict_ac_exchange['fuselage_section_width'].values()) / 2)
                         * math.atan(math.radians(dict_ac_exchange['dihedral_angle_wing']))), 4)

        else:
            if dict_ac_exchange['fuselage_type'] == 'single_aisle':
                main_gear_arrangement['main_landing_gear_wing_mounting_position_in_z_direction'] = -0.9

            if dict_ac_exchange['fuselage_type'] == 'wide_body':
                main_gear_arrangement['main_landing_gear_wing_mounting_position_in_z_direction'] = -1.3

        if 'main_gear_outer_strut_length' not in main_gear_arrangement:
            main_gear_arrangement['main_gear_outer_strut_length'] = \
                round((landing_gear_description['lever_arms'][
                        'vertical_distance_between_ground_and_aft_center_of_gravity']
                    + main_gear_arrangement['main_landing_gear_wing_mounting_position_in_z_direction'])
                    - (min_tire_diameter/2 * convert_inch_to_meter), 4)
        
        # Check nacelle ground clearance if at least one engine is exisiting.
        if dict_ac_exchange['engine_z_position'] is not None:
            # Call function to check nacelle ground clearance.
            required_main_gear_strut_delta_length, nacelle_ground_clearance_flag = \
                nacelle_ground_clearance(dict_ac_exchange, dict_mod_config, landing_gear_description,
                                         main_gear_arrangement, runtime_output)
            
            # Check if nacelle ground clearance check is failed 
            #  -> if true: -> start repositioning and resizing of main landing gear struts.
            if nacelle_ground_clearance_flag:
                if i == 1:
                    runtime_output.warning('Attention: At least one nacelle mounted to the wing '
                                           'falls below the minimum required nacelle ground clearance.')
                    runtime_output.print('Repositioning and resizing of landing gear struts started!')
                
                # Add required delta length to the length of main landing gear struts.
                main_gear_arrangement['main_gear_outer_strut_length'] += max(required_main_gear_strut_delta_length)
                # Recalculate distance between ground and fusealge center line.
            landing_gear_description['lever_arms']['vertical_distance_between_ground_and_fuselage_center_line'] = \
                (main_gear_arrangement['main_gear_outer_strut_length'] 
                    + abs(landing_gear_description['main_gear_description']['main_gear_arrangement'][
                        'main_landing_gear_wing_mounting_position_in_z_direction']) 
                    + (max_outside_diameter_in_inch/2 * convert_inch_to_meter))
                
        landing_gear_description['landing_gear_positions']['main_gear_z_position_outer_strut'] = \
            round(main_gear_arrangement['main_landing_gear_wing_mounting_position_in_z_direction'], 4)

        # Check if inner main gear struts are used -> if true: -> estimate inner main gear strut length
        if 'number_of_main_gear_inner_struts' in main_gear_arrangement:
            main_gear_arrangement['main_gear_inner_strut_length'] = \
                round(main_gear_arrangement['main_gear_outer_strut_length']
                      + abs(main_gear_arrangement['main_landing_gear_wing_mounting_position_in_z_direction'])
                      - landing_gear_description['fuselage_parameter']['fuselage_height']/2, 4)

            landing_gear_description['landing_gear_positions']['main_gear_z_position_inner_strut'] = \
                round(-landing_gear_description['fuselage_parameter']['fuselage_height'] / 2, 4)

        # nose gear strut length
        nose_gear_arrangement['nose_gear_strut_length'] = \
            round(main_gear_arrangement['main_gear_outer_strut_length']
                  + abs(main_gear_arrangement['main_landing_gear_wing_mounting_position_in_z_direction'])
                  - landing_gear_description['fuselage_parameter']['fuselage_height'] / 2
                  + (landing_gear_description['fuselage_parameter']['fuselage_height'] / 2 / 5), 4)

        ''' estimate y-position of nose and main gear struts '''
        # Check if the fuselage_width is not given in the aircraft exchange file
        #  -> if true: -> use fuselage height for further estimation
        if fuselage_section_width_value is None:
            landing_gear_description['fuselage_parameter']['fuselage_width'] \
                = landing_gear_description['fuselage_parameter']['fuselage_height']
        else:
            landing_gear_description['fuselage_parameter']['fuselage_width'] = \
                max(dict_ac_exchange['fuselage_section_width'].values())

        # nose landing gear strut y-postion
        _, fuselage_reference_point_value = \
            next(iter(dict_ac_exchange['fuselage_reference_point_in_y_direction'].items()), (None, None))
        if fuselage_reference_point_value is not None:
            if nose_gear_arrangement['number_of_nose_gear_struts'] == 1:
                landing_gear_positions['nose_gear_strut_y_position'] = fuselage_reference_point_value
            elif nose_gear_arrangement['number_of_nose_gear_struts'] == 2:
                landing_gear_positions['nose_gear_strut_y_position'] \
                    = landing_gear_description['fuselage_parameter']['fuselage_width'] / 4
        else:
            if nose_gear_arrangement['number_of_nose_gear_struts'] == 1:
                landing_gear_positions['nose_gear_strut_y_position'] = 0.0
            elif nose_gear_arrangement['number_of_nose_gear_struts'] == 2:
                landing_gear_positions['nose_gear_strut_y_position'] \
                    = landing_gear_description['fuselage_parameter']['fuselage_width'] / 4

        # main landing gear inner strut y-position
        # Check if inner main gear struts are used -> if true: -> estimate inner main gear strut y-position
        if 'number_of_main_gear_inner_struts' in main_gear_arrangement:
            if main_gear_arrangement['number_of_main_gear_inner_struts'] > 1:
                landing_gear_positions['main_gear_inner_strut_y_position'] \
                    = landing_gear_description['fuselage_parameter']['fuselage_width'] / 4
            elif fuselage_reference_point_value is not None:
                landing_gear_positions['main_gear_inner_strut_y_position'] = fuselage_reference_point_value
            else:
                landing_gear_positions['main_gear_inner_strut_y_position'] = 0.0

        # main landing gear outer strut y-position
        # (strut_length + 1/2*tire_diameter + 1/2*keel_beam_width + 0.15*tire_diameter
        # (safety margin in according to CS-25))
        landing_gear_positions['main_gear_outer_strut_y_position'] \
            = round(main_gear_arrangement['main_gear_outer_strut_length']
                    + (max_outside_diameter_in_inch/2 * convert_inch_to_meter)
                    + dict_mod_config['landing_gear_bay_keel_beam_width']/2
                    + (0.15 * max_outside_diameter_in_inch * convert_inch_to_meter), 4)

        # Check if a strut suspension system is used
        #  -> if true: -> add strut suspension travel to main gear outer strut y-position
        if dict_mod_config['strut_suspension']:
            landing_gear_positions['main_gear_outer_strut_y_position'] += dict_mod_config['strut_suspension_travel']

        main_gear_arrangement['main_landing_gear_track'] = \
            (2 * landing_gear_positions['main_gear_outer_strut_y_position'])

        # estimate turnover angle in accordance to Sforza [reference_hier] equation 7.11, and 7.12
        beta_turn_over_angle = math.degrees(math.atan(main_gear_arrangement['main_landing_gear_track'] /
                                            (2 * landing_gear_description['lever_arms'][
                                                'distance_between_nose_and_main_gear'])))

        current_turn_over_angle = round(math.degrees(math.atan(
            landing_gear_description['lever_arms']['vertical_distance_between_ground_and_aft_center_of_gravity']
            / (landing_gear_description['lever_arms']['distance_between_nose_gear_and_foremost_center_of_gravity']
               * math.atan(math.radians(beta_turn_over_angle))))), 4)

        # check if the iteration is successfully -> if true: -> break the loop
        if current_turn_over_angle <= dict_mod_config['turnover_limit'] and i > 0:
            limitation_flag = True

        # Else condition: Iteration is not successfully
        # -> add 0.05 meter to vertical distance between ground and center of gravity and restart strut estimation
        else:
            landing_gear_description['lever_arms']['vertical_distance_between_ground_and_aft_center_of_gravity'] = \
                (landing_gear_description['lever_arms']['vertical_distance_between_ground_and_aft_center_of_gravity']
                 + 0.05)
            i += 1

    runtime_output.print('Repositioning and resizing of landing gear struts successfully finished!')
    
    try:
        # check if the turnover iteration failed -> if true: -> raise an error and abort program
        if not main_gear_positioning_flag:
            raise ValueError('Error: Main gear positioning failed. Selected main gear position is wing mounted and '
                             'retractable but estimated main gear x-position is not inside the given wing geometry!\n '
                             '      Program aborted!')

        if not limitation_flag:
            raise ValueError('Error: Maximum number of iterations (1000) reached. '
                             'Turnover limit cannot be observed in the given aircraft configuration!\n '
                             '      Program aborted!')
        # Else condition: Turnover iteration was successful -> set iteration flag to limitation dictionary
        else:
            landing_gear_limitations['current_turn_over_angle'] = current_turn_over_angle
            landing_gear_limitations['limitation_flag'] = limitation_flag

    # exception handling if the turnover iteration failed
    except ValueError as e:
        runtime_output.critical('Error: ' + str(e))
        sys.exit('Iteration for landing gear positioning or turn over avoidance failed.')

    main_gear_arrangement = {key: round(value, 4) for key, value in main_gear_arrangement.items()}

    ''' estimate local tire coordinates '''
    # wheel base
    if main_gear_arrangement['number_of_main_gear_axis_per_outer_strut'] == 1:
        main_gear_arrangement['local_tire_position_in_x_for_acn'] = [minimum_required_wheel_base,
                                                                     minimum_required_wheel_base]
    elif main_gear_arrangement['number_of_main_gear_axis_per_outer_strut'] == 2:
        main_gear_arrangement['local_tire_position_in_x_for_acn'] = [0.0,
                                                                     0.0,
                                                                     -minimum_required_wheel_base,
                                                                     -minimum_required_wheel_base]
    elif main_gear_arrangement['number_of_main_gear_axis_per_outer_strut'] == 3:
        main_gear_arrangement['local_tire_position_in_x_for_acn'] = [0.0,
                                                                     0.0,
                                                                     -minimum_required_wheel_base/2,
                                                                     -minimum_required_wheel_base/2,
                                                                     -minimum_required_wheel_base,
                                                                     -minimum_required_wheel_base
                                                                     ]

    main_gear_arrangement['wheel_base'] = minimum_required_wheel_base

    # wheel track
    max_section_width_in_meter = []
    for tire, tire_dict in landing_gear_description['main_gear_description']['main_gear_tires'].items():
        for key, value in tire_dict.items():
            if key == 'max_section_width_in_inch':
                max_section_width_in_meter.append(round(float(value) * convert_inch_to_meter, 4))

    if dict_ac_exchange['fuselage_type'] == 'single_aisle':
        wheel_track = round(2 * max(max_section_width_in_meter) * 1.1, 4)
    else:
        wheel_track = round(2.75 * max(max_section_width_in_meter) * 1.1, 4)

    if wheel_track < dict_mod_config['minimum_required_wheel_track']:
        wheel_track = round(float(dict_mod_config['minimum_required_wheel_track']), 4)

    main_gear_arrangement['wheel_track'] = wheel_track
    if main_gear_arrangement['number_of_main_gear_axis_per_outer_strut'] == 1:
        main_gear_arrangement['local_tire_position_in_y_for_acn'] = [-wheel_track/2, wheel_track/2]
    if main_gear_arrangement['number_of_main_gear_axis_per_outer_strut'] == 2:
        main_gear_arrangement['local_tire_position_in_y_for_acn'] = [-wheel_track/2, wheel_track/2,
                                                                     -wheel_track/2, wheel_track/2]
    if main_gear_arrangement['number_of_main_gear_axis_per_outer_strut'] == 3:
        main_gear_arrangement['local_tire_position_in_y_for_acn'] = [-wheel_track/2, wheel_track/2,
                                                                     -wheel_track/2, wheel_track/2,
                                                                     -wheel_track/2, wheel_track/2]

    # prepare output dictionary
    nose_gear_arrangement = {key: round(value, 4) for key, value in nose_gear_arrangement.items()}
    landing_gear_positions = {key: round(value, 4) for key, value in landing_gear_positions.items()}
    landing_gear_limitations = {key: round(value, 4) for key, value in landing_gear_limitations.items()}
    landing_gear_description['landing_gear_positions'] = landing_gear_positions
    landing_gear_description['landing_gear_limitations'] = landing_gear_limitations
    landing_gear_description['main_gear_description']['main_gear_arrangement'] = main_gear_arrangement
    landing_gear_description['nose_gear_description']['nose_gear_arrangement'] = nose_gear_arrangement
    landing_gear_description['tail_tipping_point']['maximum_rotation_angle_landing'] = maximum_rotation_angle_landing
    
    # Call function to check the wing tip clearance.
    landing_gear_description = wing_tip_clearance(dict_ac_exchange, dict_mod_config, landing_gear_description, runtime_output)
    
    # debug print
    runtime_output.debug('The function "estimate_limitations.py" successfully executed!')

    return landing_gear_description
