# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
import math


def wing_tip_clearance(dict_ac_exchange, dict_mod_config, landing_gear_description, runtime_output):
    """ Estimation of wing tip clearance in accordance to Torenbeek and CS-25
            [ref. E. Torenbeek; 1982]
            [ref. CS-25 Amendment 27; 2022]

    :param dict dict_ac_exchange: Dict containing parameter and according values from aircraft exchange file
    :param dict dict_mod_config: Dict containing parameter and according values from module configuration file
    :param dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return dict landing_gear_description: Dict containing parameter and values of current landing gear estimation
    """
    ''' initialize local parameter '''
    wing_tip_clearance_angle = 0
    minimum_required_clearance_angle = 5  # according to CS25.149 -> minimum 5 degree bank angle
    
    ''' check wing tip clearance '''
    try:
        _, wing_reference_point_on_tip_in_y_value = \
            next(reversed(list(dict_ac_exchange['wing_section_position_in_y'].items())), (None, None))
        _, wing_reference_point_on_tip_in_z_value = \
            next(reversed(list(dict_ac_exchange['wing_section_position_in_z'].items())), (None, None))

        if wing_reference_point_on_tip_in_y_value is not None and wing_reference_point_on_tip_in_z_value is not None:
            distance_between_wing_tip_and_outer_strut = \
                round(wing_reference_point_on_tip_in_y_value
                      - landing_gear_description['landing_gear_positions']['main_gear_outer_strut_y_position'], 4)
            wing_tip_ground_clearance = \
                (landing_gear_description['lever_arms']['vertical_distance_between_ground_and_fuselage_center_line']
                 - wing_reference_point_on_tip_in_z_value)
            wing_tip_clearance_angle = \
                round(math.degrees(math.atan(wing_tip_ground_clearance/distance_between_wing_tip_and_outer_strut)), 4)

            if wing_tip_clearance_angle < dict_mod_config['wing_strike_limit'] \
                    and not wing_tip_clearance_angle < minimum_required_clearance_angle:
                raise ValueError('The resulting wing tip angle is below the specified clearance angle of: '
                                 + str(dict_mod_config['wing_strike_limit']) + ' degrees!')
            elif wing_tip_clearance_angle < minimum_required_clearance_angle:
                raise ValueError('The resulting wing tip angle is below the required clearance angle of the '
                                 'CS25.149 of: ' + str(minimum_required_clearance_angle) + ' degrees!')
            else:
                runtime_output.print('The resulting wing tip angle conforms the specified clearance angle of: '
                                     + str(dict_mod_config['wing_strike_limit']) + ' degrees!')

        else:
            raise ValueError('At least one wing parameter is missing in the aircraft exchange file, '
                             'wing tip clearance could not be checked!')

    # exception handling for wing tip clearance check
    except ValueError as e:
        runtime_output.error('Error: ' + str(e))

    landing_gear_description['landing_gear_limitations']['wing_tip_clearance_angles'] = wing_tip_clearance_angle

    # debug print
    runtime_output.debug('The function "wing_tip_clearance.py" successfully executed!')

    return landing_gear_description
