# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module providing plotting functionalities for current calculation method."""
# Import standard libraries.
import os
import math
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.patches import FancyBboxPatch, Polygon


def method_plot(paths_and_names, routing_dict, data_dict, method_specific_output_dict, runtime_output):
    """Plot function.

    This function is responsible for creating plots.
    [Add further information here...]

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict routing_dict: Dictionary containing routing parameters
    :param dict data_dict: Dictionary containing results of module execution
    :param dict method_specific_output_dict: Dictionary containing method specific output data
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: None
    """
    # Check path to directory for figure outputs
    project_directory = paths_and_names["project_directory"]
    plot_directory = project_directory + '/reporting/plots'
    if not os.path.isdir(plot_directory):
        os.makedirs(plot_directory)

    ''' visualisation of results '''
    # generate acn plot
    # call function to create the plots of acn methode
    # create_acn_plot(paths_and_names, data_dict['acn_estimation']['gross_mass_list'],
    #                 data_dict['acn_estimation']['aircraft_classification_number_output_list'],
    #                 data_dict['acn_estimation']['InputAlpha'], plot_directory)

    # call function to create the top and side view plots of landing gear design
    create_landing_gear_plots(paths_and_names, data_dict, plot_directory, runtime_output)

    runtime_output.print('Plots for landing gear design are successfully generated and saved.')
    return


# generate acn diagram of current aircraft project
def create_acn_plot(paths_and_names, gross_mass_list, aircraft_classification_number_output_list, alpha_factor,
                    plot_directory):
    """ This function create the acn over gross mass diagrams in accordance to:
            ICAO AERODROME DESIGN MANUAL PART 3 CHAPTER 1. SECOND EDITION 1983.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param list gross_mass_list: List containing different gross masses of current aircraft
    :param list aircraft_classification_number_output_list: List containing corresponding acn of current aircraft
    :param float alpha_factor: Float containing the used alpha factor for acn estimation
    :param string plot_directory: String contains the absolute system path to save the output figures.
    :return: none
    """

    ''' generate plot entries '''
    gross_mass_list = [x / 1000 for x in gross_mass_list]

    cbr_3 = [sublist[0] for sublist in aircraft_classification_number_output_list]
    cbr_6 = [sublist[1] for sublist in aircraft_classification_number_output_list]
    cbr_10 = [sublist[2] for sublist in aircraft_classification_number_output_list]
    cbr_15 = [sublist[3] for sublist in aircraft_classification_number_output_list]

    ''' generate plot window '''
    plot_title = ('Aircraft Classification Number for alpha factor = '
                  + str(round(alpha_factor, 2)) + ' and different subgrade strength (D, C, B, A)')
    plt.figure(figsize=(12, 7))
    plt.grid(True, linestyle='--', linewidth=0.5, color='gray')
    plt.rcParams['font.family'] = 'Calibri'
    plt.title(plot_title, fontsize=15)
    plt.xlabel('aircraft gross mass (x 1000 kg)', fontsize=15)
    plt.ylabel('Aircraft Classification Number (ACN)', fontsize=16)
    plt.xlim(min(gross_mass_list), max(gross_mass_list))
    plt.xticks(gross_mass_list)
    plt.ylim(math.floor(min(cbr_15)/10) * 10, math.ceil(max(cbr_3)/10) * 10 + 10)
    if math.floor(min(cbr_15)/10) * 10 > 0:
        y_ticks = (int((math.ceil(max(cbr_3)/10) * 10 + 10) / (math.floor(min(cbr_15)/10) * 10)))
        ticks = np.linspace(math.floor(min(cbr_15)/10) * 10, math.ceil(max(cbr_3)/10) * 10 + 10, y_ticks * 2 - 1)
        plt.yticks(ticks)

    ''' plot the acn over gross mass curves '''
    plt.plot(gross_mass_list, cbr_3[1:],
             label='D - CBR 3 (ultra low)', color=(0.5, 0.5, 0.5), linestyle='-', linewidth=1.5)
    plt.plot(gross_mass_list, cbr_6[1:],
             label='C - CBR 6 (low)', color=(0.5, 0.5, 0.5), linestyle='--', linewidth=1.5)
    plt.plot(gross_mass_list, cbr_10[1:],
             label='B - CBR 10 (medium)', color=(0.5, 0.5, 0.5), linestyle='-.', linewidth=1.5)
    plt.plot(gross_mass_list, cbr_15[1:],
             label='A - CBR 15 (high)', color=(0.5, 0.5, 0.5), linestyle=':', linewidth=1.5)

    plt.legend(fontsize=15)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)

    """ Save figures to current aircraft project directory. """
    # Save as *.svg
    tool_name = paths_and_names['tool_name']
    name_of_fig = tool_name + '_' + plot_title + '.svg'
    plt.savefig(plot_directory + '/' + name_of_fig, transparent=True)
    # Save as *.pdf
    name_of_fig = tool_name + '_' + plot_title + '.pdf'
    plt.savefig(plot_directory + '/' + name_of_fig, transparent=True)
    # plt.show()

    return


def create_landing_gear_plots(paths_and_names, data_dict, plot_directory, runtime_output):
    """ This function create the front and side view diagrams of landing gear arrangement of current aircraft project.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict data_dict: Dictionary containing system paths and ElementTrees
    :param string plot_directory: String contains the absolute system path to save the output figures.
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :return: none
    """
    ''' Plot of front view '''
    # Generate figure.
    plot_title = 'Front view of landing gear arrangement'
    fig, ax = plt.subplots(figsize=(12, 7))

    # Plot fuselage cross-section.
    ax.plot(data_dict['dict_landing_gear_output']['width_cross_section_points'],
            data_dict['dict_landing_gear_output']['upper_cross_section_points'],
            'black')
    ax.plot(data_dict['dict_landing_gear_output']['width_cross_section_points'],
            data_dict['dict_landing_gear_output']['lower_cross_section_points'],
            'black')
    ax.fill_between(data_dict['dict_landing_gear_output']['width_cross_section_points'],
                    data_dict['dict_landing_gear_output']['upper_cross_section_points'],
                    data_dict['dict_landing_gear_output']['lower_cross_section_points'],
                    color='gray', alpha=0.1)

    # Loop across all tires to plot.
    pad_size = 0.05
    i = 0
    while i < len(data_dict['dict_landing_gear_output']['tire_plot_list']):
        rounded_rect = FancyBboxPatch((data_dict['dict_landing_gear_output']['tire_plot_list'][i][0],
                                       data_dict['dict_landing_gear_output']['tire_plot_list'][i][1]),
                                      data_dict['dict_landing_gear_output']['tire_plot_list'][i][2],
                                      data_dict['dict_landing_gear_output']['tire_plot_list'][i][3],
                                      boxstyle="round4,pad=" + str(pad_size),
                                      ec="none",
                                      fc="grey",
                                      alpha=0.15)
        rounded_rect2 = FancyBboxPatch((data_dict['dict_landing_gear_output']['tire_plot_list'][i][0],
                                       data_dict['dict_landing_gear_output']['tire_plot_list'][i][1]),
                                       data_dict['dict_landing_gear_output']['tire_plot_list'][i][2],
                                       data_dict['dict_landing_gear_output']['tire_plot_list'][i][3],
                                       boxstyle="round4,pad=0.05",
                                       ec="black",
                                       fc="none")
        ax.add_patch(rounded_rect)
        ax.add_patch(rounded_rect2)
        i += 1

    # Plot measurement line and value of landing gear main gear track.
    arrow_head_length = 0.3
    p1 = data_dict['dict_landing_gear_output']['landing_gear_track'][0]
    p2 = data_dict['dict_landing_gear_output']['landing_gear_track'][1]
    dx = round(p2[0] - p1[0] + arrow_head_length, 2)
    dy = p2[1] - p1[1]
    arrow1 = ax.arrow(p1[0], p1[1], dx, dy, head_width=arrow_head_length/2,
                      head_length=arrow_head_length, fc='grey', ec='grey', linewidth=0.5)
    arrow2 = ax.arrow(p2[0], p2[1], -dx, -dy, head_width=arrow_head_length/2,
                      head_length=arrow_head_length, fc='grey', ec='grey', linewidth=0.5)
    mid_x = (arrow1.get_xy()[0, 0] + arrow2.get_xy()[0, 0]) / 2
    mid_y = p2[1] - 0.5
    ax.text(mid_x, mid_y,
            str(data_dict['dict_landing_gear_output']['landing_gear_track'][2]) + 'm',
            ha='center', va='center', color='grey')

    # Plot of strut coordinates as crosses.
    for point in data_dict['dict_landing_gear_output']['strut_positions']:
        ax.scatter(point[0], point[1], color='gray')

    # Plot wing if available.
    if 'wing_plot_list' in data_dict['dict_landing_gear_output']:
        i = 0
        while i < len(data_dict['dict_landing_gear_output']['wing_plot_list'])-1:
            ax.plot([data_dict['dict_landing_gear_output']['wing_plot_list'][i][0],
                     data_dict['dict_landing_gear_output']['wing_plot_list'][i+1][0]],
                    [data_dict['dict_landing_gear_output']['wing_plot_list'][i][1],
                     data_dict['dict_landing_gear_output']['wing_plot_list'][i+1][1]],
                    color='black')
            ax.plot([data_dict['dict_landing_gear_output']['wing_plot_list'][i][0],
                     data_dict['dict_landing_gear_output']['wing_plot_list'][i+1][0]],
                    [data_dict['dict_landing_gear_output']['wing_plot_list'][i][2],
                     data_dict['dict_landing_gear_output']['wing_plot_list'][i+1][2]],
                    color='black')
            if i == 0:
                ax.plot([data_dict['dict_landing_gear_output']['wing_plot_list'][i][0],
                         data_dict['dict_landing_gear_output']['wing_plot_list'][i][0]],
                        [data_dict['dict_landing_gear_output']['wing_plot_list'][i][1],
                         data_dict['dict_landing_gear_output']['wing_plot_list'][i][2]],
                        color='black')
            if i == len(data_dict['dict_landing_gear_output']['wing_plot_list'])-2:
                ax.plot([data_dict['dict_landing_gear_output']['wing_plot_list'][i+1][0],
                         data_dict['dict_landing_gear_output']['wing_plot_list'][i+1][0]],
                        [data_dict['dict_landing_gear_output']['wing_plot_list'][i+1][1],
                         data_dict['dict_landing_gear_output']['wing_plot_list'][i+1][2]],
                        color='black')
            i += 1

        # Plot wing clearance angle
        if data_dict['dict_landing_gear_output']['wing_clearance_list'] is not None:
            ax.plot([data_dict['dict_landing_gear_output']['wing_clearance_list'][0][0],
                     data_dict['dict_landing_gear_output']['wing_clearance_list'][1][0]],
                    [data_dict['dict_landing_gear_output']['wing_clearance_list'][0][1] - pad_size,
                     data_dict['dict_landing_gear_output']['wing_clearance_list'][1][1] - pad_size],
                    color='grey', linestyle='--', linewidth=0.5)
            ax.text(data_dict['dict_landing_gear_output']['wing_clearance_list'][1][0],
                    data_dict['dict_landing_gear_output']['wing_clearance_list'][0][1] - pad_size,
                    str(data_dict['dict_landing_gear_output']['wing_clearance_list'][-1][0]) + 'deg  ',
                    verticalalignment='bottom', horizontalalignment='right', color='gray')
            ax.plot([data_dict['dict_landing_gear_output']['wing_clearance_list'][0][0],
                     data_dict['dict_landing_gear_output']['wing_clearance_list'][1][0]],
                    [data_dict['dict_landing_gear_output']['wing_clearance_list'][0][1] - pad_size,
                     data_dict['dict_landing_gear_output']['wing_plot_list'][-1][2]],
                    color='grey', linestyle='--', linewidth=0.5)
        else:
            runtime_output.warning('Attention! The Use existing geometry mode is selected. '
                                   'Clearance and limitation angle are not plotted.')
    
    # Plot nacelle if available.
    if 'nacelle_plot_list' in data_dict['dict_landing_gear_output']:
        theta = np.linspace(0, 2*np.pi, 100)
        i = 0
        while i < len(data_dict['dict_landing_gear_output']['nacelle_plot_list']):
            y1 = (data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][0]
                  + data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][2] * np.cos(theta))
            z1 = (data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][1]
                  + data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][2] * np.sin(theta))
            y2 = (data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][0]
                  + (data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][2]
                     * (1 - data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][3]/100)) * np.cos(theta))
            z2 = (data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][1]
                  + (data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][2]
                     * (1 - data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][3]/100)) * np.sin(theta))
            ax.plot(y1, z1, color='black')
            ax.plot(y2, z2, color='black')
            vertices = np.vstack((np.column_stack((y1, z1)),
                                  np.column_stack((y2[::-1], z2[::-1]))))
            polygon = Polygon(vertices, closed=True, color='gray', alpha=0.8)
            vertices_inner = np.column_stack((y2[::-1], z2[::-1]))
            fig.gca().add_patch(polygon)
            polygon_inner = Polygon(vertices_inner, closed=True, color='gray', alpha=0.15)
            fig.gca().add_patch(polygon_inner)
            ax.scatter(data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][0],
                       data_dict['dict_landing_gear_output']['nacelle_plot_list'][i][1], color='gray')
            i += 1

    # Remove frame
    fig.gca().spines['top'].set_visible(False)
    fig.gca().spines['bottom'].set_visible(False)
    fig.gca().spines['left'].set_visible(False)
    fig.gca().spines['right'].set_visible(False)

    # Disable the axis ticks
    plt.xticks([])
    plt.yticks([])
    ax.grid(True)
    ax.axis('equal')

    #plt.show()

    """ Save figures to current aircraft project directory. """
    # Save as *.svg
    tool_name = paths_and_names['tool_name']
    name_of_fig = tool_name + '_' + plot_title + '.svg'
    plt.savefig(plot_directory + '/' + name_of_fig, transparent=True)
    # Save as *.pdf
    name_of_fig = tool_name + '_' + plot_title + '.pdf'
    plt.savefig(plot_directory + '/' + name_of_fig, transparent=True)

    return
