/**
 * \file mission_analysis.cpp
 * \author Gerrit Pauls (gerrit.pauls@tuhh.de)
 * \brief
 * \version 0.1
 * \date 2024-08-30
 *
 * @copyright Copyright (c) 2024
 *
 */
#include <memory>
#include <map>
#include <string>
#include <vector>

#include "mission_analysis.h"
#include "standard_mission/standard_mission/standard_mission.h"

/**
 * @brief Construct a new mission_analysis:: mission_analysis object
 * 
 * @param argc argument count
 * @param argv vector of C++ style argument strings
 * @param toolName entry from toolinfo.h
 * @param toolVersion entry from toolinfo.h
 */
mission_analysis::mission_analysis(const int argc, char *argv[], const std::string &toolName, const std::string &toolVersion)
        : Module(argc, argv, toolName, toolVersion) {
    std::string method = (EndnodeReadOnly<std::string>("program_settings/mode/mission_methods/fidelity_level").read(rtIO_->moduleConfig).value());
    std::string mission_type(EndnodeReadOnly<std::string>("program_settings/mode/mission_methods/mission_type").read(rtIO_->moduleConfig).value());
    if (mission_type != "design_mission" && mission_type != "study_mission" && mission_type != "requirements_mission") {
        throwError(__FILE__, __func__, __LINE__, "Specified mission '" + mission_type +
                    "' is unkown! Only use 'design_mission', 'study_mission' or 'requirements_mission'. Abort program!");
    }
    if (mission_type == "design_mission" && (!rtIO_->acxml.find("analysis/mission/design_mission") || // Catch if we are in the first loop...
            EndnodeReadOnly<bool>("program_settings/mode/mission_methods/use_fuel_estimation_method").read(rtIO_->moduleConfig).value())) { // or want to stay in that mode
        myRuntimeInfo->out << "Low-fidelity Breguet mission analysis for design mission runs because we are in the first loop or the is set in the config file!" << std::endl;
        mission_type += "_1st_loop";
    }
    route_ = {method, mission_type};
    missionStrategy_.setStrategy(missionSelection_(route_)(rtIO_));
}

strategyaccess mission_analysis::missionSelection_(const std::vector<std::string> &route) {
    /* Routing table */
    std::map<std::string, std::map<std::string, strategyaccess>> table = {
    {"low", std::map<std::string, strategyaccess>{
            {"design_mission",          [](const std::shared_ptr<RuntimeIO>& arg) { return std::make_unique<design_mission::standard_mission>(arg); }},
            {"design_mission_1st_loop", [](const std::shared_ptr<RuntimeIO>& arg) { return std::make_unique<design_mission::first_loop::standard_mission>(arg); }},
            {"study_mission",           [](const std::shared_ptr<RuntimeIO>& arg) { return std::make_unique<study_mission::standard_mission>(arg); }},
            {"requirements_mission",    [](const std::shared_ptr<RuntimeIO>& arg) { return std::make_unique<requirements_mission::standard_mission>(arg); }}
        }
    }
  };
  return table[route.at(0)][route.at(1)];
}
