/**
 * \file flight_segments.cpp
 * \author Gerrit Pauls (gerrit.pauls@tuhh.de)
 * \brief
 * \version 0.1
 * \date 2024-08-30
 *
 * @copyright Copyright (c) 2024
 *
 */


/* List of references

    [1] Airbus customer services, Flight Operations Support & Line Assistance: Getting to grips with the cost index, Issue II, May 1998, p.7
    [2] G. Bruenning, X. Hafer, G. Sachs, Flugleistungen: Grundlagen, Flugabschnitte, Third Edition, 1993

*/

#include <moduleBasics/modeSelector.h>
#include "../standard_mission/standard_mission.h"
#include <algorithm>

/* Initialization of attributes for the mission_segment object with the constructor */
standard_mission::mission_segment::mission_segment(const std::map<int, double>& an_end_total_fuel_mass,
                                                    double an_end_fuel_flow,
                                                    double an_end_time,
                                                    double an_end_range,
                                                    double an_end_altitude,
                                                    double an_end_total_mass,
                                                    double an_end_CL,
                                                    double an_end_L_over_D,
                                                    double an_end_specific_air_range,
                                                    const std::string& an_end_aero_config,
                                                    double an_end_glidepath_angle,
                                                    double an_end_TAS,
                                                    double an_end_mach,
                                                    double an_end_CAS,
                                                    double an_end_ground_speed,
                                                    double an_end_thrust,
                                                    double an_end_bleed,
                                                    double an_end_shaft_power,
                                                    const std::string& an_engine_rating,
                                                    const std::map<uint16_t, double>& an_end_n1,
                                                    double an_end_angle_of_attack,
                                                    double an_end_spoiler_factor,
                                                    const std::string& an_end_mode_name,
                                                    double an_end_reynolds_number,
                                                    double an_end_aircraft_cg,
                                                    double an_end_fuel_cg,
                                                    double an_end_stabilizer_incidence,
                                                    const std::vector<std::vector<double>>& an_end_origin_polar,
                                                    int an_energy_carrier_id)
    :
    end_total_fuel_mass(an_end_total_fuel_mass),
    end_fuel_flow(an_end_fuel_flow),
    end_time(an_end_time),
    end_range(an_end_range),
    end_altitude(an_end_altitude),
    end_total_mass(an_end_total_mass),
    end_CL(an_end_CL),
    end_L_over_D(an_end_L_over_D),
    end_specific_air_range(an_end_specific_air_range),
    end_aero_config(an_end_aero_config),
    end_glidepath_angle(an_end_glidepath_angle),
    end_TAS(an_end_TAS),
    end_mach(an_end_mach),
    end_CAS(an_end_CAS),
    end_ground_speed(an_end_ground_speed),
    end_thrust(an_end_thrust),
    end_bleed(an_end_bleed),
    end_shaft_power(an_end_shaft_power),
    end_engine_rating(an_engine_rating),
    end_n1(an_end_n1),
    end_angle_of_attack(an_end_angle_of_attack),
    end_spoiler_factor(an_end_spoiler_factor),
    end_mode_name(an_end_mode_name),
    end_reynolds_number(an_end_reynolds_number),
    end_aircraft_cg(an_end_aircraft_cg),
    end_fuel_cg(an_end_fuel_cg),
    end_stabilizer_incidence(an_end_stabilizer_incidence),
    end_origin_polar(an_end_origin_polar),
    energy_carrier_id(an_energy_carrier_id) {
}

/* Destructor of the mission_segment object */
standard_mission::mission_segment::~mission_segment() {
}

void standard_mission::calculate_ground_run(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step) {
    /* Calculate the number of steps */
    uint16_t steps = static_cast<uint16_t>(ceil(fabs((fc->end_TAS - fc->TAS) /
        EndnodeReadOnly<double>("program_settings/precision/acceleration_increment").read(this->rtIO->moduleConfig).value())));
    double delta_TAS = (fc->end_TAS - fc->TAS) / steps;
    /* Set minimum acceleration */
    for (uint16_t step_number(0); step_number < steps; ++step_number) {
        this->set_segment_initial_conditions(fc, true, this->aero.CLgroundRoll);
        fc->end_TAS = fc->TAS + delta_TAS;
        if (this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::LANDING) { //Activation of thrust reverser right after touchdown if desired
            if (EndnodeReadOnly<bool>("program_settings/general/landing/thrust_reverser/enable").read(this->rtIO->moduleConfig).value() &&
                fc->end_TAS > convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, fc->altitude, this->atm,
                EndnodeReadOnly<double>("program_settings/general/landing/thrust_reverser/deactivation_speed").read(this->rtIO->moduleConfig).value())) {
                    oc->rating = "maximum_continuous"; // Reverse thrust
                    fc->thrust_efficiency =
                        EndnodeReadOnly<double>("program_settings/general/landing/thrust_reverser/efficiency").read(this->rtIO->moduleConfig).value();
            } else {
                oc->rating = "idle"; // Idle
                fc->thrust_efficiency = 1.;
            }
        }
        this->set_segment_end_conditions(fc, fc->end_TAS,
            fc->altitude + fc->delta_h, 0., fc->acceleration, this->aero.CLgroundRoll);
        fc->end_specific_air_range = 0.; // Specific air range is set to zero, since aircraft is on the ground
        fc->end_angle_of_attack = 0.; // Determination here not via getAlphaAtFlightCondition, since CA is set manually to CLgroundRoll
        if (fc->delta_x < 0) {
            throw_mission_error(1, a_mission_step);
        } else {
            this->save_segment_data(fc, *oc, a_mission_step);
        }
        /* Clear origin vector */
        fc->origin_polar.clear();
        /* Delete processed bleed and ShaftPower-Offtakes */
        oc->delete_used_bleed_and_shaft_power_offtakes(fc->delta_t);
    }
}

void standard_mission::takeoff_calculation(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step) {
    myRuntimeInfo->out << "Take-off calculation ..." << std::endl;
    oc->set_segment_operating_conditions(a_mission_step);
    /* Set the initial conditions of the segment */
    this->set_segment_initial_conditions(fc);
    /* Takeoff distance is composed of ground run (break release until lift off) and 1st climb segment to screen height (35ft) */
    //Ground Run
    /* Calculation of vLOF (equal to v2 here)
     * Normally, vLOF is 1.1 * vMU (minimum unstick speed) for aero limited and 1.08 * vMU for geometric limited configurations. All Airbus configurations are geometric limited;
     * vMU could be determined by estimating the AoA for tailstrike and use the Cl(AOA_tailstrike) to estimate the needed velocity until lift-off
     * Using v2 here as vLOF is a conservative estimation maintaining v2 during 1st climb segment.
    */
    const double vS1g(sqrt(2. * fc->aircraft_mass * G_FORCE /
        (this->wing_surface_area * this->aero.CLmaxTakeoff * this->atm.getDensity(fc->altitude))));
    const double vLOF(1.128 * vS1g); // CS25.107-B: v2 >= 1.2 * v_Stall = 1.2 * 0.94 * v_Stall_1g = 1.128 * v_Stall_1g
    fc->end_TAS = vLOF;
    fc->delta_h = 0.;
    this->calculate_ground_run(oc, fc, a_mission_step); // Calculation of first part of takeoff distance: ground run
    /* Output of the take off parameters */
    myRuntimeInfo->out << "Lift-off speed: " << this->mission_profile.back().end_TAS << " m/s." << std::endl;
    myRuntimeInfo->out << "Lift-off thrust: " << convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, this->mission_profile.back().end_thrust) << " kN." << std::endl;
    myRuntimeInfo->out << "Lift-off distance: " << this->mission_profile.back().end_range << " m." << std::endl;
    this->set_segment_initial_conditions(fc, true, this->aero.CLgroundRoll);
    this->change_altitude_at_constant_speed(oc, fc, a_mission_step); // Takeoff distance calculation part 2: Climb to screen height (35 ft)
    this->takeoff_distance = this->mission_profile.back().end_range;
    myRuntimeInfo->out << "Takeoff distance total: " << this->takeoff_distance.value() << " m." << std::endl;
}

void standard_mission::climb_to_cruise_altitude(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step) {
    myRuntimeInfo->out << "Climb to cruise altitude ... " << std::endl;
    this->change_altitude_at_constant_speed(oc, fc, a_mission_step);
    if (this->cruise_profile_start == 0 && this->mission_profile.size()) {
        this->cruise_profile_start = this->mission_profile.size() - 1; // Set index
    }
    if (this->requirement_switch &&
        accuracyCheck(this->atm.getTemperature(0.), ISA_TEMPERATURE + this->delta_ISA, ACCURACY_HIGH)) { // Save climb values from 1500ft to ICA
            if (std::isnan(this->time_start_to_ICA)) this->time_start_to_ICA = this->mission_profile.back().end_time;
            if (std::isnan(this->time_1500ft_to_ICA)) this->time_1500ft_to_ICA = this->mission_profile.back().end_time - this->time_start_to_1500ft;
            if (std::isnan(this->range_start_to_ICA)) this->range_start_to_ICA =  this->mission_profile.back().end_range;
            if (std::isnan(this->fuel_start_to_ICA)) { // Sum all used fuel masses
                this->fuel_start_to_ICA = 0; // Initialize 0 kg perfore summing up
                for (std::pair<int, double> fuel_mass : this->mission_profile.back().end_total_fuel_mass) {
                    this->fuel_start_to_ICA -= fuel_mass.second; // Since consumed fuel is negative, subtract those masses
                }
            }
            this->initial_cruise_altitude_check = true;
            this->initial_cruise_altitude = this->mission_profile.back().end_altitude;
            this->initial_cruise_mach = this->mission_profile.back().end_mach;
    }
}

void standard_mission::climb_to_service_ceiling(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step) {
    Flight_conditions start_of_climb_fc = *fc;
    mission_segment start_of_climb_segment = this->mission_profile.back();
    double CAS = convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, fc->altitude, this->atm, fc->TAS); // Stays constant below transition altitude!
    double transition_mach = this->mission_file->mission_data.at(a_mission_step).mach; // Constant climb speed (Mach number) above transition altitude
    double altitude_increment(EndnodeReadOnly<double>("program_settings/precision/altitude_increment").read(this->rtIO->moduleConfig).value());
    if (accuracyCheck(this->atm.getTemperature(0.), ISA_TEMPERATURE + this->delta_ISA, ACCURACY_HIGH)) {
        myRuntimeInfo->out << "Climb over ICA to service ceiling (ISA + " + num2Str(static_cast<uint16_t>(this->atm.getTemperature(0) -
                                ISA_TEMPERATURE)) + ") ... " << std::endl;
    } else {
        myRuntimeInfo->out << "Flying TTC (ISA + " + num2Str(static_cast<uint16_t>(this->atm.getTemperature(0) - ISA_TEMPERATURE)) +
                              ((optimum_ROC) ? ", opt ROC) ... " : ") ... ") << std::endl;
    }
    try {
        this->climb_to_cruise_altitude(oc, fc, a_mission_step);
    } catch (const std::string& text) {
        myRuntimeInfo->warn << "ICA check not successful (" << text << "). Set value to NAN" << std::endl;
    } catch (const int& aNumber) {
        myRuntimeInfo->warn << "ICA could not be reached." << std::endl;
        return;
    }
    /* Update operating condition (increase thrust rating if possible) to show full climb potential for the following checks for service ceiling and max operating altitude */
    oc->rating = "maximum_continuous";
    fc->update_operating_condition(*oc);
    /* Do performance checks */
    if (accuracyCheck(this->atm.getTemperature(0.), ISA_TEMPERATURE + this->delta_ISA, ACCURACY_HIGH)) { //only at ISA + deltaISA conditions needed
        /* Continue climbing for service ceiling */
        /* Set the initial conditions of the segment */
        this->set_segment_initial_conditions(fc);
        fc->delta_h = this->max_operating_altitude_required * 2. - fc->altitude;
        uint16_t steps = static_cast<uint16_t>(ceil(fabs(fc->delta_h / altitude_increment * 2.)));
        fc->delta_h = (steps <= 0.) ? 0. : fc->delta_h / steps;
        try {
            myRuntimeInfo->out << std::endl << "Flying maximum operating altitude ..." << std::endl;
            myRuntimeInfo->out << "Climb Step:          " << convertUnit(METER, FOOT, fc->delta_h) << " ft."
                               << "Total:               " << convertUnit(METER, FOOT, fc->delta_h * steps) << " ft." << std::endl;
            myRuntimeInfo->out << "Start altitude:      " << convertUnit(METER, FOOT, fc->altitude) << " ft." << std::endl;
            myRuntimeInfo->out << "End altitude:        " << convertUnit(METER, FOOT, (fc->delta_h * steps + fc->altitude)) << " ft." << std::endl;
            myRuntimeInfo->out << "Transition altitude: " << convertUnit(METER, FOOT, get_transition_altitude(CAS, transition_mach, this->atm)) << " ft." << std::endl;
            myRuntimeInfo->out << "Engine Rating:       " << oc->rating << std::endl;
            for (uint16_t step_number(0); step_number < steps; ++step_number) {
                /* Set the initial conditions of the segment */
                this->set_segment_initial_conditions(fc);
                if (fc->altitude < get_transition_altitude(CAS, transition_mach, this->atm)) {
                    /* Convert calibrated airspeed into true airspeed */
                    fc->end_TAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, fc->altitude + fc->delta_h, this->atm, CAS);
                } else {
                    /* Convert Mach number into true airspeed */
                    fc->end_TAS = convertUnit(MACH, TRUEAIRSPEED, fc->altitude + fc->delta_h, this->atm, transition_mach);
                }
                this->climb_iteration(oc, fc, a_mission_step);
                this->max_operating_altitude = this->mission_profile.back().end_altitude;
            }
        } catch (const std::string& text) {
            myRuntimeInfo->warn << "maximum operating altitude check not successful (" << text << "). Set value to NAN" << std::endl;
        } catch (const int& a_Value) {
            myRuntimeInfo->out << "maximum operating altitude: " << convertUnit(METER, FOOT, this->max_operating_altitude) << " ft." << std::endl;
        }
        /* Continue climbing for max_ceiling_altitude */
        this->set_segment_initial_conditions(fc);
        fc->delta_h = this->max_operating_altitude_required * 2. - fc->altitude;
        steps = static_cast<uint16_t>(ceil(fabs(fc->delta_h / altitude_increment * 2.)));
        fc->delta_h = (steps <= 0.) ? 0. : fc->delta_h / steps;
        try {
            myRuntimeInfo->out << std::endl << "Flying absolute ceiling altitude ..." << std::endl;
            myRuntimeInfo->out << "Climb Step:         " << convertUnit(METER, FOOT, fc->delta_h) << " ft."
                               << "Total:               " << convertUnit(METER, FOOT, fc->delta_h * steps) << " ft." << std::endl;
            myRuntimeInfo->out << "Start altitude:      " << convertUnit(METER, FOOT, fc->altitude) << " ft." << std::endl;
            myRuntimeInfo->out << "End altitude:        " << convertUnit(METER, FOOT, (fc->delta_h * steps + fc->altitude)) << " ft." << std::endl;
            myRuntimeInfo->out << "Transition altitude: " << convertUnit(METER, FOOT, get_transition_altitude(CAS, transition_mach, this->atm)) << " ft." << std::endl;
            myRuntimeInfo->out << "Engine Rating:       " << oc->rating << std::endl;
            for (uint16_t step_number(0); step_number < steps; ++step_number) {
                this->set_segment_initial_conditions(fc);
                if (fc->altitude < get_transition_altitude(CAS, transition_mach, this->atm)) {
                    /* Convert calibrated airspeed into true airspeed */
                    fc->end_TAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, fc->altitude + fc->delta_h, this->atm, CAS);
                } else {
                    /* Convert Mach number into true airspeed */
                    fc->end_TAS = convertUnit(MACH, TRUEAIRSPEED, fc->altitude + fc->delta_h, this->atm, transition_mach);
                }
                this->climb_iteration(oc, fc, a_mission_step, convertUnit(FOOTPERMINUTE, METERPERSECOND, 50.));
                this->max_ceiling_altitude = this->mission_profile.back().end_altitude;
            }
        } catch (const std::string& text) {
            myRuntimeInfo->warn << "absolute ceiling altitude check not successful (" << text << "). Set value to NAN" << std::endl;
        } catch (const int& a_Value) {
            myRuntimeInfo->out << "absolute ceiling altitude: " << convertUnit(METER, FOOT, this->max_ceiling_altitude) << " ft." << std::endl;
        }
        /* Reset for one engine inoperative */
        *fc = start_of_climb_fc;
        this->mission_profile.push_back(start_of_climb_segment);
        int energy_carrier_id(fc->mission_data->at(fc->operating_condition->mission_step).energy_carrier_id);
        fc->rel_engine_availability =
            static_cast<double>(this->engine.get_number_of_engines(energy_carrier_id) - 1) / static_cast<double>(this->engine.get_number_of_engines(energy_carrier_id)); //OEI
        fc->update_operating_condition(*oc);
        this->set_segment_initial_conditions(fc);
        fc->delta_h = this->max_operating_altitude_required - fc->altitude;
        steps = static_cast<uint16_t>(ceil(fabs(fc->delta_h / altitude_increment * 2.)));
        fc->delta_h = (steps <= 0.) ? 0. : fc->delta_h / steps;
        try {
            myRuntimeInfo->out << std::endl << "Flying maximum altitude with one engine inoperative ..." << std::endl;
            myRuntimeInfo->out << "Climb Step:          " << convertUnit(METER, FOOT, fc->delta_h) << " ft."
                               << " Total:              " << convertUnit(METER, FOOT, fc->delta_h * steps) << " ft." << std::endl;
            myRuntimeInfo->out << "Start altitude:      " << convertUnit(METER, FOOT, fc->altitude) << " ft." << std::endl;
            myRuntimeInfo->out << "End altitude:        " << convertUnit(METER, FOOT, (fc->delta_h * steps + fc->altitude)) << " ft." << std::endl;
            myRuntimeInfo->out << "Transition altitude: " << convertUnit(METER, FOOT, get_transition_altitude(CAS, transition_mach, this->atm)) << " ft." << std::endl;
            myRuntimeInfo->out << "Engine Rating:       " << oc->rating << std::endl;
            for (uint16_t step_number(0); step_number < steps; ++step_number) {
                this->set_segment_initial_conditions(fc);
                if (fc->altitude < get_transition_altitude(CAS, transition_mach, this->atm)) {
                    /* Convert calibrated airspeed into true airspeed */
                    fc->end_TAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, fc->altitude + fc->delta_h, this->atm, CAS);
                } else {
                    /* Convert Mach number into true airspeed */
                    fc->end_TAS = convertUnit(MACH, TRUEAIRSPEED, fc->altitude + fc->delta_h, this->atm, transition_mach);
                }
                this->climb_iteration(oc, fc, a_mission_step);
                this->max_altitude_one_engine_inoperative = std::min(fc->end_altitude, this->max_operating_altitude_required);
            }
        } catch (const std::string& text) {
            myRuntimeInfo->warn << "maximum altitude with one engine inoperative check not successful (" << text << "). Set value to NAN" << std::endl;
        } catch (const int& a_Value) {
            myRuntimeInfo->out << "maximum altitude with one engine inoperative: " << convertUnit(METER, FOOT, this->max_altitude_one_engine_inoperative) << " ft." << std::endl;
        }
        /* Reset available engines (for OEI case) */
        fc->rel_engine_availability = 1.;
    }
}

void standard_mission::climb_iteration(Operating_conditions* oc, Flight_conditions* fc, const uint16_t &mission_step, const double& minimum_ROC,
                            const bool& save_data) {
    /* Climb Iteration until convergence of average climb angle mean_glidepath_angle */
    double residuum(0.);
    uint16_t counter(0);
    //Set start value of segment end conditions for iteration
    this->set_segment_end_conditions(fc, fc->end_TAS, fc->altitude + fc->delta_h,
                                        fc->glidepath_angle, fc->acceleration);
    do {
        /* Set iteration values */
        double old_mean_glidepath_angle(fc->mean_glidepath_angle);
        this->set_segment_end_conditions(fc, fc->end_TAS, fc->end_altitude, fc->end_glidepath_angle, fc->end_acceleration);
        /* Check conditions */
        this->check_iteration_conditions(oc, fc, mission_step, minimum_ROC);
        if (counter == 500) { // Maximum number of iterations (500 iterations) is reached, error message
            throwError(__FILE__, __func__, __LINE__, "Maximum number of iterations reached. Abort program!");
        }
        /* Calculate residuum in mean_glidepath_angle */
        residuum = fabs(old_mean_glidepath_angle - fc->mean_glidepath_angle) / fabs(old_mean_glidepath_angle);
        counter++;
    } while (residuum > ACCURACY_LOW);
    if (save_data) this->save_segment_data(fc, *oc, mission_step);
    /* Clear origin vector */
    fc->origin_polar.clear();
    /* Delete processed bleed and ShaftPower-Offtakes */
    oc->delete_used_bleed_and_shaft_power_offtakes(fc->delta_t);
}

double standard_mission::calculate_top_of_descent_aircraft_mass(const double& current_aircraft_mass, const double& current_specific_air_range, const double& range_until_descent) {
    double top_of_descent_aircraft_mass(0.);
    /* Calculate the Breguet factor for level cruise */
    double breguet_factor_level_cruise = current_specific_air_range * current_aircraft_mass;
    /* Calculate the mass at the top of descent */
    top_of_descent_aircraft_mass = current_aircraft_mass - current_aircraft_mass * (1 - exp(-range_until_descent / breguet_factor_level_cruise));
    return top_of_descent_aircraft_mass;
}

double standard_mission::calculate_climb_or_descent(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step,
                                        const bool& simulation_mode) {
    double range_for_descent(0.);
    /* Set delta_h for climb/descent */
    fc->delta_h = fc->end_altitude - fc->altitude;
    /* Set CAS, Mach, transition_mach and transitionAltitude */
    double CAS(NAN); // Stays constant below transition altitude
    double mach(NAN); // Stays constant above transition altitude
    double transition_mach(NAN);
    /*
     * Case 1: climb to cruise (possible modes: CLIMBTOCRUISE, CLIMBTOCRUISECI, CLIMBTOSERVICECEILING) - Mach needs to be fixed
     *          -> first climb with constant CAS until transition_mach is reached
     *              the transition_mach is set to endMach at cruise altitude and maintained once it is reached for further climb segments
     * Case 2: descent from cruise (possible modes: DESCENT, DESCENTCI) - CAS needs to be fixed
     *          -> the descent is performed with current Mach number until target CAS is reached and then CAS is maintained for further descent
     * default: for CLIMBCONSTCAS, APPROACH, FLCHANGE, FLCHANGESIMULATION, and LANDING, the CAS and Mach equals in every altitude
     */
    switch (this->mission_file->mission_data.at(a_mission_step).mode) {
        case Flight_conditions::CLIMBTOCRUISE:
        case Flight_conditions::CLIMBTOCRUISECI:
        case Flight_conditions::CLIMBTOSERVICECEILING: {
            CAS = convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, fc->altitude, this->atm, fc->TAS);
            mach = this->mission_file->mission_data.at(a_mission_step).mach;
            transition_mach = this->mission_file->desired_cruise_speed +
                this->delta_mach_above_crossover; //Normally, transition Mach Number is equal to initial cruise Mach number + delta
            break;
        }
        case Flight_conditions::DESCENT:
        case Flight_conditions::DESCENTCI: {
            mach = convertUnit(TRUEAIRSPEED, MACH, fc->altitude, this->atm, fc->TAS);
            CAS = this->mission_file->mission_data.at(a_mission_step).CAS;
            transition_mach = //Normally, transition_mach is equal to initial cruise Mach number + delta, but descent is defined different
                                    this->mission_file->desired_cruise_speed;
            break;
        }
        default:
            CAS = convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, fc->altitude, this->atm, fc->TAS);
            mach = convertUnit(TRUEAIRSPEED, MACH, fc->altitude, this->atm, fc->TAS);
            transition_mach = this->mission_file->desired_cruise_speed +
                this->delta_mach_above_crossover; //Normally, transition Mach Number is equal to initial cruise Mach number + delta
    }
    const double& transitionAltitude(get_transition_altitude(CAS, transition_mach, this->atm));
    /* Check if transition altitude is passed during climb/descent */
    bool change_speed_at_transition_altitude(false);
    if ((fc->end_altitude > transitionAltitude && fc->altitude < transitionAltitude) ||
        (fc->end_altitude < transitionAltitude && fc->altitude > transitionAltitude)) { //check if climb/descent through transition altitude
        change_speed_at_transition_altitude = true;
    } else {
        /* Decelerate speed to required CAS if needed
         * Acceleration is not needed. in climb, the required climb Mach above crossover is only an upper boundary --> the endCAS is also treated as an upper bound
         * if end altitude is above transition altitude, the aircraft is already below upper boundary endCAS -> no speed change
         * if current altitude is below transition altitude, the aircraft faster than the upper boundary endCAS -> deceleration needed
         * (this is not needed for climb since after climb to cruise segment there is a dedicated acceleration segment)
         */
        if (fc->delta_h < 0. && fc->altitude < transitionAltitude &&
            (this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::DESCENT ||
                this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::DESCENTCI)) {
             //Descent and current speed above upper boundary for descent speed endCAS
            myRuntimeInfo->out << "Current speed ("
                        << convertUnit(METERPERSECOND, KNOTS, convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, fc->altitude, this->atm, fc->TAS))
                        << " kts) above maximum descent speed according to descent speed schedule (" << convertUnit(METERPERSECOND, KNOTS, CAS) << " kts). "
                        << "Decelerate ..." << std::endl;
            /* To decelerate at constant altitude (ROC = 0), the mode has to be tmporarily switched from DESCENT to DECELERATE and the deltaH step for descent must be saved,
             * because it is set to 0 for speed change at constant altitude
             */
            uint16_t mode_backup(this->mission_file->mission_data.at(a_mission_step).mode);
            this->mission_file->mission_data.at(a_mission_step).mode = Flight_conditions::DECELERATE;
            double delta_h_backup = fc->delta_h;
            /* To decelerate to a constant CAS (first CAS in Descent-Speed-Schedule), the true airspeed has to be used as reference, not the Mach number */
            this->mission_file->mission_data.at(a_mission_step).mach_switch = false;
            range_for_descent += this->change_speed_at_constant_ROC(oc, fc, a_mission_step, simulation_mode);
            /* Set back the mode and the descent step to the stored values */
            this->mission_file->mission_data.at(a_mission_step).mode = mode_backup;
            fc->delta_h = delta_h_backup;
        }
    }
    /* Calculate the number of steps */
    uint16_t steps(0);
    double altitude_increment(EndnodeReadOnly<double>("program_settings/precision/altitude_increment").read(this->rtIO->moduleConfig).value());
    if (simulation_mode) {
        if (this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::DESCENT ||
            this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::DESCENTCI) {
            altitude_increment = 500.;
        } else {
            altitude_increment = 20.;
        }
    }
    steps = static_cast<uint16_t>(ceil(fabs(fc->delta_h / altitude_increment)));
    Flight_conditions fc_tmp(*fc);
    fc_tmp.delta_h /= steps;
    if (!simulation_mode) {
        myRuntimeInfo->out << ((fc_tmp.delta_h > 0.) ? "Climb step:           " : "Descent step:         ")
                            << convertUnit(METER, FOOT, fc_tmp.delta_h) << " ft." << std::endl;
        myRuntimeInfo->out << "Start altitude:       " << convertUnit(METER, FOOT, fc_tmp.altitude) << " ft." << std::endl;
        myRuntimeInfo->out << "End altitude" << ((this->mission_file->mission_data.at(a_mission_step).auto_FL_optimum) ? " (estimated): " : ":         ")
                           << convertUnit(METER, FOOT, (fc_tmp.delta_h * steps + fc_tmp.altitude)) << " ft." << std::endl;
        if (change_speed_at_transition_altitude) myRuntimeInfo->out << "Transition altitude:  " << convertUnit(METER, FOOT, transitionAltitude) << " ft." << std::endl;
    }
    for (uint16_t step_number(0); step_number < steps; ++step_number) {
        this->set_segment_initial_conditions(&fc_tmp, !simulation_mode);
        if (optimum_ROC && this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::CLIMBTOSERVICECEILING) {
            this->calculate_max_ROC_speed(&mach, fc_tmp, 1.);
            fc_tmp.TAS = fc_tmp.end_TAS;
            /* Convert Mach number into true airspeed */
            fc_tmp.end_TAS = convertUnit(MACH, TRUEAIRSPEED, fc_tmp.altitude + fc_tmp.delta_h, this->atm, mach);
        } else if (fc_tmp.altitude >= transitionAltitude) {
            /* Convert Mach number into true airspeed */
            fc_tmp.end_TAS = convertUnit(MACH, TRUEAIRSPEED, fc_tmp.altitude + fc_tmp.delta_h, this->atm, mach);
        } else {
            /* Convert calibrated airspeed into true airspeed */
            fc_tmp.end_TAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, fc_tmp.altitude + fc_tmp.delta_h, this->atm, CAS);
        }
        if (this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::FLCHANGE) {
            this->climb_iteration(oc, &fc_tmp, a_mission_step, this->mission_file->mission_data.at(a_mission_step).ROC);
        } else if (this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::FLCHANGESIMULATION) {
            this->climb_iteration(oc, &fc_tmp, a_mission_step, this->mission_file->mission_data.at(a_mission_step).ROC, !simulation_mode);
        } else if (this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::DESCENT ||
                    this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::DESCENTCI ||
                    this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::APPROACH ||
                    this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::LANDING) {
            this->climb_iteration(oc, &fc_tmp, a_mission_step, 0., !simulation_mode);
        } else {
            Flight_conditions backup_fc(fc_tmp);
            try {
                this->climb_iteration(oc, &fc_tmp, a_mission_step);
            } catch (const std::string& error_message) {
                myRuntimeInfo->warn << "Current engine rating '" << oc->rating << "' not sufficient to climb further." << std::endl;
                if (EndnodeReadOnly<bool>("program_settings/general/increase_engine_rating_during_climb").read(this->rtIO->moduleConfig).value() && oc->rating == "climb") {
                        myRuntimeInfo->out << "  - set rating to 'MaxCont'" << std::endl;
                        oc->rating = "maximum_continuous";
                        fc_tmp = backup_fc;
                        this->climb_iteration(oc, &fc_tmp, a_mission_step);
                } else {
                    throwError(__FILE__, __func__, __LINE__, error_message);
                }
            }
            /* Iterate optimum ICA-FlightLevel */
            if (this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::CLIMBTOCRUISE &&
                                                                    this->mission_file->mission_data.at(a_mission_step).auto_FL_optimum &&
                                                                    step_number == (steps - 2)) {
                myRuntimeInfo->out << "Check optimum ICA ..." << std::endl;
                double old_initial_cruise_altitude(fc_tmp.end_altitude + fc_tmp.delta_h);
                this->calculate_optimum_FL(&fc_tmp.end_altitude, this->mission_file->desired_cruise_speed, 0.,
                                            fc_tmp.aircraft_mass, oc->config, 1.,
                                            oc->rating, this->mission_file->mission_data.at(a_mission_step).round_to_regular_FL);
                if (!accuracyCheck(old_initial_cruise_altitude, fc_tmp.end_altitude, ACCURACY_LOW)) { //Optimum ICA not reached yet
                    myRuntimeInfo->out << "    - Optimum ICA increased from " << convertUnit(METER, FOOT, old_initial_cruise_altitude) << " ft to "
                                       << convertUnit(METER, FOOT, fc_tmp.end_altitude) << " ft." << std::endl;
                    this->set_segment_initial_conditions(&fc_tmp, !simulation_mode);
                    fc_tmp.delta_x = this->calculate_climb_or_descent(oc, &fc_tmp, a_mission_step, simulation_mode);
                    ++step_number;
                } else {
                    myRuntimeInfo->out << "    - ICA reached at " << convertUnit(METER, FOOT, fc_tmp.end_altitude) << "." << std::endl;
                }
            }
        }
        /* Abort climb if CL_max is exceeded */
        double current_CL_max(aero.getCLmax(
                convertUnit(TRUEAIRSPEED, MACH, fc_tmp.altitude + fc_tmp.delta_h, this->atm, fc_tmp.end_TAS), oc->config));
        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            current_CL_max = this->interpolateViaCoG(current_cg, current_CL_max,
                            getAlternativeAero(current_cg)->getCLmax(convertUnit(TRUEAIRSPEED, MACH, fc_tmp.altitude + fc_tmp.delta_h, this->atm, fc_tmp.end_TAS), oc->config));
        }
        if (fc_tmp.end_CL > current_CL_max) {
            // break;
            throwError(__FILE__, __func__, __LINE__, "Needed CL for climb exceeds CL_max. Abort climb at " +
                        num2Str(convertUnit(METER, FOOT, fc_tmp.altitude)) + " ft.");
        }
        if (simulation_mode) {
            range_for_descent += fc_tmp.delta_x;
            /* Set new conditions */
            fc_tmp.aircraft_mass -= fc_tmp.step_fuel_mass;
            fc_tmp.TAS = fc_tmp.end_TAS;
            fc_tmp.altitude += fc_tmp.delta_h;
        }
    }
    if (simulation_mode) {
        *fc = fc_tmp;
    } else {
        this->set_segment_end_conditions(fc, fc_tmp.end_TAS, fc_tmp.end_altitude, fc_tmp.end_glidepath_angle, fc_tmp.end_acceleration);
    }
    return range_for_descent;
}

void standard_mission::estimate_descent(const bool& last_segment_free_flight_switch) {
    /* Estimate distance for the descend */
    /* Initialize operating conditions object */
    Operating_conditions descent_oc(&(this->mission_file->mission_data));
    /* Set the operating conditions of the segment */
    descent_oc.set_segment_operating_conditions(this->top_of_descent_index);
    /* Initialize flight conditions object */
    Flight_conditions descent_fc(&(this->mission_file->mission_data), &descent_oc,
                                                &this->aero, &this->atm, &this->engine, this->variable_camber_segment_quantity);
    Flight_conditions descent_fc_next_FL(descent_fc);
    /* Get height at top of descent (TOD) and bottom of descent (BOD) */
    if (last_segment_free_flight_switch) { // Determine FF height and speed
        this->mission_file->mission_data.at(this->top_of_descent_index).mach = this->mission_file->desired_cruise_speed;
        this->mission_file->mission_data.at(this->top_of_descent_index).altitude = 10000.;
        /* Calculate the optimum speed and flight level */
        this->calcOptSpeedAndFL(&(this->mission_file->mission_data.at(this->top_of_descent_index).mach),
            &this->mission_file->mission_data.at(this->top_of_descent_index).altitude, 0., this->OME + this->payload_mass.value(),
            "clean", 1., "cruise", descent_fc.mission_data->at(descent_oc.mission_step).energy_carrier_id);
    }
    /* Set altitude at the top of descent */
    this->top_of_descent_altitude = this->mission_file->mission_data.at(this->top_of_descent_index).altitude; // Set altitude at the top of descent

    /* Calculation of distances */
    //Set Initial conditions
    descent_fc.end_altitude = this->top_of_descent_altitude; // Set the end altitude
    descent_fc_next_FL.end_altitude = this->top_of_descent_altitude +
        EndnodeReadOnly<double>("program_settings/precision/altitude_increment").read(this->rtIO->moduleConfig).value(); // Set the end altitude of next flight level above
    descent_fc.end_TAS = convertUnit(MACH, TRUEAIRSPEED, descent_fc.end_altitude, this->atm,
                        this->mission_file->mission_data.at(this->top_of_descent_index).mach); // Convert Mach number into true airspeed
    descent_fc_next_FL.end_TAS = convertUnit(MACH, TRUEAIRSPEED, descent_fc_next_FL.end_altitude, this->atm,
                        this->mission_file->mission_data.at(this->top_of_descent_index).mach);
    std::pair<double, double> the_descend;
    the_descend = this->calculate_descent(descent_oc, descent_fc, this->top_of_descent_mass.value()); // Calculation of descend distance (range descent) and fuel
    this->descent_range = the_descend.first;
    this->touchdown_mass = this->top_of_descent_mass.value() - the_descend.second;
    the_descend = this->calculate_descent(descent_oc, descent_fc_next_FL,
                    this->top_of_descent_mass_next_FL); // Calculation of descend distance for FL above top_of_descent_altitude and its fuel consumption
    this->descent_range_next_FL = the_descend.first;
    this->touchdown_mass_next_FL = this->top_of_descent_mass_next_FL - the_descend.second;
    double cruise_range = this->mission_file->Range - this->descent_range; // Calculation of range in cruise mode
    /* Output of descent values */
    myRuntimeInfo->out << "Range Descent: " << convertUnit(METER, NAUTICALMILE, this->descent_range) <<
                        " NM from altitude: " << convertUnit(METER, FOOT, this->top_of_descent_altitude) << " ft" << std::endl;
    myRuntimeInfo->out << "Range Cruise: " << convertUnit(METER, NAUTICALMILE, cruise_range) << " NM "
                       << "(for total distance: " << convertUnit(METER, NAUTICALMILE, (cruise_range + this->descent_range)) << " NM)." << std::endl;
    myRuntimeInfo->info << "Top of descent mass: " << this->top_of_descent_mass.value() << std::endl;
}

std::pair<double, double> standard_mission::calculate_descent(const Operating_conditions& oc,
                                                            const Flight_conditions& fc, double top_of_descent_mass_tmp) {
    //Save engine operating conditions for normal cruise before changes from flight level change simulation occur
    int energy_carrier_id(fc.mission_data->at(fc.operating_condition->mission_step).energy_carrier_id);
    std::vector<OperatingPoint> backup_operating_points = this->engine.get_operating_points(energy_carrier_id);
    Operating_conditions descent_oc = oc;
    Flight_conditions descent_fc(&this->mission_file->mission_data, &descent_oc, &this->aero, &this->atm, &this->engine, this->variable_camber_segment_quantity);
    double descent_range_tmp(0.);
    double aircraft_mass_before_descent(top_of_descent_mass_tmp);
    /* Set initial flight condition */
    descent_fc.aircraft_mass = top_of_descent_mass_tmp;
    descent_fc.altitude = std::max(this->top_of_descent_altitude, fc.end_altitude);
    descent_fc.TAS = convertUnit(MACH, TRUEAIRSPEED, descent_fc.altitude, this->atm, this->mission_file->mission_data.at(this->top_of_descent_index).mach);
    /* Calculate descend range */
    for (size_t i(this->end_of_descent_index); i < this->mission_file->mission_data.size(); ++i) {
        if (this->mission_file->mission_data.at(i).mode == Flight_conditions::DESCENT) { // Descend_to_Approach
            descent_range_tmp += this->change_altitude_at_constant_speed(&descent_oc, &descent_fc, i, true);
        } else if (this->mission_file->mission_data.at(i).mode == Flight_conditions::APPROACH) { // Descend
            descent_range_tmp += this->change_altitude_at_constant_speed(&descent_oc, &descent_fc, i, true);
        } else if (this->mission_file->mission_data.at(i).mode == Flight_conditions::DECELERATE) { // Change speed
            descent_range_tmp += this->change_speed_at_constant_ROC(&descent_oc, &descent_fc, i, true);
        }  else if (this->mission_file->mission_data.at(i).mode == Flight_conditions::GSINTERCEPTION) { // GlideslopeInterception
            descent_range_tmp += this->glideslope_interception(&descent_oc, &descent_fc, i, true);
        }  else if (this->mission_file->mission_data.at(i).mode == Flight_conditions::LANDING) { // Flare to touch down
            descent_range_tmp += this->landing(&descent_oc, &descent_fc, i, true);
        }
    }
    this->engine.set_operating_points(energy_carrier_id, backup_operating_points); // reset OperatingPoint to initial conditions
    return std::make_pair(descent_range_tmp, aircraft_mass_before_descent - descent_fc.end_aircraft_mass);
}

std::pair<double, double> standard_mission::iterate_descent(double* segmend_end_range, double* top_of_descent_aircraft_mass,
                                                            const double& current_descent_distance, const Operating_conditions& oc,
                                                            const Flight_conditions& fc, const uint16_t& a_mission_step) {
    std::pair<double, double> the_descend;
    double descend_distance(current_descent_distance);
    double iteration_accuracy(1.); //Iteration accuracy is 1NM
    double residuum(iteration_accuracy); //Initialization of residuum with the iteration accuracy
    do {
        double current_ange_until_descent(this->mission_file->Range - this->mission_profile.back().end_range - descend_distance);
        /* Calculate the new aircraft mass at top of descent */
        *top_of_descent_aircraft_mass = this->calculate_top_of_descent_aircraft_mass(fc.end_aircraft_mass, fc.end_specific_air_range, current_ange_until_descent);
        /* Calculate the new descent range */
        the_descend = this->calculate_descent(oc, fc, *top_of_descent_aircraft_mass);
        double new_descend_distance(the_descend.first);
        residuum = (new_descend_distance - descend_distance) / descend_distance;
        /* Update segment information and descent distance */
        *segmend_end_range += (descend_distance - new_descend_distance);
        descend_distance = new_descend_distance;
    } while (residuum > iteration_accuracy);
    return std::make_pair(the_descend.first, the_descend.second);
}

double standard_mission::glideslope_interception(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step,
                                       const bool& simulation_mode) {
    double range_for_descent(0.);
    if (!simulation_mode) myRuntimeInfo->out << "Glideslope Interception ... " << std::endl;
    /* Set the operating conditions of the segment */
    oc->set_segment_operating_conditions(a_mission_step);
    /* Set the initial conditions of the segment */
    this->set_segment_initial_conditions(fc, !simulation_mode);
    fc->end_TAS = fc->TAS;
    if (this->mission_file->mission_data.size() < a_mission_step) {
        throwError(__FILE__, __func__, __LINE__, "No mission data for landing after Glide Slope Interception! Abort program!");
    }
    double final_glideslope(this->mission_file->mission_data.at(a_mission_step + 1).glidepath);
    if (!simulation_mode) {
        fc->delta_x = this->mission_file->Range - fc->altitude / tan(-final_glideslope) - this->mission_profile.back().end_range;
    } else {
        fc->delta_x =  EndnodeReadOnly<double>("program_settings/general/glideslope_interception_distance").read(this->rtIO->moduleConfig).value();
    }
    myRuntimeInfo->debug << " delta_x: " << fc->delta_x << " dxGS: " << fc->altitude / tan(-final_glideslope) << std::endl;
    /* Calculate the number of steps 1 km */
    if (fc->delta_x < 0) {
        myRuntimeInfo->out << "Glide slope interception overflown!" << std::endl;
    } else {
        uint16_t steps = static_cast<uint16_t>(ceil((fc->delta_x / 500.))); // 10 m accuracy
        fc->delta_x = fc->delta_x / steps;
        myRuntimeInfo->debug << "fc->delta_x: " << fc->delta_x << std::endl;
        for (uint16_t step_number(0); step_number < steps; ++step_number) {
            /* Set the initial conditions of the segment */
            this->set_segment_initial_conditions(fc, !simulation_mode);
            this->cruise_iteration(oc, fc, a_mission_step, !simulation_mode);
            if (simulation_mode) {
                range_for_descent += fc->delta_x;
                /* Set new conditions */
                fc->aircraft_mass -= fc->step_fuel_mass;
                fc->TAS = fc->end_TAS;
                fc->altitude += fc->delta_h;
            }
        }
    }
    return range_for_descent;
}

double standard_mission::landing(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step, const bool& simulation_mode) {
    /**< \todo [T148] (https://unicado.ilr.rwth-aachen.de/T148): Consider using this method - http://walter.bislins.ch/aviatik/index.asp?page=Bremsweg+eines+Verkehrsflugzeugs
     */
    if (!simulation_mode) myRuntimeInfo->out << "Landing ... " << std::endl;
    double descent_to_touchdown_range(0.);
    oc->set_segment_operating_conditions(a_mission_step);
    /* Set the initial conditions for the segment */
    this->set_segment_initial_conditions(fc, !simulation_mode);
    if (!simulation_mode) this->landing_distance = this->mission_profile.back().end_range; //get current range when landing is started
    /* Calculation of landing route: Approach from 50ft at threshold with flare to touch down + ground run */
    //Approach to touch down
    // [2] p.313-324
    descent_to_touchdown_range = this->change_altitude_at_constant_speed(oc, fc, a_mission_step, simulation_mode);
    if (!simulation_mode) {
        myRuntimeInfo->out << "Touchdown after " << convertUnit(METER, NAUTICALMILE, this->mission_profile.back().end_range) << " NM" << std::endl;
        //Ground run
        myRuntimeInfo->out << "Rolling on the runway during the reaction time ..." << std::endl;
        this->mission_file->mission_data.at(a_mission_step).glidepath = 0.; // set track angle of last segment to zero for ground run
        this->mission_file->mission_data.at(a_mission_step).mode = Flight_conditions::TAKEOFF;
        this->set_segment_initial_conditions(fc, true, this->aero.CLgroundRoll);
        fc->end_altitude = 0.;
        fc->end_glidepath_angle = 0.;
        fc->end_acceleration = fc->acceleration;
        fc->delta_t = EndnodeReadOnly<double>("program_settings/general/landing/rotation_time").read(this->rtIO->moduleConfig).value();
        fc->end_TAS = fc->TAS + fc->end_acceleration * fc->delta_t;
        this->set_segment_end_conditions(fc, fc->end_TAS, fc->end_altitude,
                                        fc->end_glidepath_angle, fc->end_acceleration, fc->end_CL);
        /* Set new speed and way for ground run during reaction/rotation time after touchdown */
        double residuum(0.);
        do {
            double oldTAS(fc->end_TAS);
            fc->delta_t = EndnodeReadOnly<double>("program_settings/general/landing/rotation_time").read(this->rtIO->moduleConfig).value();
            fc->end_TAS = fc->TAS + fc->mean_acceleration * fc->delta_t;
            this->set_segment_end_conditions(fc, fc->end_TAS, fc->end_altitude, fc->end_glidepath_angle, fc->end_acceleration,
                                                      fc->end_CL);
            /* Calculate residuum in end_TAS */
            residuum = fabs(oldTAS - fc->end_TAS) / fabs(oldTAS);
        } while (residuum > ACCURACY_LOW);
        fc->end_CL = this->aero.CLgroundRoll;
        fc->end_L_over_D = this->aero.getCLLoverD(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS), fc->end_CL, fc->operating_condition->config);
        fc->end_angle_of_attack = 0.;
        fc->end_specific_air_range = 0.;
        this->mission_file->mission_data.at(a_mission_step).mode = Flight_conditions::LANDING;
        this->save_segment_data(fc, *oc, a_mission_step);
        /* Clear origin vector */
        fc->origin_polar.clear();
        /* Delete processed bleed and ShaftPower-Offtakes */
        oc->delete_used_bleed_and_shaft_power_offtakes(fc->delta_t);
        //Deceleration from touchdown speed to runway exit speed
        double range_at_deceleration_start(this->mission_profile.back().end_range);
        if (fc->end_TAS > EndnodeReadOnly<double>("program_settings/general/landing/runway_exit_speed").read(this->rtIO->moduleConfig).value()) {
            fc->end_TAS = EndnodeReadOnly<double>("program_settings/general/landing/runway_exit_speed").read(this->rtIO->moduleConfig).value();
            myRuntimeInfo->out << "Braking to taxi speed (" << convertUnit(METERPERSECOND, KNOTS, fc->end_TAS) << " kts) ..." << std::endl;
            this->calculate_ground_run(oc, fc, a_mission_step);
        }
        this->landing_distance = this->mission_profile.back().end_range - this->landing_distance.value();
        myRuntimeInfo->out << "Ground Run Distance: " << (this->mission_profile.back().end_range - range_at_deceleration_start) << " m." << std::endl;
        myRuntimeInfo->out << "Landing Distance: " << this->landing_distance.value() << " m." << std::endl;
    }
    return descent_to_touchdown_range;
}

double standard_mission::change_speed_at_constant_ROC(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step,
                                         const bool& simulation_mode) {
    if (!simulation_mode) myRuntimeInfo->out << "Change speed with constant ROC ... " << std::endl;
    double speed_change_range(0.);
    double speed(0.);
    double delta_speed(0.);
    double acceleration_increment(0.);
    double minimum_acceleration(0.);
    /* Set the operating conditions of the segment */
    oc->set_segment_operating_conditions(a_mission_step);
    /* Set the initial conditions of the segment */
    this->set_segment_initial_conditions(fc, !simulation_mode);
    if (this->mission_file->mission_data.at(a_mission_step).mach_switch) {
        /* Convert true airspeed into Mach number */
        speed = convertUnit(TRUEAIRSPEED, MACH, fc->altitude, this->atm, fc->TAS);
        /* Calculate the number of steps, acceleration delta 0.005 Mach */
        delta_speed = this->mission_file->mission_data.at(a_mission_step).mach - speed;
        /* Set acceleration increment */
        acceleration_increment = EndnodeReadOnly<double>("program_settings/precision/mach_acceleration_increment").read(this->rtIO->moduleConfig).value();
        /* Set minimum acceleration */
        minimum_acceleration = 0.01 * signum(delta_speed);
    } else {
        /* Convert true airspeed into calibrated airspeed */
        speed = convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, fc->altitude, this->atm, fc->TAS);
        /* Calculate the number of steps, acceleration delta 1 m/s */
        delta_speed = this->mission_file->mission_data.at(a_mission_step).CAS - speed;
        /* Set acceleration increment */
        acceleration_increment = EndnodeReadOnly<double>("program_settings/precision/acceleration_increment").read(this->rtIO->moduleConfig).value();
        if (simulation_mode) acceleration_increment *= 5.; // for estimation use five times the original step size for fast estimation
        /* Set minimum acceleration */
        minimum_acceleration = 0.02 * signum(delta_speed);
    }
    uint16_t steps(0); // Number of steps
    /* Calculate the number of steps */
    if (fabs(delta_speed) > ACCURACY_HIGH) {
        steps = static_cast<uint16_t>(ceil(fabs(delta_speed / acceleration_increment)));
        delta_speed /= steps;
    } else {
        delta_speed = 0.;
    }
    if (!simulation_mode) {
        if (this->mission_file->mission_data.at(a_mission_step).mach_switch) {
            /* Output of Mach number parameters */
            myRuntimeInfo->out << "Acceleration step Mach: " << delta_speed << std::endl;
            myRuntimeInfo->out << "Start Mach number: " << speed << std::endl;
            myRuntimeInfo->out << "End Mach number: " << delta_speed * steps + speed << std::endl;
        } else {
            /* Output of calibrated airspeed parameters */
            myRuntimeInfo->out << "Acceleration step CAS: " << convertUnit(METERPERSECOND, KNOTS, delta_speed) << " kts" << std::endl;
            myRuntimeInfo->out << "Start speed CAS: " << convertUnit(METERPERSECOND, KNOTS, speed) << " kts" << std::endl;
            myRuntimeInfo->out << "End speed CAS: " << convertUnit(METERPERSECOND, KNOTS, (delta_speed * steps + speed)) << " kts" << std::endl;
        }
    }
    for (uint16_t step_number(0); step_number < steps; ++step_number) {
        /* Set the initial conditions of the segment */
        this->set_segment_initial_conditions(fc, !simulation_mode);
        speed += delta_speed; // Change Mach number by delta
        if (this->mission_file->mission_data.at(a_mission_step).mach_switch) {
            /* Convert Mach number into true airspeed */
            fc->end_TAS = convertUnit(MACH, TRUEAIRSPEED, fc->altitude, this->atm, speed);
        } else {
            /* Convert calibrated airspeed into true airspeed */
            fc->end_TAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, fc->altitude, this->atm, speed);
        }
        Flight_conditions backup_fc(*fc);
        try {
            this->fixed_ROC_acceleration_iteration(oc, fc, minimum_acceleration, a_mission_step, !simulation_mode);
        } catch (const std::string& error_message) {
            myRuntimeInfo->warn << "Current engine rating '" << oc->rating << "' not sufficient to climb further." << std::endl;
            if (EndnodeReadOnly<bool>("program_settings/general/increase_engine_rating_during_climb").read(this->rtIO->moduleConfig).value() &&
                oc->rating == "climb") {
                    myRuntimeInfo->out << "  - set rating to 'MaxCont'" << std::endl;
                    oc->rating = "maximum_continuous";
                    *fc = backup_fc;
                    this->fixed_ROC_acceleration_iteration(oc, fc, minimum_acceleration, a_mission_step, !simulation_mode);
            } else {
                throwError(__FILE__, __func__, __LINE__, error_message);
            }
        }
        if (simulation_mode) {
            speed_change_range += fc->delta_x;
            /* Set new conditions */
            fc->aircraft_mass -= fc->step_fuel_mass;
            fc->TAS = fc->end_TAS;
            fc->altitude += fc->delta_h;
        }
    }
    return speed_change_range;
}

double standard_mission::change_altitude_at_constant_speed(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step,
                                              const bool& simulation_mode) {
    if (!simulation_mode) {
        myRuntimeInfo->out << "Climb or descent with constant CAS or Mach number "
                           << ((this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::FLCHANGE) ? "(with constant ROC) ..." : "...")
                           << std::endl;
    }
    /* Set the operating conditions of the segment */
    oc->set_segment_operating_conditions(a_mission_step);
    if (this->mission_file->mission_data.at(a_mission_step).mode ==
            Flight_conditions::FLCHANGESIMULATION ||
            this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::FLCHANGE) {
        /* Set engine rating for flight level change to that preset in the mission.xml; if no flight level change segment is available, the rating Climb is used */
        std::vector<mission_file_data_handler::mission_data_set>::iterator theFLchangeMode(find_if(this->mission_file->mission_data.begin(),
                                                    this->mission_file->mission_data.end(),
                                                    [](const mission_file_data_handler::mission_data_set& a_mission_step){return a_mission_step.mode_name == "ChangeFLconstROC";}));
        if (theFLchangeMode != this->mission_file->mission_data.end()) {
            oc->rating = (*theFLchangeMode).rating;
        } else {
            oc->rating = "climb"; //Set engine rating to Climb as default for FLchange
        }
        fc->update_operating_condition(*oc);
    }
    /* Set the initial conditions of the segment */
    this->set_segment_initial_conditions(fc, !simulation_mode);
    uint16_t mode_backup(this->mission_file->mission_data.at(a_mission_step).mode);
    if (this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::TAKEOFF) {
        /* Climb from LOF to Screen Height (35 ft) with constant CAS */
        fc->end_altitude = convertUnit(FOOT, METER, 35.); // Conversion of end_altitude from [ft] to [m]
        this->mission_file->mission_data.at(a_mission_step).mode = Flight_conditions::CLIMBCONSTCAS;
    } else if (this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::LANDING) {
        fc->end_altitude = convertUnit(FOOT, METER, 0.); // Conversion of end_altitude from [ft] to [m] (kept in with conversion to mark the Unit)
    } else if (this->mission_file->mission_data.at(a_mission_step).mode ==
                        Flight_conditions::CLIMBTOCRUISE && this->mission_file->mission_data.at(a_mission_step).auto_FL_optimum) {
        this->calculate_optimum_FL(&fc->end_altitude, this->mission_file->desired_cruise_speed, 0., fc->aircraft_mass, oc->config,
                            1., oc->rating, this->mission_file->mission_data.at(a_mission_step).round_to_regular_FL);
    } else {
        fc->end_altitude = std::min(this->mission_file->mission_data.at(a_mission_step).altitude, this->max_operating_altitude_required);
    }
    double delta_range(this->calculate_climb_or_descent(oc, fc, a_mission_step, simulation_mode));
    /* Reset mode */
    this->mission_file->mission_data.at(a_mission_step).mode = mode_backup;
    if (this->requirement_switch && this->mission_file->mission_data.at(a_mission_step).mode == Flight_conditions::CLIMBCONSTCAS) { // save TTC
        // Save time up to 1500 ft
        if (fabs(this->mission_profile.back().end_altitude - convertUnit(FOOT, METER, 1500.)) < ACCURACY_LOW &&
            accuracyCheck(this->atm.getTemperature(0.), ISA_TEMPERATURE + this->delta_ISA, ACCURACY_HIGH) &&
            std::isnan(this->time_start_to_1500ft)) { // write ISA_deltaISA values
                this->time_start_to_1500ft = this->mission_profile.back().end_time;
        }
    }
    return delta_range;
}

void standard_mission::cruise_at_constant_mach(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step) {
    myRuntimeInfo->out << "Cruise at constant Mach number ... " << std::endl;
    bool iterate_top_of_descent_mass(EndnodeReadOnly<bool>("program_settings/general/iterate_top_of_descent_mass").read(this->rtIO->moduleConfig).value());
    int energy_carrier_id(fc->mission_data->at(fc->operating_condition->mission_step).energy_carrier_id);

    /* Set the operating conditions of the segment */
    oc->set_segment_operating_conditions(a_mission_step);
    /* Set the initial conditions of the segment */
    this->set_segment_initial_conditions(fc);
    double x_end = this->mission_file->mission_data.at(a_mission_step).rel_segment_length *
                    (this->mission_file->Range - this->descent_range - this->mission_profile.at(this->cruise_profile_start).end_range) +
                    this->mission_profile.at(this->cruise_profile_start).end_range;
                    // this->mission_profile.at(this->cruise_profile_start).end_range is range at ICA (before acceleration)
    /* Calculate the number of steps 1 km */
    fc->delta_x = x_end - this->mission_profile.back().end_range;
    if (fc->delta_x < 0. || (this->mission_file->Range - this->descent_range) - this->mission_profile.back().end_range < 0.) {
        myRuntimeInfo->out << "Cruise distance already covered!" << std::endl;
    } else {
        while ((x_end - this->mission_profile.back().end_range) / this->mission_profile.back().end_range > ACCURACY_LOW &&
               (this->mission_file->Range - this->descent_range) - this->mission_profile.back().end_range > 0.) {
            /* Iterate descent distance */
            if (iterate_top_of_descent_mass) {
                double tmp_TOD = this->top_of_descent_mass.value();
                std::pair<double, double> the_descend(this->iterate_descent(
                    &x_end, &tmp_TOD, this->descent_range, *oc, *fc, a_mission_step));
                this->top_of_descent_mass = tmp_TOD;
                this->descent_range = the_descend.first;
                this->touchdown_mass = this->top_of_descent_mass.value() - the_descend.second;
            }
            fc->delta_x = x_end - this->mission_profile.back().end_range;
            uint16_t rangeSteps =
                static_cast<uint16_t>(ceil((fc->delta_x /
                EndnodeReadOnly<double>("program_settings/precision/way_increment").read(this->rtIO->moduleConfig).value())));
            uint16_t rangeStepsFLchange = static_cast<uint16_t>(ceil(fc->delta_x /
                (EndnodeReadOnly<double>("program_settings/precision/specific_air_range_check_increment").read(this->rtIO->moduleConfig).value() *
                                            this->mission_file->Range)));
            fc->delta_x /= rangeSteps;
            for (uint16_t step_number(0); step_number < rangeSteps; ++step_number) {
                /* Set the initial conditions of the segment */
                this->set_segment_initial_conditions(fc);
                fc->end_TAS = fc->TAS;
                if (this->mission_file->mission_data.at(a_mission_step).auto_FL_optimum &&
                    ((step_number % static_cast<uint16_t>(rangeSteps / rangeStepsFLchange)) == 0 || step_number == 0)) {
                    /*******FLIGHT LEVEL CHANGE SIMULATION ***************/
                    /* Calculate conditions with flight level change */
                    Operating_conditions backup_oc(*oc);
                    Flight_conditions backup_fc(*fc);
                    Flight_conditions FL_change_fc(*fc);
                    //Save engine operating conditions for normal cruise before changes from flight level change simulation occur
                    std::vector<OperatingPoint> backup_operating_points = this->engine.get_operating_points(energy_carrier_id);
                    try {
                        this->simulate_FL_change(&FL_change_fc, *oc, this->mission_file->mission_data.at(a_mission_step).auto_altitude_steps, a_mission_step);
                    } catch (const std::string& error_message) {
                        /* Reset all values to backup */
                        this->engine.set_operating_points(energy_carrier_id, backup_operating_points); //Reset engine operating conditions to normal cruise conditions
                        *oc = backup_oc;
                        *fc = backup_fc;
                        this->mission_file->mission_data.at(a_mission_step).mode = Flight_conditions::CRUISE;
                        /* Perform cruise iteration */
                        this->cruise_iteration(oc, fc, a_mission_step, true);
                        continue;
                    }
                    /* Update descent distance from higher flight level after first check */
                    if (iterate_top_of_descent_mass && (step_number / static_cast<uint16_t>(rangeSteps / rangeStepsFLchange) == 1)) {
                            double currentEndCruiseSegment(x_end);
                            std::pair<double, double> the_descend;
                            the_descend = this->iterate_descent(&currentEndCruiseSegment, &this->top_of_descent_mass_next_FL, this->descent_range_next_FL,
                                            *oc, FL_change_fc, a_mission_step);
                            this->descent_range_next_FL = the_descend.first;
                            this->touchdown_mass_next_FL = this->top_of_descent_mass_next_FL - the_descend.second;
                    }
                    this->engine.set_operating_points(energy_carrier_id, backup_operating_points); //Reset engine operating conditions to normal cruise conditions
                    /*******CONTINUOUS CRUISE SIMULATION ***************/
                    /* Calculate conditions for level cruise without flight level change */
                    double cruise_range(0.);
                    double cruise_fuel(0.);
                    while ((this->mission_profile.back().end_range + cruise_range) <= (this->mission_profile.back().end_range + FL_change_fc.delta_x)) {
                        this->cruise_iteration(oc, fc, a_mission_step, false);
                        fc->aircraft_mass = fc->end_aircraft_mass;
                        fc->TAS = fc->end_TAS;
                        fc->altitude = fc->end_altitude;
                        this->set_segment_initial_conditions(fc, false);
                        cruise_range += fc->delta_x;
                        cruise_fuel += fc->step_fuel_mass;
                    }
                    /*******STATE COMPARISON ***************/
                    /* Compare conditions and decide for doing a flight level change or not */
                    bool doFLChange(this->decide_FL_change(*fc, FL_change_fc,
                                        this->mission_file->mission_data.at(a_mission_step).CI, cruise_range, cruise_fuel));
                    if (doFLChange) {
                        // Check for max_operating_altitude with ACCURACY_LOW to avoid not allowing step due to numerical uncertainties
                        if (fc->altitude + this->mission_file->mission_data.at(a_mission_step).auto_altitude_steps
                                > this->max_operating_altitude_required + ACCURACY_LOW) {
                            myRuntimeInfo->err << "Cannot change flight level because maximum operating altitude would be exceeded (max_operating_altitude = "
                                               << convertUnit(METER, FOOT, this->max_operating_altitude_required) << " ft)." << std::endl;
                            this->mission_file->mission_data.at(a_mission_step).auto_FL_optimum = false;
                        } else {
                            myRuntimeInfo->out << "---> Automatic FL change" << std::endl;
                            /* Reset all values */
                            this->engine.set_operating_points(energy_carrier_id, backup_operating_points); //Reset engine operating conditions to normal cruise conditions
                            *oc = backup_oc;
                            *fc = backup_fc;
                            /* Set new mission data */
                            this->mission_file->mission_data.at(a_mission_step).mode = Flight_conditions::FLCHANGE;
                            this->mission_file->mission_data.at(a_mission_step).altitude =
                                fc->altitude + this->mission_file->mission_data.at(a_mission_step).auto_altitude_steps;
                            /* Change FL */
                            this->set_segment_initial_conditions(fc);
                            this->change_altitude_at_constant_speed(oc, fc, a_mission_step);
                            Flight_conditions finalSegmentFLchangeReal(*fc);
                            /* Reset values */
                            myRuntimeInfo->out << "<--- continue cruising..." << std::endl;
                            this->mission_file->mission_data.at(a_mission_step).mode = Flight_conditions::CRUISE;
                            oc->rating = "cruise";
                            fc->update_operating_condition(*oc);
                            this->set_segment_initial_conditions(fc);
                            this->cruise_iteration(oc, fc, a_mission_step, true);
                            break;
                        }
                    }
                    /* Reset all values */
                    this->engine.set_operating_points(energy_carrier_id, backup_operating_points); //Reset engine operating conditions to normal cruise conditions
                    *oc = backup_oc;
                    *fc = backup_fc;
                }
                this->cruise_iteration(oc, fc, a_mission_step, true);
                if (accuracyCheck(fc->delta_x, 0., ACCURACY_LOW)) {
                    throwError(__FILE__, __func__, __LINE__, "Thrust too low to continue cruise! Abort program!");
                }
            }
            /* Iterate descent distance */
            if (iterate_top_of_descent_mass) {
                double tmp_TOD = this->top_of_descent_mass.value();
                std::pair<double, double> the_descend(this->iterate_descent(&x_end, &tmp_TOD, this->descent_range,
                                                        *oc, *fc, a_mission_step));
                this->top_of_descent_mass = tmp_TOD;
                this->descent_range = the_descend.first;
                this->touchdown_mass = this->top_of_descent_mass.value() - the_descend.second;
            }
        }
    }
}

void standard_mission::air_to_air_fuelling(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step) {
    myRuntimeInfo->out << "Air-to-air Fueling ... " << std::endl;
    oc->set_segment_operating_conditions(a_mission_step);
    /* Set the initial conditions of the segment */
    this->set_segment_initial_conditions(fc);
    /* Set the true airspeed at the end of the segment */
    fc->end_TAS = fc->TAS;
    /* Calculate the number of steps 1 km */
    double fuel_transfer_to_manoeuvre_time(fabs(this->mission_file->mission_data.at(a_mission_step).rel_delta_fuel_mass *
                                        this->mission_file->air_to_air_fuel_mass) / this->mission_file->mission_data.at(a_mission_step).transfer_rate /
                                        this->mission_file->mission_data.at(a_mission_step).transfer_time);
    if (fuel_transfer_to_manoeuvre_time > .9) {
        myRuntimeInfo->err << "Relative Air-to-air fuel amount " << fuel_transfer_to_manoeuvre_time << " > 90%! Abort program!" << std::endl;
        exit(1);
    }
    fc->delta_x = convertUnit(MINUTE, SECOND, this->mission_file->mission_data.at(a_mission_step).transfer_time) * fc->TAS;
    if (this->mission_file->mission_data.at(a_mission_step).rel_delta_fuel_mass > 1.) {
        myRuntimeInfo->err << "Relative Air-to-air fuel amount " << this->mission_file->mission_data.at(a_mission_step).rel_delta_fuel_mass
                            << " > 1! Abort program!" << std::endl;
        exit(1);
    }
    uint16_t steps =
        static_cast<uint16_t>(ceil((fc->delta_x / EndnodeReadOnly<double>("program_settings/precision/way_increment").read(this->rtIO->moduleConfig).value())));
    uint16_t docking_steps = static_cast<uint16_t>(floor((steps * fuel_transfer_to_manoeuvre_time)));
    fc->delta_x = fc->delta_x / steps;
    for (uint16_t step_number(0); step_number < steps; ++step_number) {
        /* Set the initial conditions of the segment */
        this->set_segment_initial_conditions(fc);
        this->cruise_iteration(oc, fc, a_mission_step, true);
        if (steps > docking_steps) {
            this->mission_profile.back().end_total_mass +=
                this->mission_file->mission_data.at(a_mission_step).rel_delta_fuel_mass * this->mission_file->air_to_air_fuel_mass / (steps - docking_steps);
            this->mission_profile.back().end_total_mass +=
                this->mission_file->mission_data.at(a_mission_step).delta_payload_mass / (steps - docking_steps);
        }
    }
}

void standard_mission::free_flight(Operating_conditions* oc, Flight_conditions* fc, const uint16_t& a_mission_step) {
    myRuntimeInfo->out << "Climbing to Free-Flight-Optimum ... " << std::endl;
    /* Set the operating conditions of the segment */
    oc->set_segment_operating_conditions(a_mission_step);
    /* Set the initial conditions of the segment */
    this->set_segment_initial_conditions(fc);
    /* Convert true airspeed into Mach number */
    double mach_optimum(convertUnit(TRUEAIRSPEED, MACH, fc->altitude, this->atm, fc->TAS));
    double altitude_optimum(fc->altitude);
    /**< \todo [T147] (https://unicado.ilr.rwth-aachen.de/T147): Determine target altitude by iteration for free flight-optimum
     */
    this->calcOptSpeedAndFL(&mach_optimum, &altitude_optimum, 0., fc->aircraft_mass, oc->config, 1., oc->rating, fc->mission_data->at(oc->mission_step).energy_carrier_id);
    /* Convert true airspeed into calibrated airspeed */
    double CAS = convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, fc->altitude, this->atm, fc->TAS); // Stays constant below transition altitude!
    double transition_mach = mach_optimum; // Constant climb speed (Mach number) above transition altitude
    /* Calculate the number of steps, max 10 m climb */
    fc->delta_h = altitude_optimum - fc->altitude;
    uint16_t flightLevelSteps =
        static_cast<uint16_t>(ceil(fabs(fc->delta_h / EndnodeReadOnly<double>(
        "program_settings/precision/altitude_increment").read(this->rtIO->moduleConfig).value())));
    fc->delta_h = fc->delta_h / flightLevelSteps;
    myRuntimeInfo->out << "Climb Step:          " << convertUnit(METER, FOOT, fc->delta_h) << " ft." << std::endl;
    myRuntimeInfo->out << "Start altitude:      " << convertUnit(METER, FOOT, fc->altitude) << " ft." << std::endl;
    myRuntimeInfo->out << "End altitude:        " << convertUnit(METER, FOOT, (fc->delta_h * flightLevelSteps + fc->altitude)) << " ft." << std::endl;
    myRuntimeInfo->out << "Transition altitude: " << convertUnit(METER, FOOT, get_transition_altitude(CAS, transition_mach, this->atm)) << " ft." << std::endl;
    for (uint16_t step_number(0); step_number < flightLevelSteps; ++step_number) {
        /* Set the initial conditions of the segment */
        this->set_segment_initial_conditions(fc);
        if (fc->altitude < get_transition_altitude(CAS, transition_mach, this->atm)) {
            /* Convert calibrated airspeed into true airspeed */
            fc->end_TAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, fc->altitude + fc->delta_h, this->atm, CAS);
        } else {
            /* Convert Mach number into true airspeed */
            fc->end_TAS = convertUnit(MACH, TRUEAIRSPEED, fc->altitude + fc->delta_h, this->atm, transition_mach);
        }
        Flight_conditions backup_fc(*fc);
        try {
            this->climb_iteration(oc, fc, a_mission_step);
        } catch (const std::string& error_message) {
            myRuntimeInfo->warn << "Current engine rating '" << oc->rating << "' not sufficient to climb further." << std::endl;
            if (EndnodeReadOnly<bool>("program_settings/general/increase_engine_rating_during_climb").read(this->rtIO->moduleConfig).value() &&
                oc->rating == "climb") {
                    myRuntimeInfo->out << "  - set rating to 'MaxCont'" << std::endl;
                    oc->rating = "maximum_continuous";
                    *fc = backup_fc;
                    this->climb_iteration(oc, fc, a_mission_step);
            } else {
                throwError(__FILE__, __func__, __LINE__, error_message);
            }
        }
        if (this->cruise_profile_start != 0) {
            this->cruise_profile_start = this->mission_profile.size(); // set index
        }
    }
    /* Free-Flight-Cruise */
    myRuntimeInfo->out << "Free-Flight-Cruise ... " << std::endl;
    /* Set the operating conditions of the segment */
    oc->set_segment_operating_conditions(a_mission_step);
    /* Set the initial conditions of the segment */
    this->set_segment_initial_conditions(fc);

    double x_end = this->mission_file->mission_data.at(a_mission_step).rel_segment_length *
                    (this->mission_file->Range - this->descent_range - this->mission_profile.at(this->cruise_profile_start).end_range) +
                    this->mission_profile.at(this->cruise_profile_start).end_range;
    /* Calculate the number of steps 1 km */
    fc->delta_x = x_end - this->mission_profile.back().end_range;
    if (fc->delta_x < 0) {
        myRuntimeInfo->out << "Cruise distance already covered!" << std::endl;
    } else {
        uint16_t rangeSteps =
            static_cast<uint16_t>(ceil((fc->delta_x /
                EndnodeReadOnly<double>("program_settings/precision/way_increment").read(this->rtIO->moduleConfig).value())));
        fc->delta_x = fc->delta_x / rangeSteps;
        for (uint16_t step_number(0); step_number < rangeSteps; ++step_number) {
            /* Set the initial conditions of the segment */
            this->set_segment_initial_conditions(fc);
            free_flight_iteration(oc, fc, a_mission_step, true);
        }
    }
}

void standard_mission::calculate_econ_cruise_speed(double* econ_mach_number, const double& altitude, const double CI, const double& aircraft_mass,
                                                const std::string& config, const std::string& rating) {
    double rel_step_size(1.e-2); // Relative step size
    double lowerComparisonMachNumber(0.);
    double upperComparisionMachNumber(0.);
    double lower_cost_function(0.); //Cost function with lowerComparisionMachNumber in [kg/m]
    double upper_cost_function(0.); //Cost function with upperComparisionMachNumber in [kg/m]
    double maximum_mach(std::min(convertUnit(CALIBRATEDAIRSPEED, MACH, altitude, this->atm, this->max_operating_speed), this->max_operating_mach));
    int energy_carrier_id =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/cruise")).value();

    // Search maximum
    //Save engine operating conditions for normal cruise before changes from flight level change simulation occur
    std::vector<OperatingPoint> backup_operating_points = this->engine.get_operating_points(energy_carrier_id);
    while (*econ_mach_number < maximum_mach && *econ_mach_number > 0.2) {
        lowerComparisonMachNumber = *econ_mach_number * (1 - rel_step_size);
        upperComparisionMachNumber = *econ_mach_number * (1 + rel_step_size);
        double econMachCostFunction = this->calculate_cost_function(energy_carrier_id, *econ_mach_number, altitude, aircraft_mass, config, rating, 0., 0., CI);
        lower_cost_function = this->calculate_cost_function(energy_carrier_id, lowerComparisonMachNumber, altitude, aircraft_mass, config, rating, 0., 0., CI);
        upper_cost_function = this->calculate_cost_function(energy_carrier_id, upperComparisionMachNumber, altitude, aircraft_mass, config, rating, 0., 0., CI);
        if (econMachCostFunction < lower_cost_function && econMachCostFunction < upper_cost_function) {
            break;
        } else if (lower_cost_function > econMachCostFunction && upper_cost_function < econMachCostFunction) {
            *econ_mach_number = upperComparisionMachNumber;
        } else {
            *econ_mach_number = lowerComparisonMachNumber;
        }
    }
    if (*econ_mach_number >= maximum_mach) {
        *econ_mach_number = maximum_mach;
        this->engine.set_operating_points(energy_carrier_id, backup_operating_points);
        return;
    }
    /* Golden Section Search */
    std::vector<double> abszissa_values(this->get_golden_section_starting_points(
                                        lowerComparisonMachNumber, upperComparisionMachNumber, *econ_mach_number)); //Get start values
    while (fabs(abszissa_values.at(3) - abszissa_values.at(0)) > ACCURACY_MEDIUM * (fabs(abszissa_values.at(1)) + fabs(abszissa_values.at(2)))) { //Get values of cost function
        lower_cost_function = this->calculate_cost_function(energy_carrier_id, abszissa_values.at(1), altitude, aircraft_mass, config, rating, 0., 0., CI);
        upper_cost_function = this->calculate_cost_function(energy_carrier_id, abszissa_values.at(2), altitude, aircraft_mass, config, rating, 0., 0., CI);
        this->update_abszissa_values(&abszissa_values, {lower_cost_function, upper_cost_function});
    }
    if (lower_cost_function < upper_cost_function) { // We are done. Output the best of the two current values.
        *econ_mach_number = abszissa_values.at(1);
    } else {
        *econ_mach_number = abszissa_values.at(2);
    }
    this->engine.set_operating_points(energy_carrier_id, backup_operating_points);
    return;
}

void standard_mission::simulate_FL_change(Flight_conditions* fc, const Operating_conditions& oc, const double& altitude_change_step,
                                        const uint16_t& a_mission_step) {
    /* Backup */
    double aircraft_massBeforeClimb(fc->end_aircraft_mass);
    double delta_xClimb(0.);
    Operating_conditions theFLchangeOC(oc); //OperatingCondition
    /* Simulate flight level change */
    this->mission_file->mission_data.at(a_mission_step).mode = Flight_conditions::FLCHANGESIMULATION;
    this->mission_file->mission_data.at(a_mission_step).altitude = fc->altitude + altitude_change_step;
    this->set_segment_initial_conditions(fc, false);
    delta_xClimb = (this->change_altitude_at_constant_speed(&theFLchangeOC, fc, a_mission_step, true));
    double step_fuel_massClimb(aircraft_massBeforeClimb - fc->end_aircraft_mass);
    /* Set new initial conditions in new altitude */
    fc->aircraft_mass = fc->end_aircraft_mass;
    fc->TAS = fc->end_TAS;
    fc->altitude = fc->end_altitude;
    this->mission_file->mission_data.at(a_mission_step).mode = Flight_conditions::CRUISE;
    theFLchangeOC.rating = "cruise";
    fc->update_operating_condition(theFLchangeOC);
    this->set_segment_initial_conditions(fc, false);
    this->cruise_iteration(&theFLchangeOC, fc, a_mission_step, false);
    /* Add climb way and fuel to current flight condition */
    fc->delta_x += delta_xClimb;
    fc->step_fuel_mass += step_fuel_massClimb;
    /* Reset values */
    this->mission_file->mission_data.at(a_mission_step).altitude = fc->altitude - altitude_change_step;
}

void standard_mission::fixed_ROC_acceleration_iteration(Operating_conditions* oc, Flight_conditions* fc, const double& min_acceleration,
                                            const uint16_t &mission_step, const bool& save_data) {
    /* Fixed ROC Iteration until convergence of delta_h */
    double residuum(0.);
    uint16_t counter(0);
    //Set start value of segment end conditions for iteration
    this->set_segment_end_conditions(fc, fc->end_TAS, fc->altitude, fc->glidepath_angle, fc->acceleration);
    do {
        double oldend_altitudeeight(fc->end_altitude);
        /* set new values for delta_h and with this for end_TAS */
        fc->delta_h = sin(fc->mean_glidepath_angle) / 2. / fc->mean_acceleration * (pow(fc->end_TAS, 2.)
                                        - pow(fc->TAS, 2.));
        fc->end_altitude = fc->altitude + fc->delta_h;
        double CAS(convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, oldend_altitudeeight, this->atm, fc->end_TAS));
        fc->end_TAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, fc->end_altitude, this->atm, CAS);
        this->set_segment_end_conditions(fc, fc->end_TAS, fc->end_altitude, fc->end_glidepath_angle, fc->end_acceleration);
        /* Check acceleration potential */
        this->check_iteration_conditions(oc, fc, mission_step, 0.);
        if (counter == 500) { // Maximum number of iterations (500 iterations) is reached, error message
            throwError(__FILE__, __func__, __LINE__, "Maximum number of iterations reached. Abort program!");
        }
        /* Calculate residuum in mean_glidepath_angle */
        residuum = fabs(oldend_altitudeeight - fc->end_altitude) / fabs(oldend_altitudeeight);
        counter++;
    } while (residuum > ACCURACY_LOW);
    if (save_data) this->save_segment_data(fc, *oc, mission_step);
    /* Clear origin vector */
    fc->origin_polar.clear();
    /* Delete processed bleed and ShaftPower-Offtakes */
    oc->delete_used_bleed_and_shaft_power_offtakes(fc->delta_t);
}

void standard_mission::check_iteration_conditions(Operating_conditions* oc, Flight_conditions* fc,
                                                    const uint16_t &mission_step, const double& minimum_ROC) {
    /* Check if flight direction is forward */
    if (fc->delta_x < 0.) {
        std::stringstream error_message;
        error_message << "Negative flight distance. Rating '" + oc->rating + "' seems not sufficient to maintain forward flight direction." << std::endl;
        error_message << "fc->delta_x " << fc->delta_x << std::endl;
        error_message << "fc->delta_t " << fc->delta_t << std::endl;
        error_message << "fc->drag " << fc->end_drag << std::endl;
        error_message << "fc->Thrust " << fc->end_thrust << std::endl;
        error_message << "fc->mean_acceleration " << fc->mean_acceleration << std::endl;
        error_message << "fc->mean_glidepath_angle " << fc->mean_glidepath_angle << std::endl;
        error_message << "fc->end_altitude " << fc->end_altitude << std::endl;
        // error_message << "N1: " << this->engine.get_operating_point().N << std::endl; //TODO(Gerrit): Splitten und rein
        throwError(__FILE__, __func__, __LINE__,  error_message.str());
    }
    /* Check if minimum ROC is maintained */
    if (fabs(fc->ROC) < minimum_ROC) {
        myRuntimeInfo->warn << "At " << convertUnit(METER, FOOT, fc->altitude) << " ft, the maximum possible climb rate "
                            << convertUnit(METERPERSECOND, FOOTPERMINUTE, fc->ROC) << " ft/min is less than the specified minimum requirement: "
                            << convertUnit(METERPERSECOND, FOOTPERMINUTE, minimum_ROC) << " ft/min." << std::endl;
        if (this->requirement_switch) {
            myRuntimeInfo->out << "Finish climb" << std::endl;
            if (fc->ROC < 0.) {
                throw std::string("Insufficient climb power");
            } else {
                throw 1;
            }
        }
    }
    /* Check climb/descent angle:
     * The problem is badly conditioned at values close to zero, it can lead to divergence.
     */
    if (!accuracyCheck(fc->delta_h, 0., ACCURACY_HIGH) && fabs(fc->mean_glidepath_angle) < ACCURACY_LOW) {
        throw_mission_error(6, mission_step);
    }
}

void standard_mission::cruise_iteration(Operating_conditions* oc, Flight_conditions* fc, const uint16_t &mission_step, bool save_data) {
    /* Cruise Iteration until convergence of average fuel flow mean_fuel_mass_flow */
    double residuum(0.);
    uint16_t counter(0);
    //Set start value of segment end conditions for iteration
    fc->end_TAS = fc->TAS;
    fc->delta_h = 0.;
    this->set_segment_end_conditions(fc, fc->end_TAS, fc->altitude + fc->delta_h,
                                        fc->glidepath_angle, fc->acceleration, NAN,
                                        EndnodeReadOnly<bool>("program_settings/general/use_breguet_estimation_in_cruise").read(this->rtIO->moduleConfig).value());
    do {
        /* Set iteration values */
        double previous_fuel_mass_flow(fc->mean_fuel_mass_flow);
        this->set_segment_end_conditions(fc, fc->end_TAS, fc->end_altitude, fc->end_glidepath_angle, fc->end_acceleration, NAN,
                                             EndnodeReadOnly<bool>("program_settings/general/use_breguet_estimation_in_cruise").read(this->rtIO->moduleConfig).value());
        /* Check conditions */
        this->check_iteration_conditions(oc, fc, mission_step, 0.);
        if (counter == 500) { // Maximum number of iterations (500 iterations) is reached, error message
            throwError(__FILE__, __func__, __LINE__, "Maximum number of iterations reached. Abort program!");
        }
        /* Calculate residuum in mean_fuel_mass_flow */
        residuum = fabs(previous_fuel_mass_flow - fc->mean_fuel_mass_flow) / fabs(previous_fuel_mass_flow);
        counter++;
    } while (residuum > ACCURACY_LOW);
    if (save_data) this->save_segment_data(fc, *oc, mission_step);
    /* Delete processed bleed and ShaftPower-Offtakes */
    oc->delete_used_bleed_and_shaft_power_offtakes(fc->delta_t);
}

void standard_mission::free_flight_iteration(Operating_conditions* oc, Flight_conditions* fc, const uint16_t &mission_step, bool save_data) {
    int energy_carrier_id(fc->mission_data->at(fc->operating_condition->mission_step).energy_carrier_id);
    /* Add bleed and ShaftPower-Offtakes */
    fc->bleed = oc->background_bleed;
    if (!oc->bleed_times.empty()) {
        fc->bleed += oc->bleed_values.front();
    }
    fc->shaft_power = oc->background_shaft_power;
    if (!oc->shaft_power_times.empty()) {
        fc->shaft_power += oc->shaft_power_values.front();
    }
    fc->end_altitude = fc->altitude;
    fc->end_TAS = fc->TAS;
    /* Calculate drag, start value */
    fc->drag = this->aero.getCruiseDrag(convertUnit(TRUEAIRSPEED, MACH, fc->altitude,
                                                        this->atm, fc->TAS), fc->altitude, 0.,
                                                        fc->aircraft_mass, oc->config, this->atm);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        fc->drag = this->interpolateViaCoG(current_cg, fc->drag,
                        getAlternativeAero(current_cg)->getCruiseDrag(convertUnit(TRUEAIRSPEED, MACH, fc->altitude,
                                                            this->atm, fc->TAS), fc->altitude, 0.,
                                                            fc->aircraft_mass, oc->config, this->atm));
    }
    fc->end_drag = fc->drag;
    /* Calculate drag */
    this->engine.calculate_N1_with_thrustlimit(energy_carrier_id, fc->altitude, convertUnit(TRUEAIRSPEED, MACH, fc->altitude,
                                                this->atm, fc->TAS), this->atm, oc->derate, oc->rating,
                                                fc->bleed, fc->shaft_power, fc->drag);
    fc->thrust = this->engine.get_thrust_aircraft(energy_carrier_id);
    fc->fuel_mass_flow = this->engine.get_aircraft_fuelflow(energy_carrier_id);
    fc->energyflow = {energy_carrier_id, this->engine.get_aircraft_energyflow(energy_carrier_id, fc->fuel_mass_flow)};
    fc->end_thrust = fc->thrust;
    /* Convert true airspeed into Mach number */
    double mach(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS));
    this->calcOptSpeedAndFL(&mach, &fc->end_altitude, 0., fc->aircraft_mass, oc->config, 1., oc->rating, fc->mission_data->at(oc->mission_step).energy_carrier_id);
    /* Convert Mach number into true airspeed */
    fc->end_TAS = convertUnit(MACH, TRUEAIRSPEED, fc->end_altitude, this->atm, mach);
    fc->delta_h = fc->end_altitude - fc->altitude;
    fc->glidepath_angle = atan(fc->delta_h / fc->delta_x);
    fc->acceleration = (fc->thrust - fc->drag -
                fc->aircraft_mass * sin(fc->glidepath_angle) * G_FORCE) /
                fc->aircraft_mass; // Acceleration according to Newton's law. Acceleration force F_a = (thrust.netThrust - Drag - m*cos(glidepath_angle)*g)
    // Drag in path direction
    fc->mean_acceleration = fc->acceleration;
    fc->end_acceleration = fc->acceleration;
    double err(0.);
    unsigned int counter(0); //iteration counter
    do {
        fc->delta_t = fc->delta_x / cos(fc->glidepath_angle) / (fc->TAS + fc->end_TAS) * 2.;
        this->engine.calculate_N1_with_thrustlimit(energy_carrier_id, fc->end_altitude, convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                                                    this->atm, oc->derate, oc->rating, fc->bleed,
                                                    fc->shaft_power, fc->end_drag);
        fc->end_fuel_mass_flow = this->engine.get_aircraft_fuelflow(energy_carrier_id);
        fc->end_energyflow = {energy_carrier_id, this->engine.get_aircraft_energyflow(energy_carrier_id, fc->end_fuel_mass_flow)};
        /* Calculate fuel */
        fc->mean_fuel_mass_flow = (fc->fuel_mass_flow + fc->end_fuel_mass_flow) / 2.;
        fc->step_fuel_mass = fc->delta_t * fc->mean_fuel_mass_flow;
        fc->end_drag = this->aero.getCruiseDrag(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                                                                fc->end_altitude, 0.,
                                                                (fc->aircraft_mass - fc->step_fuel_mass) * cos(fc->glidepath_angle),
                                                                oc->config, this->atm);

        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            fc->end_drag = this->interpolateViaCoG(current_cg, fc->end_drag,
                getAlternativeAero(current_cg)->getCruiseDrag(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                        fc->end_altitude, 0., (fc->aircraft_mass - fc->step_fuel_mass) * cos(fc->glidepath_angle),
                        oc->config, this->atm));
        }
        double required_thrust(sin(fc->glidepath_angle) * (fc->aircraft_mass - fc->step_fuel_mass) * G_FORCE +
                              (fc->aircraft_mass - fc->step_fuel_mass) * fc->end_acceleration + fc->end_drag);
        this->engine.calculate_N1_with_thrustlimit(energy_carrier_id, fc->end_altitude,
                            convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                            this->atm, oc->derate, oc->rating, fc->bleed,
                            fc->shaft_power, required_thrust);
        fc->end_thrust = this->engine.get_thrust_aircraft(energy_carrier_id);
        fc->end_acceleration = (fc->end_thrust - fc->end_drag -
                                (fc->aircraft_mass - fc->step_fuel_mass) *
                                sin(fc->glidepath_angle) * G_FORCE) / (fc->aircraft_mass - fc->step_fuel_mass);
        /* Check drag */
        if (fabs(fc->end_thrust - sin(fc->glidepath_angle) * (fc->aircraft_mass - fc->step_fuel_mass)*G_FORCE -
                fc->end_drag - (fc->aircraft_mass - fc->step_fuel_mass)*fc->end_acceleration) > 1e-6) {
            throw_mission_error(2, mission_step);
        }
        if (counter > 10 && fabs(fc->mean_acceleration - (fc->acceleration + fc->end_acceleration) / 2.) /
                fabs(fc->mean_acceleration) >= err) {
            myRuntimeInfo->out << "Divergence free_flight_iteration!" << std::endl;
            err = fabs(fc->mean_acceleration - (fc->acceleration + fc->end_acceleration) / 2.) /
                fabs(fc->mean_acceleration);
            if (err < ACCURACY_LOW * 20.) {
                break;
            } else {
                throw_mission_error(5, mission_step);
            }
        } else {
            err = fabs(fc->mean_acceleration -
                    (fc->acceleration + fc->end_acceleration) / 2.) / fabs(fc->mean_acceleration);
            fc->mean_acceleration = fc->mean_acceleration -
                    (fc->mean_acceleration - (fc->acceleration + fc->end_acceleration) / 2.) * 0.1;
        }
        counter++;
    } while (err > ACCURACY_LOW && counter < 500);
    if (counter >= 500) { // Maximum number of iterations (500 iterations) is reached, error message
        myRuntimeInfo->err << "Maximum number of iterations reached. Abort program!" << std::endl;
        exit(1);
    }
    /* Check acceleration potential */
    if (fc->delta_x < 0) throw_mission_error(1, mission_step);
    fc->end_CL = this->aero.getCL(fc->end_altitude,
        convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
        0., fc->aircraft_mass - fc->step_fuel_mass, this->atm);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        fc->end_CL = this->interpolateViaCoG(current_cg, fc->end_CL,
                    getAlternativeAero(current_cg)->getCL(fc->end_altitude,
                        convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                        0., fc->aircraft_mass - fc->step_fuel_mass, this->atm));
    }

    fc->end_L_over_D = this->aero.getCLLoverD(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS), fc->end_CL, fc->operating_condition->config);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        fc->end_L_over_D = this->interpolateViaCoG(current_cg, fc->end_L_over_D,
            getAlternativeAero(current_cg)->getCLLoverD(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                                                        fc->end_CL, fc->operating_condition->config));
    }

    double sar_tmp;
    try {
            sar_tmp =
                this->static_flight_specific_air_range(energy_carrier_id, convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                    fc->end_altitude, fc->aircraft_mass -
                    fc->step_fuel_mass, oc->config, oc->rating);
        } catch (const int& engine_status) {
            throwError(__FILE__, __func__, __LINE__,
                    ("Required thrust (" + num2Str(this->aero.getCruiseDrag(mach, fc->end_altitude, 0., fc->aircraft_mass -
                    fc->step_fuel_mass, oc->config, this->atm)) + " kN) cannot be delivered by engines in current status."));
        }
    fc->end_specific_air_range = sar_tmp;

    fc->end_angle_of_attack =
            this->aero.getAlphaAtFlightCondition(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
            oc->config, fc->end_CL);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        fc->end_angle_of_attack = this->interpolateViaCoG(current_cg, fc->end_angle_of_attack,
            getAlternativeAero(current_cg)->getAlphaAtFlightCondition(
                convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                oc->config, fc->end_CL));
    }
    if (save_data) this->save_segment_data(fc, *oc, mission_step);
    /* Clear origin vector */
    fc->origin_polar.clear();
    /* Delete processed bleed and ShaftPower-Offtakes */
    oc->delete_used_bleed_and_shaft_power_offtakes(fc->delta_t);
}

void standard_mission::set_segment_initial_conditions(Flight_conditions* fc, const bool& use_mission_condition, const double& an_end_CL) {
    /* Segment initial conditions have to be iterated until convergence due to internal dependencies
     * iteration values are: CL, Drag, and thrust
     * note: the weight (mass) as fourth force has not to be integrated here because it is fixed as segment start value as well as the true airspeed (TAS) and flight altitude (h)
     */
    double residuumLift(0.), residuumDrag(0.), residuumThrust(0.);
    uint16_t counter(0);
    int energy_carrier_id(fc->mission_data->at(fc->operating_condition->mission_step).energy_carrier_id);

    /* Set fixed values */
    if (use_mission_condition && !this->mission_profile.empty()) {
        fc->aircraft_mass = this->mission_profile.back().end_total_mass;
        fc->TAS = this->mission_profile.back().end_TAS;
        fc->altitude = this->mission_profile.back().end_altitude;
    }
    /* Add bleed and ShaftPower-Offtakes */
    fc->bleed = fc->operating_condition->background_bleed;
    if (fc->operating_condition->bleed_times.size()) {
        fc->bleed += fc->operating_condition->bleed_values.front();
    }
    fc->shaft_power = fc->operating_condition->background_shaft_power;
    if (fc->operating_condition->shaft_power_times.size()) {
        fc->shaft_power += fc->operating_condition->shaft_power_values.front();
    }
    /* Set iteration start values */
    this->set_initial_iteration_start_values(fc, an_end_CL);
    /* Iteration loop */
    do {
        myRuntimeInfo->debug << "Iteration (" << counter + 1 << "): " << std::endl;
        double oldLift(fc->CL);
        double oldDrag(fc->drag);
        double oldThrust(fc->thrust);
        /* Set glidepath_angle and ROC */
        fc->set_initial_flight_track();
        /* Acceleration */
        fc->set_initial_acceleration_factors(this->braking_coefficient, this->friction_coefficient, this->wing_surface_area);
        /* Lift and Drag calculation */
        this->set_initial_aerodynamic_values(fc, an_end_CL);
        /* Thrust calculation */
        fc->set_initial_thrust();
        /* Check number of iterations */
        if (counter == 500) { // Maximum number of iterations (500 iterations) is reached, error message
            throwError(__FILE__, __func__, __LINE__, "Maximum number of iterations reached. Abort program!");
        }
        /* Calculate residui */
        residuumLift = fabs(oldLift - fc->CL) / fabs(oldLift);
        residuumDrag = fabs(oldDrag - fc->drag) / fabs(oldDrag);
        residuumThrust = fabs(oldThrust - fc->thrust) / fabs(oldThrust);
        counter++;
        myRuntimeInfo->debug << "CL: " << fc->CL << "; Residuum: " << residuumLift << std::endl;
        myRuntimeInfo->debug << "Drag: " << fc->drag << " N; Residuum: " << residuumDrag << std::endl;
        myRuntimeInfo->debug << "thrust: " << fc->thrust << " N; Residuum: " << residuumThrust << std::endl;
        myRuntimeInfo->debug << "---------------------------" << std::endl;
    } while (residuumLift > ACCURACY_MEDIUM || residuumDrag > ACCURACY_MEDIUM || residuumThrust > ACCURACY_MEDIUM);
    /* Final additional needed values */
    fc->reynolds = this->atm.getDensity(fc->altitude) * this->MAC * fc->TAS /
                                            this->atm.getViscosity(fc->altitude);
    fc->fuel_mass_flow = this->engine.get_aircraft_fuelflow(energy_carrier_id) * fc->rel_engine_availability;
    fc->energyflow = {energy_carrier_id, this->engine.get_aircraft_energyflow(energy_carrier_id, fc->fuel_mass_flow)};
}

void standard_mission::set_initial_iteration_start_values(Flight_conditions* fc, const double& an_end_CL) {
    /* Flight track values */
    fc->glidepath_angle = fc->end_glidepath_angle;
    /* Acceleration */
    fc->acceleration = fc->end_acceleration;
    /* Lift and Drag */
    this->set_initial_aerodynamic_values(fc, an_end_CL);
	/* Thrust
     * IMPORTANT NOTE: the switch engine_limit_reached must be controlled here, because once the spoiler is activated the status must be maintained during iteration
     *                 otherwise there will be oscillations in drag
     */
    fc->engine_limit_reached = false;
    fc->set_initial_thrust();
}

void standard_mission::set_initial_aerodynamic_values(Flight_conditions* fc, const double& an_end_CL) {
    if (std::isnan(an_end_CL)) {
        fc->CL = this->aero.getCL(fc->altitude,
                                                                convertUnit(TRUEAIRSPEED, MACH, fc->altitude, this->atm, fc->TAS),
                                                                0., fc->aircraft_mass * cos(fc->glidepath_angle), atm);
        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            fc->CL = this->interpolateViaCoG(current_cg, fc->CL,
                getAlternativeAero(current_cg)->getCL(fc->altitude,
                    convertUnit(TRUEAIRSPEED, MACH, fc->altitude, this->atm, fc->TAS), 0.,
                                    fc->aircraft_mass * cos(fc->glidepath_angle), atm));
        }
    } else {
        fc->CL = an_end_CL;
    }
    fc->drag = this->aero.getCLDrag(convertUnit(TRUEAIRSPEED, MACH, fc->altitude, this->atm, fc->TAS),
                                                    fc->altitude, fc->CL,
                                                    fc->operating_condition->config, atm);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        fc->drag = this->interpolateViaCoG(current_cg, fc->drag,
            getAlternativeAero(current_cg)->getCLDrag(convertUnit(TRUEAIRSPEED, MACH, fc->altitude, this->atm, fc->TAS),
                                                        fc->altitude, fc->CL,
                                                        fc->operating_condition->config, atm));
    }
    if (this->mission_file->mission_data.at(fc->operating_condition->mission_step).mode == Flight_conditions::LANDING) {
        fc->drag *= this->aero.maxSpoilerFactor; //activate ground spoilers
    }
}

void standard_mission::set_iteration_start_values(Flight_conditions* fc, const double& end_track_angle, const double& end_acceleration, const double& an_end_CL) {
    /* Flight track values */
    fc->end_glidepath_angle = end_track_angle;
    fc->mean_glidepath_angle = (fc->glidepath_angle + fc->end_glidepath_angle) / 2.;
    fc->ROC = sin(fc->mean_glidepath_angle) * (fc->TAS + fc->end_TAS) / 2.;
    /* Acceleration */
    fc->end_acceleration = end_acceleration;
    fc->mean_acceleration = (fc->acceleration + fc->end_acceleration) / 2.;
    /* Lift and Drag */
    double mach(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS));
    if (std::isnan(an_end_CL)) {
        fc->end_CL = this->aero.getCL(fc->end_altitude, mach, 0., fc->aircraft_mass * cos(fc->end_glidepath_angle), atm);
        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            fc->end_CL = this->interpolateViaCoG(current_cg, fc->end_CL,
                this->getAlternativeAero(current_cg)->getCL(fc->end_altitude, mach, 0., fc->aircraft_mass * cos(fc->end_glidepath_angle), atm));
        }
    } else {
        fc->end_CL = an_end_CL;
    }
    fc->end_L_over_D = this->aero.getCLLoverD(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS), fc->end_CL, fc->operating_condition->config);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        fc->end_L_over_D = this->interpolateViaCoG(current_cg, fc->end_L_over_D,
            getAlternativeAero(current_cg)->getCLLoverD(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                                                        fc->end_CL, fc->operating_condition->config));
    }
    fc->end_drag = this->aero.getCLDrag(mach, fc->end_altitude, fc->end_CL, fc->operating_condition->config, atm);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        fc->end_drag = this->interpolateViaCoG(current_cg, fc->end_drag,
            this->getAlternativeAero(current_cg)->getCLDrag(mach, fc->end_altitude, fc->end_CL, fc->operating_condition->config, atm));
    }
    if (std::isnan(fc->end_drag)) {
        throwError(__FILE__, __func__, __LINE__, "Drag NaN.");
    }
    /* Mass */
    fc->end_aircraft_mass = fc->aircraft_mass;
	/* Thrust
     * IMPORTANT NOTE: the switch engine_limit_reached must be controlled here, because once the spoiler is activated the status must be maintained during iteration
     *                 otherwise there will be oscillations in drag
     */
    fc->engine_limit_reached = false;
    fc->set_thrust();
}

void standard_mission::set_aerodynamic_values(Flight_conditions* fc, const double& an_end_CL) {
    double mach(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS));
    fc->end_reynolds = atm.getDensity(fc->end_altitude) *
                                        this->MAC * fc->end_TAS / atm.getViscosity(fc->end_altitude);

    // Angle of attack
    fc->end_angle_of_attack = this->aero.getAlphaAtFlightCondition(mach, fc->operating_condition->config, fc->end_CL);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        fc->end_angle_of_attack = this->interpolateViaCoG(current_cg, fc->end_angle_of_attack,
            this->getAlternativeAero(current_cg)->getAlphaAtFlightCondition(
                mach, fc->operating_condition->config, fc->end_CL));
    }

    // Lift coefficient
    if (std::isnan(an_end_CL)) { //only update lift if not preset
        fc->end_CL = this->aero.getCL(fc->end_altitude, mach, 0.,
                                        fc->end_aircraft_mass * cos(fc->end_glidepath_angle), this->atm);
        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            fc->end_CL = this->interpolateViaCoG(current_cg, fc->end_CL,
                this->getAlternativeAero(current_cg)->getCL(fc->end_altitude, mach, 0.,
                                                            fc->end_aircraft_mass * cos(fc->end_glidepath_angle), this->atm));
        }
    }

    // L over D
    fc->end_L_over_D = this->aero.getCLLoverD(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS), fc->end_CL, fc->operating_condition->config);
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        fc->end_L_over_D = this->interpolateViaCoG(current_cg, fc->end_L_over_D,
            getAlternativeAero(current_cg)->getCLLoverD(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS),
                                                        fc->end_CL, fc->operating_condition->config));
    }

    // Drag
    fc->end_drag = this->aero.getCLDrag(mach, fc->end_altitude, fc->end_CL,
                                                            fc->operating_condition->config, this->atm);
        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            fc->end_drag = this->interpolateViaCoG(current_cg, fc->end_drag,
                getAlternativeAero(current_cg)->getCLDrag(mach, fc->end_altitude, fc->end_CL,
                                                            fc->operating_condition->config, this->atm));
        }
    if (std::isnan(fc->end_drag)) {
        throwError(__FILE__, __func__, __LINE__, "Drag NaN.");
    }
    /* Activate spoilers during Landing or if thrust is larger than required thrust for stationary flight */
    if (this->mission_file->mission_data.at(fc->operating_condition->mission_step).mode == Flight_conditions::LANDING ||
        (fc->engine_limit_reached &&
            (fc->end_thrust - (sin(fc->mean_glidepath_angle) * fc->end_aircraft_mass * G_FORCE +
                fc->end_drag + fc->end_aircraft_mass * fc->mean_acceleration)) / fc->end_thrust > ACCURACY_HIGH)) {
        myRuntimeInfo->info << "Activate spoiler ..." << std::endl;
        fc->activate_spoilers();
    }
}

void standard_mission::set_specific_air_range(Flight_conditions* fc) {
    double mach(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS));
    double LoD(this->aero.getCLLoverD(mach, fc->end_CL, fc->operating_condition->config));
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        LoD = this->interpolateViaCoG(current_cg, LoD,
            this->getAlternativeAero(current_cg)->getCLLoverD(mach, fc->end_CL, fc->operating_condition->config));
    }
    int energy_carrier_id(fc->mission_data->at(fc->operating_condition->mission_step).energy_carrier_id);
    const double TSFC(this->engine.get_aircraft_fuelflow(energy_carrier_id) / this->engine.get_thrust_aircraft(energy_carrier_id));
    fc->end_specific_air_range = LoD * fc->end_TAS / (TSFC * fc->end_aircraft_mass * G_FORCE);
}

void standard_mission::set_segment_end_conditions(Flight_conditions* fc, const double& end_TAS, const double& end_altitude,
                                                    const double& end_track_angle, const double& end_acceleration,
                                                    const double& an_end_CL, const bool& use_breguet) {
    /* Segment end conditions have to be iterated until convergence due to internal dependencies
     * iteration values are: end_aircraft_mass, CLend, DragEnd, end_thrust to consider all four forces
     * mean values: delta_x, delta_t, mean_acceleration, mean_glidepath_angle, ROC, mean_fuel_mass_flow, step_fuel_mass
     */
    double residuumMass(0.), residuumLift(0.), residuumDrag(0.), residuumThrust(0.);
    uint16_t counter(0);
    /* Set fixed values */
    fc->end_TAS = end_TAS;
    fc->end_altitude = end_altitude;
    fc->delta_h = fc->end_altitude - fc->altitude;
    /* Set iteration start values */
    this->set_iteration_start_values(fc, end_track_angle, end_acceleration, an_end_CL);
    /* Iteration loop
     * Note: Acceleration factors have to be set before calculation of way and time
     */
    do {
        myRuntimeInfo->debug << "Iteration (" << counter + 1 << "): " << fc->end_aircraft_mass << " kg." << std::endl;
        double oldEndMass(fc->end_aircraft_mass);
        double oldEndLift(fc->end_CL);
        double oldEndDrag(fc->end_drag);
        double oldthrust(fc->end_thrust);
        /* Set glidepath_angle and ROC */
        fc->set_flight_track();
        /* Acceleration */
        fc->set_acceleration_factors(this->braking_coefficient, this->friction_coefficient, this->wing_surface_area);
        /* Thrust */
        fc->set_thrust();
        /* Lift and Drag calculation */
        this->set_aerodynamic_values(fc, an_end_CL);
        /* Way and time */
        fc->set_way_and_time();
        /* SAR */
        this->set_specific_air_range(fc);
        /* Masses */
        fc->update_end_mass(use_breguet);
        /* Check number of iterations */
        if (counter == 500) { // Maximum number of iterations (500 iterations) is reached, error message
            throwError(__FILE__, __func__, __LINE__, "Maximum number of iterations reached. Abort program!");
        }
        /* Calculate residui */
        residuumMass = fabs(oldEndMass - fc->end_aircraft_mass) / fabs(oldEndMass);
        residuumLift = fabs(oldEndLift - fc->end_CL) / fabs(oldEndLift);
        residuumDrag = fabs(oldEndDrag - fc->end_drag) / fabs(oldEndDrag);
        residuumThrust = fabs(oldthrust - fc->end_thrust) / fabs(oldthrust);
        counter++;
        myRuntimeInfo->debug << "end_aircraft_mass: " << fc->end_aircraft_mass << " kg; Residuum: " << residuumMass << std::endl;
        myRuntimeInfo->debug << "end_CL: " << fc->end_CL << "; Residuum: " << residuumLift << std::endl;
        myRuntimeInfo->debug << "DragEnd: " << fc->end_drag << " N; Residuum: " << residuumDrag << std::endl;
        myRuntimeInfo->debug << "end_thrust: " << fc->end_thrust << " N; Residuum: " << residuumThrust << std::endl;
        myRuntimeInfo->debug << "---------------------------" << std::endl;
    } while (residuumMass > ACCURACY_MEDIUM || residuumLift > ACCURACY_MEDIUM || residuumDrag > ACCURACY_MEDIUM || residuumThrust > ACCURACY_MEDIUM);
    double mach(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, this->atm, fc->end_TAS));
    if (fc->operating_condition->config == "clean" && this->variable_camber_switch) {
        fc->origin_polar = aero.getCruiseOriginPolar(mach, fc->end_altitude , 0., fc->end_aircraft_mass,
                                                                    fc->operating_condition->config, this->atm);
    }
    fc->stabilizer_incidence = aero.getIStabPolar(mach, fc->end_CL, fc->operating_condition->config);
}

void standard_mission::set_segment_end_cg(Flight_conditions *fc) {
    if (sophistication_level != UNSOPHISTICATED) { // TODO(Gerrit): end_total_fuel_mass und previous_loaded_fuel_mass anpassen!
        double new_fuel_mass = this->previous_loaded_fuel_mass[0] + (this->mission_profile.back().end_total_fuel_mass[0] - fc->step_fuel_mass);
        if (fc->first_run_switch) {
            fc->new_aircraft_cg = cg_design;
            fc->prior_aircraft_cg = fc->new_aircraft_cg;
            fc->new_fuel_cg = this->calculate_fuel_cg(new_fuel_mass);
            fc->prior_fuel_cg = fc->new_fuel_cg;
            fc->prior_fuel_mass = loaded_fuel_mass[0];
            fc->first_run_switch = false;
        } else {
            fc->prior_fuel_mass = new_fuel_mass + fc->step_fuel_mass;
            fc->prior_aircraft_cg = fc->new_aircraft_cg;
            fc->prior_fuel_cg = fc->new_fuel_cg;
            fc->new_fuel_cg = this->calculate_fuel_cg(new_fuel_mass);
            fc->new_aircraft_cg = (fc->prior_aircraft_cg * fc->aircraft_mass +
                new_fuel_mass * fc->new_fuel_cg - fc->prior_fuel_cg * fc->prior_fuel_mass) / fc->end_aircraft_mass;
        }
        this->interpolate_stabilizer_incidence(fc);
    } else {
        fc->new_aircraft_cg = cg_design;
        fc->new_fuel_cg = 0.;
    }
}

void standard_mission::save_segment_data(Flight_conditions* fc, const Operating_conditions &oc, unsigned int mission_step) {
    double aircraft_mass_tmp = this->mission_profile.back().end_total_mass - fc->step_fuel_mass;
    int energy_carrier_id(fc->mission_data->at(fc->operating_condition->mission_step).energy_carrier_id);
    if (aircraft_mass_tmp <= 0. || std::isnan(aircraft_mass_tmp) || std::isinf(aircraft_mass_tmp)) {
        myRuntimeInfo->err << "Invalid value for aircraft mass: m = " << aircraft_mass_tmp << " kg" << std::endl;
        myRuntimeInfo->err << "mission_profile.back().end_total_mass = " << this->mission_profile.back().end_total_mass << " kg" << std::endl;
        myRuntimeInfo->err << "fc->step_fuel_mass = " << fc->step_fuel_mass << " kg" << std::endl;
        myRuntimeInfo->err << "mission_profile.back().end_range = " << this->mission_profile.back().end_range << " kg" << std::endl;
        myRuntimeInfo->err << "mission_profile.back().end_altitude = " << convertUnit(METER, FOOT, this->mission_profile.back().end_altitude) << " ft" << std::endl;
        myRuntimeInfo->err << "mission_profile.back().end_CL = " << this->mission_profile.back().end_CL << std::endl;
        myRuntimeInfo->err << "Abort program!" << std::endl;
        exit(1);
    }
    double FL_mom = this->mission_profile.back().end_altitude + fc->delta_h;
    /* Convert true airspeed into Mach number */
    double Mach_mom = convertUnit(TRUEAIRSPEED, MACH, this->mission_profile.back().end_altitude + fc->delta_h, this->atm, fc->end_TAS);
    try {
        this->engine.calculate_N1_with_thrustlimit(energy_carrier_id, FL_mom, Mach_mom, this->atm, oc.derate, oc.rating,
                                                    fc->bleed, fc->shaft_power, fabs(fc->end_thrust));
    } catch (const int& engine_status) {
        switch (engine_status) {
            case 2: {
                std::stringstream error_message;
                error_message << "Required thrust (" << convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, fabs(fc->end_thrust))
                        << " kN) cannot be delivered by engines in current status. ";
                std::string newRating("climb");
                if (oc.rating == "climb") {
                    newRating = "maximum_continuous";
                } else if (oc.rating == "maximum_continuous") {
                    error_message << "Reduce drag.";
                    throwError(__FILE__, __func__, __LINE__, error_message.str());
                } else {}
                Operating_conditions oc_tmp(oc);
                oc_tmp.rating = newRating;
                myRuntimeInfo->info << error_message.str() << "Automatically increasing thrust rating tmporarily from '" << oc.rating << "' to '"
                                    << oc_tmp.rating << "'" << std::endl;
                this->save_segment_data(fc, oc_tmp, mission_step);
                return;
            }
            default:
                throwError(__FILE__, __func__, __LINE__, "Engine could not be set to requested status.");
        }
    }
    this->set_segment_end_cg(fc);
    int energy_carrier_id_tmp = oc.mission_data->at(oc.mission_step).energy_carrier_id;
    // engine.drain_tanks(energy_carrier_id_tmp, fc->delta_t, fc->step_fuel_mass); // TODO(Gerrit): ggf. fuer CoG
    std::map<int, double> fuel_mass_map_tmp = this->mission_profile.back().end_total_fuel_mass;
    fuel_mass_map_tmp[energy_carrier_id_tmp] -= fc->step_fuel_mass; // only subtract mass from the used energy carrier

    /* Create a map to push N1 values for all propulsors */
    std::map<uint16_t, double> n1_map; // Fill map with N1 for each used propulsor and its ID
    for (node* propulsor : rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion").getVector("propulsor")) {
        n1_map[propulsor->getIntAttrib("ID")] = 0.;
    }
    for (auto prop : engine.get_propulsors(energy_carrier_id_tmp)) {
        double n1_temp = prop.engine.get_operating_point().N;
        for (uint16_t id : prop.engine_id) {
            n1_map.at(id) = n1_temp;
        }
    }

    /* Add landing or takeoff fuel if the step is within one of these segments */
    if (fc->altitude <= (convertUnit(FOOT, METER, 3000.))) { // If altitude is below FL30, LTO reached
        if (this->cruise_profile_start == 0) { // If no cruise profile was reached yet, assume takeoff is in progress
            this->takeoff_fuel_mass[energy_carrier_id_tmp] += fc->step_fuel_mass;
        } else { // if cruise was flown, assume landing is in progress
            this->landing_fuel_mass[energy_carrier_id_tmp] += fc->step_fuel_mass;
        }
    }

    // Push the result
    this->mission_profile.push_back(mission_segment(fuel_mass_map_tmp,
                                   fc->mean_fuel_mass_flow,
                                   this->mission_profile.back().end_time + fc->delta_t,
                                   this->mission_profile.back().end_range + fc->delta_x,
                                   this->mission_profile.back().end_altitude + fc->delta_h,
                                   this->mission_profile.back().end_total_mass - fc->step_fuel_mass,
                                   fc->end_CL,
                                   fc->end_L_over_D,
                                   fc->end_specific_air_range,
                                   oc.config,
                                   fc->mean_glidepath_angle,
                                   fc->end_TAS,
                                   convertUnit(TRUEAIRSPEED, MACH, this->mission_profile.back().end_altitude + fc->delta_h, this->atm, fc->end_TAS),
                                   convertUnit(TRUEAIRSPEED, CALIBRATEDAIRSPEED, this->mission_profile.back().end_altitude +
                                                fc->delta_h, this->atm, fc->end_TAS),
                                   fc->end_TAS * cos(fc->mean_glidepath_angle),
                                   fc->end_thrust,
                                   fc->bleed,
                                   fc->shaft_power,
                                   oc.rating,
                                   n1_map,
                                   fc->end_angle_of_attack,
                                   fc->end_spoiler_factor,
                                   this->mission_file->mission_data.at(mission_step).mode_name,
                                   fc->end_reynolds,
                                   fc->new_aircraft_cg,
                                   fc->new_fuel_cg,
                                   fc->stabilizer_incidence,
                                   fc->origin_polar,
                                   energy_carrier_id_tmp));
    Modeselector mode_selector(EndnodeReadOnly<std::string>(
        "/module_configuration_file/program_settings/general/polar_switch_mission_point/polar_switch_selector").read(this->rtIO->moduleConfig).value());
    if (mode_selector.get_mode_numeric() != 0 && !this->aero.altPolarsInUse && oc.config == "clean") {
        bool polars_just_switched(false);
        double total_mission_time(this->previous_flight_time); // [sec]
        if (total_mission_time < 0.001) {
            /* */
            /**< \todo [T1100] (https://unicado.ilr.rwth-aachen.de/T1100): Maybe replace estimation according to t=s/v by Breguet [s]
             */
            total_mission_time = this->mission_file->Range /
                                convertUnit(MACH, TRUEAIRSPEED, this->initial_cruise_altitude_required, this->atm, this->initial_cruise_mach_required);
        }
        /* RuntimeInfos for mission progress */
        if (mode_selector.get_mode_numeric() == 1) {
            if (EndnodeReadOnly<double>("program_settings/general/polar_switch_mission_point/absolute_range_flown").read(this->rtIO->moduleConfig).value() <=
                    convertUnit(METER, NAUTICALMILE, this->mission_profile.back().end_range)) {
                /* Mode: Flown Range absolute */
                this->aero.switchToAlternateDragPolars();
                polars_just_switched = true;
            }
        } else if (mode_selector.get_mode_numeric() == 2) {
            if (EndnodeReadOnly<double>("program_settings/general/polar_switch_mission_point/relative_range_flown").read(this->rtIO->moduleConfig).value() <=
                    this->mission_profile.back().end_range / this->mission_file->Range) {
                /* Mode: Flown Range relative */
                this->aero.switchToAlternateDragPolars();
                polars_just_switched = true;
            }
        } else if (mode_selector.get_mode_numeric() ==  3) {
            if (EndnodeReadOnly<double>("program_settings/general/polar_switch_mission_point/absolute_time_passed").read(this->rtIO->moduleConfig).value() <=
                    convertUnit(SECOND, MINUTE, (this->mission_profile.back().end_time))) {
                /* Mode: Remaining Time absolute */
                this->aero.switchToAlternateDragPolars();
                polars_just_switched = true;
            }
        } else if (mode_selector.get_mode_numeric() == 4) {
            if (EndnodeReadOnly<double>("program_settings/general/polar_switch_mission_point/relative_time_passed").read(this->rtIO->moduleConfig).value() <=
                    this->mission_profile.back().end_time / total_mission_time) {
                /* Mode: Remaining Time relative */
                this->aero.switchToAlternateDragPolars();
                polars_just_switched = true;
            }
        } else {
            myRuntimeInfo->err << "Unforseen Mode for PolarSwitchMissionPoint (" << mode_selector.get_mode_string() << ")! Program Abort!" << std::endl;
            exit(1);
        }
        if (polars_just_switched) {
            myRuntimeInfo->out << "******* Polars have been switched to alternative set! ***" << std::endl
                               << "Current Range [nm]: " << convertUnit(METER, NAUTICALMILE, this->mission_profile.back().end_range) << std::endl
                               << "Total mission Range [nm]: " << convertUnit(METER, NAUTICALMILE, this->mission_file->Range) << std::endl
                               << "Rel. Range: " << this->mission_profile.back().end_range / this->mission_file->Range << std::endl
                               << "Current mission time [min]: " << convertUnit(SECOND, MINUTE, this->mission_profile.back().end_time) << std::endl
                               << "Total mission time [min]: " << convertUnit(SECOND, MINUTE, total_mission_time) << std::endl
                               << "remaing mission time [min]: " << convertUnit(SECOND, MINUTE, (total_mission_time - this->mission_profile.back().end_time)) << std::endl
                               << "Rel. time: " << 1. - this->mission_profile.back().end_time / total_mission_time << std::endl;
            myRuntimeInfo->out << "PolarSwitchMissionPoint " << mode_selector.get_mode_string() << "): " << std::endl;// << mySettingsPt->polarSwitchMissionPoint << std::endl;
        }
    }
}

void standard_mission::interpolate_stabilizer_incidence(Flight_conditions *fc) {
    double current_cg;
    double design_value = fc->stabilizer_incidence;
    if (fc->new_aircraft_cg != 0) {
        current_cg = fc->new_aircraft_cg;
    } else {
        current_cg = cg_design;
    }
    double partner_value;
    if (current_cg >= cg_design) { //aft and design
        partner_value = aero_aft->getIStabPolar(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, *fc->atm, fc->end_TAS),
                                                             fc->end_CL, fc->operating_condition->config);
    } else { //forward and design
        partner_value = aero_forward->getIStabPolar(convertUnit(TRUEAIRSPEED, MACH, fc->end_altitude, *fc->atm, fc->end_TAS),
                                                             fc->end_CL, fc->operating_condition->config);
    }
    fc->stabilizer_incidence = this->interpolateViaCoG(current_cg, design_value, partner_value);
}

bool standard_mission::decide_FL_change(const Flight_conditions& cruise_fc, const Flight_conditions& next_FL_fc,
                                        const double& cost_index, const double& cruise_range, const double& cruise_fuel) {
    /* Calculate ranges until descent for constant flight level (rest_range_cruise) and for flight level change (rest_range_climb) using the same starting point x */
    double rest_range(this->mission_file->Range - this->mission_profile.back().end_range);
    double rest_range_cruise(rest_range - cruise_range - this->descent_range);
    double rest_range_climb(rest_range - next_FL_fc.delta_x - this->descent_range_next_FL);
    double breguet_factor_level_cruise(cruise_fc.end_specific_air_range * cruise_fc.end_aircraft_mass);
    double breguet_factor_climb_step(next_FL_fc.end_specific_air_range * next_FL_fc.end_aircraft_mass);
    double cost_function_level_cruise((cruise_fc.end_aircraft_mass * (1 - exp(- rest_range_cruise / breguet_factor_level_cruise)) + cruise_fuel) / rest_range_cruise +
                                   (this->top_of_descent_mass.value() - this->touchdown_mass) / this->descent_range +
                                   (cost_index / convertUnit(MINUTE, SECOND, 1.) / cruise_fc.end_TAS));
    double cost_function_level_change((next_FL_fc.end_aircraft_mass * (1 - exp(- rest_range_climb / breguet_factor_climb_step))
                                            + next_FL_fc.step_fuel_mass) / rest_range_climb
                                            + (this->top_of_descent_mass_next_FL - this->touchdown_mass_next_FL) / this->descent_range_next_FL
                                            + (cost_index / convertUnit(MINUTE, SECOND, 1.) / next_FL_fc.end_TAS));
    if (cost_function_level_cruise > cost_function_level_change) {
        if (next_FL_fc.end_specific_air_range > cruise_fc.end_specific_air_range) {
            return true;
        } else {
            return false;
        }
    } else {
        return false;
    }
}

void standard_mission::calculate_optimum_FL(double* optimum_altitude, const double& mach, const double& CI,
                                            const double& aircraft_mass, const std::string& config, const double& derate,
                                            const std::string& rating, const bool& rounding_switch) {
    double rel_step_size(2.e-2); // Relative step size
    double lower_altitude(0.);
    double upper_altitude(0.);
    double lower_cost_function(0.); //Cost function with lowerComparisionMachNumber in [kg/m]
    double upper_cost_function(0.); //Cost function with upperComparisionMachNumber in [kg/m]
    int energy_carrier_id =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/cruise")).value();
    /* Search start values */
    //Save engine operating conditions for normal cruise before changes from flight level change simulation occur
    std::vector<OperatingPoint> backup_operating_points = this->engine.get_operating_points(energy_carrier_id);
    while (*optimum_altitude < this->max_operating_altitude_required + ACCURACY_LOW && *optimum_altitude >= 0.) {
        lower_altitude = *optimum_altitude * (1 - rel_step_size);
        upper_altitude = *optimum_altitude * (1 + rel_step_size);
        double econ_altitude_cost_function = this->calculate_cost_function(energy_carrier_id, mach, *optimum_altitude, aircraft_mass, config, rating, 0., 0., CI);
        lower_cost_function = this->calculate_cost_function(energy_carrier_id, mach, lower_altitude, aircraft_mass, config, rating, 0., 0., CI);
        upper_cost_function = this->calculate_cost_function(energy_carrier_id, mach, upper_altitude, aircraft_mass, config, rating, 0., 0., CI);
        if (econ_altitude_cost_function < lower_cost_function && econ_altitude_cost_function < upper_cost_function) {
            break;
        } else if (lower_cost_function > econ_altitude_cost_function && upper_cost_function < econ_altitude_cost_function) {
            *optimum_altitude = upper_altitude;
        } else {
            *optimum_altitude = lower_altitude;
        }
    }
    /* Golden Section Search */
    std::vector<double> abszissa_values(this->get_golden_section_starting_points(lower_altitude, upper_altitude, *optimum_altitude)); //Get start values
    while (fabs(abszissa_values.at(3) - abszissa_values.at(0)) > ACCURACY_MEDIUM * (fabs(abszissa_values.at(1)) + fabs(abszissa_values.at(2)))) { //Get values of cost function
        lower_cost_function = this->calculate_cost_function(energy_carrier_id, mach, abszissa_values.at(1), aircraft_mass, config, rating, 0., 0., CI);
        upper_cost_function = this->calculate_cost_function(energy_carrier_id, mach, abszissa_values.at(2), aircraft_mass, config, rating, 0., 0., CI);
        this->update_abszissa_values(&abszissa_values, {lower_cost_function, upper_cost_function});
    }
    if (lower_cost_function < upper_cost_function) { // We are done. Output the best of the two current values.
        *optimum_altitude = abszissa_values.at(1);
    } else {
        *optimum_altitude = abszissa_values.at(2);
    }
    /* Round to regular flight level */
    if (rounding_switch) {
        double optiumum_FL;
        uint16_t round_down_FL, round_up_FL;
        optiumum_FL = convertUnit(NOPREFIX, METER, HECTO, FOOT, *optimum_altitude);
        round_down_FL = RoundDown(optiumum_FL / 10., 0) * 10.;
        round_up_FL = RoundUp(optiumum_FL / 10., 0) * 10.;
        /* Re-use lower_altitude, upper_altitude, lower_cost_function and upper_cost_function */
        lower_altitude  = convertUnit(HECTO, FOOT, NOPREFIX, METER, round_down_FL);
        upper_altitude = convertUnit(HECTO, FOOT, NOPREFIX, METER, round_up_FL);
        lower_cost_function = this->calculate_cost_function(energy_carrier_id, mach, lower_altitude, aircraft_mass, config, rating, 0., 0., CI);
        upper_cost_function = this->calculate_cost_function(energy_carrier_id, mach, upper_altitude, aircraft_mass, config, rating, 0., 0., CI);
        /* Decision whether to round up or down based on range factor. Preferred to round up due to more realistic mission profiles in testing. */
        if (lower_cost_function < upper_cost_function) {
            myRuntimeInfo->out << "    - Rounded down optimum FL (" << optiumum_FL << ") to regular FL " << round_down_FL << "." << std::endl;
            *optimum_altitude = lower_altitude;
        } else {
            myRuntimeInfo->out << "    - Rounded up optimum FL (" << optiumum_FL << ") to regular FL " << round_up_FL << "." << std::endl;
            *optimum_altitude = upper_altitude;
        }
    }
    this->engine.set_operating_points(energy_carrier_id, backup_operating_points);
    return;
}

void standard_mission::calculate_max_ROC_speed(double* mach, const Flight_conditions& fc, const double& derate) {
    double rel_step_size(1.e-2); // Relative step size
    double mach_1(0.);
    double mach_2(0.);
    double ROC_1(0.);
    double ROC_2(0.);
    int energy_carrier_id(fc.mission_data->at(fc.operating_condition->mission_step).energy_carrier_id);
    //Save engine operating conditions for normal cruise before changes from flight level change simulation occur
    std::vector<OperatingPoint> backup_operating_points = this->engine.get_operating_points(energy_carrier_id);
    double maximum_mach(std::min(convertUnit(CALIBRATEDAIRSPEED, MACH, fc.altitude, this->atm, this->max_operating_speed), this->max_operating_mach));
    /* Search maximum */
    while (*mach < maximum_mach || *mach > 0.4) {
        mach_1 = *mach * (1 + rel_step_size);
        mach_2 = *mach * (1 - rel_step_size);
        this->engine.calculate_N1_with_penalties(energy_carrier_id, fc.altitude, *mach, this->atm, 1., fc.operating_condition->rating, 0., 0.);
        double cruise_drag(this->aero.getCruiseDrag(*mach, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            cruise_drag = this->interpolateViaCoG(current_cg, cruise_drag,
                            getAlternativeAero(current_cg)->getCruiseDrag(*mach, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
        }
        double ROC = -(std::min(1., (this->engine.get_thrust_aircraft(energy_carrier_id) - cruise_drag) / fc.aircraft_mass / G_FORCE)) *
                      convertUnit(MACH, TRUEAIRSPEED, fc.altitude, this->atm, *mach);
        this->engine.calculate_N1_with_penalties(energy_carrier_id, fc.altitude, mach_1, this->atm, 1., fc.operating_condition->rating, 0., 0.);
        double cruise_drag_1(this->aero.getCruiseDrag(mach_1, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            cruise_drag_1 = this->interpolateViaCoG(current_cg, cruise_drag_1,
                            getAlternativeAero(current_cg)->getCruiseDrag(mach_1, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
        }
        ROC_1 = -(std::min(1., (this->engine.get_thrust_aircraft(energy_carrier_id) - cruise_drag_1) / fc.aircraft_mass / G_FORCE)) *
                                              convertUnit(MACH, TRUEAIRSPEED, fc.altitude, this->atm, mach_1);
        this->engine.calculate_N1_with_penalties(energy_carrier_id, fc.altitude, mach_2, this->atm, 1., fc.operating_condition->rating, 0., 0.);
        double cruise_drag_2(this->aero.getCruiseDrag(mach_2, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
        if (sophistication_level != UNSOPHISTICATED) {
            double current_cg(get_current_cg());
            cruise_drag_2 = this->interpolateViaCoG(current_cg, cruise_drag_2,
                            getAlternativeAero(current_cg)->getCruiseDrag(mach_2, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
        }
        ROC_2 = -(std::min(1., (this->engine.get_thrust_aircraft(energy_carrier_id) - cruise_drag_2) / fc.aircraft_mass / G_FORCE)) *
                                              convertUnit(MACH, TRUEAIRSPEED, fc.altitude, this->atm, mach_2);
        if (ROC < ROC_1 && ROC < ROC_2) {
            break;
        } else if (ROC_1 > ROC && ROC_2 < ROC) {
            *mach = mach_2;
        } else {
            *mach = mach_1;
        }
    }
    /* Given a function f, and given a bracketing triplet of abscissas ax, bx, cx (such that bx is
    between ax and cx,and f(bx) is less than both f(ax) and f(cx)), this routine performs a
    golden section search for the minimum, isolating it to a fractional precision of about tol. The
    abscissa of the minimum is returned as xmin, and the minimum function value is returned as
    golden, the returned function value. */
    const double ratio(2. / (1. + sqrt(5.))); // The golden ratio = 0.61803399.
    const double tol(1.e-7); // Error
    double x0, x1, x2, x3;
    x0 = mach_1; // At any given time we will keep track of four points, x0,x1,x2,x3.
    x3 = mach_2;
    if (fabs(mach_2 - *mach) > fabs(*mach - mach_1)) { // Make x0 to x1 the smaller segment,
        x1 = *mach;
        x2 = *mach + (1. - ratio) * (mach_2 - *mach); // and fill in the new point to be tried.
    } else {
        x2 = *mach;
        x1 = *mach - (1. - ratio) * (*mach - mach_1);
    }
    this->engine.calculate_N1_with_penalties(energy_carrier_id, fc.altitude, x1, this->atm, 1., fc.operating_condition->rating, 0., 0.);
    double cruise_drag_1(this->aero.getCruiseDrag(x1, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        cruise_drag_1 = this->interpolateViaCoG(current_cg, cruise_drag_1,
                        getAlternativeAero(current_cg)->getCruiseDrag(x1, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
    }
    ROC_1 = -(std::min(1., (this->engine.get_thrust_aircraft(energy_carrier_id) - cruise_drag_1) / fc.aircraft_mass / G_FORCE)) *
                                          convertUnit(MACH, TRUEAIRSPEED, fc.altitude, this->atm, x1);
    this->engine.calculate_N1_with_penalties(energy_carrier_id, fc.altitude, x2, this->atm, 1., fc.operating_condition->rating, 0., 0.);
    double cruise_drag_2(this->aero.getCruiseDrag(x2, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
    if (sophistication_level != UNSOPHISTICATED) {
        double current_cg(get_current_cg());
        cruise_drag_2 = this->interpolateViaCoG(current_cg, cruise_drag_2,
                        getAlternativeAero(current_cg)->getCruiseDrag(x2, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
    }
    ROC_2 = -(std::min(1., (this->engine.get_thrust_aircraft(energy_carrier_id) - cruise_drag_2) / fc.aircraft_mass / G_FORCE)) *
                                          convertUnit(MACH, TRUEAIRSPEED, fc.altitude, this->atm, x2);
    while (fabs(x3 - x0) > tol * (fabs(x1) + fabs(x2))) {
        if (ROC_2 < ROC_1) { //
            /* One possible outcome, its housekeeping, */
            x0 = x1;
            x1 = x2;
            x2 = ratio * x1 + (1. - ratio) * x3;
            /* and a new function evaluation. */
            ROC_1 = ROC_2;
            this->engine.calculate_N1_with_penalties(energy_carrier_id, fc.altitude, x2, this->atm, 1., fc.operating_condition->rating, 0., 0.);
            cruise_drag_2 = this->aero.getCruiseDrag(x2, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm);
            if (sophistication_level != UNSOPHISTICATED) {
                double current_cg(get_current_cg());
                cruise_drag_2 = this->interpolateViaCoG(current_cg, cruise_drag_2,
                                getAlternativeAero(current_cg)->getCruiseDrag(x2, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
            }
            ROC_2 = (std::min(1., (this->engine.get_thrust_aircraft(energy_carrier_id) - cruise_drag_2) / fc.aircraft_mass / G_FORCE)) *
                                                 convertUnit(MACH, TRUEAIRSPEED, fc.altitude, this->atm, x2);
        } else {
            /* The other outcome, */
            x3 = x2;
            x2 = x1;
            x1 = ratio * x2 + (1. - ratio) * x0;
            /* and its new function evaluation. */
            ROC_2 = ROC_1;
            this->engine.calculate_N1_with_penalties(energy_carrier_id, fc.altitude, x1, this->atm, 1., fc.operating_condition->rating, 0., 0.);
            cruise_drag_1 = this->aero.getCruiseDrag(x1, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm);
            if (sophistication_level != UNSOPHISTICATED) {
                double current_cg(get_current_cg());
                cruise_drag_1 = this->interpolateViaCoG(current_cg, cruise_drag_1,
                                getAlternativeAero(current_cg)->getCruiseDrag(x1, fc.altitude, 0., fc.aircraft_mass, fc.operating_condition->config, this->atm));
            }
            ROC_1 = (std::min(1., (this->engine.get_thrust_aircraft(energy_carrier_id) - cruise_drag_1) / fc.aircraft_mass / G_FORCE)) *
                                                 convertUnit(MACH, TRUEAIRSPEED, fc.altitude, this->atm, x1);
        }
    }
    /* Back to see if we are done. */
    if (ROC_1 < ROC_2) { // We are done. Output the best of the two current values.
        *mach = x1;
    } else {
        *mach = x2;
    }
    this->engine.set_operating_points(energy_carrier_id, backup_operating_points);
    return;
}

void standard_mission::calcOptSpeedAndFL(double* econ_mach_number, double* altitude, const double CI, const double& aircraft_mass, const std::string& config, const double& derate,
                                const std::string& rating, const int& energy_carrier_id) {
    double rel_step_size(2.e-2); // Relative step size
    double tolerance(1.e-7);
    double P[2][3] = {{0, *econ_mach_number, 0}, {0, *altitude, 0}};
    double Y[3];
    bool converged(0);
    bool found_start[2] = {0, 0};
    int iterations(0);
    double PR[2], PRR[2], PBAR[2];
    double ALPHA = 0.015, BETA = 0.007, GAMMA = 1.005;
    int I, IHI, ILO, INHI, J, MPTS;
    double RTOL, YPR, YPRR;
    MPTS = 3;
    int NDIM(2);
    double maximum_mach(std::min(convertUnit(CALIBRATEDAIRSPEED, MACH, P[1][1], this->atm, this->max_operating_speed), this->max_operating_mach));
    if (*econ_mach_number > maximum_mach) *econ_mach_number = maximum_mach * (1. - rel_step_size);
    //Save engine operating conditions for normal cruise before changes from flight level change simulation occur
    std::vector<OperatingPoint> backup_operating_points = this->engine.get_operating_points(energy_carrier_id);
    while (iterations < 1000) {
        if (!found_start[0] || !found_start[1]) {
            for (int i(0); i < 2; i++) {
                P[i][0] = P[i][1] * (1. - rel_step_size);
                P[i][2] = P[i][1] * (1. + rel_step_size);
                P[abs(i - 1)][0] = P[abs(i - 1)][1];
                P[abs(i - 1)][2] = P[abs(i - 1)][1];
                /* Mach number limitation */
                Y[0] = this->calculate_cost_function(energy_carrier_id, P[0][0], P[1][0], aircraft_mass, config, rating, 0., 0., CI);
                Y[1] = this->calculate_cost_function(energy_carrier_id, P[0][1], P[1][1], aircraft_mass, config, rating, 0., 0., CI);
                Y[2] = this->calculate_cost_function(energy_carrier_id, P[0][2], P[1][2], aircraft_mass, config, rating, 0., 0., CI);
                if (Y[1] < Y[0] && Y[1] < Y[2]) {
                    found_start[i] = 1;
                } else if (Y[0] < Y[1]) {
                    P[i][1] = P[i][0];
                } else {
                    P[i][1] = P[i][2];
                }
            }
        } else {
            ILO = 0;
            if (Y[0] > Y[1]) {
                IHI = 0;
                INHI = 1;
            } else {
                IHI = 1;
                INHI = 0;
            }
            for (I = 0; I < MPTS; I++) {
                if (Y[I] < Y[ILO]) {
                    ILO = I;
                }
                if (Y[I] > Y[IHI]) {
                    INHI = IHI;
                    IHI = I;
                } else if (Y[I] > Y[INHI]) {
                    if (I != IHI)  INHI = I;
                }
            }
            // Sequence of y-values stored in ILO, INHI, IHI
            // Compute the fractional range from highest to lowest and return if satisfactory.
            RTOL = 2.0 * fabs(Y[IHI] - Y[ILO]) / (fabs(Y[IHI]) + fabs(Y[ILO])); // Review of the results
            if (RTOL < tolerance)  converged = 1; // normal exit
            for (J = 0; J < NDIM; J++) {
                PBAR[J] = 0.0;
            }
            for (I = 0; I < MPTS; I++) {
                if (I != IHI) {
                    for (J = 0; J < NDIM; J++) {
                        PBAR[J] += P[J][I];
                    }
                }
            }
            for (J = 0; J < NDIM; J++) {
                PBAR[J] /= 1.0 * NDIM;
                PR[J] = (1.0 + ALPHA) * PBAR[J] - ALPHA * P[J][IHI];
            }
            YPR = this->calculate_cost_function(energy_carrier_id, PR[0], PR[1], aircraft_mass, config, rating, 0., 0., CI);
            if (YPR <= Y[ILO]) {
                for (J = 0; J < NDIM; J++) {
                    PRR[J] = GAMMA * PR[J] + (1.0 - GAMMA) * PBAR[J];
                }
                YPRR = this->calculate_cost_function(energy_carrier_id, PRR[0], PRR[1], aircraft_mass, config, rating, 0., 0., CI);
                if (YPRR < Y[ILO]) {
                    for (J = 0; J < NDIM; J++) {
                        P[J][IHI] = PRR[J];
                    }
                    Y[IHI] = YPRR;
                } else {
                    for (J = 0; J < NDIM; J++) {
                        P[J][IHI] = PR[J];
                    }
                    Y[IHI] = YPR;
                }
            } else if (YPR >= Y[INHI]) {
                if (YPR < Y[IHI]) {
                    for (J = 0; J < NDIM; J++) {
                        P[J][IHI] = PR[J];
                    }
                    Y[IHI] = YPR;
                }
                for (J = 0; J < NDIM; J++) {
                    PRR[J] = BETA * P[J][IHI] + (1.0 - BETA) * PBAR[J];
                }
                YPRR = this->calculate_cost_function(energy_carrier_id, PRR[0], PRR[1], aircraft_mass, config, rating, 0., 0., CI);
                if (YPRR < Y[IHI]) {
                    for (J = 0; J < NDIM; J++) {
                        P[J][IHI] = PRR[J];
                    }
                    Y[IHI] = YPRR;
                } else {
                    for (I = 0; I < MPTS; I++) {
                        if (I != ILO) {
                            for (J = 0; J < NDIM; J++) {
                                PR[J] = 0.5 * (P[J][I] + P[J][ILO]);
                                P[J][I] = PR[J];
                            }
                            Y[I] = this->calculate_cost_function(energy_carrier_id, PR[0], PR[1], aircraft_mass, config, rating, 0., 0., CI);
                        }
                    }
                }
            } else {
                for (J = 0; J < NDIM; J++) {
                    P[J][IHI] = PR[J];
                }
                Y[IHI] = YPR;
            }
        }
        iterations++;
        if (iterations == 1000) { // The maximum number of iterations (1000 iterations) is reached, error message
            myRuntimeInfo->err << "Maximum number of iterations reached!" << std::endl;
            exit(1);
        }
        if (converged) {
            *econ_mach_number = P[0][1];
            *altitude = P[1][1];
            break;
        }
    }
    this->engine.set_operating_points(energy_carrier_id, backup_operating_points);
}
