/**
 * \file standard_mission_output.cpp
 * \author Gerrit Pauls (gerrit.pauls@tuhh.de)
 * \brief
 * \version 0.1
 * \date 2024-08-30
 *
 * @copyright Copyright (c) 2024
 *
 */
#include <algorithm>
#include <iomanip>
#include <utility>
#include <tuple>

#include <moduleBasics/colors.h>
#include <moduleBasics/html.h>
#include <moduleBasics/module.h>
#include <moduleBasics/plot.h>

#include "standard_mission_output.h"
#include "../../toolinfo.h"

#ifdef _WIN32
    #include <fileapi.h>
#endif // _WIN32

/* Initialization of attributes for the standard_mission_output object with the constructor */
standard_mission_output::standard_mission_output(standard_mission* mission_ptr)
    :
    /* Initialization of pointers */
    my_mission_ptr(mission_ptr),
    rtIO(mission_ptr->rtIO),
    report_(rtIO),
    /* Initialization of switch for requirements checked */
    requirements_checked(false),
    /* Initialization of SAR plot altitudes */
    start_altitude(0.),
    end_altitude(0.),
    SAR_plot_on(EndnodeReadOnly<bool>("program_specific/specific_air_range_plot").read(rtIO->moduleConfig).value()),
    SAR_mach(),
    SAR_FL(),
    SAR_total_mass(),
    specific_air_ranges(),
    /* localize variable camber */
    first_filled_position(0) {
    this->initialize();
    if (my_mission_ptr->requirement_switch) { // Requirements mission does not write a full report
        if (rtIO->reportOn) {
            this->generate_requirements_html_report();
            report_.generateReports(rtIO->programname + "_" + my_mission_ptr->mission_type);
        }
    } else {
        /*Write mission output csv files */
        this->generate_mission_profile_data();
        if (rtIO->plotOn) {
            this->generate_mission_plots();
            if (SAR_plot_on) {
                this->generate_SAR_plot_data();
                this->generate_SAR_plots();
            }
        }
        if (rtIO->reportOn) {
            this->generate_html_report();
            this->generate_tex_report_body();
            report_.generateReports(rtIO->programname + "_" + my_mission_ptr->mission_type);
        }
    }
}

/* Destructor of the standard_mission_output object */
standard_mission_output::~standard_mission_output() {
}

void standard_mission_output::initialize() {
    /* Set file names */
    std::string outputFilename = "/" + std::string(rtIO->moduleConfig.at("control_settings/aircraft_exchange_file_name/value"));
    outputFilename.resize(outputFilename.rfind('.')); // Cut off any extension, if it exists
    this->output_name << outputFilename << "_" << this->my_mission_ptr->mission_type
        << "_R" << static_cast<int>(ceil(convertUnit(METER, NAUTICALMILE, this->my_mission_ptr->mission_file->Range)))
        << "_PL" << static_cast<int>(ceil(this->my_mission_ptr->payload_mass.value()));
    if (this->my_mission_ptr->mission_file->air_to_air_fuel_mass != 0.)
        this->output_name << "_AAF" << this->my_mission_ptr->mission_file->air_to_air_fuel_mass;
    this->csv_file_name = rtIO->getMissionDataDir() + this->output_name.str() + "_out.csv";
    if (!fileExists(rtIO->getMissionDataDir())) {
        myRuntimeInfo->out << "Create directory for mission data output in project folder: " << rtIO->getMissionDataDir() << std::endl;
        CreateDirectory(rtIO->getMissionDataDir().c_str(), NULL);
    }
    this->base_name_output_files = rtIO->programname + "_" + this->my_mission_ptr->mission_type;
    /* Check if the calculated reserve fuel + extra fuel is negative after taxiing at destination airport. If so, warning message */
    double fuelAtDestinationGate = this->my_mission_ptr->get_fuel_sum(this->my_mission_ptr->loaded_fuel_mass) -
                                    this->my_mission_ptr->get_fuel_sum(this->my_mission_ptr->trip_fuel_mass) -
                                    this->my_mission_ptr->taxi_fuel_mass_origin.second - this->my_mission_ptr->taxi_fuel_mass_destination.second;
    if (fuelAtDestinationGate < 0.) {
        myRuntimeInfo->warn << "Reserve fuel and extra fuel are not enough to taxi to gate at destination airport." << std::endl;
    }

    std::string subpath("requirement_compliance/top_level_aircraft_requirements");
    bool checks_exist(rtIO->acxml.find(subpath + "/initial_cruise_altitude") &&
                        rtIO->acxml.find(subpath + "/maximum_operating_altitude") &&
                        rtIO->acxml.find(subpath + "/maximum_altitude_one_engine_inoperative") &&
                        rtIO->acxml.find(subpath + "/initial_cruise_mach_number") &&
                        rtIO->acxml.find(subpath + "/design_time_to_climb"));

    if (checks_exist && // If checks were made, control if they were finished correctly
        EndnodeReadOnly<bool>(subpath + "/initial_cruise_altitude/checked").read(rtIO->acxml).value() &&
        EndnodeReadOnly<bool>(subpath + "/maximum_operating_altitude/checked").read(rtIO->acxml).value() &&
        EndnodeReadOnly<bool>(subpath + "/maximum_altitude_one_engine_inoperative/checked").read(rtIO->acxml).value() &&
        EndnodeReadOnly<bool>(subpath + "/initial_cruise_mach_number/checked").read(rtIO->acxml).value() &&
        EndnodeReadOnly<bool>(subpath + "/design_time_to_climb/checked").read(rtIO->acxml).value()) {
        this->requirements_checked = true;
    } else {
        this->requirements_checked = false;
    }
}

void standard_mission_output::generate_mission_profile_data() {
    std::ofstream csvStream;
    csvStream.open(this->csv_file_name.c_str());
    if (!csvStream) {
        /* Error message: file could not be opened */
        myRuntimeInfo->err << "Could not open " << this->csv_file_name << std::endl;
        exit(1);
    }
    /* Adding zeros to empty fields of the flap setting vector, e.g. non-cruise flight segments (variable camber) */
    for (size_t i(0); i < my_mission_ptr->mission_profile.size(); ++i) {
        if (my_mission_ptr->mission_profile.at(i).end_origin_polar.empty()) {
            my_mission_ptr->mission_profile.at(i).end_origin_polar.resize(2);
            for (size_t j(0); j < my_mission_ptr->variable_camber_segment_quantity; ++j) {
                if (first_filled_position != 0) { // workaround to fill origin_polar[0] with the correct FlapSegmentID.
                    my_mission_ptr->mission_profile.at(i).end_origin_polar[0].push_back(my_mission_ptr->mission_profile.at(first_filled_position).end_origin_polar[0].at(j));
                } else { // adding zeros to origin_polar[0] for non-cruise flight segments
                    my_mission_ptr->mission_profile.at(i).end_origin_polar[0].push_back(0);
                } // adding zeros for flap settings in non-cruise flight segments
                my_mission_ptr->mission_profile.at(i).end_origin_polar[1].push_back(0.);
            }
        } else if (first_filled_position == 0) {
            first_filled_position = i;
        }
    }

    // Define header
    csvStream << "Time [s]; " << "Range [m]; " << "Altitude [m]; " << "FL [100 ft]; " << "Mode name [-]; "
                << "Total mass [kg]; " << "Energy carrier (ID); " << "Thrust [N]; " << "Fuelflow [kg/s]; ";
    for (std::pair<const int, std::tuple<std::string, double, double, double>> energy_carrier : this->rtIO->aircraft_energy_carriers()) {
       csvStream << (std::format("Fuel consumed ({}, ID = {}) [kg]; ", rtIO->get_fuel_type(energy_carrier.first), energy_carrier.first))
                    << (std::format("Energy consumed ({}, ID = {}) [J]; ", rtIO->get_fuel_type(energy_carrier.first), energy_carrier.first));
    }
    csvStream << "Mach [-]; " << "CAS [m/s]; " << "TAS [m/s]; " << "TAS [kts]; " << "ROC [fpm]; " << "SAR [m/kg]; " << "Aero Config [-]; " << "C_L [-]; "
                 << "L over D [-]; " << "Spoiler Factor [-]; " << "Reynolds Number [-]; " <<  "Engine Rating [-]; ";
    for (node* propulsor : rtIO->acxml.at("component_design/propulsion/specific").getVector("propulsion")) {
        csvStream << "Engine N1 (" << propulsor->at("engine/model/value") << ", ID = " << propulsor->getIntAttrib("ID") << ") [-]; ";
    }
    csvStream << "Shaft power offtake [W]; " << "Bleed [kg/s]; ";
    if (this->my_mission_ptr->sophistication_level != this->my_mission_ptr->UNSOPHISTICATED) { // Center of Gravity shift not needed if not calculated
        csvStream << "CoG aircraft [m]; " << "CoG fuel [m]; ";
    }
    csvStream << "Angle of attack [deg]; " << "Glidepath angle [deg]; " << "Incidence angle (stabilizer) [deg]";
    for (uint8_t k(0); k < this->my_mission_ptr->mission_profile.at(first_filled_position).end_origin_polar[0].size(); ++k) {
        std::string flap_setting_name("segment ID " + num2Str(this->my_mission_ptr->mission_profile.at(first_filled_position).end_origin_polar[0].at(k)) + " flap setting");
        csvStream << ";(" << 32+k << ") " << flap_setting_name << " [deg]; ";
        if (rtIO->plotOn && this->my_mission_ptr->variable_camber_switch) { // Collect names for plot data
            matplot_axis_data tmp;
            tmp.name = flap_setting_name;
            this->flap_settings.push_back(tmp);
        }
    }

    // Fill csv
    for (unsigned int i = 0; i < my_mission_ptr->mission_profile.size(); i++) {
        csvStream << std::endl
        << std::setprecision(10) << my_mission_ptr->mission_profile.at(i).end_time << "; "
        << my_mission_ptr->mission_profile.at(i).end_range << "; "
        << my_mission_ptr->mission_profile.at(i).end_altitude << "; "
        << convertUnit(NOPREFIX, METER, HECTO, FOOT, my_mission_ptr->mission_profile.at(i).end_altitude)
                                    << "; " // Conversion of end_altitude from [m] to [ft]
        << my_mission_ptr->mission_profile.at(i).end_mode_name << "; "
        << my_mission_ptr->mission_profile.at(i).end_total_mass << "; "
        << std::get<0>(rtIO->aircraft_energy_carriers().at(my_mission_ptr->mission_profile.at(i).energy_carrier_id))
            << " (ID = " << my_mission_ptr->mission_profile.at(i).energy_carrier_id << "); "
        << my_mission_ptr->mission_profile.at(i).end_thrust << "; "
        << my_mission_ptr->mission_profile.at(i).end_fuel_flow << "; ";
        for (const std::pair<const int, double>& fuel_mass : my_mission_ptr->mission_profile.at(i).end_total_fuel_mass) {
            csvStream << fuel_mass.second << "; "
                << fuel_mass.second * rtIO->get_fuel_gravimetric_energy_density(fuel_mass.first) << "; ";
        }
        csvStream << my_mission_ptr->mission_profile.at(i).end_mach << "; "
        << my_mission_ptr->mission_profile.at(i).end_CAS << "; "
        << my_mission_ptr->mission_profile.at(i).end_TAS << "; "
        << convertUnit(NOPREFIX, METERPERSECOND, NOPREFIX, KNOTS,
                                               my_mission_ptr->mission_profile.at(i).end_TAS) << "; "; // Conversion of TAS from [m/s] to [kts]
        if (i > 0) {
            csvStream << convertUnit(METERPERSECOND, FOOTPERMINUTE, (my_mission_ptr->mission_profile.at(i).end_altitude - my_mission_ptr->mission_profile.at(i - 1).end_altitude) /
                                                               (my_mission_ptr->mission_profile.at(i).end_time - my_mission_ptr->mission_profile.at(i - 1).end_time)) << "; ";
        } else {
            csvStream << 0.0 << "; ";
        }
        csvStream << my_mission_ptr->mission_profile.at(i).end_specific_air_range << "; "
        << my_mission_ptr->mission_profile.at(i).end_aero_config << "; "
        << my_mission_ptr->mission_profile.at(i).end_CL << "; "
        << my_mission_ptr->mission_profile.at(i).end_L_over_D << "; "
        << my_mission_ptr->mission_profile.at(i).end_spoiler_factor << "; "
        << my_mission_ptr->mission_profile.at(i).end_reynolds_number << "; "
        << my_mission_ptr->mission_profile.at(i).end_engine_rating << "; ";
        for (node* propulsor : rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion").getVector("propulsor")) {
            csvStream << my_mission_ptr->mission_profile.at(i).end_n1.at(propulsor->getIntAttrib("ID")) << "; "; // Add N1 for each engine
        }
        csvStream << my_mission_ptr->mission_profile.at(i).end_shaft_power << "; "
        << my_mission_ptr->mission_profile.at(i).end_bleed << "; ";
        if (this->my_mission_ptr->sophistication_level != this->my_mission_ptr->UNSOPHISTICATED) { // Center of Gravity shift not needed if not calculated
            csvStream << my_mission_ptr->mission_profile.at(i).end_aircraft_cg << "; "
                        << my_mission_ptr->mission_profile.at(i).end_fuel_cg << "; ";
        }
        csvStream << my_mission_ptr->mission_profile.at(i).end_angle_of_attack << "; "
        << convertUnit(RADIAN, DEGREE, my_mission_ptr->mission_profile.at(i).end_glidepath_angle) << "; "
        << my_mission_ptr->mission_profile.at(i).end_stabilizer_incidence;
        for (uint8_t k(0); k < my_mission_ptr->mission_profile.at(i).end_origin_polar[1].size(); ++k) {
            csvStream << "; " << my_mission_ptr->mission_profile.at(i).end_origin_polar[1].at(k);
            if (rtIO->plotOn && this->my_mission_ptr->variable_camber_switch) { // Fill the empty flap_settings with data
                this->flap_settings[k].data.push_back(my_mission_ptr->mission_profile.at(i).end_origin_polar[1].at(k));
            }
        }
        if (rtIO->plotOn) {
            this->range_profile.data.push_back(convertUnit(METER, NAUTICALMILE, my_mission_ptr->mission_profile.at(i).end_range));
            this->altitude_profile.data.push_back(convertUnit(NOPREFIX, METER, HECTO, FOOT, my_mission_ptr->mission_profile.at(i).end_altitude));
            this->C_L_profile.data.push_back(my_mission_ptr->mission_profile.at(i).end_CL);
            this->thrust_profile.data.push_back(convertUnit(NOPREFIX, NEWTON, KILO, NEWTON, my_mission_ptr->mission_profile.at(i).end_thrust));
            double fuel_consumption_tmp = std::accumulate(my_mission_ptr->mission_profile.at(i).end_total_fuel_mass.begin(), // Sum all consumed fuel masses
                                                            my_mission_ptr->mission_profile.at(i).end_total_fuel_mass.end(), 0.0,
                                                            [](double total, const std::pair<int, double>& fuel_mass_tmp) {
                                                                return total - fuel_mass_tmp.second / 1000.0;});
            this->fuel_consumption_profile.data.push_back(fuel_consumption_tmp);
            this->mach_profile.data.push_back(my_mission_ptr->mission_profile.at(i).end_mach);
            this->CAS_profile.data.push_back(convertUnit(METERPERSECOND, KNOTS, my_mission_ptr->mission_profile.at(i).end_CAS));
            this->TAS_profile.data.push_back(convertUnit(METERPERSECOND, KNOTS, my_mission_ptr->mission_profile.at(i).end_TAS));
            this->bleed_air_profile.data.push_back(my_mission_ptr->mission_profile.at(i).end_bleed);
            this->shaft_power_profile.data.push_back(my_mission_ptr->mission_profile.at(i).end_TAS);
            if (my_mission_ptr->sophistication_level != my_mission_ptr->UNSOPHISTICATED) { // No CoG shift
                this->CoG_pos_profile.data.push_back(my_mission_ptr->mission_profile.at(i).end_aircraft_cg);
                this->i_stab_profile.data.push_back(my_mission_ptr->mission_profile.at(i).end_stabilizer_incidence);
            }
        }
    }
    csvStream.close();
}

void standard_mission_output::generate_SAR_plot_data() {
    /* Open SAR plot of the CSV file */
    std::ofstream plot;
    std::string csvFile = rtIO->getCsvFilesDir() + "/" + this->base_name_output_files + "_SAR.csv";
    int energy_carrier_id =
        EndnodeReadOnly<int>("energy_carrier_ID").read(rtIO->acxml.at("requirements_and_specifications/design_specification/propulsion/energy_provider/cruise")).value();
    plot.open(csvFile.c_str());
    /* Check if plot could not be opened. If so, error message */
    if (!plot) {
        myRuntimeInfo->err << csvFile << " could not be opened!" << std::endl;
        exit(1);
    } else {
        plot << "# " + csvFile << std::endl;
        /* Customize data structure: */
        plot << "# Diagram about specific air range" << std::endl;
        plot << "# Ma-Number [-]; Altitude [FL]; AC mass [kg]; SAR [m/kg];" << std::endl;
        std::string Config_set = "clean";
        std::string Rating_set = "cruise";
         // begin at 80 % of ICA rounded to lower FL which is a multiple of 5
        start_altitude = floor(convertUnit(NOPREFIX, METER, HECTO, FOOT, 0.8 * my_mission_ptr->initial_cruise_altitude_required)/5)*5;

        if (my_mission_ptr->max_operating_altitude > ACCURACY_LOW) { // Get maximum altitude from requirements mission if available
            end_altitude = my_mission_ptr->max_operating_altitude;
        } else {
            end_altitude = my_mission_ptr->max_operating_altitude_required; // Use the maximum altitude requirement
        }
        end_altitude = ceil(convertUnit(NOPREFIX, METER, HECTO, FOOT, end_altitude)/5)*5; // round to higher FL which is a multiple of 5
        double altitudeStep(5.); // [FL]
        double weightSteps(30.); // [kg]
        double acc(10e-5);
        // Temporarily issue exit when CL_max is exceeded
        my_mission_ptr->aero.switchOnReqSwitch();
        for (double Ma(my_mission_ptr->mission_file->desired_cruise_speed - 0.04); Ma <= (my_mission_ptr->mission_file->desired_cruise_speed + 0.04); Ma += 0.02) {
            SAR_mach.push_back(Ma);
            for (double aircraft_mass(0.5 * my_mission_ptr->MTOM); aircraft_mass <= my_mission_ptr->MTOM; aircraft_mass += 0.5 * my_mission_ptr->MTOM / weightSteps) {
                if ((Ma - (my_mission_ptr->mission_file->desired_cruise_speed - 0.04)) < acc) {
                    SAR_total_mass.push_back(aircraft_mass/1000.);
                }
                for (double Alt(start_altitude); Alt <= end_altitude; Alt += altitudeStep) {
                    if (Ma == my_mission_ptr->mission_file->desired_cruise_speed - 0.04 && aircraft_mass == 0.5 * my_mission_ptr->MTOM) {
                        SAR_FL.push_back(Alt); // Catch FL only in first MA and AC weight loop (the values are the same for each loop)
                    }
                    double SAR_tmp(NAN);
                    try {
                        SAR_tmp = my_mission_ptr->static_flight_specific_air_range(energy_carrier_id, Ma, convertUnit(HECTO, FOOT, NOPREFIX, METER, Alt),
                                                                                        aircraft_mass, Config_set, Rating_set);
                    } catch (const std::string& text)  {
                        myRuntimeInfo->out << "... but this was intended: SAR limit for plotting reached!" << std::endl;
                    } catch (...) {
                    }
                    plot << std::setprecision(7) << Ma << "; ";
                    plot << std::setprecision(7) << Alt << "; "; // in meters
                    plot << std::setprecision(6) << aircraft_mass/1000. << "; "; // in kg
                    plot << std::setprecision(6) << SAR_tmp << "; " << std::endl;
                    specific_air_ranges.push_back(SAR_tmp);
                }
            }
        }
    }
    plot.close();
    my_mission_ptr->aero.switchOffReqSwitch();
}

void standard_mission_output::generate_mission_plots() {
    this->range_profile.name = "Range";
    this->altitude_profile.name = "Altitude";
    this->C_L_profile.name = "C_L";
    this->thrust_profile.name = " Thrust";
    this->fuel_consumption_profile.name = "Consumed fuel";
    this->CAS_profile.name = "Calibrated Airspeed";
    this->TAS_profile.name = "True Airspeed";
    this->mach_profile.name = "Mach number";
    this->bleed_air_profile.name = "Bleed Air";
    this->shaft_power_profile.name = "Shaft Power";

    /* Plot with matplot++*/
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 600);

    // upper
    matplot::axes_handle ax1 = fig->add_subplot(2, 1, 0);
    fig->title("Mission Profile (m_{takeoff} = " + num2Str(my_mission_ptr->takeoff_mass.value()) + " kg; m_{payload} = " +
        num2Str(my_mission_ptr->payload_mass.value()) + " kg)");
    fig->title_font_size_multiplier(1.5);
    // Left y axis
    set_matplot_axes(ax1, {this->altitude_profile.data}, "Altitude [FL]", false, this->range_profile.data);
    // Right y axis
    set_matplot_axes(ax1, {this->C_L_profile.data}, "C_L [-]", true);
    ax1->y2lim({std::accumulate(this->C_L_profile.data.begin(), this->C_L_profile.data.end(), 0.0)/this->C_L_profile.data.size()*0.5, // Get rid off C_L peaks at start and landing
                std::accumulate(this->C_L_profile.data.begin(), this->C_L_profile.data.end(), 0.0)/this->C_L_profile.data.size()*1.5});
    // Fill the charts
    add_matplot_to_axes(ax1, this->range_profile, {this->altitude_profile}, true, {this->C_L_profile});
    // lower plot
    matplot::axes_handle ax2 = fig->add_subplot(2, 1, 1);
    // Left y axis
    set_matplot_axes(ax2, {this->thrust_profile.data}, "Thrust [kN]", false, this->range_profile.data, "Range [NM]");
    // Right y axis
    set_matplot_axes(ax2, {this->fuel_consumption_profile.data}, "Consumed fuel [metric tons]", true);
    // Fill the charts
    add_matplot_to_axes(ax2, this->range_profile, {this->thrust_profile}, true, {this->fuel_consumption_profile});

    // Shift plots closer to each other
    matplot::color_array pos1 = ax1->position(); // position({left, bottom, width, height}) [0; 1] fraction of fig height and width
    matplot::color_array pos2 = ax2->position();
    pos1[1] -= 0.075; // lower starting height of subplot 1 by 7.5 %
    pos1[3] += 0.075; // increase heigth of subplot 1 by 7.5 %
    pos2[1] -= 0.075; // lower starting height of subplot 2 by 7.5 %
    pos2[3] += 0.07; // increase heigth of subplot 2 by 7.5 %
    ax2->position(pos2);
    ax1->position(pos1);

    // Save the plot
    matplot::save(fig, rtIO->getPlotDir() + "/" + this->base_name_output_files + "_Alt-CL.svg");
    ax1->clear();
    ax2->clear();
    matplot::cla();

    if (my_mission_ptr->variable_camber_switch) {
        /* Plot with matplot++*/
        fig = matplot::figure(true); // overwrite old figure
        fig->quiet_mode(true);
        fig->size(600, 300);

        std::vector<std::vector<double>> tmp_flap_data; // get data vector from matplot_axis_data vector
        for (uint8_t i(0); i < this->flap_settings.size(); i++) {
            tmp_flap_data.push_back(flap_settings[i].data);
        }

        // Set up axes and plot data
        matplot::axes_handle ax = fig->current_axes();
        fig->title("Flap Settings");
        fig->title_font_size_multiplier(1.5);
        set_matplot_axes(ax, tmp_flap_data, "Flap settings [deg]", false, range_profile.data, "Range [NM]");
        add_matplot_to_axes(ax, range_profile, flap_settings, true);

        // Save the plot
        matplot::save(fig, rtIO->getPlotDir() + "/" + this->base_name_output_files + "_trim_setting.svg");
        ax->clear();
        matplot::cla();
    }

    /* Plot the center of gravity shift */
    if (!(this->my_mission_ptr->aero_aft == nullptr) && !(this->my_mission_ptr->aero_forward == nullptr)) { // No alternative aerodynamics = no CoG shift!
        this->CoG_pos_profile.name = "Center of Gravity";
        this->i_stab_profile.name = "Stabilizer flap angle";
        /* Plot with matplot++*/
        fig = matplot::figure(true); // overwrite old figure
        fig->quiet_mode(true);
        fig->size(600, 300);

        // Set up axes and plot data
        matplot::axes_handle ax = fig->current_axes();
        fig->title("Trimming");
        fig->title_font_size_multiplier(1.5);
        set_matplot_axes(ax, {this->CoG_pos_profile.data}, "Center of gravity global x-position [m]", false, this->range_profile.data, "Range [NM]");
        set_matplot_axes(ax, {this->i_stab_profile.data}, "Stabilizer flap angle [deg]", true);
        add_matplot_to_axes(ax, this->range_profile, {this->CoG_pos_profile}, true, {this->i_stab_profile});

        // Save the plot
        matplot::save(fig, rtIO->getPlotDir() + "/" + this->base_name_output_files + "_trim_setting.svg");
        ax->clear();
        matplot::cla();
    }

    /* Mission speed and Mach profile */
    // Set up matplot++ figure
    fig = matplot::figure(true); // overwrite old figure
    fig->quiet_mode(true);
    fig->size(600, 500);

    // Set up axes and plot data
    matplot::axes_handle ax = fig->current_axes();
    fig->title("Speed characteristics");
    fig->title_font_size_multiplier(1.5);
    set_matplot_axes(ax, {this->mach_profile.data}, "Mach number [-]", false, this->range_profile.data, "Range [NM]");
    set_matplot_axes(ax, {this->CAS_profile.data, this->TAS_profile.data}, "Airspeed [kts]", true);
    add_matplot_to_axes(ax, this->range_profile, {this->mach_profile}, true, {this->CAS_profile, this->TAS_profile});
    ax->y2lim({0, get_min_max_value(this->TAS_profile.data, false)*1.05}); // Reset y2 for better scaling
    // Set up legend
    matplot::legend()->num_rows(3);
    matplot::legend()->location(matplot::legend::general_alignment::bottom);

    // Save the plot
    matplot::save(fig, rtIO->getPlotDir() + "/" + this->base_name_output_files + "_speed.svg");
    ax->clear();
    matplot::cla();

    /* Mission Plotscript Offtakes */
    // Set up matplot++ figure
    fig = matplot::figure(true); // overwrite old figure
    fig->quiet_mode(true);
    fig->size(600, 500);

    // Set up axes and plot data
    ax = fig->current_axes();
    fig->title("Mission offtakes");
    fig->title_font_size_multiplier(1.5);
    set_matplot_axes(ax, {this->shaft_power_profile.data}, "Shaft power [kW]", false, this->range_profile.data, "Range [NM]");
    set_matplot_axes(ax, {this->bleed_air_profile.data}, "Bleed air [kg/s]", true);
    add_matplot_to_axes(ax, this->range_profile, {this->shaft_power_profile}, true, {this->bleed_air_profile});
    // Save the plot
    matplot::save(fig, rtIO->getPlotDir() + "/" + this->base_name_output_files + "_offtake.svg");
    ax->clear();
    matplot::cla();
}

void standard_mission_output::generate_SAR_plots() {
    /* Plot with matplot++ */
    matplot::figure_handle fig = matplot::figure(true);
    fig->quiet_mode(true);
    fig->size(600, 300);
    std::string title("Specific air range (SAR) | Ma = ");
    fig->title_font_size_multiplier(1.5);

    // Define axes
    matplot::axes_handle axes = fig->current_axes();
    axes->font_size(14);
    axes->xlabel("Aircraft mass [metric tons]");
    axes->x_axis().label_font_size(14);
    axes->ylabel("Flight level [FL = 100 ft]");
    axes->y_axis().label_font_size(14);
    axes->cblabel("SAR [m/kg]");
    axes->cb_axis().label_font_size(14);
    matplot::color_array pos = axes->position(); // position({left, bottom, width, height}) [0; 1] fraction of fig height and width
    pos[1] += .075; // Push plot closer to title
    axes->position(pos);
    matplot::colormap(matplot::palette::rdylgn());

    // Loop through plot data
    for (uint16_t SAR_plot_counter(0); SAR_plot_counter < SAR_mach.size(); SAR_plot_counter++) { // One plot for each Mach number
        fig->title(title + num2Str(SAR_mach[SAR_plot_counter]));

        // Get mesh for colorplot
        auto [X, Y] = matplot::meshgrid(SAR_total_mass, SAR_FL);
        matplot::vector_2d Z;

        // initilize min and max value with first value of the plot's data set
        double z_max(specific_air_ranges[specific_air_ranges.size()/SAR_mach.size()*SAR_plot_counter]);
        double z_min(specific_air_ranges[specific_air_ranges.size()/SAR_mach.size()*SAR_plot_counter]);

        // Map specific_air_ranges to a SAR_FL : SAR_total_mass grid
        for (uint16_t i(0); i < SAR_total_mass.size(); i++) { // iterate through x-axis
            matplot::vector_1d Z_tmp;
            for (uint16_t j(0); j < SAR_FL.size(); j++) { // iterate through y-axis
                Z_tmp.push_back( // Add total number of entries for each inner and outer loop
                    specific_air_ranges[specific_air_ranges.size()/SAR_mach.size()*SAR_plot_counter + i*(SAR_FL.size()) + j]);
                // Gather minimum and maximum to scale the colormap
                if (Z_tmp.back() > z_max) {
                    z_max = Z_tmp.back();
                } else if (Z_tmp.back() < z_min) {
                    z_min = Z_tmp.back();
                }
                // filter NaN values (0 = no range possible = edge of plot); must be done AFTER z_min was set!
                if (Z_tmp != Z_tmp) {
                    Z_tmp.back() = 0.;
                }
            }
            Z.push_back(Z_tmp);
        }
        // double step_length(floor(log10(z_min)))
        double lower_cb_level(RoundDown(z_min, -floor(log10(z_min) - 1)));
        double upper_cb_level(RoundUp(z_max, -floor(log10(z_max) - 1)));
        // double cb_steps((upper_cb_level - lower_cb_level)/20);
        matplot::vector_1d levels = matplot::linspace(lower_cb_level, upper_cb_level, 11);
        axes->cblim({lower_cb_level, upper_cb_level});
        axes->cbticks(levels);
        axes->contourf(X, Y, matplot::transpose(Z), levels);
        // Save the plot
        matplot::save(fig, rtIO->getPlotDir() + "/" + this->base_name_output_files + "_SAR_mach_"  + num2Str(SAR_mach[SAR_plot_counter]) + ".svg");
        matplot::cla();
        axes->clear();
    }
}

void standard_mission_output::generate_html_report() { // cppcheck-suppress unusedFunction
    std::string fuelPlanning;
    switch (my_mission_ptr->fuel_planning) {
        case my_mission_ptr->JAR: fuelPlanning = "JAR-OPS 1.255 Subpart D + AMC-OPS 1.255"; break;
        case my_mission_ptr->FAR_DOMESTIC: fuelPlanning = "FAR 121.639 Subpart U"; break;
        case my_mission_ptr->FAR_FLAG: fuelPlanning = "FAR 121.645 Subpart U"; break;
        default: break;
    }

    std::string flightTime(timeFormater(this->my_mission_ptr->mission_profile.back().end_time));
    std::string blockTime(timeFormater(this->my_mission_ptr->mission_profile.back().end_time +
                    this->my_mission_ptr->mission_file->taxi_time_origin + this->my_mission_ptr->mission_file->taxi_time_destination));
    report_.htmlReportStream() << html::div_start("class=\"box\"")
        << html::h2("Mission specifications")
        << html::table_begin({"Parameter", "Comment", "Unit", "Value"})
        << html::table_row("Range", "", "NM",  Rounding(convertUnit(METER, NAUTICALMILE, this->my_mission_ptr->mission_file->Range), 2))
        << html::table_row("Payload (PL)", "Passengers + Cargo", "kg",  this->my_mission_ptr->payload_mass.value())
        << html::table_row("Passengers", "", "-", my_mission_ptr->mission_file->PAX)
        << html::table_row("Cargo mass", "", "kg", Rounding(my_mission_ptr->mission_file->cargo_mass, 0))
        << html::table_row("Cruise Mach number", "", "-",  Rounding(this->my_mission_ptr->mission_file->desired_cruise_speed, 2))
        << html::table_row("Cruise climb steps", "", "ft",  Rounding(this->my_mission_ptr->step_height_cruise, 0))
        << html::table_row("Climb Mach number", "above crossover altitude", "-",
                                this->my_mission_ptr->mission_file->desired_cruise_speed + this->my_mission_ptr->delta_mach_above_crossover)
        << html::table_row("Climb CAS above FL100", "", "KCAS",  Rounding(this->my_mission_ptr->climb_speed_above_FL100, 0))
        << html::table_row("Climb CAS below FL100", "", "KCAS",  Rounding(this->my_mission_ptr->climb_speed_below_FL100, 0))
        << html::table_row("Descent CAS above FL100", "", "KCAS",  Rounding(this->my_mission_ptr->descent_speed_above_FL100, 0))
        << html::table_row("Descent CAS below FL100", "", "KCAS",  Rounding(this->my_mission_ptr->descent_speed_below_FL100, 0))
        << html::table_row("Taxi-out time", "", "min",  Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->mission_file->taxi_time_origin), 0))
        << html::table_row("Taxi-in time", "", "min",  Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->mission_file->taxi_time_destination), 0))
        << html::table_row("Alternate distance", "", "NM",  Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->mission_file->alternate_distance), 0))
        << html::table_row("Fuel reserves method", fuelPlanning, "",  "")
        << html::table_end
        << html::h2("Mission simulation results")
        << html::table_begin({"Parameter", "Comment", "Unit", "Value"})
        << html::table_row("Take-off mass", "TOM = OME + PL + MF - TXF<sub>out</sub>", "kg", Rounding(this->my_mission_ptr->takeoff_mass.value(), 0))
        << html::table_row("Operating mass empty", "OME", "kg", Rounding(this->my_mission_ptr->OME, 0))
        << html::table_row("Mission (loaded) fuel", "MF = TF + RF + EF + TXF<sub>out</sub>", "kg",
                            Rounding(this->my_mission_ptr->get_fuel_sum(this->my_mission_ptr->loaded_fuel_mass), 0))
        << html::table_row("Trip fuel", "TF", "kg", Rounding(my_mission_ptr->get_fuel_sum(this->my_mission_ptr->trip_fuel_mass), 0))
        << html::table_row("Reserve fuel", "RF", "kg", Rounding(this->my_mission_ptr->get_fuel_sum(this->my_mission_ptr->loaded_fuel_mass) -
                                                        my_mission_ptr->get_fuel_sum(this->my_mission_ptr->trip_fuel_mass) -
                                                        this->my_mission_ptr->taxi_fuel_mass_origin.second - this->my_mission_ptr->extra_fuel_mass.second, 0))
        << html::table_row("Extra fuel", "EF", "kg", Rounding(this->my_mission_ptr->extra_fuel_mass.second, 0))
        << html::table_row("Taxi-out fuel", "TXF<sub>out</sub>", "kg", Rounding(this->my_mission_ptr->taxi_fuel_mass_origin.second, 0))
        << html::table_row("Taxi-in fuel", "TXF<sub>in</sub>", "kg", Rounding(this->my_mission_ptr->taxi_fuel_mass_destination.second, 0))
        << html::table_row("Landing mass", "MTOM - TF", "kg", Rounding(this->my_mission_ptr->mission_profile.back().end_total_mass, 0))
        << html::table_row("Block time", "", "h", blockTime)
        << html::table_row("Flight time", "", "h", flightTime)
        << html::table_end;
        if (requirements_checked) {
            report_.htmlReportStream() << html::h2("Calculated performance values for TLARs")
                << html::table_begin({"Parameter", "Comment", "Unit", "Value"})
                << html::table_row("Max. operating altitude", "climb with 100ft/min, ISA + " + num2Str(this->my_mission_ptr->delta_ISA) +
                    " *", "ft", Rounding(convertUnit(METER, FOOT, this->my_mission_ptr->max_operating_altitude), 0))
                << html::table_row("One-engine out (OEI) net ceiling", "after T/O @ MTOM, 100ft/min, ISA + "+ num2Str(this->my_mission_ptr->delta_ISA) +
                    " *", "ft", Rounding(convertUnit(METER, FOOT, this->my_mission_ptr->max_altitude_one_engine_inoperative), 0))
                << html::table_row("Time-to-1500ft", "from break release to 1500 ft **", "min",
                                    Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_start_to_1500ft), 0))
                << html::table_row("Time-to-climb", "from 1500 ft to ICA **", "min",
                                    Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_1500ft_to_ICA), 0))
                << html::table_row("Time-to-ICA", "from break release to ICA **", "min",
                                    Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_start_to_ICA), 0))
                << html::table_row("Range-to-ICA", "from break release to ICA **", "NM", Rounding(convertUnit(METER, NAUTICALMILE, this->my_mission_ptr->range_start_to_ICA), 0))
                << html::table_row("Fuel-to-ICA", "from break release to ICA **", "kg", Rounding(this->my_mission_ptr->fuel_start_to_ICA, 0))
                << html::table_end
                << "<p style=\"text-align: right; margin-right: 3rem;\">* Engine rating: Max. Continuous<br>** Engine rating: Climb</p>" << std::endl;
            /* Cruise steps */
            if (!this->my_mission_ptr->cruise_profile.empty()) {
                report_.htmlReportStream() << html::h2("Estimated optimum cruise profile")
                    << html::table_begin({"Cruise Step", "Flight Level [100 ft]", "Rel. Cruise Step Length [%]"});
                for (uint16_t i(0); i < this->my_mission_ptr->cruise_profile.size(); i++) {
                    report_.htmlReportStream() << html::table_row(i + 1, Rounding(this->my_mission_ptr->cruise_profile.at(i).FL, 1),
                                                                    Rounding(this->my_mission_ptr->cruise_profile.at(i).rel_step_length, 2) * 100.);
                }
                report_.htmlReportStream() << html::table_end;
            }
        }
        report_.htmlReportStream() << html::div_end;
    /* Graphics */
    if (rtIO->plotOn) {
        report_.htmlReportStream() << html::div_start("class=\"box\"")
            << "<hr>" << std::endl
            << html::image("../plots/" + this->base_name_output_files + "_Alt-CL.svg", "alt=\"Broken Path\"")
            << "<hr>" << std::endl
            << html::image("../plots/" + this->base_name_output_files + "_speed.svg", "alt=\"Broken Path\"")
            << "<hr>" << std::endl
            << html::image("../plots/" + this->base_name_output_files + "_offtake.svg", "alt=\"Broken Path\"");
        if (std::string(rtIO->moduleConfig.at("program_settings/mode/mission_methods/center_of_gravity_method/value")) != "mode_0") {
            report_.htmlReportStream() << "<hr>" << std::endl
                << html::image("../plots/" + this->base_name_output_files + "_trim_setting.svg", "alt=\"Broken Path\"");
        }
        if (my_mission_ptr->variable_camber_switch) {
            report_.htmlReportStream() << "<hr>" << std::endl
                << html::image("../plots/" + this->base_name_output_files + "_flap_setting.svg", "alt=\"Broken Path\"");
        }
        if (SAR_plot_on) {
            report_.htmlReportStream() << "<hr>" << std::endl;
            for (uint16_t i(0); i < SAR_mach.size(); i++) {
                if (i != 0) {
                    report_.htmlReportStream() << "<br>" << std::endl;
                }
                report_.htmlReportStream() << html::image("../plots/" + this->base_name_output_files + "_SAR_mach_" + num2Str(SAR_mach[i]) + ".svg", "alt=\"Broken Path\"");
            }
        }
        report_.htmlReportStream() << html::div_end;
    }
}

void standard_mission_output::generate_tex_report_body() { // cppcheck-suppress unusedFunction
    std::string missionString;
    if (this->my_mission_ptr->mission_type == "design_mission") {
        missionString = "Design Mission";
    } else {
        missionString = "Study Mission";
    }

    std::string flightTime(timeFormater(this->my_mission_ptr->mission_profile.back().end_time));
    std::string blockTime(timeFormater(this->my_mission_ptr->mission_profile.back().end_time +
                this->my_mission_ptr->mission_file->taxi_time_origin + this->my_mission_ptr->mission_file->taxi_time_destination));

    /* Start report body */
    report_.texReportStream() << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << std::endl
        << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << std::endl
        << "\\begin{frame}" << std::endl
        << " \\frametitle{Mission Analysis (" << missionString << ")}" << std::endl
        << std::endl
        << " \\begin{table}" << std::endl
        << "     \\centering" << std::endl
        << "     \\scriptsize" << std::endl
        << "     \\begin{tabularx}{250pt}{Xllr}" << std::endl
        << "         \\toprule" << std::endl
        << "         \\textcolor{ilrdark}{Parameter} & \\textcolor{ilrdark}{Comment} & \\textcolor{ilrdark}{Unit} & \\textcolor{ilrdark}{Value} \\\\" << std::endl
        << "         \\midrule[0.5pt]" << std::endl
        << "         \\multicolumn{4}{l}{\\textbf{Mission specifications}}\\\\" << std::endl
        << "         \\midrule[0.5pt]" << std::endl
        << "             Range                 & & $\\nauticalmiles$        & " << std::setprecision(0) << std::fixed
                                                                                    << Rounding(convertUnit(METER, NAUTICALMILE, this->my_mission_ptr->mission_file->Range), 0)
            << " \\\\ " << std::endl
        << "             Payload (PL)          & & $\\kilogram$                & "
            << this->my_mission_ptr->payload_mass.value() << " \\\\ " << std::endl
        << "             Cruise Mach number &             & $-$                & " << std::setprecision(2)
            << std::fixed << this->my_mission_ptr->mission_file->desired_cruise_speed << " \\\\ " << std::endl;
        if (this->my_mission_ptr->auto_climb_switch)
            report_.texReportStream() << "         Cruise climb steps &             & $\\feet$            & "
                << std::setprecision(0) << std::fixed << this->my_mission_ptr->step_height_cruise << " \\\\ " << std::endl
        << "             Climb Mach number & \\tiny{above crossover altitude} & $-$ & "  << std::setprecision(2) << std::fixed
            << this->my_mission_ptr->mission_file->desired_cruise_speed + my_mission_ptr->delta_mach_above_crossover << " \\\\ " << std::endl// $M_{crossover}$
        << "             Climb CAS above FL100    &       & $\\KCAS$           & " << std::setprecision(0) << std::fixed
            << this->my_mission_ptr->climb_speed_above_FL100 << " \\\\ " << std::endl // $CAS_{design,Climb}$
        << "             Climb CAS below FL100    &       & $\\KCAS$           & "
            << this->my_mission_ptr->climb_speed_below_FL100 << " \\\\ " << std::endl // $CAS_{ATCLimit,Climb}$
        << "             Descent CAS above FL100  &       & $\\KCAS$           & "
            << this->my_mission_ptr->descent_speed_above_FL100 << " \\\\ " << std::endl // $CAS_{design,Descent}$
        << "             Descent CAS below FL100  &       & $\\KCAS$           & "
            << this->my_mission_ptr->descent_speed_below_FL100 << " \\\\ " << std::endl // $CAS_{ATCLimit,Descent}$
        << "             Taxi-out time            &       & $\\min$            & " << this->my_mission_ptr->mission_file->taxi_time_origin << " \\\\ " << std::endl
        << "             Taxi-in time             &       & $\\min$            & " << this->my_mission_ptr->mission_file->taxi_time_destination << " \\\\ " << std::endl
        << "             Alternate distance       &       & $\\nauticalmiles$  & "
            << convertUnit(METER, NAUTICALMILE, this->my_mission_ptr->mission_file->alternate_distance) << " \\\\ " << std::endl;
    std::string fuelPlanning;
    switch (my_mission_ptr->fuel_planning) {
        case my_mission_ptr->JAR: fuelPlanning = "JAR-OPS 1.255 Subpart D + AMC-OPS 1.255"; break;
        case my_mission_ptr->FAR_DOMESTIC: fuelPlanning = "FAR 121.639 Subpart U"; break;
        case my_mission_ptr->FAR_FLAG: fuelPlanning = "FAR 121.645 Subpart U"; break;
        default: break;
    }
    report_.texReportStream() << "             Fuel reserves method &      &  \\multicolumn{2}{c}{\\textit{" << fuelPlanning << "}} \\\\" << std::endl
        << "         \\midrule[0.5pt]" << std::endl
        << "         \\multicolumn{4}{l}{\\textbf{Mission simulation results}}\\\\" << std::endl
        << "         \\midrule[0.5pt]" << std::endl
        << "             Take-off mass & \\tiny{$TOM = OME + PL + MF - TXF_{out}$} & $\\kilogram$ & " << Rounding(my_mission_ptr->takeoff_mass.value(), 0) << " \\\\ " << std::endl
        << "             Mission (loaded) fuel  &\\tiny{$MF = TF + RF + EF + TXF_{out}$}  & $\\kilogram$   & "
            << Rounding(this->my_mission_ptr->get_fuel_sum(this->my_mission_ptr->loaded_fuel_mass), 0) << " \\\\ " << std::endl
        << "             Trip fuel      & \\tiny{$TF$}        & $\\kilogram$   & "
            << Rounding(my_mission_ptr->get_fuel_sum(this->my_mission_ptr->trip_fuel_mass), 0) << " \\\\ " << std::endl
        << "             Reserve fuel   & \\tiny{$RF$}        & $\\kilogram$   & "
            << Rounding(this->my_mission_ptr->get_fuel_sum(this->my_mission_ptr->loaded_fuel_mass) - my_mission_ptr->get_fuel_sum(my_mission_ptr->trip_fuel_mass) -
                        my_mission_ptr->taxi_fuel_mass_origin.second - my_mission_ptr->extra_fuel_mass.second, 0) << " \\\\ " << std::endl
        << "             Extra fuel     & \\tiny{$EF$}        & $\\kilogram$   & " << Rounding(my_mission_ptr->extra_fuel_mass.second, 0) << " \\\\ " << std::endl
        << "             Taxi-out fuel  & \\tiny{$TXF_{out}$} & $\\kilogram$   & " << Rounding(my_mission_ptr->taxi_fuel_mass_origin.second, 0) << " \\\\ " << std::endl
        << "             Taxi-in fuel   & \\tiny{$TXF_{in}$}  & $\\kilogram$   & " << Rounding(my_mission_ptr->taxi_fuel_mass_destination.second, 0) << " \\\\ " << std::endl
        << "             Block time             && $\\hour$              & " << blockTime << " \\\\ " << std::endl
        << "             Flight time              && $\\hour$                  & " << flightTime << " \\\\ " << std::endl
        << "        \\bottomrule" << std::endl
        << "    \\end{tabularx}" << std::endl
        << " \\end{table}" << std::endl
        << std::endl
        << "\\end{frame}" << std::endl
        << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << std::endl
        << "\\begin{frame}" << std::endl
        << " \\frametitle{" << missionString << " Simulation Results (1)}" << std::endl
        << std::endl
        << "\\hfill" << std::endl
        << std::endl;
    if (rtIO->plotOn) {
        report_.texReportStream() << " \\begin{figure}" << std::endl
        << "     \\tiny" << std::endl
        << "     \\def\\svgwidth{7cm}" << std::endl
        << "     \\import{" + win2lin(relativePath(rtIO->getPlotDir(), rtIO->getReportDir())) + "}{"+
            std::string(TOOL_NAME) + "_" +  my_mission_ptr->mission_type + "_Alt-CL.pdf_tex}" << std::endl
        << "     \\caption{\\scriptsize{Mission profiles: Altitude, lift coeff., thrust, fuel}}" << std::endl
        << " \\end{figure}" << std::endl
        << std::endl;
    }
    report_.texReportStream() << "\\end{frame}" << std::endl
    << std::endl
    << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << std::endl
    << "\\begin{frame}" << std::endl
    << " \\frametitle{" << missionString << " Simulation Results (2)}" << std::endl
    << std::endl
    << "\\hfill" << std::endl
    << std::endl;
    if (rtIO->plotOn) {
        report_.texReportStream() << " \\begin{figure}[htb]" << std::endl
        << " \\begin{center}" << std::endl
        << " \\subfigure{" << std::endl // [\\scriptsize{Mach number and airspeeds}]
        << "     \\tiny" << std::endl
        << "     \\def\\svgwidth{7cm}" << std::endl
        << "     \\import{" + win2lin(relativePath(rtIO->getPlotDir(), rtIO->getReportDir())) + "}{"+
            std::string(TOOL_NAME) + "_" +  my_mission_ptr->mission_type + "_speed.pdf_tex}" << std::endl
        << "  }" << std::endl
        << " \\hfill" << std::endl
        << " \\subfigure{" << std::endl
        << "     \\tiny" << std::endl
        << "     \\def\\svgwidth{7cm}" << std::endl
        << "     \\import{" + win2lin(relativePath(rtIO->getPlotDir(), rtIO->getReportDir())) + "}{"+
            std::string(TOOL_NAME) + "_" +  my_mission_ptr->mission_type + "_offtake.pdf_tex}" << std::endl
        << "  }" << std::endl
        << " \\end{center}" << std::endl
        << " \\caption{\\scriptsize{Mission profiles: Speeds and offtakes}}" << std::endl
        << " \\end{figure}" << std::endl
        << std::endl;
    }
    report_.texReportStream() << "\\end{frame}" << std::endl
    << std::endl
    << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << std::endl;
    if (this->requirements_checked) {
        report_.texReportStream() << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << std::endl
        << "%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%" << std::endl
        << "\\begin{frame}" << std::endl
        << " \\frametitle{Climb Performance}" << std::endl
        << std::endl
        << " \\begin{table}" << std::endl
        << "     \\centering" << std::endl
        << "     \\scriptsize" << std::endl
        << "     \\begin{tabularx}{250pt}{Xllr}" << std::endl
        << "         \\toprule" << std::endl
        << "             \\textcolor{ilrdark}{Parameter} & \\textcolor{ilrdark}{Comment} & \\textcolor{ilrdark}{Unit} & \\textcolor{ilrdark}{Value} \\\\" << std::endl
        << "         \\midrule[0.5pt]" << std::endl
        << "         \\multicolumn{4}{l}{\\textbf{Calculated performance values for TLARs}}\\\\" << std::endl
        << "         \\midrule[0.5pt]" << std::endl
        << " AEO net ceiling     & \\tiny{climb with 100 ft/min, ISA + " << this->my_mission_ptr->delta_ISA << ", after T/O @ MTOM}  & $\\feet$ & "
               << std::setprecision(0) << std::fixed << Rounding(convertUnit(METER, FOOT, this->my_mission_ptr->max_operating_altitude), 0) << " \\\\ " << std::endl
        << " AEO net ceiling     & \\tiny{climb with 50 ft/min, ISA + " << this->my_mission_ptr->delta_ISA << ", after T/O @ MTOM}   & $\\feet$ & "
               << std::setprecision(0) << std::fixed << Rounding(convertUnit(METER, FOOT, this->my_mission_ptr->max_ceiling_altitude), 0) << " \\\\ " << std::endl
        << " OEI net ceiling     & \\tiny{climb with 100 ft/min, ISA + " << this->my_mission_ptr->delta_ISA << ", after T/O @ MTOM}  & $\\feet$ & "
               << std::setprecision(0) << std::fixed << Rounding(convertUnit(METER, FOOT, this->my_mission_ptr->max_altitude_one_engine_inoperative), 0) << " \\\\ " << std::endl
        << " Time-to-climb       & \\tiny{from 1500ft to ICA, after T/O @ MTOM, ISA + " << this->my_mission_ptr->delta_ISA << "}  & $\\min$  & "
               << std::setprecision(1) << std::fixed << Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_1500ft_to_ICA), 2) << " \\\\ " << std::endl
        << "         \\midrule[0.5pt]" << std::endl
        << "         \\multicolumn{4}{l}{\\textbf{Climb to initial cruise altitude (after T/O @ MTOM, ISA + " << this->my_mission_ptr->delta_ISA << ")}}\\\\" << std::endl
        << "         \\midrule[0.5pt]" << std::endl
        << " Time-to-climb  & \\tiny{from 1500ft to ICA} & $\\min$           & "
               << std::setprecision(1) << std::fixed << Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_1500ft_to_ICA), 2) << " \\\\ " << std::endl
        << " Time-to-1500ft & \\tiny{from break release to 1500ft}  & $\\min$           & "
               << std::setprecision(1) << std::fixed << Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_start_to_1500ft), 2) << " \\\\ " << std::endl
        << " Time-to-ICA    & \\tiny{from break release to ICA}     & $\\min$           & "
               << std::setprecision(1) << std::fixed << Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_start_to_ICA), 2) << " \\\\ " << std::endl
        << " Range-to-ICA   & \\tiny{from break release to ICA}     & $\\nauticalmiles$ & " << std::setprecision(0)
            << std::fixed << Rounding(this->my_mission_ptr->range_start_to_ICA, 2) << " \\\\ " << std::endl
        << " Fuel-to-ICA    & \\tiny{from break release to ICA}     & $\\kilogram$      & " << std::setprecision(0)
            << std::fixed << Rounding(this->my_mission_ptr->fuel_start_to_ICA, 2) << " \\\\ " << std::endl
        << "        \\bottomrule" << std::endl
        << "    \\end{tabularx}" << std::endl
        << " \\end{table}" << std::endl
        << std::endl
        << "\\end{frame}" << std::endl;
    }
}

void standard_mission_output::generate_requirements_html_report() {
    report_.htmlReportStream() << html::div_start("class=\"box\"");
    /* Fill HTML */
        if (requirements_checked) { // If-catch whether requirements were really checked
        /* Reference top of descent from design mission */
        std::string subpath("analysis/mission/design_mission/cruise/cruise_steps");
        double design_mission_top_of_descent_altitude =
            rtIO->acxml.at(subpath +"/cruise_step@" + num2Str(rtIO->acxml.at(subpath).getVector("cruise_step").size() - 1) + "/altitude/value");
        /* Read in requirements for altitudes, time to climb and Mach number */
        subpath = "requirements_and_specifications/requirements/top_level_aircraft_requirements";
        double max_operating_altitude_required = rtIO->acxml.at(subpath + "/flight_envelope/maximum_operating_altitude/value");
        double max_altitude_one_engine_inoperative_required = rtIO->acxml.at(subpath + "/flight_envelope/maximum_altitude_one_engine_inoperative/value");
        double time_to_climb_requirement = rtIO->acxml.at(subpath + "/design_mission/time_to_climb/value");
        bool initial_mach_check(this->my_mission_ptr->initial_cruise_mach + ACCURACY_LOW >= // Added ACCURACY_LOW to catch floating point errors when the values are actually equal
                                            this->my_mission_ptr->initial_cruise_mach_required);
        bool max_altitude_check(this->my_mission_ptr->max_operating_altitude >= this->my_mission_ptr->max_operating_altitude_required &&
                                                                this->my_mission_ptr->max_operating_altitude_required >= design_mission_top_of_descent_altitude);
        bool max_altitude_oei_check(max_altitude_one_engine_inoperative_required <= this->my_mission_ptr->max_altitude_one_engine_inoperative);
        bool climb_time_check(this->my_mission_ptr->initial_cruise_altitude_check && (time_to_climb_requirement >= this->my_mission_ptr->time_1500ft_to_ICA));
        std::string success("<span style=\"color:green;\">&#10004;</span>");
        std::string fail("<span style=\"color:red;\">&#10006;</span>");
            report_.htmlReportStream() << html::h2("Calculated performance values for TLARs")
                << html::table_begin({"Parameter", "Comment", "Unit", "Value", "Required value", "Check"})
                << html::table_row("Initial cruise altitude (ICA)", "ISA + " + num2Str(this->my_mission_ptr->delta_ISA),
                    "ft", Rounding(convertUnit(METER, FOOT, this->my_mission_ptr->initial_cruise_altitude),
                        -1), // precision -1 to catch floating point errors, since both values should be equal
                    Rounding(convertUnit(METER, FOOT, this->my_mission_ptr->initial_cruise_altitude_required), 0),
                    (this->my_mission_ptr->initial_cruise_altitude_check ? success : fail))
                << html::table_row("Initial cruise Mach number", "ISA + " + num2Str(this->my_mission_ptr->delta_ISA),
                    "-", static_cast<double>(Rounding(this->my_mission_ptr->initial_cruise_mach, 2)),
                    static_cast<double>(Rounding(this->my_mission_ptr->initial_cruise_mach_required, 2)), (initial_mach_check ? success : fail))
                << html::table_row("Max. operating altitude", "climb with 100ft/min, ISA + " + num2Str(this->my_mission_ptr->delta_ISA) +
                    " *", "ft", Rounding(convertUnit(METER, FOOT, this->my_mission_ptr->max_operating_altitude), 0),
                    Rounding(convertUnit(METER, FOOT, max_operating_altitude_required), 0), (max_altitude_check ? success : fail))
                << html::table_row("One-engine out (OEI) net ceiling", "after T/O @ MTOM, 100ft/min, ISA + "+ num2Str(this->my_mission_ptr->delta_ISA) +
                    " *", "ft", Rounding(convertUnit(METER, FOOT, this->my_mission_ptr->max_altitude_one_engine_inoperative), 0),
                    Rounding(convertUnit(METER, FOOT, max_altitude_one_engine_inoperative_required), 0), (max_altitude_oei_check ? success : fail))
                << html::table_row("Time-to-1500ft", "from break release to 1500 ft **", "min",
                                    Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_start_to_1500ft), 0), "Not specified", "-")
                << html::table_row("Time-to-climb", "from 1500 ft to ICA **", "min",
                                    Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_1500ft_to_ICA), 0),
                                    Rounding(convertUnit(SECOND, MINUTE, time_to_climb_requirement), 0),
                                    (climb_time_check ? success : (fail + (this->my_mission_ptr->initial_cruise_altitude_check ? "" : " (ICA not reached)"))))
                << html::table_row("Time-to-ICA", "from break release to ICA **", "min",
                                    Rounding(convertUnit(SECOND, MINUTE, this->my_mission_ptr->time_start_to_ICA), 0), "Not specified", "-")
                << html::table_row("Range-to-ICA", "from break release to ICA **", "NM",
                        Rounding(convertUnit(METER, NAUTICALMILE, this->my_mission_ptr->range_start_to_ICA), 0), "Not specified", "-")
                << html::table_row("Fuel-to-ICA", "from break release to ICA **", "kg",
                                    Rounding(this->my_mission_ptr->fuel_start_to_ICA, 0), "Not specified", "-")
                << html::table_end
                << "<p style=\"text-align: right; margin-right: 3rem;\">* Engine rating: Max. Continuous<br>** Engine rating: Climb</p>" << std::endl;
        }
    report_.htmlReportStream() << html::div_end;
}

std::string standard_mission_output::timeFormater(double time) {
    double double_hours = convertUnit(SECOND, HOUR, time);
    int int_hours = double_hours;
    int minutesRemainder = (double_hours - int_hours) * 60;
    return (num2Str(int_hours) + ":" + num2Str(minutesRemainder));
}

void standard_mission_output::set_matplot_axes(std::shared_ptr<matplot::axes_type> axes, const std::vector<std::vector<double>>& y_data,
                                                const std::string& y_label, bool y2_switch, const std::vector<double>& x_data, const std::string& x_label) {
    try {
        /* Get y axis limits */
        double min_limit = get_min_max_value(y_data[0], true);
        double max_limit = get_min_max_value(y_data[0], false);
        if (y_data.size() > 1) { // If multiple y data sets must be checked, do so here
        for (uint16_t i(1); i < y_data.size(); i++) {
                double max_limit_tmp = get_min_max_value(y_data[i], false);
                if (max_limit < max_limit_tmp) {
                    max_limit = max_limit_tmp;
                }
                double min_limit_tmp = get_min_max_value(y_data[i], true);
                if (min_limit > min_limit_tmp) {
                    min_limit = min_limit_tmp;
                }
            }
        }
        if (!y2_switch) {// Left y axis
            matplot::grid(axes, matplot::on); // Turn on grid when left axis is set up
            // Expand x limits for better looks
            double min_x_limit = get_min_max_value(x_data, true) - (get_min_max_value(x_data, false) - get_min_max_value(x_data, true))*0.025;
            double max_x_limit = get_min_max_value(x_data, false) + (get_min_max_value(x_data, false) - get_min_max_value(x_data, true))*0.025;
            axes->xlim({min_x_limit, max_x_limit});
            if (!x_label.empty()) { // x only initialized on left y axis
                axes->xlabel(x_label);
                axes->x_axis().label_font_size(14);
            } else {
                axes->xtickformat({}); // Empty format to delete x lables while keeping its ticks
            }
            axes->ylabel(y_label);
            axes->y_axis().label_font_size(14);
            axes->ylim({min_limit, max_limit*1.25}); // Increase upper limit for legend spacing
        } else { // Right y axis
            axes->y2label(y_label);
            axes->y2_axis().label_font_size(14);
            axes->y2lim({min_limit, max_limit*1.25}); // Increase upper limit for legend spacing
        }
    } catch (const std::string& error_message) {
        throwError(__FILE__, __func__, __LINE__, error_message);
    }
}

void standard_mission_output::add_matplot_to_axes(std::shared_ptr<matplot::axes_type> axes, matplot_axis_data x_data,
                                        std::vector<matplot_axis_data> y1_data, bool legend_switch,
                                        std::vector<matplot_axis_data> y2_data) {
    try {
        axes->font_size(14);
        axes->colororder(colors::unicado_to_matplot_colororder());
        std::vector<std::string_view> line_styles({"-", "--", "-.", "-", ":", "-o", "-x"});
        uint16_t plot_counter(0); // Count plots per y axis to vary line style
        std::vector<std::string> legend_entries;
        while (plot_counter < y1_data.size()) {
            axes->plot(x_data.data, y1_data[plot_counter].data, line_styles[plot_counter])->line_width(1.5).color(axes->colororder()[0]);
            if (plot_counter == 0) {
                axes->hold(true);
            }
            if (legend_switch) {
                legend_entries.push_back(y1_data[plot_counter].name);
            }
            plot_counter++;
        }
        plot_counter = 0;
        for (uint16_t i(0); i < y2_data.size(); i++) {
            axes->plot(x_data.data, y2_data[i].data, line_styles[plot_counter])->line_width(1.5).use_y2(true).color(axes->colororder()[1]);
            if (legend_switch) {
                legend_entries.push_back(y2_data[i].name);
            }
            plot_counter++;
        }

        /* Increase starting height of plot by 5 % to cope for increase title height + decrease offset to it */
        matplot::color_array pos = axes->position(); // position({left, bottom, width, height}) [0; 1] fraction of fig height and width
        pos[1] += .1;
        axes->position(pos);

        /* Set up legend */
        uint16_t legend_size = legend_entries.size();
        if (legend_size > 1 && legend_size <= 2) { // legend for 1 or 0 entries not needed
            matplot::legend(axes, legend_entries);
        } else if (legend_size == 3) {
            matplot::legend(axes, legend_entries)->num_rows(1);
        } else {
            matplot::legend(axes, legend_entries)->num_rows(2);
        }
        axes->hold(false);
    } catch (const std::string& error_message) {
        throwError(__FILE__, __func__, __LINE__, error_message);
    }
}

double standard_mission_output::get_min_max_value(const std::vector<double>& data, bool min_max_switch) {
    std::vector<double>::const_iterator vec_iterator;
    if (min_max_switch) {
        vec_iterator = std::min_element(data.begin(), data.end());
    } else {
        vec_iterator = std::max_element(data.begin(), data.end());
    }
    if (vec_iterator != data.end()) {
        return *vec_iterator;
    } else {
        myRuntimeInfo->err << "Dataset broken or empty. Abort!" << std::endl;
        exit(1);
    }
}
