/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef performance_assessment_SRC_TAW_DEFAULTMETHODS_FUELPLANNING_LOW_FIDELITY_FUELFROMMASSFRACTION_H_
#define performance_assessment_SRC_TAW_DEFAULTMETHODS_FUELPLANNING_LOW_FIDELITY_FUELFROMMASSFRACTION_H_

#include <runtimeInfo/runtimeInfo.h>

#include <cmath>
#include <vector>


/**\class fuelFromMassFraction Class contains the mass fractions necessary for the flight
*
*/
class fuelFromMassFraction {
 public:
    double theBreguetFactor;  /**< Breguet factor (1/m) */  // Breguet Factors = g*SFC / V*L/D

    double breguetEfficiencyBlockFuel; /**< Relation between target and actual block fuel (-) */
    double breguetEfficiencyTripFuel; /**< Relation between target and actual trip fuel (-) */

    double missionFuel; /**< Required fuel for the block time (kg) */
    double tripFuel; /**< Required fuel from brake release till landing at the destination (kg) */


    /**\brief Calibration of Breguet fuel estimation
    *\param targetMissionFuel Supposed aircraft mission fuel needed (kg)
    *\param targetTripFuel Supposed aircraft trip fuel needed (kg)
    *\param designPayload Aircraft design payload (kg)
    *\param designRange Aircraft design range (NM)
    */
    void calibrateBreguet(double targetMissionFuel, double targetTripFuel, double designPayload, double designRange);

    /**\brief Computes iteratively the range after Breguet
    *\param payloadForRange Aircraft payload (kg)
    *\param takeOffMass Aircraft take-off mass (kg)
    *\param theMTOM Maximum aircraft take off mass (kg)
    *\param theDesignRange Aircraft design range (NM)
    *\return range of the aircraft after Breguet
    */
    double iterativeBreguetForRange(double payloadForRange, double takeOffMass, double theMTOM, double theDesignRange);

    /**\brief Function calculates the block and the trip fuel for the given payload and range
    *\param missionPayload Payload during the mission (kg)
    *\param missionRange Aircraft range of the mission (NM)
    */
    void calcFuelFromMassFractions(double missionPayload, double missionRange);

    /**\brief Constructor fuelFromMassFraction
    *\param thisOME Operating empty mass (kg)
    *\param cruiseMach Aircraft cruise Mach number (-)
    */
    fuelFromMassFraction(double thisOME, double cruiseMach);

    /**\brief Destructor fuelFromMassFraction
    *
    */
    virtual ~fuelFromMassFraction();

 private:
    double theOME; /**< Operating empty mass (kg) */
    double theCruiseMach; /**< Cruise Mach number (-) */
};

#endif  // performance_assessment_SRC_TAW_DEFAULTMETHODS_FUELPLANNING_LOW_FIDELITY_FUELFROMMASSFRACTION_H_
