/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "fuelPlanning.h"


namespace low {

fuelPlanning::fuelPlanning(std::shared_ptr<aircraft> anAcftPtr, const std::shared_ptr<performance_assessment_settings> aSettingsPt, const std::shared_ptr<RuntimeIO>& rtIO)
    :
    rtIO(rtIO),
    theAcftPt(anAcftPtr),
    theSettingsPt(aSettingsPt),
    myFlightSim(anAcftPtr),
    myBreguet(theAcftPt->OME, theAcftPt->MInitialCruise),
    relConFuel(theSettingsPt->myFuelPlan.contingencyFuel),
    P(pointData()) {
    tripTime = 0.;
    blockTime = 0.;
    myFuel.taxiFuelTakeoff     = 0.;
    myFuel.taxiFuelLanding     = 0.;
    myFuel.missionFuel           = 0.;  // flight fuel: fuel consumed without taxiing! -> better: mission fuel?
    myFuel.contigencyFuel      = 0.;
    myFuel.alternateFuel       = 0.;
    myFuel.finalReserveFuel    = 0.;
    myFuel.tripFuel            = 0.;
    // ctor
}

void fuelPlanning::doFuelPlanning() {
    this->calcStandardFuelPlanning(&myFuel.taxiFuelTakeoff, &myFuel.taxiFuelLanding, &myFuel.missionFuel, &myFuel.contigencyFuel, &myFuel.alternateFuel, &myFuel.finalReserveFuel,
                                   &myFuel.tripFuel, theAcftPt->myMission.missionRange, theAcftPt->myMission.missionPayload, true, theAcftPt->aero);
    if (theSettingsPt->designForMTOM == true) {
        theAcftPt->MTOM = Rounding(theAcftPt->OME + theAcftPt->myMission.missionPayload + myFuel.missionFuel - myFuel.taxiFuelTakeoff, 2);
    } else {
// TODO(franz#1#): ueberpruefen ob so ok
        if (theSettingsPt->useStudyMissionForAnalysis) {
            if (theAcftPt->MTOM < theAcftPt->OME + theAcftPt->myMission.missionPayload + myFuel.tripFuel - myFuel.taxiFuelTakeoff) {
                myRuntimeInfo->err << "The mission cannot be flown.!" << std::endl;
                myRuntimeInfo->err << "MTOM exceedance!" << std::endl;
                myRuntimeInfo->err << "MTOM: " << theAcftPt->MTOM << " kg" << std::endl;
                myRuntimeInfo->err << "TOM:  " << theAcftPt->OME + theAcftPt->myMission.missionPayload + myFuel.tripFuel - myFuel.taxiFuelTakeoff << " kg" << std::endl;
                exit(1);
            }
        }
    }
    if (theAcftPt->MFM < myFuel.tripFuel) {
        myRuntimeInfo->err << "The mission cannot be flown.!" << std::endl;
        myRuntimeInfo->err << "MFM exceedance!" << std::endl;
        myRuntimeInfo->err << "MFM: " << theAcftPt->MFM << " kg" << std::endl;
        myRuntimeInfo->err << "Trip Fuel:  " << myFuel.tripFuel << " kg" << std::endl;
        exit(1);
    }
    /* Determining the flight time */
    for (unsigned int i = 0; i <= theAcftPt->myMission.Steps.size() - 1; i++) {
        tripTime += convertUnit(SECOND, MINUTE, theMissionFuel.at(i + 1).Time);
    }
    blockTime = tripTime;
    tripTime += theAcftPt->myMission.taxiTimeOrigin + theAcftPt->myMission.taxiTimeDestination;
}

void fuelPlanning::doPayloadRangeOLD() {
    /* Determining the Breguet Factors = g*SFC / V*L/D */
    //double breguetFactor(0.);
    double cruiseRange(0.);
    double tempBreguetFactor(0.);
    // where is theMissionFuel.at(i+1).consumedFuel (see below) set/calculated?
    for (unsigned int i(0); i <= theAcftPt->myMission.Steps.size() - 1; i++) {
        if (theAcftPt->myMission.Steps.at(i).modeStep == "cruise") {
            if ((1. / (1. - theMissionFuel.at(i + 1).consumedFuel / theMissionFuel.at(i).Mass)) < 10e-8) {
                myRuntimeInfo->err << "Abort: Log-Argument becomes too small!!!" << std::endl;
                myRuntimeInfo->err << "1 / (1-f_consumed/f_mission) = " << (1. / (1. - theMissionFuel.at(i + 1).consumedFuel / theMissionFuel.at(i).Mass)) << std::endl;
                exit(1);
            } else {
                tempBreguetFactor += theMissionFuel.at(i + 1).Range * (1. / theMissionFuel.at(i + 1).Range)
                                     * log(1. / (1. - theMissionFuel.at(i + 1).consumedFuel / theMissionFuel.at(i).Mass));
                cruiseRange += theMissionFuel.at(i + 1).Range;
            }
        }
    }
    myBreguet.theBreguetFactor = tempBreguetFactor / cruiseRange;
    myBreguet.calibrateBreguet(myFuel.missionFuel, myFuel.tripFuel, theAcftPt->designPayload, theAcftPt->designRange);
    /* Determination of the 0.Segments */
    //1. Point
    myPayloadRange.push_back(payloadRangeData());
    myPayloadRange.push_back(payloadRangeData());
    myPayloadRange.back().Payload   = theAcftPt->maxPayload;
    myPayloadRange.back().Fuel      = theAcftPt->MTOM - theAcftPt->OME - myPayloadRange.back().Payload;
    myPayloadRange.back().Range     = myBreguet.iterativeBreguetForRange(theAcftPt->maxPayload, theAcftPt->MTOM, theAcftPt->MTOM, theAcftPt->designRange);
    myPayloadRange.back().TOM       = theAcftPt->MTOM;
    // 2 Point ->Linear equation for Fuel: f(r) = m*r+a
    vector<double> theRange;
    vector<double> theFuel;
    for (double i = 0.5; i <= 1.0; i += 0.05) {
        theRange.push_back(i * myPayloadRange.back().Range);
        myBreguet.calcFuelFromMassFractions(theAcftPt->maxPayload, theRange.back());
        theFuel.push_back(myBreguet.tripFuel);
    }
    vector<double> dataCoeff;
    dataCoeff = calcRegressionCoefficientsUsingQRdecomp(theRange, theFuel, 1);
    theFuel.clear();
    theRange.clear();
    // 0. Point
    // myPayloadRange.insert( myPayloadRange.begin(), payloadRangeData() ); //see line 131!
    myPayloadRange.at(0).Payload   = theAcftPt->maxPayload;
    myPayloadRange.at(0).Range     = 0.0;
    myPayloadRange.at(0).Fuel      = dataCoeff.at(0);
    myPayloadRange.at(0).TOM       = theAcftPt->OME + theAcftPt->maxPayload + myPayloadRange.at(0).Fuel;
    dataCoeff.clear();
    /* Determination of the 1. Segments */
    // 2. Point
    myPayloadRange.push_back(payloadRangeData());
    myPayloadRange.back().TOM       = theAcftPt->MTOM;
    if (theAcftPt->OME + theAcftPt->MFM < theAcftPt->MTOM) {
        myPayloadRange.back().Payload   = theAcftPt->MTOM - theAcftPt->OME - theAcftPt->MFM;
        if (myPayloadRange.back().Payload < 0.) {
            myRuntimeInfo->err << "Payload cannot be negative!" << std::endl;
            myRuntimeInfo->err << "Abort!!!" << std::endl;
            exit(1);
        }
        myPayloadRange.back().Fuel = theAcftPt->MFM;
    } else {
        myPayloadRange.back().Payload = 0.;
        myPayloadRange.back().Fuel = theAcftPt->MTOM - theAcftPt->OME;
        if (myPayloadRange.back().Fuel > theAcftPt->MFM) {
            myRuntimeInfo->err << "Fuel cannot be larger than MFM!" << std::endl;
            myRuntimeInfo->err << "Abort!!!" << std::endl;
            exit(1);
        }
    }
    myPayloadRange.back().Range     = myBreguet.iterativeBreguetForRange(myPayloadRange.back().Payload, theAcftPt->MTOM, theAcftPt->MTOM, theAcftPt->designRange);
    /* 3. Segment: no payload and maximum refueled */
    if (theAcftPt->OME + theAcftPt->MFM < theAcftPt->MTOM) {
        myPayloadRange.push_back(payloadRangeData());
        myPayloadRange.back().TOM       = theAcftPt->OME + theAcftPt->MFM;
        myPayloadRange.back().Payload   = 0.;
        myPayloadRange.back().Fuel      = theAcftPt->MFM;
        myPayloadRange.back().Range     = myBreguet.iterativeBreguetForRange(0., myPayloadRange.back().TOM, theAcftPt->MTOM,  theAcftPt->designRange);
    } else {
        myPayloadRange.push_back(payloadRangeData());
        int ID = myPayloadRange.size() - 2;
        myPayloadRange.back().TOM       = myPayloadRange.at(ID).TOM;
        myPayloadRange.back().Payload   = myPayloadRange.at(ID).Payload;
        myPayloadRange.back().Fuel      = myPayloadRange.at(ID).Fuel;
        myPayloadRange.back().Range     = myPayloadRange.at(ID).Range;
    }
}

void fuelPlanning::doFuelPayloadRangeOLD() {
    if (!fileExists(rtIO->getCsvFilesDirTool())) {
        myRuntimeInfo->out << "Create csvOutput folder in program folder: " << rtIO->getCsvFilesDirTool() << std::endl;
        CreateDirectory(rtIO->getCsvFilesDirTool().c_str(), NULL);  // ...Create folder
    }
    std::ofstream plot;
    string plotFile(rtIO->getCsvFilesDirTool() + "performance_assessment" + "_fuelPayloadRange.csv");  // Replaced global variable "TOOLNAME" (?)
    plot.open(plotFile.c_str());
    plot << "# Range [NM]; Payload [kg]; Consumed Fuel [kg]; Time [h]" << std::endl;

    /* 1. Range at maxPayload */
    for (int i(0); i <= 25; i++) {
        for (unsigned int j = 0; j <= 50; j++) {
            double range = static_cast<double>(i) / 25. * myPayloadRange.at(1).Range;
            double payload = static_cast<double>(j) / 50. * theAcftPt->maxPayload;
            myBreguet.calcFuelFromMassFractions(payload, range);
            double fuel = myBreguet.missionFuel;
            double time = 1.10 * range / (theAcftPt->MInitialCruise * theAcftPt->atm.getSpeedOfSoundISA(35000. * 0.3048) * 3.6);
            plot << range / 1.852 << "; ";
            plot << payload << "; ";
            plot << fuel << "; ";
            plot << time << std::endl;
        }
        plot << std::endl;
    }
    /* 2. Range at maxPayload */
    for (unsigned int i(1); i <= 25; i++) {
        for (unsigned int j(0); j <= 50; j++) {
            double range = myPayloadRange.at(1).Range + static_cast<double>(i) / 25. * (myPayloadRange.at(2).Range - myPayloadRange.at(1).Range);
            double m = (myPayloadRange.at(1).Payload - myPayloadRange.at(2).Payload) / (myPayloadRange.at(1).Range - myPayloadRange.at(2).Range);
            double a = myPayloadRange.at(1).Payload - m * myPayloadRange.at(1).Range;
            double payload = static_cast<double>(j) / 50. * (m * range + a);
            myBreguet.calcFuelFromMassFractions(payload, range);
            double fuel = myBreguet.missionFuel;
            double time = 1.10 * range / (theAcftPt->MInitialCruise * theAcftPt->atm.getSpeedOfSoundISA(35000. * 0.3048) * 3.6);
            plot << range / 1.852 << "; ";
            plot << payload << "; ";
            plot << fuel << "; ";
            plot << time << std::endl;
        }
        plot << std::endl;
    }
    /* 3. Range at maxPayload */
    for (int i = 1; i <= 24; i++) {
        for (int j(0); j <= 50; j++) {
            double range = myPayloadRange.at(2).Range + static_cast<double>(i) / 25. * (myPayloadRange.at(3).Range - myPayloadRange.at(2).Range);
            double m = (myPayloadRange.at(2).Payload - myPayloadRange.at(3).Payload) / (myPayloadRange.at(2).Range - myPayloadRange.at(3).Range);
            double a = myPayloadRange.at(2).Payload - m * myPayloadRange.at(2).Range;
            double payload = static_cast<double>(j) / 50. * (m * range + a);
            myBreguet.calcFuelFromMassFractions(payload, range);
            double fuel = myBreguet.missionFuel;
            double time = 1.10 * range / (theAcftPt->MInitialCruise * theAcftPt->atm.getSpeedOfSoundISA(35000. * 0.3048) * 3.6);
            plot << range / 1.852 << "; ";
            plot << payload << "; ";
            plot << fuel << "; ";
            plot << time << std::endl;
        }
        plot << std::endl;
    }
    plot.close();
}

void fuelPlanning::payloadRangeFromAcftXML() {
    //0.point
    myPayloadRange.push_back(payloadRangeData());
    myPayloadRange.back().Payload   = theAcftPt->maxPayload;
    myPayloadRange.back().Fuel      = 0.;
    myPayloadRange.back().Range     = 0.;
    myPayloadRange.back().TOM       = 0.;
    //0.point
    myPayloadRange.push_back(payloadRangeData());
    myPayloadRange.back().Payload   = theAcftPt->maxPayload;
    myPayloadRange.back().Fuel      = theAcftPt->MTOM - theAcftPt->OME - theAcftPt->maxPayload;
    myPayloadRange.back().Range     = theAcftPt->rangeMaxPayload;
    myPayloadRange.back().TOM       = theAcftPt->MTOM;
    //0.point
    myPayloadRange.push_back(payloadRangeData());
    myPayloadRange.back().Payload   = theAcftPt->payloadRangeMaxfuel;
    myPayloadRange.back().Fuel      = theAcftPt->MFM;
    myPayloadRange.back().Range     = theAcftPt->rangeMaxFuel;
    myPayloadRange.back().TOM       = theAcftPt->MTOM;
    //0.point
    myPayloadRange.push_back(payloadRangeData());
    myPayloadRange.back().Payload   = 0.;
    myPayloadRange.back().Fuel      = 0.;
    myPayloadRange.back().Range     = theAcftPt->rangeFerry;
    myPayloadRange.back().TOM       = 0.;
}

void fuelPlanning::calcStandardFuelPlanning(double *theTaxiFuelTakeoff, double *theTaxiFuelLanding, double *theMissionFuel, double *theContigencyFuel,
                                            double *theAlternateFuel, double *theFinalReserveFuel, double *theTripFuel,
                                            double theRange, double thePayload, bool writingMissionData, const aerodynamics &myAERO) {
    if (thePayload > theAcftPt->maxPayload) {
        myRuntimeInfo->err << "The Payload mission is larger than the maximum payload.!" << std::endl;
        myRuntimeInfo->err << "Mission Payload: " << thePayload << " kg" << std::endl;
        myRuntimeInfo->err << "Max. Payload:    " << theAcftPt->maxPayload << " kg" << std::endl;
        myRuntimeInfo->err << "The Mission cannot be flown! Abort!" << std::endl;
        exit(1);
    }
    *theTripFuel = this->fastFuelEstimate(myAERO);
    *theMissionFuel = *theTripFuel;  // initialization of the mission fuel
    *theTaxiFuelTakeoff = myFlightSim.calcTaxiFuel(theAcftPt->myMission.taxiTimeOrigin);
    *theTaxiFuelLanding = myFlightSim.calcTaxiFuel(theAcftPt->myMission.taxiTimeDestination);
    double oldMissionFuel(0.);
    int ID(0);
    do {
        oldMissionFuel = *theMissionFuel;
        double acMass = theAcftPt->OME + thePayload + *theMissionFuel;
        acMass -= *theTaxiFuelTakeoff;  // Take-off mass
        *theTripFuel = this->calcMissionFuel(acMass, theRange, writingMissionData, myAERO);
        acMass -= *theTripFuel;
        *theAlternateFuel = this->calcAlternateFuel(acMass, myAERO);
        acMass -= *theAlternateFuel;
        *theFinalReserveFuel = this->calcFinalReserveFuel(acMass, myAERO);
        *theContigencyFuel = relConFuel * *theTripFuel;
        // no taxi fuel at landing necessary if reserve fuel is used
        *theMissionFuel = *theTaxiFuelTakeoff + *theTripFuel + *theAlternateFuel + *theFinalReserveFuel + *theContigencyFuel;
        // *theTripFuel = *theTaxiFuel + *theMissionFuel + *theAlternateFuel + *theFinalReserveFuel + *theContigencyFuel;
        ID++;
        if (ID >= 30) {
            myRuntimeInfo->out << "Attention: No tank convergence could be achieved." << std::endl;
            myRuntimeInfo->out << "convergence criterion (tripFuel_0-tripFuel_1)/tripFuel_1 <! 0.0001" << std::endl;
            myRuntimeInfo->out << "Convergence achieved after 30 iterations: " << fabs(oldMissionFuel - *theMissionFuel) / *theMissionFuel << std::endl;
            break;
        }
    } while (fabs(oldMissionFuel - *theMissionFuel) / *theMissionFuel >= 0.0001);
}

double fuelPlanning::calcMissionFuel(double acMass, double Range, bool writeMission, aerodynamics theAERO) {
    if (writeMission == true) {
        theMissionFuel.clear();
        theMissionFuel.push_back(missionFuelData());
        theMissionFuel.back().Time = 0.;
        theMissionFuel.back().Range = 0.;
        theMissionFuel.back().Altitude = 0.;
        theMissionFuel.back().TAS = 0.;
        theMissionFuel.back().Mach = 0.;
        theMissionFuel.back().Mass = acMass;
        theMissionFuel.back().consumedFuel = 0.;
        theMissionFuel.back().Thrust = 0.;
        theMissionFuel.back().fuelFlow = 0.;
        theMissionFuel.back().ROC = 0.;
    }
    vector<double> missionFuel;
    double distance = 0.;
    double flownDistance = 0.;
    double distanceCruise(0.);
    double flownCruise = 0.0;
    bool firstCruise(true);
    unsigned int i(0);
    while (i <= theAcftPt->myMission.Steps.size() - 1) {
        setP(i);
        bool checkSteadyLevelFlight(false);
        bool checkAcceleration(false);
        double time(0.);
        double Drag(0.);
        /* TakeOff, Climb, Accelerate, Cruise, Descend, Landing */
        if (theAcftPt->myMission.Steps.at(i).Point.simplifiedMODE == "takeoff") {
            time = convertUnit(MINUTE, SECOND, 1.);  // Assumption that take-off lasts 1 minute
            // Drag = 0.0;  // TODO(hansmann): Wechsel von Drag=0.0 & calculate_N1_with_thrustlimit auf calculate_N1_with_penalties mit vollschub abklären
            // theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_takeoff, P.Altitude, P.Mach, theAcftPt->atm, P.derate, P.rating,
            //     P.bleedOff, P.shaftOff, convertUnit(NEWTON, KILO, NEWTON, Drag));
            theAcftPt->myEngines.calculate_N1_with_penalties(theAcftPt->myEnergyProviders.energyCarrierID_takeoff, P.Altitude, P.Mach, theAcftPt->atm, P.derate, P.rating,
                P.bleedOff, P.shaftOff);
            missionFuel.push_back(time * theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_takeoff));
        } else if (theAcftPt->myMission.Steps.at(i).Point.simplifiedMODE == "climb") {
            time = 0.0;
            distance = 0.;
            if (theAcftPt->myMission.Steps.at(i - 1).endAltitude < theAcftPt->myMission.Steps.at(i).endAltitude) {
                missionFuel.push_back(0.);
                myFlightSim.getClimbFuel(acMass,  theAcftPt->myMission.Steps.at(i - 1).endAltitude,
                                         theAcftPt->myMission.Steps.at(i).endAltitude, theAcftPt->myMission.Steps.at(i - 1).endMach,
                                         theAcftPt->myMission.Steps.at(i).endMach, P.config, P.rating, P.bleedOff, P.shaftOff, theAERO, &missionFuel.back(), &distance, &time);
            } else {
                myRuntimeInfo->out << "... No change in altitude despite climb segment." << std::endl;
                missionFuel.push_back(0.0);
            }
            flownDistance += distance;
        } else if (theAcftPt->myMission.Steps.at(i).Point.simplifiedMODE == "accelerate") {
            Drag = 0.0;
            time = 0.0;
            distance = 0.;
            try {
                checkAcceleration = myFlightSim.checkForAcceleration(acMass, P.Altitude, P.Mach,
                                    P.config, P.rating, P.derate, P.bleedOff, P.shaftOff, theAERO);
            } catch (const std::string& errMsg) {
                myRuntimeInfo->out << errMsg << std::endl;
                myRuntimeInfo->out << "Try to increase rating..." << std::endl;
            }
            if (checkAcceleration == false && P.rating == "cruise") {
                P.rating = "climb";
                try {
                    checkAcceleration = myFlightSim.checkForAcceleration(acMass, P.Altitude, P.Mach,
                                        P.config, P.rating, P.derate, P.bleedOff, P.shaftOff, theAERO);
                } catch (const std::string& errMsg) {
                    myRuntimeInfo->out << errMsg << std::endl;
                    myRuntimeInfo->out << "Try to increase rating..." << std::endl;
                }
            }
            if (checkAcceleration == false && P.rating == "climb") {
                P.rating = "maximum_continuous";
                try {
                    checkAcceleration = myFlightSim.checkForAcceleration(acMass, P.Altitude, P.Mach,
                                        P.config, P.rating, P.derate, P.bleedOff, P.shaftOff, theAERO);
                } catch (const std::string& errMsg) {
                    myRuntimeInfo->out << errMsg << std::endl;
                    myRuntimeInfo->err << "Rating cannot be further increased..." << std::endl;
                }
            }
            if (checkAcceleration == false) {
                exitAcceleration(i, P, acMass);
            }
            missionFuel.push_back(0.);
            myFlightSim.getAccelerateFuel(acMass, P.Altitude,
                                          theAcftPt->myMission.Steps.at(i - 1).endMach, theAcftPt->myMission.Steps.at(i).endMach,
                                          P.config, P.rating,
                                          P.bleedOff, P.shaftOff,
                                          theAERO,
                                          &missionFuel.back(), &distance, &time);
            flownDistance += distance;
        } else if (theAcftPt->myMission.Steps.at(i).Point.simplifiedMODE == "cruise") {
            if (firstCruise == true) {
                distanceCruise = Range - flownDistance - 10000. / tan(convertUnit(DEGREE, RADIAN, 5.));
                flownCruise = 0.;
                firstCruise = false;
            }
            try {
                checkSteadyLevelFlight = myFlightSim.checkForSteadyLevelFlight(acMass, P.Altitude, P.Mach, P.config, P.rating, P.bleedOff, P.shaftOff, theAERO);
            } catch (const std::string& errMsg) {
                myRuntimeInfo->out << errMsg << std::endl;
                myRuntimeInfo->out << "Try to increase rating..." << std::endl;
            }
            if (checkSteadyLevelFlight == false && P.rating == "cruise") {
                P.rating = "climb";
                try {
                    checkSteadyLevelFlight = myFlightSim.checkForSteadyLevelFlight(acMass, P.Altitude, P.Mach, P.config, P.rating, P.bleedOff, P.shaftOff, theAERO);
                } catch (const std::string& errMsg) {
                    myRuntimeInfo->out << errMsg << std::endl;
                    myRuntimeInfo->out << "Try to increase rating..." << std::endl;
                }
            }
            if (checkSteadyLevelFlight == false && P.rating == "climb") {
                P.rating = "maximum_continuous";
                try {
                    checkSteadyLevelFlight = myFlightSim.checkForSteadyLevelFlight(acMass, P.Altitude, P.Mach, P.config, P.rating, P.bleedOff, P.shaftOff, theAERO);
                } catch (const std::string& errMsg) {
                    myRuntimeInfo->out << errMsg << std::endl;
                    myRuntimeInfo->err << "Rating cannot be further increased..." << std::endl;
                }
            }
            if (checkSteadyLevelFlight == false) {
                exitSteadyLevelFlight(i, P, Drag);
            }
            double cruiseRange(0.0);
            cruiseRange = theAcftPt->myMission.Steps.at(i).relSegmentLength * distanceCruise - flownCruise;
            if (cruiseRange < 0.) {
                myRuntimeInfo->err << "Negative cruise route!" << std::endl;
                myRuntimeInfo->err << "Abort!" << std::endl;
                myRuntimeInfo->err << "Segment: " << i << std::endl;
                myRuntimeInfo->err << "Altitude:" << theAcftPt->myMission.Steps.at(i).Point.Altitude << std::endl;
                myRuntimeInfo->err << "Altitude:" << theAcftPt->myMission.Steps.at(i).Point.Mach << std::endl;
                myRuntimeInfo->err << "-> Possible reason: too slow climb to the altitude, so that the climb is longer than the actual mission segment." << std::endl;
                exit(1);
            }
            distance = cruiseRange;
            flownCruise += cruiseRange;
            flownDistance += cruiseRange;
            missionFuel.push_back(0.);
            time = 0.;
            myFlightSim.getCruiseFuel(acMass, P.Altitude, P.Mach, P.config, P.rating, P.bleedOff, P.shaftOff, theAERO,
                                      &missionFuel.back(), cruiseRange, &time);
        } else if (theAcftPt->myMission.Steps.at(i).Point.simplifiedMODE == "descend") {
            time = 0.;
            missionFuel.push_back(0.);
            myFlightSim.getDescendFuel(theAcftPt->myMission.Steps.at(i - 1).endAltitude,
                                       theAcftPt->myMission.Steps.at(i).endAltitude, theAcftPt->myMission.Steps.at(i - 1).endMach,
                                       theAcftPt->myMission.Steps.at(i).endMach, P.bleedOff, P.shaftOff, &missionFuel.back(), &distance, &time);
            flownDistance += distance;
        } else if (theAcftPt->myMission.Steps.at(i).Point.simplifiedMODE == "landing") {
            Drag = 0.0;
            time = convertUnit(MINUTE, SECOND, 2.0);  // Assume the final approach and landing will take 2 minutes.
            distance = 0.;
            theAcftPt->myEngines.calculate_N1_with_penalties(theAcftPt->myEnergyProviders.energyCarrierID_landing, P.Altitude, P.Mach, theAcftPt->atm, P.derate, P.rating,
                P.bleedOff, P.shaftOff);
            missionFuel.push_back(time * theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_landing));
        } else {
            myRuntimeInfo->err << "unknown simplified Mode !" << std::endl;
            myRuntimeInfo->err << theAcftPt->myMission.Steps.at(i).Point.simplifiedMODE << std::endl;
            exit(1);
        }
        acMass -= missionFuel.back();
        if (writeMission == true) {
            double maxROC(0.);
            theMissionFuel.push_back(missionFuelData());
            theMissionFuel.back().Time = time;
            theMissionFuel.back().Range = distance;
            theMissionFuel.back().Altitude = P.Altitude;
            theMissionFuel.back().TAS = P.TAS;
            theMissionFuel.back().Mach = P.Mach;
            theMissionFuel.back().Mass = acMass;
            theMissionFuel.back().consumedFuel = missionFuel.back();
            theMissionFuel.back().Thrust = theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise);  // TODO(hansmann): Take operating state into account
            theMissionFuel.back().fuelFlow = theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_cruise);
            theMissionFuel.back().ROC = maxROC;  // this is obviously wrong. Should be fixed when parts are merged with missionAnalysis
        }
        i++;
    }
    return accumulate(missionFuel.begin(), missionFuel.end(), 0.0);
}

double fuelPlanning::calcAlternateFuel(double acMass, aerodynamics theAERO) {
    /* Alternate Fuel */
    double Fuel(0.0);
    double alternateAlt = convertUnit(FOOT, METER, 27000.);
    double alternateMach = 0.95 * theAcftPt->MInitialCruise;
    double tempAlternateRange = theAcftPt->myMission.alternateRange - 1. * alternateAlt / tan(convertUnit(DEGREE, RADIAN, 5.));
    /* Climb */
    double tempTime(0.);
    double climbFuel(0.);
    double climbDistance(0.);
    // 0 ft up to 1,500 ft in Climb configuration with v2 + 10 kts
    double climbTAS = theAcftPt->v2(acMass, convertUnit(FOOT, METER, 750.)) + convertUnit(KNOTS, METERPERSECOND, 10.);
    myFlightSim.getClimbFuel(acMass, 0.0, convertUnit(FOOT, METER, 1500.),
                             climbTAS / theAcftPt->atm.getSpeedOfSound(0.), climbTAS / theAcftPt->atm.getSpeedOfSound(convertUnit(FOOT, METER, 1500.)),
                             "climb", "takeoff", 0.0, 0.0, theAERO,
                             &climbFuel, &climbDistance, &tempTime);
    Fuel += climbFuel;
    acMass -= climbFuel;
    tempAlternateRange -= climbDistance;
    // Acceleration from v2+10kt to 250 kt
    tempTime = 0.;
    climbFuel = 0.;
    climbDistance = 0.;
    myFlightSim.getAccelerateFuel(acMass, convertUnit(FOOT, METER, 1500.),
                                  climbTAS / theAcftPt->atm.getSpeedOfSound(convertUnit(FOOT, METER, 1500.)),
                                  convertUnit(KNOTS, METERPERSECOND, 250.) / theAcftPt->atm.getSpeedOfSound(convertUnit(FOOT, METER, 1500.)),
                                  "clean", "takeoff", 0.0, 0.0, theAERO,
                                  &climbFuel, &climbDistance, &tempTime);
    Fuel += climbFuel;
    acMass -= climbFuel;
    tempAlternateRange -= climbDistance;
    // Climb from 1500 ft up to 10000 ft with 250 kt
    tempTime = 0.;
    climbFuel = 0.;
    climbDistance = 0.;
    climbTAS = convertUnit(KNOTS, METERPERSECOND, 250);
    myFlightSim.getClimbFuel(acMass, convertUnit(FOOT, METER, 1500.), convertUnit(FOOT, METER, 10000.),
                        climbTAS / theAcftPt->atm.getSpeedOfSound(convertUnit(FOOT, METER, 1500.)), climbTAS / theAcftPt->atm.getSpeedOfSound(convertUnit(FOOT, METER, 10000.)),
                        "clean", "maximum_continuous", 0.0, 0.0, theAERO,
                        &climbFuel, &climbDistance, &tempTime);
    Fuel += climbFuel;
    acMass -= climbFuel;
    tempAlternateRange -= climbDistance;
    // accelerating to 300 kt
    tempTime = 0.;
    climbFuel = 0.;
    climbDistance = 0.;
    myFlightSim.getAccelerateFuel(acMass, convertUnit(FOOT, METER, 10000.),
                                  climbTAS / theAcftPt->atm.getSpeedOfSound(convertUnit(FOOT, METER, 10000.)),
                                  convertUnit(KNOTS, METERPERSECOND, 300) / theAcftPt->atm.getSpeedOfSound(convertUnit(FOOT, METER, 10000.)),
                                  "clean", "maximum_continuous", 0.0, 0.0, theAERO,
                                  &climbFuel, &climbDistance, &tempTime);
    Fuel += climbFuel;
    acMass -= climbFuel;
    tempAlternateRange -= climbDistance;
    // Climb to alternative altitude with 300 kt
    tempTime = 0.;
    climbFuel = 0.;
    climbDistance = 0.;
    climbTAS = 300. * (1.852 / 3.6);
    myFlightSim.getClimbFuel(acMass, 10000.*0.3048, alternateAlt,
                             climbTAS / theAcftPt->atm.getSpeedOfSound(10000.*0.3048), climbTAS / theAcftPt->atm.getSpeedOfSound(alternateAlt),
                             "clean", "climb", 0.0, 0.0, theAERO,
                             &climbFuel, &climbDistance, &tempTime);
    Fuel += climbFuel;
    acMass -= climbFuel;
    tempAlternateRange -= climbDistance;
    // Acceleration to alternative Mach
    if (alternateMach > climbTAS / theAcftPt->atm.getSpeedOfSound(alternateAlt)) {
        tempTime = 0.;
        climbFuel = 0.;
        climbDistance = 0.;
        myFlightSim.getAccelerateFuel(acMass, alternateAlt,
                                      climbTAS / theAcftPt->atm.getSpeedOfSound(10000.*0.3048),
                                      alternateMach,
                                      "clean", "maximum_continuous", 0.0, 0.0, theAERO,
                                      &climbFuel, &climbDistance, &tempTime);
        Fuel += climbFuel;
        acMass -= climbFuel;
        tempAlternateRange -= climbDistance;
    }
    if (tempAlternateRange < 0.) {
        tempAlternateRange = 50.;
    }
    /* Alternate Cruise */
    tempTime = 0.;
    climbFuel = 0.;
    climbDistance = 0.;
    myFlightSim.getCruiseFuel(acMass, alternateAlt, alternateMach, "clean", "maximum_continuous", 0.0, 0.0, theAERO,
                              &climbFuel, tempAlternateRange, &tempTime);
    Fuel += climbFuel;
    return Fuel;
}

double fuelPlanning::calcFinalReserveFuel(double acMass, aerodynamics theAERO) {
    // Determination of Final Reserves: 30min @ 1,500ft
    // SARPerformance SAR(*theAcftPt);
    // Determining the most effective speed via SAR
    double holdingM(0.);
    double holding_altitude = convertUnit(FOOT, METER, 1500.);
    double holdingSpeed = 1.05 * theAcftPt->v1gS_Clean(acMass, holding_altitude);
    holdingM = calcBestSARSpeed(holding_altitude, acMass, holdingSpeed, convertUnit(KNOTS, METERPERSECOND, 250.));
    double holdingDrag = theAERO.getCruiseDrag(holdingM, holding_altitude, 0., acMass, "clean", theAcftPt->atm);
    theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_cruise, holding_altitude, holdingM, theAcftPt->atm, 1.,
                                                        "maximum_continuous", 0.0, 0.0, holdingDrag);
    double LoverD = (G_FORCE * acMass) / holdingDrag;
    double SFC = theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_cruise)
               / theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise);
    return acMass * (expm1(convertUnit(MINUTE, SECOND, 30.) * G_FORCE * SFC * 1. / LoverD));  // return calcBreguet(acMass, 30.*60., SFC, LoverD);
}

double fuelPlanning::calcBestSARSpeed(double alt, double mass, double lowerSpeed, double upperSpeed) {
    low::SARPerformance SAR(theAcftPt);  // TODO(Hansmann): Reconsider hardcoding namespace / fidelity level
    double tempSAR(0.);
    double tempMach(lowerSpeed / theAcftPt->atm.getSpeedOfSound(alt));
    while (lowerSpeed < upperSpeed) {
        if (tempSAR <= SAR.calculateSAR(mass, alt, lowerSpeed / theAcftPt->atm.getSpeedOfSound(alt), "climb", "clean")) {
            tempMach = lowerSpeed / theAcftPt->atm.getSpeedOfSound(convertUnit(FOOT, METER, 1500.));
        }
        lowerSpeed += 0.5;
    }
    return tempMach;
}

double fuelPlanning::calcBreguet(double W0, double Time, double theSFC, double theLoD) {
    return W0 * (1. - 1. / exp(Time * G_FORCE * theSFC * 1. / theLoD));
}

//double fuelPlanning::calcRangeForMass(double Range0, double Range1, double Mass0, double Mass1, double targetMass) {
//    double m(0.);
//    double a(0.);
//    m = (Mass0 - Mass1) / (Range0 - Range1);
//    a = Mass0 - m * Range0;
//    return (targetMass - a) / m;
//}

void fuelPlanning::setP(int ID) {
    P.Mach = theAcftPt->myMission.Steps.at(ID).Point.Mach;
    P.Altitude = theAcftPt->myMission.Steps.at(ID).Point.Altitude;
    P.TAS = theAcftPt->myMission.Steps.at(ID).Point.TAS;
    P.ROC = theAcftPt->myMission.Steps.at(ID).Point.ROC;
    P.derate = theAcftPt->myMission.Steps.at(ID).derating;
    P.bleedOff = theAcftPt->myMission.Steps.at(ID).Point.bleedOffTakes;
    P.shaftOff = theAcftPt->myMission.Steps.at(ID).Point.shaftOffTakes;
    P.config = theAcftPt->myMission.Steps.at(ID).configuration;
    P.rating = theAcftPt->myMission.Steps.at(ID).rating;
}

//void fuelPlanning::exitROC(int ID, pointData theP, double massAircraft, aerodynamics theAERO) {
//    theAcftPt->myEngine.setEngineRating(theP.Altitude, theP.Mach, theAcftPt->atm, theP.derate, theP.rating, theP.bleedOff, theP.shaftOff);
//    double maxROC = theAcftPt->ROC(massAircraft, theP.Altitude, theP.Mach, 0.0,
//                                   theP.config, theP.rating, theP.bleedOff, theP.shaftOff, false, theAERO);
//    myRuntimeInfo->err << "Error in fuel estimation!" << std::endl;
//    myRuntimeInfo->err << "Thrust does not allow the required ROC!" << std::endl;
//    myRuntimeInfo->err << "mission segment: " << ID << std::endl;
//    myRuntimeInfo->err << "Mission mode: " << theAcftPt->myMission.Steps.at(ID).modeStep << std::endl;
//    myRuntimeInfo->err << "max. ROC: " << maxROC << " ft/min" << std::endl;
//    myRuntimeInfo->err << "required ROC: " << theP.ROC << " ft/min" << std::endl;
//    myRuntimeInfo->err << "flight altitude: " << theP.Altitude / 0.3048 << " ft" << std::endl;
//    myRuntimeInfo->err << "Mach number: " << theP.Mach << "" << std::endl;
//    myRuntimeInfo->err << "Engine-Rating: " << theP.rating << std::endl;
//    exit(1);
//}

void low::fuelPlanning::exitAcceleration(int ID, pointData theP, double massAircraft) {
    double Drag = theAcftPt->aero.getCruiseDrag(theP.Mach, theP.Altitude, 0., massAircraft, theP.config, theAcftPt->atm);
    theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_cruise, theP.Altitude, theP.Mach, theAcftPt->atm, theP.derate, theP.rating, theP.bleedOff, theP.shaftOff, Drag);
    myRuntimeInfo->err << "Error in fuel estimation!" << std::endl;
    myRuntimeInfo->err << "Thrust does not allow acceleration!" << std::endl;
    myRuntimeInfo->err << "mission segment: " << ID << std::endl;
    myRuntimeInfo->err << "Mission mode: " << theAcftPt->myMission.Steps.at(ID).modeStep << std::endl;
    myRuntimeInfo->err << "available thrust: " << theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise) << " N" << std::endl;
    myRuntimeInfo->err << "drag: " << Drag << " N" << std::endl;
    myRuntimeInfo->err << "flight altitude: " << theP.Altitude << " m" << std::endl;
    myRuntimeInfo->err << "Mach number: " << theP.Mach << "" << std::endl;
    myRuntimeInfo->err << "Engine-Rating: " << theP.rating << std::endl;
    exit(1);
}

void low::fuelPlanning::exitSteadyLevelFlight(int ID, pointData theP, double theDrag) {
    myRuntimeInfo->err << "Error in fuel estimation!" << std::endl;
    myRuntimeInfo->err << "Thrust does not allow Steady Level Flight!" << std::endl;
    myRuntimeInfo->err << "Mission segment: " << ID << std::endl;
    myRuntimeInfo->err << "Mission mode: " << theAcftPt->myMission.Steps.at(ID).modeStep << std::endl;
    myRuntimeInfo->err << "available thrust: " << theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise) << " N" << std::endl;
    myRuntimeInfo->err << "drag: " << theDrag << " N" << std::endl;
    myRuntimeInfo->err << "flight altitude: " << theP.Altitude << " m" << std::endl;
    myRuntimeInfo->err << "Mach number: " << theP.Mach << "" << std::endl;
    myRuntimeInfo->err << "Engine-Rating: " << theP.rating << std::endl;
    exit(1);
}

double low::fuelPlanning::fastFuelEstimate(aerodynamics theAERO) {
    double Time = theAcftPt->myMission.missionRange / (theAcftPt->myMission.crMach * theAcftPt->atm.getSpeedOfSoundISA(10000.));
    double acMass = theAcftPt->MTOM;
    double theLoverD = theAERO.getCruiseLoverD(theAcftPt->myMission.crMach, 10000., 0.0, acMass, "clean", theAcftPt->atm);
    double theDrag = theAERO.getCruiseDrag(theAcftPt->myMission.crMach, 10000., 0.0, acMass, "clean", theAcftPt->atm);
    theAcftPt->myEngines.calculate_N1_with_thrustlimit(theAcftPt->myEnergyProviders.energyCarrierID_cruise, 10000., theAcftPt->myMission.crMach, theAcftPt->atm, 1.0,
        "maximum_continuous", 0., 0., theDrag);
    double theSFC = theAcftPt->myEngines.get_aircraft_fuelflow(theAcftPt->myEnergyProviders.energyCarrierID_cruise)
        / theAcftPt->myEngines.get_thrust_aircraft(theAcftPt->myEnergyProviders.energyCarrierID_cruise);
    return this->calcBreguet(theAcftPt->MTOM, Time, theSFC, theLoverD);
}

}  // namespace low
