/*  Copyright (C) 2012-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "missionDefinition.h"

using std::endl;
using std::stringstream;

missionDefinition::missionDefinition(const string &fileNameAndDir, double delta_isa)
    :
    bleedOffTO(0.0),
    shaftOffTO(0.0),
    bleedOffCr(0.0),
    shaftOffCr(0.0),
    deltaISA(delta_isa),
    missionXML(aixml::openDocument(fileNameAndDir)) {
    myRuntimeInfo->out << "Reading the Mission: " << fileNameAndDir << endl;
    missionPayload = EndnodeReadOnly<double>("/mission/payload").read(missionXML).value();
    missionPAX = EndnodeReadOnly<int>("mission/number_of_pax").read(missionXML).value();
    missionCargo = EndnodeReadOnly<double>("/mission/cargo_mass").read(missionXML).value();
    missionRange = EndnodeReadOnly<double>("mission/range").read(missionXML).value();
    alternateRange = EndnodeReadOnly<double>("mission/alternate_distance").read(missionXML).value();

    crMach = EndnodeReadOnly<double>("mission/desired_cruise_speed").read(missionXML).value();
    taxiTimeOrigin = EndnodeReadOnly<double>("mission/taxi_time_origin").read(missionXML).value();
    taxiTimeDestination = EndnodeReadOnly<double>("mission/taxi_time_destination").read(missionXML).value();
    numberDepartureSteps = missionXML.getVector("/mission/departure/departure_step", 1).size();
    numberCruiseSteps = missionXML.getVector("/mission/cruise/cruise_step", 1).size();
    numberApproachSteps = missionXML.getVector("/mission/approach/approach_step", 1).size();

    /* Set atmosphere */
    myISA.setAtmosphere(0., ISA_TEMPERATURE + this->deltaISA, ISA_PRESSURE);
    /* Read mission */
    this->readMission();
    this->getTyicalOffTakes();
    myRuntimeInfo->out << "... Mission successfully read!" << endl;
}

missionDefinition::~missionDefinition() {
    //dtor
}

void missionDefinition::readMission() {
    this->readSteps("departure", this->numberDepartureSteps);
    this->readSteps("cruise", this->numberCruiseSteps);
    this->readSteps("approach", this->numberApproachSteps);
}

void missionDefinition::readSteps(const string& missionPhase, const uint16_t& numberTotalSteps) {
    for (uint16_t aStepNumber(0); aStepNumber < numberTotalSteps; ++aStepNumber) {
        string xmlPath("mission/");
        if (missionPhase == "departure") {
            xmlPath += "departure/departure_step@";
        } else if (missionPhase == "cruise") {
            xmlPath += "cruise/cruise_step@";
        } else {
            xmlPath += "approach/approach_step@";
        }
        xmlPath += num2Str(aStepNumber);
        this->Steps.push_back(missionStep());
        this->Steps.back().missionPhase = missionPhase;
        this->getMissionSegment(xmlPath);
        this->getMissionPoint();
    }
}

void missionDefinition::getMissionSegment(string XML) {
    /* General parameters that are always available */
    Steps.back().modeStep = string(missionXML.at(XML+"/mode/value"));
    Steps.back().configuration = string(missionXML.at(XML + "/configuration/value"));
    Steps.back().rating = string(missionXML.at(XML + "/rating/value"));
    Steps.back().derating = EndnodeReadOnly<double>(XML + "/derate").read(missionXML).value();
    if (missionXML.find(XML+"/altitude")) {
        Steps.back().endAltitude = EndnodeReadOnly<double>(XML + "/altitude").read(missionXML).value();
    } else {
        Steps.back().endAltitude = NAN;
    }
    if (missionXML.find(XML+"/calibrated_airspeed")) {
        Steps.back().endCAS = EndnodeReadOnly<double>(XML + "/calibrated_airspeed").read(missionXML).value();
    } else {
        Steps.back().endCAS = NAN;
    }
    /* Off-Takes */
    if (missionXML.find(XML+"/shaft_power_takeoff_schedule/background_power_level")) {
        Steps.back().shaftOffTakes = EndnodeReadOnly<double>(XML + "/shaft_power_takeoff_schedule/background_power_level").read(missionXML).value();
    } else {
        Steps.back().shaftOffTakes = 0;
    }

    if (missionXML.find(XML+"/bleed_air_takeoff_schedule/background_bleed_level")) {
        Steps.back().bleedOffTakes = EndnodeReadOnly<double>(XML + "/bleed_air_takeoff_schedule/background_bleed_level").read(missionXML).value();
    } else {
        Steps.back().bleedOffTakes = 0;
    }

    /* Mission phase dependent parameters */
    if (Steps.back().missionPhase == "departure") {
        /* ROC */
        if (missionXML.find(XML + "/ROC")) {
            Steps.back().endROC = EndnodeReadOnly<double>(XML + "/ROC").read(missionXML).value();
        } else {
            Steps.back().endROC = 0;
        }
        /* Filling the unoccupied entries depending on the step */
        if (Steps.back().modeStep == "takeoff") {
            Steps.back().endAltitude = 0.0;
            Steps.back().endFL = 0.0;
            Steps.back().endCAS = 82.3;  // =160.0 kts;    // TODO(hansmann): Mildly illegal to assume 160 flat?
        } else if (Steps.back().modeStep == "climb") {
            if (Steps.back().endAltitude <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "The climb phase is not the final flight altitude! Abort!" << endl;
                exit(1);
            }
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
            //Steps.back().endROC = missionXML.at(XML + "/ROC");
        } else if (Steps.back().modeStep == "accelerate") {
            if (Steps.back().endCAS <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Acceleration flight phase is not specified as final speed! Abort!" << endl;
                exit(1);
            }
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
        } else {
            myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
            myRuntimeInfo->err << "Step in this flight phase is unknown! Abort!" << endl;
            exit(1);
        }
        Steps.back().endFL = Steps.back().endAltitude / 100;
        Steps.back().endTAS =  convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
        Steps.back().endMach = convertUnit(TRUEAIRSPEED, MACH, Steps.back().endAltitude, myISA, Steps.back().endTAS);
        Steps.back().relSegmentLength = 0.0;
    } else if (Steps.back().missionPhase == "cruise") {
        if (missionXML.find(XML+"/mach_number")) {
            Steps.back().endMach = EndnodeReadOnly<double>(XML + "/mach_number").read(missionXML).value();
        }
        if (missionXML.find(XML+"/altitude")) {
            Steps.back().endFL = EndnodeReadOnly<double>(XML+"/altitude").read(missionXML).value() / 100;
        }
        if (missionXML.find(XML+"/relative_segment_length")) {
            Steps.back().relSegmentLength = EndnodeReadOnly<double>(XML + "/relative_segment_length").read(missionXML).value();
        }
        /* Filling the unoccupied entries depending on the step */
        if (Steps.back().modeStep == "change_speed_to_CAS") {
            if (Steps.back().endCAS <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Acceleration flight phase is not specified as final speed! Abort!" << endl;
                exit(1);
            }
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
            Steps.back().endFL = Steps.back().endAltitude / 100;
            Steps.back().endTAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
            Steps.back().endMach = convertUnit(TRUEAIRSPEED, MACH, Steps.back().endAltitude, myISA, Steps.back().endTAS);
        } else if (Steps.back().modeStep == "climb_to_cruise") {
            if (Steps.back().endFL <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Climb phase is not defined as an end flight level.! Abort!" << endl;
                exit(1);
            }
            if (Steps.back().endMach <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Climbing phase is not defined as final mach number! Abort!" << endl;
                exit(1);
            }
            Steps.back().endAltitude = 100 * Steps.back().endFL;
            Steps.back().endTAS = convertUnit(MACH, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endMach);
            Steps.back().endCAS = convertUnit(MACH, CALIBRATEDAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endMach);
            cruiseAlt.push_back(Steps.back().endAltitude);
            //Steps.back().endROC = missionXML.at(XML + "/ROC");
        } else if (Steps.back().modeStep == "change_speed_to_Mach") {
            if (Steps.back().endMach <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Acceleration phase is not defined as a final mach number! Abort!" << endl;
                exit(1);
            }
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
            Steps.back().endFL = Steps.at(Steps.size() - 2).endFL;
            Steps.back().endTAS = convertUnit(MACH, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endMach);
            Steps.back().endCAS = convertUnit(MACH, CALIBRATEDAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endMach);
        } else if (Steps.back().modeStep == "cruise") {
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
            Steps.back().endFL = Steps.at(Steps.size() - 2).endFL;
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
            Steps.back().endTAS = Steps.at(Steps.size() - 2).endTAS;
            Steps.back().endMach = Steps.at(Steps.size() - 2).endMach;
            cruiseSegLength.push_back(Steps.back().relSegmentLength);
        } else if (Steps.back().modeStep == "change_flight_level") {
            if (Steps.back().endFL <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Climb phase is not defined as an end flight level.! Abort!" << endl;
                exit(1);
            }
            Steps.back().endAltitude = 100 * Steps.back().endFL;
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
            Steps.back().endTAS = Steps.at(Steps.size() - 2).endTAS;
            Steps.back().endMach = Steps.at(Steps.size() - 2).endMach;
            cruiseAlt.push_back(Steps.back().endAltitude);
        } else if (Steps.back().modeStep == "change_flight_level_constant_ROC") {
            if (Steps.back().endFL <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Climb phase is not defined as an end flight level.! Abort!" << endl;
                exit(1);
            }
            Steps.back().endAltitude = 100 * Steps.back().endFL;
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
            Steps.back().endTAS = Steps.at(Steps.size() - 2).endTAS;
            Steps.back().endMach = Steps.at(Steps.size() - 2).endMach;
            cruiseAlt.push_back(Steps.back().endAltitude);
        } else if (Steps.back().modeStep == "descend_to_approach") {
            if (Steps.back().endCAS <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Descend is not defined as a final speed! Abort!" << endl;
                exit(1);
            }
            if (Steps.back().endAltitude <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Descend is not to defined as a final flight altitude! Abort!" << endl;
                exit(1);
            }
            Steps.back().endFL = Steps.back().endAltitude / 100;
            Steps.back().endTAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
            Steps.back().endMach = convertUnit(TRUEAIRSPEED, MACH, Steps.back().endAltitude, myISA, Steps.back().endTAS);
        } else {
            myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
            myRuntimeInfo->err << "Step in this flight phase is unknown! Abort!" << endl;
            exit(1);
        }
    } else if (Steps.back().missionPhase == "approach") {
        /* Filling the unoccupied entries depending on the step */
        if (Steps.back().modeStep == "change_speed") {
            if (Steps.back().endCAS <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Delay flight phase is not defined as final speed! Abort!" << endl;
                exit(1);
            }
            Steps.back().endAltitude = Steps.at(Steps.size() - 2).endAltitude;
        } else if (Steps.back().modeStep == "descend") {
            if (Steps.back().endAltitude <= 0.0) {
                myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
                myRuntimeInfo->err << "Descend is not defined as a final flight altitude! Abort!" << endl;
                exit(1);
            }
            Steps.back().endCAS = Steps.at(Steps.size() - 2).endCAS;
        } else if (Steps.back().modeStep == "landing") {
            Steps.back().endAltitude = 0.0;
            Steps.back().endCAS = 0.0;
        } else if (Steps.back().modeStep == "level_glide_slope_interception") {
            Steps.back() = Steps.at(Steps.size() - 2);
        } else {
            myRuntimeInfo->err << "flight phase: " << Steps.back().missionPhase << " | Step: " << Steps.back().modeStep << endl;
            myRuntimeInfo->err << "Step in this flight phase is unknown! Abort!" << endl;
            exit(1);
        }
        Steps.back().endFL = Steps.back().endAltitude / 100;
        Steps.back().endTAS = convertUnit(CALIBRATEDAIRSPEED, TRUEAIRSPEED, Steps.back().endAltitude, myISA, Steps.back().endCAS);
        Steps.back().endMach = convertUnit(TRUEAIRSPEED, MACH, Steps.back().endAltitude, myISA, Steps.back().endTAS);
        Steps.back().relSegmentLength = 0.0;
    } else {
        myRuntimeInfo->err << "The mission segment cannot be determined!" << endl;
        myRuntimeInfo->err << "The Mission phase is unknown: " << Steps.back().missionPhase << endl;
        exit(1);
    }
}

void missionDefinition::getMissionPoint() {
    if (Steps.size() == 1) {
        Steps.back().Point.Altitude = Steps.back().endAltitude;
        Steps.back().Point.Temperature = myISA.getTemperature(Steps.back().Point.Altitude);
        Steps.back().Point.CAS = Steps.back().endCAS;
        Steps.back().Point.TAS = Steps.back().endTAS;
        Steps.back().Point.Mach = Steps.back().endMach;
        Steps.back().Point.shaftOffTakes = Steps.back().shaftOffTakes;
        Steps.back().Point.bleedOffTakes = Steps.back().bleedOffTakes;
    } else {
        Steps.back().Point.Altitude = Steps.at(Steps.size() - 2).endAltitude
                                      + 0.5 * (Steps.back().endAltitude - Steps.at(Steps.size() - 2).endAltitude);
        Steps.back().Point.Temperature = myISA.getTemperature(Steps.back().Point.Altitude);
        Steps.back().Point.CAS = Steps.at(Steps.size() - 2).endCAS
                                 + 0.5 * (Steps.back().endCAS - Steps.at(Steps.size() - 2).endCAS);
        Steps.back().Point.TAS = Steps.at(Steps.size() - 2).endTAS
                                 + 0.5 * (Steps.back().endTAS - Steps.at(Steps.size() - 2).endTAS);
        Steps.back().Point.Mach = Steps.at(Steps.size() - 2).endMach
                                  + 0.5 * (Steps.back().endMach - Steps.at(Steps.size() - 2).endMach);
        Steps.back().Point.shaftOffTakes = Steps.at(Steps.size() - 2).shaftOffTakes
                                           + 0.5 * (Steps.back().shaftOffTakes - Steps.at(Steps.size() - 2).shaftOffTakes);
        Steps.back().Point.bleedOffTakes = Steps.at(Steps.size() - 2).bleedOffTakes
                                           + 0.5 * (Steps.back().bleedOffTakes - Steps.at(Steps.size() - 2).bleedOffTakes);
    }
    Steps.back().Point.ROC = Steps.back().endROC;
    Steps.back().Point.density = myISA.getDensity(Steps.back().Point.Altitude);
    Steps.back().Point.dynamicPressure = 0.5 * myISA.getDensity(Steps.back().Point.Altitude) * pow(Steps.back().Point.TAS, 2.);
    if (Steps.back().modeStep == "takeoff") {
        Steps.back().Point.simplifiedMODE = "takeoff";
    } else if (Steps.back().modeStep == "climb") {
        Steps.back().Point.simplifiedMODE = "climb";
    } else if (Steps.back().modeStep == "accelerate") {
        Steps.back().Point.simplifiedMODE = "accelerate";
    } else if (Steps.back().modeStep == "cruise") {
        Steps.back().Point.simplifiedMODE = "cruise";
    } else if (Steps.back().modeStep == "climb_to_cruise") {
        Steps.back().Point.simplifiedMODE = "climb";
    } else if (Steps.back().modeStep == "change_speed_to_CAS") {
        Steps.back().Point.simplifiedMODE = "accelerate";
    } else if (Steps.back().modeStep == "change_speed_to_Mach") {
        Steps.back().Point.simplifiedMODE = "accelerate";
    } else if (Steps.back().modeStep == "change_flight_level") {
        Steps.back().Point.simplifiedMODE = "climb";
    } else if (Steps.back().modeStep == "change_flight_level_constant_ROC") {
        Steps.back().Point.simplifiedMODE = "climb";
    } else if (Steps.back().modeStep == "descend_to_approach") {
        Steps.back().Point.simplifiedMODE = "descend";
    } else if (Steps.back().modeStep == "approach") {
        Steps.back().Point.simplifiedMODE = "descend";
    } else if (Steps.back().modeStep == "descend") {
        Steps.back().Point.simplifiedMODE = "descend";
    } else if (Steps.back().modeStep == "change_speed") {
        Steps.back().Point.simplifiedMODE = "descend";
    } else if (Steps.back().modeStep == "landing") {
        Steps.back().Point.simplifiedMODE = "landing";
    } else if (Steps.back().modeStep == "level_glide_slope_interception") {
        Steps.back().Point.simplifiedMODE = "descend";
    } else {
        myRuntimeInfo->err << "Mode unknown!" << endl;
        myRuntimeInfo->err << Steps.back().modeStep << endl;
        exit(1);
    }
    if (Steps.back().Point.ROC <= 0.) {
        Steps.back().Point.ROC = 250.;
    }
}

void missionDefinition::getTyicalOffTakes() {
    bleedOffTO = Steps.at(0).bleedOffTakes;
    shaftOffTO = Steps.at(0).shaftOffTakes;
    int counter(0);
    for (unsigned int i(0); i <= Steps.size() - 1; i++) {
        if (Steps.at(i).modeStep == "cruise") {
            bleedOffCr += Steps.at(i).bleedOffTakes;
            shaftOffCr += Steps.at(i).shaftOffTakes;
            counter++;
        }
    }
    bleedOffCr = bleedOffCr / counter;
    shaftOffCr = shaftOffCr / counter;
}
