/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

/* The following statements import necessary header files */
#include "ATAXX.h"                                // Import of header file from src/aircraftSystems folder for ATAXX
#include <runtimeInfo/runtimeInfo.h>                          // Import of header file from external runtime library for using runtimeInfo

/* Constructor for ATAXX */
ATAXX::ATAXX(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO),
    refPtX_cockpit(0.),
    refPtX_avionicsBay(0.) {
    ATA = "ATA-XX";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

/* Destructor for ATAXX */
ATAXX::~ATAXX() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void ATAXX::readParameters(const node& config) {
    std::string remainingConsumerConfig = "/module_configuration_file/program_settings/systems_constants/remaining_consumers";
    avBayLocation = config.at("/module_configuration_file/program_settings/systems_constants/avionics/location_of_aviationics_bay/value");
    data_->configuration.checkForValidity(avBayLocation, 0, 2, "avBayLocation");
    percentageBleedAir   = config.at(remainingConsumerConfig + "/percentage_bleed_air/value");
    percentageElectric   = config.at(remainingConsumerConfig + "/percentage_electric/value");
    percentageHydraulic  = config.at(remainingConsumerConfig + "/percentage_hydraulic/value");
    scalingFactorATAXX =  checkBoundaries(config.at(remainingConsumerConfig + "/mass/scaling_factor/value"), 0, true, 1, true);
    percInstrOfATAXX = checkBoundaries(config.at(remainingConsumerConfig + "/mass/distribution/percentage_instrumentation_of_ATAXX/value"), 0, true, 1, true);
    percAutoFlightOfATAXX = checkBoundaries(config.at(remainingConsumerConfig + "/mass/distribution/percentage_auto_flight_of_ATAXX/value"), 0, true, 1, true);
    percNavigOfATAXX = checkBoundaries(config.at(remainingConsumerConfig + "/mass/distribution/percentage_navigation_of_ATAXX/value"), 0, true, 1, true);
    percCommunicOfATAXX = checkBoundaries(config.at(remainingConsumerConfig + "/mass/distribution/percentage_communication_of_ATAXX/value"), 0, true, 1, true);
    if ((percInstrOfATAXX + percAutoFlightOfATAXX + percNavigOfATAXX + percCommunicOfATAXX) != 1.) {
        myRuntimeInfo->err << "Sum of percentages for remaining consumers is not 1. Use scaling factor of remaining consumers instead. Abort program." << std::endl;
        exit(1);
    }
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/remaining_consumers").read(config).value();
}

/* Updating system specific parameters in aircraft xml file */
void ATAXX::updateXML() {
    updateMass("instruments", percInstrOfATAXX * Mass, "ATA31");
    //AutomaticFlightSystem
    updateMass("automatic_flight_system", percAutoFlightOfATAXX * Mass, "ATA22");
    //Navigation
    updateMass("navigation", percNavigOfATAXX * Mass, "ATA34");
    //Communication
    updateMass("communication", percCommunicOfATAXX * Mass, "ATA23");
}

/* Calculation of the system power */
void ATAXX::calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
                                  const std::vector<powerProfile>& allPreviousSystemsHydraulic,
                                  const std::vector<powerProfile>& allPreviousSystemsElectric,
                                  const powerProfile& allPreviousSystemsHeat) {
    initializePowerProfile();
    getPower(allPreviousSystemsBleed, allPreviousSystemsHydraulic, allPreviousSystemsElectric, allPreviousSystemsHeat);
    calculateMaxMissionPower();
}

/* Calculation of the skin mass */
void ATAXX::calculateSinkMass(const powerProfile& sourceBleed,
                              const std::vector<powerProfile>& sourceElectric,
                              const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    /* Calculate Systems Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}

/* Determination of the power */
void ATAXX::getPower(const powerProfile& previousSystemsBleed,
                     const std::vector<powerProfile>& previousSystemsHydraulic,
                     const std::vector<powerProfile>& previousSystemsElectric,
                     const powerProfile& previousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(previousSystemsHeat);
    /* Calculate Systems Power */
    /* Remaining Consumers including AVIONICS */
    /* New estimation for electric from F. Peter 18.6.14 after [Bue14] p. 80 */
//    cout << std::endl;
    bleedAir.maxMissionPower  = percentageBleedAir * previousSystemsBleed.maxMissionPower;
    /*Loop over electric circuits*/
    for (size_t n(0); n < Electric.size(); n++) {
        Electric.at(n).maxMissionPower  = percentageElectric * previousSystemsElectric.at(n).maxMissionPower;
//        cout << "myElectric->at(n).maxMissionPower is: " << myElectric->maxMissionPower << std::endl;
    }
    /*Loop over hydraulic circuits*/
    for (size_t n(0); n < Hydraulic.size(); n++) {
        Hydraulic.at(n).maxMissionPower = percentageHydraulic * previousSystemsHydraulic.at(n).maxMissionPower;
    }
    bleedAir.designPower  = percentageBleedAir * previousSystemsBleed.designPower;
    /*Loop over electric circuits*/
    for (size_t n(0); n < Electric.size(); n++) {
        Electric.at(n).designPower  = percentageElectric * previousSystemsElectric.at(n).designPower;
//        cout << "myElectric->at(n).designPower is: " << myElectric->designPower << std::endl;
    }
    /*Loop over hydraulic circuits*/
    for (size_t n(0); n < Hydraulic.size(); n++) {
        Hydraulic.at(n).designPower = percentageHydraulic * previousSystemsHydraulic.at(n).designPower / static_cast<double>(Hydraulic.size());
    }
    /*Loop over electric circuits*/
    for (size_t n(0); n < Electric.size(); n++) {
        for (unsigned i = 0; i < data_->mission.Steps.size(); i++) {
            Electric.at(n).missionPower.at(i).baseLoad = percentageElectric * previousSystemsElectric.at(n).missionPower.at(i).baseLoad / static_cast<double>(Electric.size());
//            cout << "myElectric->at(n).missionPower.at(i).baseLoad is: " << myElectric->missionPower.at(i).baseLoad << std::endl;
        }
    }
    /*Loop over hydraulic circuits*/
    for (size_t n(0); n < Hydraulic.size(); n++) {
        for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
            Hydraulic.at(n).missionPower.at(i).baseLoad = percentageHydraulic * previousSystemsHydraulic.at(n).missionPower.at(i).baseLoad;
        }
    }
    /*For heatLoad electrical powerProfile is accumulated*/
    powerProfile myElectricSum(getPowerProfileSum(Electric));
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        bleedAir.missionPower.at(i).baseLoad = percentageBleedAir * previousSystemsBleed.missionPower.at(i).baseLoad;
        heatLoad.missionPower.at(i).baseLoad = myElectricSum.missionPower.at(i).baseLoad;
    }
}

/* Determination of the mass */
double ATAXX::getMass() {
    /* From [Tor13], S. 289 (pdf S. 308) -> method requires the maximum range, but at this point only the DesignRange in MICADO is known */
    double designRange = data_->data.Performance.designRange / 1000; // conversion from m to km
    return  static_cast<double>(scalingFactorATAXX * 0.347 * pow(data_->data.Mass.OME, 0.556) * pow(designRange, 0.25));
}

/* Determination of the reference point */
Vec3 ATAXX::getRefPoint() {
    /* Instruments: Placement at the end of the cockpit segment */
    this->refPtX_cockpit = 0.0;
    this->refPtX_cockpit = data_->data.Accommodation.payload_tube_x_ref;
    /* Auto Flight System, Navigation & Communication: Placement in Avionics Bay */
    if (avBayLocation == 0) {
        this->refPtX_avionicsBay = data_->theWing.at(0).aerodynamicSurface.sections.front().origin[0] -
                                   0.25 * data_->theWing.at(0).aerodynamicSurface.sections.front().get_chord_length(); // Initial method
    } else if (avBayLocation == 1) {
        this->refPtX_avionicsBay = this->refPtX_cockpit * 0.25; // Location of the avionics bay in the nose for business-jet-like AC; approximation: 25% of the cockpit length
    } else if (avBayLocation == 2) {
        this->refPtX_avionicsBay = this->refPtX_cockpit;    // Location of the avionics bay behind the cockpit segment for A320-like AC
    } else {
        myRuntimeInfo->err << "Error: invalid location of the avionics bay. Program exit!" << std::endl;
        exit(1);
    }
    /* Determination of the total CG point in x-direction */
    double refX_Pt_tmp = percInstrOfATAXX * this->refPtX_cockpit + (percAutoFlightOfATAXX + percNavigOfATAXX + percCommunicOfATAXX)* this->refPtX_avionicsBay;
    return Vec3(data_->theFuselage.back().origin[0] + refX_Pt_tmp, data_->theFuselage.back().origin[1], data_->theFuselage.back().origin[2]);
}
