/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/

/* The following statements import necessary header files */
#include "conventionalATA26.h"                                       // Import of header file from src/aircraftSystems folder for conventionalATA26
#include <runtimeInfo/runtimeInfo.h>                                 // Import of header file from external runtime library for using runtimeInfo

/* Constructor for conventionalATA26 */
conventionalATA26::conventionalATA26(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO) {
    ATA = "ATA-26";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

/* Destructor for conventionalATA26 */
conventionalATA26::~conventionalATA26() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA26::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/fire_protection_system").read(config).value();
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA26::updateXML() {
    updateMass("fire_protection", Mass, ATA);
}

/* Calculation of the system power */
void conventionalATA26::calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
        const std::vector<powerProfile>& allPreviousSystemsHydraulic,
        const std::vector<powerProfile>& allPreviousSystemsElectric,
        const powerProfile& allPreviousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(allPreviousSystemsBleed);
    UNUSED(allPreviousSystemsHydraulic);
    UNUSED(allPreviousSystemsElectric);
    UNUSED(allPreviousSystemsHeat);
    /* Calculate Systems Power */
    initializePowerProfile();
    if (Operating == true) {
        getPower();
    }
    calculateMaxMissionPower();
}

/* Calculation of the skin mass */
void conventionalATA26::calculateSinkMass(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    /* Calculate Systems Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}

/* Determination of the power */
void conventionalATA26::getPower() {
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        bleedAir.missionPower.at(i).baseLoad     = 0.0;
        for (size_t n(0); n < Hydraulic.size(); n++) {
            Hydraulic.at(n).missionPower.at(i).baseLoad    = 0.0;
        }
        for (size_t n(0); n < Electric.size(); n++) {
            Electric.at(n).missionPower.at(i).baseLoad    = 0.0;
        }
        heatLoad.missionPower.at(i).baseLoad     = 0.0;
    }
}

/* Determination of the mass */
double conventionalATA26::getMass() {
    // From [Tor13] Tab. 8-12 (Furnishings), page 291 (pdf page 310)
    /* Fire Detection and Extinguishing System */
    double fireDetectionMass;
    // method is based on engine type of the first propulsor
    std::string propulsionType = data_->data.Propulsion.Propulsor.at(0).type;
    if (propulsionType == "FAN" || propulsionType == "JET" || propulsionType == "GTF") {
        fireDetectionMass = 0.0012 * data_->data.Mass.MTOM;
    } else {
        fireDetectionMass = 0.0030 * data_->data.Mass.MTOM;
        myRuntimeInfo->info << "Since propulsion type is not FAN, JET, or GTF, ATA26 mass method for turbo prop is used." << std::endl;
    }
    return fireDetectionMass;
}

/* Determination of the reference point */
Vec3 conventionalATA26::getRefPoint() {
    /* Assumption: center of gravity at 45% fuselage length */
    double fuselage_length = geom2::measure::length(data_->theFuselage.back());
    return Vec3(data_->theFuselage.back().origin[0] + 0.45 * fuselage_length, data_->theFuselage.back().origin[1], data_->theFuselage.back().origin[2]);
}
