/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "conventionalATA33.h"

#include <runtimeInfo/runtimeInfo.h>

conventionalATA33::conventionalATA33(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO)
    :
    aircraftSystem(systemName, systemIsOperating, data, rtIO),
    wingLightPowerSW(false),
    RTOLightPowerSW(false),
    taxiLightPowerSW(false),
    landingLightPowerSW(false),
    logoLightPowerSW(false),
    strobesLightPowerSW(false),
    cabinLightPowerSW(false),
    cabinLightPower(0.) {
    ATA = "ATA-33";
    myRuntimeInfo->out << "  Create: " << Name << "  ->  " << ATA << std::endl;
    //ctor
}

conventionalATA33::~conventionalATA33() {
    //dtor
}

/* Reading system specific parameters of configuration xml file */
void conventionalATA33::readParameters(const node& config) {
    scaling_factor = EndnodeReadOnly<double>("/module_configuration_file/program_settings/scaling_factors/lighting_system").read(config).value();
    std::string lightsConfig = "module_configuration_file/program_settings/systems_constants";
    navigationLightPower   = config.at(lightsConfig + "/navigation_light_power/value");
    data_->configuration.checkForValidity(navigationLightPower, 10.0, 1000.0, "Lights.navigationLightPower");
    rotatingBeaconLightPower   = config.at(lightsConfig + "/rotating_beacon_light_power/value");
    data_->configuration.checkForValidity(rotatingBeaconLightPower, 10.0, 1000.0, "Lights.rotatingBeaconLightPower");
    wingLightPower   = config.at(lightsConfig + "/wing_light_power/value");
    data_->configuration.checkForValidity(wingLightPower, 10.0, 5000.0, "Lights.wingLightPower");
    RTOLightPower   = config.at(lightsConfig + "/rto_light_power/value");
    data_->configuration.checkForValidity(RTOLightPower, 10.0, 1000.0, "Lights.RTOLightPower");
    taxiLightPower   = config.at(lightsConfig + "/taxi_light_power/value");
    data_->configuration.checkForValidity(taxiLightPower, 10.0, 5000.0, "Lights.taxiLightPower");
    landingLightPower   = config.at(lightsConfig + "/landing_light_power/value");
    data_->configuration.checkForValidity(landingLightPower, 100.0, 5000.0, "Lights.landingLightPower");
    logoLightPower   = config.at(lightsConfig + "/logo_light_power/value");
    data_->configuration.checkForValidity(logoLightPower, 10.0, 1000, "Lights.logoLightPower");
    strobesLightPower   = config.at(lightsConfig + "/strobe_light_power/value");
    data_->configuration.checkForValidity(strobesLightPower, 10.0, 1000, "Lights.strobesLightPower");
    specificEmergencyLightPower   = config.at(lightsConfig + "/specific_emergency_light_power/value");
    data_->configuration.checkForValidity(specificEmergencyLightPower, 0.1, 50.0, "Lights.specificEmergencyLightPower");
    specificCabinLightPower   = config.at(lightsConfig + "/specific_cabin_light_power/value");
    data_->configuration.checkForValidity(specificCabinLightPower, 0.1, 50.0, "Lights.specificCabinLightPower");
    flightDeckLightPower   = config.at(lightsConfig + "/flight_deck_light_power/value");
    data_->configuration.checkForValidity(flightDeckLightPower, 100.0, 5000.0, "Lights.flightDeckLightPower");
    numberOfPowerSources = config.at(lightsConfig + "/shaft_power_sources/number_of_power_sources/value");
    double operationFactor_gesamt(0.);
    for (int i(0); i < numberOfPowerSources; i++) {
        powerSources.push_back(PowerSource());
        powerSources.back().type =
            std::string(config.at(lightsConfig + "/shaft_power_sources/power_source@"
                                     + num2Str(i) + "/type/value"));
        powerSources.back().sourceID =
            config.at(lightsConfig + "/shaft_power_sources/power_source@"
                         + num2Str(i) + "/source_ID/value");
        powerSources.back().operationFactor =
            config.at(lightsConfig + "/shaft_power_sources/power_source@"
                         + num2Str(i) + "/operation_factor/value");
        operationFactor_gesamt += Rounding(powerSources.back().operationFactor, 10);
    }
    if (operationFactor_gesamt < 1.) {
        myRuntimeInfo->err << "The work distribution of the \"PowerSources\" for the lighting results together in less than 100% ("
                           << operationFactor_gesamt * 100 << "%). adjust in config. Abort!" <<
                           std::endl;
        exit(1);
    } else if (operationFactor_gesamt > 1.) {
        myRuntimeInfo->err << "The work distribution of the \"PowerSources\" for the lighting results together in more than 100% ("
                           << operationFactor_gesamt * 100 << "%). adjust in config. Abort!" <<
                           std::endl;
        exit(1);
    }
}

/* Updating system specific parameters in aircraft xml file */
void conventionalATA33::updateXML() {
    updateMass("lighting", Mass, ATA);
}

void conventionalATA33::calculateSystemsPower(const powerProfile& allPreviousSystemsBleed,
        const std::vector<powerProfile>& allPreviousSystemsHydraulic,
        const std::vector<powerProfile>& allPreviousSystemsElectric,
        const powerProfile& allPreviousSystemsHeat) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(allPreviousSystemsBleed);
    UNUSED(allPreviousSystemsHydraulic);
    UNUSED(allPreviousSystemsElectric);
    UNUSED(allPreviousSystemsHeat);
    /* Calculate Systems Power */
    initializePowerProfile();
    if (Operating == true) {
        getPower();
    }
    calculateMaxMissionPower();
}
void conventionalATA33::calculateSinkMass(const powerProfile& sourceBleed,
        const std::vector<powerProfile>& sourceElectric,
        const std::vector<powerProfile>& sourceHydraulic) {
    /* Define unused function parameters to avoid linter warnings */
    UNUSED(sourceBleed);
    UNUSED(sourceElectric);
    UNUSED(sourceHydraulic);
    /* Calculate Systems Mass */
    Mass = scaling_factor * getMass();
    refPoint = getRefPoint();
}
double conventionalATA33::getMissionSegmentLightPower(double *myLightHeatLoad) {
    double theLightPower(0.);
    if (wingLightPowerSW)
        theLightPower += wingLightPower;// cout << "wingLightPowerSW is on; ";
    if (RTOLightPowerSW)
        theLightPower += RTOLightPower;
    if (taxiLightPowerSW)
        theLightPower += taxiLightPower;
    if (landingLightPowerSW)
        theLightPower += landingLightPower;
    if (logoLightPowerSW)
        theLightPower += logoLightPower;
    if (strobesLightPowerSW)
        theLightPower += strobesLightPower;
    if (cabinLightPowerSW) {
        theLightPower += cabinLightPower;
        *myLightHeatLoad = cabinLightPower;
    }
    return theLightPower;
}

void conventionalATA33::getPower() {
//    vector<double> lenghtOfCabin=data_->data.Accommodation.lengthDeck;
    /* ATA33 no consumption of bleed or hydraulic **/
    double permanentExterior(navigationLightPower + rotatingBeaconLightPower);
    double permanentInterior(specificEmergencyLightPower * data_->data.Accommodation.cabinVolumeTotal + flightDeckLightPower);
    cabinLightPower = specificCabinLightPower * data_->data.Accommodation.cabinVolumeTotal;
    for (size_t n(0); n < Electric.size(); n++) {
        double designPower_tmp(0.);
        designPower_tmp =    permanentExterior
                             + wingLightPower
                             + RTOLightPower
                             + taxiLightPower
                             + landingLightPower
                             + logoLightPower
                             + strobesLightPower
                             + permanentInterior
                             + cabinLightPower;
        for (size_t j(0); j < powerSources.size(); j++) {
            double operationFactor_tmp(powerSources.at(j).operationFactor);
            if (powerSources.at(j).type == "Electric" && powerSources.at(j).sourceID == static_cast<int>(n + 1)) {
                Electric.at(n).designPower += designPower_tmp * operationFactor_tmp;
            }
        }
    }
    for (size_t i = 0; i < data_->mission.Steps.size(); i++) {
        wingLightPowerSW = false;
        RTOLightPowerSW = false;
        taxiLightPowerSW = false;
        landingLightPowerSW = false;
        logoLightPowerSW = false;
        strobesLightPowerSW = false;
        cabinLightPowerSW = false;
        for (size_t n(0); n < Electric.size(); n++) {
            Electric.at(n).missionPower.at(i).baseLoad = 0.0;
        }
        heatLoad.missionPower.at(i).baseLoad = 0.0;
//        cout << std::endl << "The myMissionPt->Steps.at(" << i << ").missionPhase is: " << myMissionPt->Steps.at(i).missionPhase << std::endl;
        if (data_->mission.Steps.at(i).missionPhase == "Departure") {
            if (data_->mission.Steps.at(i).modeStep == "Takeoff") {
                wingLightPowerSW = true;
                RTOLightPowerSW = false;
                taxiLightPowerSW = true;
                landingLightPowerSW = true;
                logoLightPowerSW = true;
                strobesLightPowerSW = true;
                cabinLightPowerSW = false;
            } else if (data_->mission.Steps.at(i).modeStep == "Accelerate" || convertUnit(METER, FOOT, data_->mission.Steps.at(i).endAltitude) < 10000.) {
                wingLightPowerSW = true;
                RTOLightPowerSW = false;
                taxiLightPowerSW = false;
                landingLightPowerSW = true;
                logoLightPowerSW = true;
                strobesLightPowerSW = true;
                cabinLightPowerSW = true;
            } else {
                wingLightPowerSW = true;
                RTOLightPowerSW = false;
                taxiLightPowerSW = false;
                landingLightPowerSW = false;
                logoLightPowerSW = false;
                strobesLightPowerSW = false;
                cabinLightPowerSW = true;
            }
        } else if (data_->mission.Steps.at(i).missionPhase == "Cruise") {
            wingLightPowerSW = true;
            RTOLightPowerSW = false;
            taxiLightPowerSW = false;
            landingLightPowerSW = false;
            logoLightPowerSW = false;
            strobesLightPowerSW = false;
            cabinLightPowerSW = true;
        } else if (data_->mission.Steps.at(i).missionPhase == "Approach") {
            if (data_->mission.Steps.at(i).configuration == "Clean") {
                wingLightPowerSW = true;
                RTOLightPowerSW = false;
                taxiLightPowerSW = false;
                landingLightPowerSW = false;
                logoLightPowerSW = false;
                strobesLightPowerSW = false;
                cabinLightPowerSW = true;
            } else if (data_->mission.Steps.at(i).modeStep != "Landing") {
                wingLightPowerSW = true;
                RTOLightPowerSW = false;
                taxiLightPowerSW = false;
                landingLightPowerSW = true;
                logoLightPowerSW = true;
                strobesLightPowerSW = true;
                cabinLightPowerSW = true;
            } else {
                wingLightPowerSW = true;
                RTOLightPowerSW = false;
                taxiLightPowerSW = true;
                landingLightPowerSW = true;
                logoLightPowerSW = true;
                strobesLightPowerSW = true;
                cabinLightPowerSW = false;
            }
        } else {
            myRuntimeInfo->err << "Unknown mission phase in electrical energy calculation for Lighting. Program abort!" << std::endl;
            myRuntimeInfo->err << "mission phase: " << data_->mission.Steps.at(i).missionPhase << std::endl;
            exit(1);
        }
        for (size_t n(0); n < Electric.size(); n++) {
            double baseLoad_tmp(0.);
            baseLoad_tmp = permanentExterior + permanentInterior + getMissionSegmentLightPower(&heatLoad.missionPower.at(i).baseLoad);
            for (size_t j(0); j < powerSources.size(); j++) {
                double operationFactor_tmp(powerSources.at(j).operationFactor);
                if (powerSources.at(j).type == "Electric" && powerSources.at(j).sourceID == static_cast<int>(n + 1)) {
                    Electric.at(n).missionPower.at(i).baseLoad += baseLoad_tmp * operationFactor_tmp;
                }
            }
//            cout << "myElectric->missionPower.at(i).baseLoad is: " << myElectric->at(n).missionPower.at(i).baseLoad << std::endl;
        }
        heatLoad.missionPower.at(i).baseLoad += permanentInterior;
//        cout << "myHeatLoad->missionPower.at(i).baseLoad is: " << myHeatLoad->missionPower.at(i).baseLoad << std::endl;
    }
}

double conventionalATA33::getMass() {
    /* Source [Sch04]: mass of ATA33 is about. 2% of the total system mass **/
    /* Source [Sch04]: system weight depends on MTOM**/
    return 0.02 * 0.92 * pow(data_->data.Mass.MTOM, 0.85);
    //Info: A320 Landing Lights weights according to honeywell = 2* 7.484kg
    //Info: A320 LED Navigation Lights Goodrich Lighting Systems (green/red) = 2* 0.408kg
    //Info: A320 Logo Lights weights according to honeywell = 2* 1.4kg
}

Vec3 conventionalATA33::getRefPoint() {
    /* Assumption: center of gravity at 45% fuselage length **/
    double fuselage_length = geom2::measure::length(data_->theFuselage.back());
    return Vec3(data_->theFuselage.back().origin[0] + 0.45 * fuselage_length, data_->theFuselage.back().origin[1], data_->theFuselage.back().origin[2]);
}
