/*  Copyright (C) 2011-2022 Institute of Aerospace Systems, RWTH Aachen University - All rights reserved.
    This file is part of UNICADO.

    UNICADO is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    UNICADO is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with UNICADO.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA70_H_
#define SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA70_H_

/* The following statements import necessary libraries */
#include <string>                                                                  // Import c++ standard library for using string
#include <vector>                                                                  // Import c++ standard library for using vector
#include <memory>                                                                  // Import c++ standard library for using memory

/* The following statements import necessary header files */
#include <svl/SVL.h>                                                               // Import of header file from src folder for SVL
#include "../aircraftSystem.h"                                                     // Import of header file from src folder for aircraft system
#include "../powerProfile.h"                                                       // Import of header file from src folder for power profile
#include "../systemsIOData.h"                                                      // Import of header file from systemsIOData for accessing data_. Can't be included in
                                                                                   // aircraftSystem.h because that would lead to circular dependencies


/** \class conventionalATA70
 * \brief class that contains all properties of the engine system (ATA70)
 */
class conventionalATA70: public aircraftSystem {
 public:
    double percentageBleedAir;/**< [-] percentage of bleed air provided by the engines*/
    double percentageElectricPower;/**< [-] percentage of electrical power provided by the engines*/
    double percentageHydraulicPower;/**< [-] percentage of hydraulic power provided by the engines*/

    double efficiencyFactorBleedAir;/**< [-] efficiency factor of the bleed air system */
    /**
     * \brief function for reading the system specific parameters from the config
     * 
     * \param config const node&: systemsDesign configuration file
    */
    void readParameters(const node& config) override;
    /**
     * \brief function for updating the system specific values in the aircraft xml
    */
    void updateXML() override;
    void calculateSourcePower(const powerProfile& sourceBleed,
                              const std::vector<powerProfile>& sourceElectric,
                              const std::vector<powerProfile>& sourceHydraulic) override;
    void calculateSourceMass(const powerProfile& sourceBleed,
                             const std::vector<powerProfile>& sourceElectric,
                             const std::vector<powerProfile>& sourceHydraulic) override;

    /** \brief calculates the power of the system (ATA70)
     * \param theBleed powerProfile
     * \param theElectric vector<powerProfile>
     * \param theHydraulic vector<powerProfile>
     * \return void
     */
    void getPower(powerProfile theBleed, std::vector<powerProfile> theElectric, std::vector<powerProfile> theHydraulic);

    /** \brief calculates the mass of the system (ATA70)
     * \return double: mass of the system [kg]
     */
    double getMass();

    /** \brief calculates the reference point of the system (ATA70)
     * \param forwardCenterGravity Vec3: forward center of gravity in global x,y,z-coordinates [m]
     * \param aftCenterGravity Vec3: aft center of gravity in global x,y,z-coordinates [m]
     * \return Vec3: reference point of the system in global x,y,z-coordinates [m]
     */
    Vec3 getRefPoint(Vec3 forwardCenterGravity, Vec3 aftCenterGravity);

    /** \brief Constructor
     * \param systemName string
     * \param systemIsOperating bool: switch whether system is operating (1) or not (0)
     * \param theSettings const systemsDesignSettings&
     * \param myAircraft aircraft
     */
    conventionalATA70(const std::string& systemName, bool systemIsOperating, std::shared_ptr<systemsIOData> data, const RuntimeIO& rtIO);
    virtual ~conventionalATA70();

 private:
    /** \brief negates all values in the power profile
     * \param Power powerProfile*: pointer to the power profile
     * \return void
     */
    void powerNegation(powerProfile *Power);
};

#endif // SYSTEMS_DESIGN_SRC_AIRCRAFTSYSTEMS_CONVENTIONALATA70_H_
