/**
 * \file standardConfig.cpp
 * \author 
 * \brief standard systems design configuration class for reading from standardConf.xml 
 * \version 0.1
 * \date 2024-06-27
 * 
 * @copyright Copyright (c) 2024
 * 
 */
#include "standardConfig.h"
#include <algorithm>
#include <aixml/node.h>
#include <standardFiles/functions.h>
#include <unitConversion/unitConversion.h>

/* Methods for reading aircraft systems section of config */

void standardConfig::readEnergySinks(const node& config) {
    int numberSystems(0);
    numberSystems = config.at("/module_configuration_file/program_settings/aircraft_systems/energy_sinks/number_of_sinks/value");
    for (int i = 1; i <= numberSystems; i++) {
        Sinks.push_back(systemDescription());
        Sinks.back().systemName = std::string(config.at("/module_configuration_file/program_settings/aircraft_systems/energy_sinks/system@" + num2Str(i-1) +
                                                        "/system_description/value"));
        Sinks.back().systemOperating =
            EndnodeReadOnly<bool>("/module_configuration_file/program_settings/aircraft_systems/energy_sinks/system@" + num2Str(i-1)+ "/operating_switch").read(config).value();
    }
}

void standardConfig::readEnergySources(const node& config) {
    int numberSystems(0);
    numberSystems = config.at("/module_configuration_file/program_settings/aircraft_systems/energy_sources/number_of_sources/value");
    for (int i = 1; i <= numberSystems; i++) {
        Sources.push_back(systemDescription());
        Sources.back().systemName = std::string(config.at("/module_configuration_file/program_settings/aircraft_systems/energy_sources/system@"
                                                          + num2Str(i-1) + "/system_description/value"));
        Sources.back().systemOperating = EndnodeReadOnly<bool>("/module_configuration_file/program_settings/aircraft_systems/energy_sources/system@"
                                                   + num2Str(i-1) + "/operating_switch").read(config).value();
    }
}
void standardConfig::readECS(const node& config) {
    theECS.systemName = std::string(config.at("/module_configuration_file/program_settings/aircraft_systems/environmental_control_system/system_description/value"));
    theECS.systemOperating = EndnodeReadOnly<bool>("/module_configuration_file/program_settings/aircraft_systems/environmental_control_system/operating_switch")
                                                   .read(config).value();
}
void standardConfig::readEnergyConductors(const node& config) {
    int numberSystems(0);
    numberSystems = config.at("/module_configuration_file/program_settings/aircraft_systems/energy_conductors/number_of_conductors/value");
    for (int i = 1; i <= numberSystems; i++) {
        Conductors.push_back(systemDescription());
        Conductors.back().systemName = std::string(config.at("/module_configuration_file/program_settings/aircraft_systems/energy_conductors/system@"
                                                             + num2Str(i-1) + "/system_description/value"));
        Conductors.back().systemOperating = EndnodeReadOnly<bool>("/module_configuration_file/program_settings/aircraft_systems/energy_conductors/system@"
                                                      + num2Str(i-1) + "/operating_switch").read(config).value();
    }
}
void standardConfig::readVirtualSystems(const node& config) {
    /* read system name, operating status, mass, reference point and power demand for each virtual system */
    int numberSystems(0);
    numberSystems = config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/number_of_virtual_systems/value");
    for (int i = 1; i <= numberSystems; i++) {
        virtualSystem.push_back(virtSystemProperties());
        virtualSystem.back().systemName         = std::string(config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                                                                        + num2Str(i-1) + "/system_description/value"));
        virtualSystem.back().systemOperating    =
            EndnodeReadOnly<bool>("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/operating_switch").read(config).value();
        virtualSystem.back().Mass               =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/mass/value");
        virtualSystem.back().refPoint[0]        =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/center_of_gravity/x/value");
        virtualSystem.back().refPoint[1]        =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/center_of_gravity/y/value");
        virtualSystem.back().refPoint[2]        =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/center_of_gravity/z/value");
        virtualSystem.back().powerBleed[0]      =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/departure/bleed_air/value");
        virtualSystem.back().powerHydraulic[0]  =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/departure/hydraulic/value");
        virtualSystem.back().powerElectric[0]   =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/departure/electric/value");
        virtualSystem.back().heatLoad[0]        =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/departure/heat_load/value");
        virtualSystem.back().powerBleed[1]      =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/cruise/bleed_air/value");
        virtualSystem.back().powerHydraulic[1]  =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/cruise/hydraulic/value");
        virtualSystem.back().powerElectric[1]   =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/cruise/electric/value");
        virtualSystem.back().heatLoad[1]        =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/cruise/heat_load/value");
        virtualSystem.back().powerBleed[2]      =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/approach/bleed_air/value");
        virtualSystem.back().powerHydraulic[2]  =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/approach/hydraulic/value");
        virtualSystem.back().powerElectric[2]   =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/approach/electric/value");
        virtualSystem.back().heatLoad[2]        =
            config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/system@"
                         + num2Str(i-1) + "/power_demand/approach/heat_load/value");
        /* check that values for mass and energy consumtion are positive */
        if (virtualSystem.back().Mass < 0.) {
            myRuntimeInfo->err << "A negative mass was assigned to the virtual system named: " << virtualSystem.back().systemName << std::endl;
        }
        if (virtualSystem.back().powerBleed[0] < 0.
                || virtualSystem.back().powerHydraulic[0] < 0.
                || virtualSystem.back().powerElectric[0] < 0.
                || virtualSystem.back().heatLoad[0] < 0.
                || virtualSystem.back().powerBleed[1] < 0.
                || virtualSystem.back().powerHydraulic[1] < 0.
                || virtualSystem.back().powerElectric[1] < 0.
                || virtualSystem.back().heatLoad[1] < 0.
                || virtualSystem.back().powerBleed[2] < 0.
                || virtualSystem.back().powerHydraulic[2] < 0.
                || virtualSystem.back().powerElectric[2] < 0.
                || virtualSystem.back().heatLoad[2] < 0.) {
            myRuntimeInfo->err << "A negative energy consumption was assigned to the virtual system named: "
                               << virtualSystem.back().systemName << std::endl;
        }
    }
}

/* Methods for reading program settings */
void standardConfig::readProgramSettings(const node& config) {
    texReportWithoutMasses = EndnodeReadOnly<bool>("/module_configuration_file/program_settings/tex_report_without_masses").read(config).value();
    controlDeviceWarning = EndnodeReadOnly<bool>("/module_configuration_file/program_settings/control_device_warning").read(config).value();
    if (designCase == true) {
        theProgramName = std::string(TOOL_NAME) + "_design";
    } else if (reqMission == true) {
        theProgramName = std::string(TOOL_NAME) + "_requirment";
    } else {
        theProgramName = std::string(TOOL_NAME) + "_study";
    }
    /* Operator items scaling factor (other scaling factors are read by respective system) */
    OpItemsFactor = config.at("/module_configuration_file/program_settings/scaling_factors/operator_items/value");
    /* Number of hydraulic and electric circuits is needed for initializing aircraft system */
    Hydraulic.numberOfhydraulicCircuits = config.at("/module_configuration_file/program_settings/systems_constants/hydraulic_system/hydraulic_circuits/number_of_circuits/value");
    Electric.numberOfElectricCircuits = config.at("/module_configuration_file/program_settings/systems_constants/electric_system/electric_circuits/number_of_circuits/value");
    numberOfVirtualSystems = config.at("/module_configuration_file/program_settings/aircraft_systems/virtual_systems/number_of_virtual_systems/value");
    numberOfConductors = config.at("/module_configuration_file/program_settings/aircraft_systems/energy_conductors/number_of_conductors/value");
    numberOfSources = config.at("/module_configuration_file/program_settings/aircraft_systems/energy_sources/number_of_sources/value");
}

/* Methods for checking validity of values */

void standardConfig::checkForValidity(double Value, double lowerBound, double upperBound, const std::string &Text) {
    if (std::isnan(Value)) { // Test for NaN
        myRuntimeInfo->err << "Value is NaN!" << std::endl;
        myRuntimeInfo->err << "Parameter from:" << std::endl;
        myRuntimeInfo->err << Text << std::endl;
        myRuntimeInfo->err << "Abort!" << std::endl;
        exit(1);
    }
    if (std::isinf(Value)) { // Test for real number
        myRuntimeInfo->err << "Value is Infinite!" << std::endl;
        myRuntimeInfo->err << "Parameter from:" << std::endl;
        myRuntimeInfo->err << Text << std::endl;
        myRuntimeInfo->err << "Abort!" << std::endl;
        exit(1);
    }
    if (Value > upperBound || Value < lowerBound) {
        myRuntimeInfo->err << "Invalid value range!" << std::endl;
        myRuntimeInfo->err << "Parameter from:" << std::endl;
        myRuntimeInfo->err << Text << std::endl;
        myRuntimeInfo->err << "With the value:" << std::endl;
        myRuntimeInfo->err << Value << " (allowed is " << lowerBound << " <= X <= " << upperBound << ")" << std::endl;
        myRuntimeInfo->err << "Abort!" << std::endl;
        exit(1);
    }
}
