# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

"""Module containing functions for the validity check of the tank configuration."""
# Import standard libraries.
import sys


def check_tank_configuration(paths_and_names, dict_ac_data, dict_tank_design, runtime_output):
    """Check for valid tank configuration in aircraft exchange file.
    
    Checks the validity of the tank configuration in the aircraft exchange file based on the energy carrier type.
    Depending on the configuration, the tank design dictionary is updated with the number of tanks, tank configuration
    type, and validity status.

    :param dict paths_and_names: Dictionary containing system paths and ElementTrees
    :param dict dict_ac_data: Dict containing parameter and values from aircraft exchange and module configuration file
    :param dict dict_tank_design: Dictionary containing tank design parameter
    :param logging.Logger runtime_output: Logging object used for capturing log messages in the module
    :raises KeyError: Raised if key not found in 'combinations' dictionary (invalid tank location/position combination)
    :return dict dict_tank_design: Dictionary containing tank design parameter
    """
    # Extract values from 'dict_ac_data' and 'dict_tank_design'.
    energy_carrier = dict_ac_data['energy_carrier']
    tank_entity_list = [key for key in dict_tank_design if key.startswith('tank_') and key[5:].isdigit()]
    wing_mounting = dict_ac_data['wing_mounting']
    number_of_tanks = 0

    """Kerosene."""
    if energy_carrier == 'kerosene':
        # Define tank combination dict based on energy carrier and wing mounting.
        if wing_mounting == 'low':
            combinations = {
                # Standard integral wing tanks.
                'wing': 
                    {'inner_left': False,
                    'outer_left': False,
                    'inner_right': False,
                    'outer_right': False,
                    'center': False},
                # Trim tank.
                'horizontal_stabilizer':
                    {'total': False},
                # Additional center tank.
                'fuselage':
                    {'center': False}
            }
        elif wing_mounting == 'high':
            combinations = {
                # Standard integral wing tanks.
                'wing': 
                    {'inner_left': False,
                    'inner_right': False,
                    'center': False},
                # Trim tank.
                'horizontal_stabilizer':
                    {'total': False},
                # Additional center tank.
                'fuselage':
                    {'center': False}
            }
        else:
            # If wing mounting is not high or low, wing definition is invalid. Abort program.
            runtime_output.critical('Error: Wing mounting not "high" or "low". Invalid wing definition. '
                                    + 'Program aborted.')
            sys.exit('Exit information: Invalid wing definition in aircraft exchange file.')

        # Iterate over tank entities of 'tank_entity_list'.
        for tank_entity in tank_entity_list:
            # Temporary variables to shorten lines.
            tmp_location = dict_tank_design[tank_entity]['tank_location']
            tmp_position = dict_tank_design[tank_entity]['tank_position']

            # Test if current combination of tank location and position is available in 'combinations' dictionary.
            try:
                tmp_combination = combinations[tmp_location][tmp_position]

                if tmp_combination is False:
                    # If value is 'False', set value to 'True'.
                    combinations[tmp_location][tmp_position] = True
                else:
                    # If value is 'True', combination already exists and tank definition is invalid. Abort program.
                    runtime_output.critical('Error: Tank "' + str(tank_entity) + '" with location "'
                                            + str(tmp_location) + '" and position "' + str(tmp_position)
                                            + '" already defined. ' + 'Invalid tank definition. Program aborted.')
                    sys.exit('Exit information: Invalid tank definition in aircraft exchange file.')

            # Raise KeyError if combination of tank location and position is not available in 'combinations' dict.
            except KeyError:
                location_exists = tmp_location in combinations
                position_exists = tmp_position in combinations[tmp_location] if location_exists else False
                if location_exists and not position_exists:
                    tmp_str = f"Tank location '{tmp_location}' is valid, but position '{tmp_position}' is not. "
                else:
                    tmp_str = f"Tank location '{tmp_location}' is invalid. "
                runtime_output.critical('Error: KeyError found in "check_tank_configuration" function. '
                                        + tmp_str + 'Program aborted.')
                sys.exit('Exit information: Invalid tank definition in aircraft exchange file.')

        # Valid combinations.
        wing_positions = list(combinations['wing'].keys())
        valid_wing_tanks = all(combinations['wing'][key] for key in wing_positions)
        trim_tank = combinations['horizontal_stabilizer']['total']
        additional_center_tank = combinations['fuselage']['center']
        valid_wing_with_trim_tank = valid_wing_tanks and trim_tank
        valid_wing_with_additional_center_tank = valid_wing_tanks and additional_center_tank
        valid_wing_with_additional_center_and_trim_tank = valid_wing_with_trim_tank and additional_center_tank

        # Define number of tanks and set tank configuration name for valid tank configurations.
        if valid_wing_tanks and not (trim_tank or additional_center_tank):
            if wing_mounting == 'low':
                number_of_tanks = 5
            elif wing_mounting == 'high':
                number_of_tanks = 3
            tank_configuration = 'wing_all_tanks'
            # Print.
            runtime_output.print('Valid tank configuration found in acXML: ' + tank_configuration)
        elif valid_wing_with_trim_tank and not additional_center_tank:
            if wing_mounting == 'low':
                number_of_tanks = 6
            elif wing_mounting == 'high':
                number_of_tanks = 4
            tank_configuration = 'wing_with_trim_tank'
            # Print.
            runtime_output.print('Valid tank configuration found in acXML: ' + tank_configuration)
        elif valid_wing_with_additional_center_tank and not trim_tank:
            if wing_mounting == 'low':
                number_of_tanks = 6
            elif wing_mounting == 'high':
                number_of_tanks = 4
            tank_configuration = 'wing_with_additional_center_tank'
            # Print.
            runtime_output.print('Valid tank configuration found in acXML: ' + tank_configuration)
        elif valid_wing_with_additional_center_and_trim_tank:
            if wing_mounting == 'low':
                number_of_tanks = 7
            elif wing_mounting == 'high':
                number_of_tanks = 5
            tank_configuration = 'wing_with_additional_center_and_trim_tank'
            # Print.
            runtime_output.print('Valid tank configuration found in acXML: ' + tank_configuration)
        # Invalid tank configuration.
        else:
            number_of_tanks = 0
            tank_configuration = ''
            runtime_output.critical('Error: Invalid tank definition. Program aborted.')
            sys.exit('Exit information: Invalid tank definition in aircraft exchange file.')

        # Prepare output.
        dict_tank_design['number_of_tanks'] = number_of_tanks
        dict_tank_design['tank_configuration'] = tank_configuration

    """Liquid hydrogen."""
    # TODO Implement liquid hydrogen.
    if dict_ac_data['energy_carrier'] == 'liquid_hydrogen':
        combinations = {
            # Standard integral wing tanks.
            'fuselage': 
                {'tailcone': False,
                'rear': False,
                'front': False,
                'top': False,
                'bottom': False},
            # External wing pods.
            'wing':
                {'pods': False}
        }

        runtime_output.critical('Error: Check for valid tank configuration not yet implemented for liquid hydrogen '
                             + 'configuration. Program aborted.')
        sys.exit('Exit information: Missing function for liquid hydrogen configuration.')

    return dict_tank_design
