# UNICADO - UNIversity Conceptual Aircraft Design and Optimization
#
# Copyright (C) 2024 UNICADO consortium
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# Description:
# This file is part of UNICADO.

# Import standard modules.
from math import pi as PI


def calculate_parameter_cylinder(diameter, length, thickness_wall=0.003, thickness_insulation=0.20,
                                 density_material_wall=2850, density_material_insulation=32):
    """Calculate parameters for cylinder.

    This function calculates the following parameters of the cylindrical section of the tank:
        * Volume
        * Mass of the tank wall
        * Mass of the insulation
    The calculation of the volume is based on geometric volume estimates. The volume of the tank wall
    and insulation are determined based on a hollow cylinder. The mass can then be determined using the
    corresponding material density.

    :param float diameter: Diameter of the cylindrical tank section in m
    :param float length: Length of the cylindrical tank section in m
    :param float thickness_wall: Wall thickness in m, defaults to 0.003
    :param float thickness_insulation: Insulation thickness in m, defaults to 0.20
    :param int density_material_wall: Density of tank wall material in kg/m^3, defaults to 2850
    :param int density_material_insulation: Density of insulation material in kg/m^3, defaults to 32
    :return various: Volume of tank in m^3, mass of wall in kg, mass of insulation in kg
    """
    radius_excl_wall = diameter*0.5
    radius_incl_wall = radius_excl_wall + thickness_wall
    radius_incl_insulation = radius_incl_wall + thickness_insulation
    # Volume calculation for tank, based on standard cylinder volume.
    volume_tank = PI*radius_excl_wall**2*length
    # Volume calculation for wall, based on volume calculation of hollow cylinder.
    volume_wall = PI*length*(radius_incl_wall**2 - radius_excl_wall**2)
    # Volume calculation for insulation, based on volume calculation of hollow cylinder.
    volume_insulation = PI*length*(radius_incl_insulation**2 - radius_incl_wall**2)
    # Mass calculation of wall and insulation.
    mass_wall = volume_wall*density_material_wall
    mass_insulation = volume_insulation*density_material_insulation

    return volume_tank, mass_wall, mass_insulation
