/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "controlDeviceMethods.h"

#include <aircraftGeometry2/processing/measure.h>
#include <aircraftGeometry2/processing/transform.h>

std::tuple<::geom2::MultisectionSurface<geom2::PolygonSection>, std::tuple<double, double>> control_device(
    const std::string& name, double inner_chord_rel_pos_from, double inner_chord_rel_pos_to,
    double inner_spanwise_rel_pos, double outer_chord_rel_pos_from, double outer_chord_rel_pos_to,
    double outer_spanwise_rel_pos, double max_negative_deflection, double max_positive_deflection);

std::tuple<::geom2::MultisectionSurface<geom2::PolygonSection>, std::tuple<double, double>> control_device(
    const std::string& name, double inner_chord_rel_pos_from, double inner_chord_rel_pos_to,
    double inner_spanwise_rel_pos, double outer_chord_rel_pos_from, double outer_chord_rel_pos_to,
    double outer_spanwise_rel_pos, double max_negative_deflection, double max_positive_deflection) {
  geom2::MultisectionSurface<geom2::PolygonSection> control_device;

  geom2::Polygon_2 inner_chord;
  inner_chord.push_back(geom2::Point_2(inner_chord_rel_pos_from, 0.0));
  inner_chord.push_back(geom2::Point_2(inner_chord_rel_pos_to, 0.0));

  geom2::Polygon_2 outer_chord;
  outer_chord.push_back(geom2::Point_2(outer_chord_rel_pos_from, 0.0));
  outer_chord.push_back(geom2::Point_2(outer_chord_rel_pos_to, 0.0));

  control_device.sections.emplace_back(inner_chord);
  control_device.sections.emplace_back(outer_chord);

  control_device.sections.front().origin = geom2::Point_3(0.0, 0.0, inner_spanwise_rel_pos);
  control_device.sections.back().origin = geom2::Point_3(0.0, 0.0, outer_spanwise_rel_pos);

  control_device.name = name;

  std::tuple<double, double> control_device_deflections = {max_negative_deflection, max_positive_deflection};
  return {control_device, control_device_deflections};
}

std::tuple<std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>, std::vector<std::tuple<double, double>>>
user_defined::control_devices(std::vector<ControlDevice> devices) {
  std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> controls;
  std::vector<std::tuple<double, double>> controls_deflection;
  for (auto item : devices) {
    auto [control_device_geometry, control_device_deflections] =
        control_device(item.type.value(), item.position.inner_position_chord_from.value(),
                       item.position.inner_position_chord_to.value(), item.position.inner_position_spanwise.value(),
                       item.position.outer_position_chord_from.value(), item.position.outer_position_chord_to.value(),
                       item.position.outer_position_spanwise.value(), item.deflection_full_negative.value(),
                       item.deflection_full_positive.value());
    controls.push_back(control_device_geometry);
    controls_deflection.push_back(control_device_deflections);
  }
  return {controls, controls_deflection};
}

std::tuple<std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>, std::vector<std::tuple<double, double>>>
empirical::control_devices(const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& spars,
                           const geom2::MultisectionSurface<geom2::AirfoilSection>& wing,
                           const geom2::MultisectionSurface<geom2::PolygonSection>& fuselage, const bool is_kinked,
                           const double eta_kink, const std::string& config_leading_edge_high_lift_device_type,
                           const std::string& config_trailing_edge_high_lift_device_type) {
  std::vector<geom2::MultisectionSurface<geom2::PolygonSection>> controls;
  std::vector<std::tuple<double, double>> controls_deflection;

  /* Calculate leading edge high lift devices */
  /* Assumption spars relative chord position is constant */
  const double eta_fuselage = geom2::measure::width_max(fuselage) / geom2::measure::span(wing);
  const double half_span = std::abs(geom2::measure::span(wing) * 0.5);

  const double front_spar_rel_chord_pos_inner = spars.front().sections.front().get_contour().vertices().at(0).x();
  const double front_spar_rel_chord_pos_outer = spars.front().sections.back().get_contour().vertices().at(0).x();
  std::tuple<geom2::MultisectionSurface<geom2::PolygonSection>, std::tuple<double, double>> device;
  /* Calculate slats */
  if (is_kinked) {
    device = control_device(config_leading_edge_high_lift_device_type, 0.0, front_spar_rel_chord_pos_inner,
                                 1.02 * eta_fuselage, 0.0, front_spar_rel_chord_pos_outer, 0.98*eta_kink, 0.0,
                                 40 * geom2::detail::to_radians);
    controls.push_back(std::get<0>(device));
    controls_deflection.push_back(std::get<1>(device));

    device = control_device(config_leading_edge_high_lift_device_type, 0.0, front_spar_rel_chord_pos_inner,
                                 1.02 * eta_kink, 0.0, front_spar_rel_chord_pos_outer, 0.95, 0.0,
                                 40 * geom2::detail::to_radians);
    controls.push_back(std::get<0>(device));
    controls_deflection.push_back(std::get<1>(device));
  } else {
    device = control_device(config_leading_edge_high_lift_device_type, 0.0, front_spar_rel_chord_pos_inner,
                                 1.02 * eta_fuselage, 0.0, front_spar_rel_chord_pos_outer, 0.95, 0.0,
                                 40 * geom2::detail::to_radians);
    controls.push_back(std::get<0>(device));
    controls_deflection.push_back(std::get<1>(device));
  }
  /* Calculate trailing edge high lift devices */
  /* Assumption spars relative chord position is constant */
  /* Calculate flaps */
  if (is_kinked) {
    /* Device from fuselage to kink*/
    device = control_device(config_trailing_edge_high_lift_device_type,0.75, 1.0,
                            1.02 * eta_fuselage,0.75, 1.0, 0.99 * eta_kink, 0.0,
                            45 * geom2::detail::to_radians);
    controls.push_back(std::get<0>(device));
    controls_deflection.push_back(std::get<1>(device));

    /* Device from kink to eta = 0.78 */
    device =
        control_device(config_trailing_edge_high_lift_device_type, 0.75, 1.0, 1.01 * eta_kink,
                       0.75, 1.0, 0.78, 0.0, 45 * geom2::detail::to_radians);
    controls.push_back(std::get<0>(device));
    controls_deflection.push_back(std::get<1>(device));

  } else {
    /* Devices from 1.02 eta_fuselage to eta = 0.78 */
    device = control_device(config_trailing_edge_high_lift_device_type, 0.75 , 1.0,
                            1.02 * eta_fuselage, 0.75 , 1.0,
                            0.98 * 0.5 * (1.02 * eta_fuselage + .78), 0.0, 45 * geom2::detail::to_radians);
    controls.push_back(std::get<0>(device));
    controls_deflection.push_back(std::get<1>(device));

    device = control_device(config_trailing_edge_high_lift_device_type, 0.75 * 1.1, 1.0,
                            1.02 * 0.5 * (1.02 * eta_fuselage + .78), 0.75, 1.0, 0.78, 0.0,
                            45 * geom2::detail::to_radians);
    controls.push_back(std::get<0>(device));
    controls_deflection.push_back(std::get<1>(device));
  }

  /* Calculate aileron devices */
  device = control_device("aileron", 0.7, 1.0, 0.8, 0.7, 1.0, 0.98, -20. * geom2::detail::to_radians,
                          20. * geom2::detail::to_radians);
  controls.push_back(std::get<0>(device));
  controls_deflection.push_back(std::get<1>(device));

  /* Calculate spoilers */
  /* spoiler ground */
  if (is_kinked) {
    device =
        control_device("spoiler_ground", 0.7, 0.5 * (1 + 0.7),
                       0.5 * (0.98 * eta_kink + 1.02 * eta_fuselage), 0.7,
                       0.5 * (1+0.7), 0.98 * eta_kink, 0.0, 45 * geom2::detail::to_radians);
    controls.push_back(std::get<0>(device));
    controls_deflection.push_back(std::get<1>(device));

    /* spoiler air */
    int number_of_outboard_spoilers = 4;
    const double delta_eta_out_fraction = (0.95 * 0.8 - 1.02 * eta_kink) / number_of_outboard_spoilers;
    double eta_from = 1.02 * eta_kink;
    double eta_to = eta_from + delta_eta_out_fraction;
    for (int i = 0; i < number_of_outboard_spoilers; ++i) {
      device = control_device("spoiler_air", 0.7, 0.5 * (1 + 0.7),
                              eta_from, 0.7, 0.5 * (1 + 0.7),
                              eta_to, 0.0, 45 * geom2::detail::to_radians);
      controls.push_back(std::get<0>(device));
      controls_deflection.push_back(std::get<1>(device));

      eta_from = eta_to;
      eta_to += delta_eta_out_fraction;
    }
  } else {
    int number_of_spoilers = 5;
    const double delta_eta_out_fraction = (0.95 * 0.8 - 1.02 * eta_fuselage) / number_of_spoilers;
    double eta_from = 1.02 * eta_fuselage;
    double eta_to = eta_from + delta_eta_out_fraction;
    for (int i = 0; i < number_of_spoilers; ++i) {
      device = control_device("spoiler", 0.7, 0.5 * (1 + 0.7),
                              eta_from, 0.7, 0.5 * (1 + 0.7),
                              eta_to, 0.0, 45 * geom2::detail::to_radians);
      controls.push_back(std::get<0>(device));
      controls_deflection.push_back(std::get<1>(device));
    }
  }
  return {controls, controls_deflection};
}

std::tuple<std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>, std::vector<std::tuple<double, double>>>
get_control_devices(modes control_device_mode, std::vector<ControlDevice> devices,
                    const std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>& spars,
                    const geom2::MultisectionSurface<geom2::AirfoilSection>& wing,
                    const geom2::MultisectionSurface<geom2::PolygonSection>& fuselage, const bool is_kinked,
                    const double eta_kink, const std::string& config_leading_edge_high_lift_device_type,
                    const std::string& config_trailing_edge_high_lift_device_type) {
  std::map<
      std::string,
      std::tuple<std::string, std::function<std::tuple<std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>,
                                                       std::vector<std::tuple<double, double>>>(void)>>>
      control_device_selector = {{"mode_0", {"user_defined", [=]() { return user_defined::control_devices(devices); }}},
                                 {"mode_1", {"empirical", [=]() {
                                               return empirical::control_devices(
                                                   spars, wing, fuselage, is_kinked, eta_kink,
                                                   config_leading_edge_high_lift_device_type,
                                                   config_trailing_edge_high_lift_device_type);
                                             }}}};
  const auto [method, control_devices_method] = control_device_selector[control_device_mode];

  std::tuple<std::vector<geom2::MultisectionSurface<geom2::PolygonSection>>, std::vector<std::tuple<double, double>>>
      output_control_devices = control_devices_method();
  std::cout << "Control devices [method]: " << method << "\n";
  std::cout << "Control devices [values]: "
            << "\n";
  auto geometry = std::get<0>(output_control_devices);
  auto deflections = std::get<1>(output_control_devices);
  for (size_t i = 0; i < geometry.size(); ++i) {
    const auto [min_defl, max_defl] = deflections.at(i);
    std::cout << "\tName              : " << geometry.at(i).name << "\n";
    std::cout << "\tDeflections       : " << min_defl * geom2::detail::to_degrees << "\t"
              << max_defl * geom2::detail::to_degrees << "\n";
  }

  return output_control_devices;
}
