/*
 * UNICADO - UNIversity Conceptual Aircraft Design and Optimization
 *
 * Copyright (C) 2024 UNICADO consortium
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 *
 * Description:
 * This file is part of UNICADO.
 */

#include "wingProfileMethods.h"

#include <atmosphere/atmosphere.h>
#include <unitConversion/constants.h>

#include <algorithm>
#include <cmath>
#include <functional>
#include <iostream>
#include <map>

bool check_user_inputs(std::vector<std::tuple<double, double, std::string>>& to_check);
bool check_minimum_amount(const std::vector<std::tuple<double, double, std::string>>& to_check);

std::vector<std::tuple<double, double, std::string>> user_defined::wing_profile_and_thickness_distribution(
    const std::vector<std::tuple<double, double, std::string>> user_wing_profiles_and_thickness_to_chord,
    const double eta_fuselage, std::vector<double> eta_positions_additional) {
  std::vector<std::tuple<double, double, std::string>> output = user_wing_profiles_and_thickness_to_chord;
  if (!check_user_inputs(output)) {
    throw("Error - No root or tip elements set ... abort!\n");
  }

  if (!check_minimum_amount(output)) {
    throw("Error - Not enough user inputs ... abort!\n");
  }

  // Adding additional parts
  for (auto eta : eta_positions_additional) {
    if (std::find_if_not(output.begin(), output.end(), [eta](std::tuple<double, double, std::string>& to_check) {
          return (fabs(std::get<0>(to_check) - eta) > 1E-4);
        }) == output.end()) {
      // Add additional eta
      output.push_back({eta, 0.0, ""});
    }
  }

  // Sort output vector
  std::sort(output.begin(), output.end(),
            [](std::tuple<double, double, std::string>& left, std::tuple<double, double, std::string>& right) {
              return std::get<0>(left) < std::get<0>(right);
            });

  // Constant thickness within fuselage
  double root_thickness = std::get<1>(output.front());
  for (size_t i = 1; i < output.size(); ++i) {
    if (std::get<0>(output[i]) < eta_fuselage) {
      output[i] = {std::get<0>(output[i]), root_thickness, std::get<2>(output[i])};
    }
  }
  // Compute thickness
  for (size_t i = 1; i < output.size() - 1; ++i) {
    // If no profile -> additional eta found!
    if (std::get<2>(output[i]).compare("") == 0) {
      for (size_t j = i + 1; j < output.size(); ++j) {
        // If next profile found -> therefore thickness can be computed
        if (std::get<2>(output[j]).compare("") != 0) {
          // Get normalized distance between left and right thickness
          double x_normalized = (std::get<0>(output[i]) - std::get<0>(output[i - 1])) /
                                (std::get<0>(output[j]) - std::get<0>(output[i - 1]));

          // Compute distance by linear interpolation
          double thickness = std::lerp(std::get<1>(output[i - 1]), std::get<1>(output[j]), x_normalized);

          // Set current eta, thickness value and profile
          output[i] = {std::get<0>(output[i]), thickness, std::get<2>(output[i - 1])};

          // Break current iteration and look for next element
          break;
        }
      }
    }
  }
  // Return rearranged and cleaned vector
  return output;
}

std::vector<std::tuple<double, double, std::string>> torenbeek_jenkinson::wing_profile_and_thickness_distribution(
    const std::string& wing_profile, const double maximum_thickness_to_chord_ratio,
    const double technology_factor_airfoil, const double mach_design, const double design_altitude,
    const double quarter_chord_sweep, const double mach_drag_divergence, const double wing_loading,
    const double eta_fuselage, std::vector<double> eta_positions) {
  constexpr double climb_mass_loss = 0.02;
  constexpr double kappa_air = 1.4;
  atmosphere isa;
  const double pressure_design_altitude = isa.getPressureISA(design_altitude);
  const double mach_drag_divergence_effective = (mach_design + mach_drag_divergence) * sqrt(cos(quarter_chord_sweep));
  const double lift_coefficient = 2 * G_FORCE * wing_loading * (1 - climb_mass_loss) /
                                  (kappa_air * pressure_design_altitude * pow(mach_design, 2.));

  const double thickness_to_chord_ratio_mean =
      0.3 * cos(quarter_chord_sweep) *
      pow(((1 - pow((5 + pow(mach_drag_divergence_effective, 2)) /
                        (5 + pow(technology_factor_airfoil - 0.25 * lift_coefficient, 2)),
                    3.5))) *
              (sqrt(1 - pow(mach_drag_divergence_effective, 2.)) / (pow(mach_drag_divergence_effective, 2))),
          2. / 3.);

  const double root_thickness_to_chord = maximum_thickness_to_chord_ratio;
  const double tip_thickness_to_chord = (4. * thickness_to_chord_ratio_mean - root_thickness_to_chord) / 3.;

  auto thickness_distribution_function = [=](const double eta) {
    if (eta < eta_fuselage) {
      return root_thickness_to_chord;
    } else {
      double m = (tip_thickness_to_chord - root_thickness_to_chord) / (1. - eta_fuselage);
      double b = root_thickness_to_chord;
      return m * (eta - eta_fuselage) + b;
    }
  };
  // Add elements in output vector eta, t/c, profile name
  std::vector<std::tuple<double, double, std::string>> output;
  for (auto eta : eta_positions) {
    output.push_back({eta, thickness_distribution_function(eta), wing_profile});
  }

  // Sort output
  std::sort(output.begin(), output.end(),
            [](std::tuple<double, double, std::string>& left, std::tuple<double, double, std::string>& right) {
              return std::get<0>(left) < std::get<0>(right);
            });
  return output;
}

// Helper Methods
/**
 * \brief Checks if root and tip profile exist
 *
 * \param to_check vector container of tuples <eta, t/c, Profile name>
 * \return true tip & root exist
 * \return false tip or root does not exist
 */
bool check_user_inputs(std::vector<std::tuple<double, double, std::string>>& to_check) {
  // Check if root exists
  if (std::find_if(to_check.begin(), to_check.end(), [](std::tuple<double, double, std::string>& root) {
        return (fabs(std::get<0>(root) - 0.0) < 1E-4);
      }) == to_check.end()) {
    return false;
  }

  // Check if tip exists
  if (std::find_if(to_check.begin(), to_check.end(), [](std::tuple<double, double, std::string>& tip) {
        return (fabs(std::get<0>(tip) - 1.0) < 1E-4);
      }) == to_check.end()) {
    return false;
  }
  return true;
}

/**
 * \brief Check if minimum two profiles exist
 *
 * \param to_check vector container of tuples <eta, t/c, Profile name>
 * \return true two profiles exists
 * \return false less than two profiles exist
 */
bool check_minimum_amount(const std::vector<std::tuple<double, double, std::string>>& to_check) {
  if (to_check.size() < 2) {
    return false;
  }
  return true;
}

// == SELECTION ================================================================
std::vector<std::tuple<double, double, std::string>> get_wing_profile_and_thickness(
    modes wing_profile_and_thickness,
    const std::vector<std::tuple<double, double, std::string>> user_wing_profile_and_thickess,
    const std::string& wing_profile, const double maximum_thickness_to_chord_ratio,
    const double technology_factor_airfoil, const double mach_design, const double design_altitude,
    const double quarter_chord_sweep, const double mach_drag_divergence, const double wing_loading,
    const double eta_fuselage, std::vector<double> eta_positions) {
  std::map<modes, std::tuple<std::string, std::function<std::vector<std::tuple<double, double, std::string>>(void)>>> wing_profile_and_thickness_selection = {{"mode_0",
                                                                                                                                                               {"user_defined",
                                                                                                                                                                [=]() {
                                                                                                                                                                  return user_defined::
                                                                                                                                                                      wing_profile_and_thickness_distribution(
                                                                                                                                                                          user_wing_profile_and_thickess,
                                                                                                                                                                          eta_fuselage,
                                                                                                                                                                          eta_positions);
                                                                                                                                                                }}},
                                                                                                                                                              {"mode_1", {"torenbeek_jenkinson", [=]() {
                                                                                                                                                                            return torenbeek_jenkinson::wing_profile_and_thickness_distribution(
                                                                                                                                                                                wing_profile,
                                                                                                                                                                                maximum_thickness_to_chord_ratio,
                                                                                                                                                                                technology_factor_airfoil,
                                                                                                                                                                                mach_design,
                                                                                                                                                                                design_altitude,
                                                                                                                                                                                quarter_chord_sweep,
                                                                                                                                                                                mach_drag_divergence,
                                                                                                                                                                                wing_loading,
                                                                                                                                                                                eta_fuselage,
                                                                                                                                                                                eta_positions);
                                                                                                                                                                          }}}};
  const auto [method, wing_profile_and_thickness_vector_method] =
      wing_profile_and_thickness_selection[wing_profile_and_thickness];
  std::vector<std::tuple<double, double, std::string>> wing_profile_and_thickness_vector =
      wing_profile_and_thickness_vector_method();
  std::cout << "Wing profile and thickness [method]: " << method << "\n";
  std::cout << "Wing profile and thickness [values]: "
            << "\n";
  for (auto item : wing_profile_and_thickness_vector) {
    const auto [eta, t_to_c, profile_name] = item;
    std::cout << "\tEta               : " << eta << "\n";
    std::cout << "\tThickness to chord: " << t_to_c << "\n";
    std::cout << "\tProfile           : " << profile_name << "\n";
  }
  return wing_profile_and_thickness_vector;
}